#!/bin/bash

# This script builds the tranalyzer2 bin and the default plugins. If it is
# executed with the argument "install", the tranalyzer binary is installed into
# "/usr/local/bin"
#
# If you want to build additional plugins  you need to go into their source
# folders and execute ./autogen.sh


DEBUG_MODE=0
DO_INSTALL=0
NOCONFIG=0

function USAGE ()
{
    echo ""
    echo "USAGE:"
    echo "    autogen.sh [-idht]"
    echo ""
    echo "OPTIONS:"
    echo "    -i  install binary in system PATH directory (Must be first argument if specified!)"
    echo "    -d  compile in debug and profiling mode"
    echo "    -h  show this usage information"
    echo "    -t  tranalyzer install directory"
    echo ""
    exit $E_OPTERROR    # Exit and explain usage, if no argument(s) given.
}

#PROCESS ARGS
while getopts "indht:" Option
do
    case $Option in
        i    ) DO_INSTALL=1;;
        n    ) NOCONFIG=1;;
        d    ) DEBUG_MODE=1;;
        t    ) TRANALYZER_INSTALL_DIR=$OPTARG;;
        h    ) USAGE
               exit 0;;
        *    ) echo ""
               echo "Unimplemented or incomplete option chosen!"
               USAGE   # DEFAULT
    esac
done

if [ "$DO_INSTALL" == 1 ]; then
	# Shift arguments one left to remove install argument which is not recognised by the subscripts
	shift
fi

echo "autogen: building Tranalyzer2"

cd tranalyzer2
./autogen.sh $@
make

if [ "$DO_INSTALL" == 1 ]; then
	if [[ $EUID -ne 0 ]]; then
		echo "********************************************************************************"
		echo "* The install option can only be used as root! Try 'sudo ./autogen.sh -i' *"
		echo "********************************************************************************"
		exit 1
	else
		echo "autogen: installing tranalyzer2 in /usr/local/bin"
		make install
	fi
fi
cd ..

#echo "autogen: building plugin 'standardFileSink'"
#cd standardFileSink
#./autogen.sh $@
#cd ..

echo "autogen: building plugin 'textFileSink'"
cd textFileSink
./autogen.sh $@
cd ..

echo "autogen: building plugin 'basicFlowOutput'"
cd basicFlowOutput
./autogen.sh $@
cd ..

echo "autogen: building plugin 'basicLayer4CalcStatistics'"
cd basicLayer4CalcStatistics
./autogen.sh $@
cd ..

echo "autogen: building plugin 'connectionCounter'"
cd connectionCounter
./autogen.sh $@
cd ..

echo "autogen: building plugin 'icmpDecode'"
cd icmpDecode
./autogen.sh $@
cd ..

echo "autogen: building plugin 'protocolStatistics'"
cd protocolStatistics
./autogen.sh $@
cd ..

echo "autogen: building plugin 'tcpFlags'"
cd tcpFlags
./autogen.sh $@
cd ..

echo "autogen: building plugin 'tcpStates'"
cd tcpStates
./autogen.sh $@
cd ..

echo "autogen: building plugin 'macRecorder'"
cd macRecorder
./autogen.sh $@
cd ..

echo "autogen: building plugin 'portBasedClassifier'"
cd portBasedClassifier
./autogen.sh $@
cd ..
