#!/bin/bash

# This script builds the tranalyzer2 and all plugins in the tranalyzer2/trunk directory. 
# If it is executed with the argument "install", the tranalyzer binary is installed into
# "/usr/local/bin"
#


DEBUG_MODE=0
DO_INSTALL=0
NOCONFIG=0

function USAGE ()
{
    echo ""
    echo "USAGE:"
    echo "    autogen.sh [-idht]"
    echo ""
    echo "OPTIONS:"
    echo "    -i  install binary in system PATH directory (Must be first argument if specified!)"
    echo "    -d  compile in debug and profiling mode"
    echo "    -h  show this usage information"
    echo "    -t  tranalyzer install directory"
    echo ""
    exit $E_OPTERROR    # Exit and explain usage, if no argument(s) given.
}

#PROCESS ARGS
while getopts "indht:" Option
do
    case $Option in
        i    ) DO_INSTALL=1;;
        n    ) NOCONFIG=1;;
        d    ) DEBUG_MODE=1;;
        t    ) TRANALYZER_INSTALL_DIR=$OPTARG;;
        h    ) USAGE
               exit 0;;
        *    ) echo ""
               echo "Unimplemented or incomplete option chosen!"
               USAGE   # DEFAULT
    esac
done

if [ "$DO_INSTALL" == 1 ]; then
	# Shift arguments one left to remove install argument which is not recognised by the subscripts
	shift
fi

echo "autogen: building Tranalyzer2" 

cd tranalyzer2
./autogen.sh $@
make

if [ "$DO_INSTALL" == 1 ]; then
	if [[ $EUID -ne 0 ]]; then
		echo "********************************************************************************"
		echo "* The install option can only be used as root! Try 'sudo ./autogen.sh -i' *"
		echo "********************************************************************************"
		exit 1
	else
		echo "autogen: installing tranalyzer2 in /usr/local/bin"
		make install
	fi
fi
cd ..

list=`ls | awk -v a=$PWD '{ if ($0 != "" && $0 != "tranalyzer_plugin_skeleton" && $0 != "tranalyzer2" && $0 != "scripts" && !index($0, "autogen")) print a"/"$0;}'`

for plugin in $list; do
	echo 
	echo "building plugin: "$plugin
	cd $plugin
	./autogen.sh $@
	cd ..
done
