/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "tranalyzer.h"
#include "main.h"
#include "basicFlowOutput.h"
#include "packetCapture.h"
#include <net/ethernet.h>


char* get_plugin_name() {
	return "basicFlowOutput";
}


char* get_plugin_version() {
	return "0.5.8";
}


unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}


unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}


void initialize(unsigned long pPluginIndex) {
	pluginIndexBasicFlowOutput = pPluginIndex;
#ifdef __APPLE__
	mainHashMap = getMainHashMap(); 
#endif // __APPLE__
#if ((BASICFLOWOUTPUT_MPLS == 1 && MAX_MPLS_DEPTH > 0) || BASICFLOWOUTPUT_PPP == 1 ||  BASICFLOWOUTPUT_L2TP == 1 || BASICFLOWOUTPUT_GRE == 1 || (BASICFLOWOUTPUT_ETHERTYPE == 1 && BF_MAX_ETHTYPE_DEPTH > 0)) 
	size_t data_size = sizeof(encapsulateProtPerFlow_t) * mainHashMap->hashChainTableSize;
	encapsulateProtPerFlow = (encapsulateProtPerFlow_t*) malloc(data_size);
	memset(encapsulateProtPerFlow, '\0', data_size); // redundant, is intitialized at onflowgenerated
#endif // BASICFLOWOUTPUT_MPLS ...

#if (BASICFLOWOUTPUT_SUBNET_TEST == 1 || BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1 || BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1)
	char user_plugin_folder[256];
	memset(user_plugin_folder,'\0', 256);

	if (pluginFolder) {
		strncat(user_plugin_folder, pluginFolder, 255 - strlen(user_plugin_folder));
		strncat(user_plugin_folder, "/", 1);
	} else {
		strncpy(user_plugin_folder, getenv("HOME"),255);
		strncat(user_plugin_folder, "/", 1);
		strncat(user_plugin_folder, PLUGIN_FOLDER, 255 - strlen(user_plugin_folder));
	}

	strncat(user_plugin_folder, SUBNETFILE, 255 - strlen(user_plugin_folder));

	bfo_subnet_table = malloc(sizeof(subnettable_t));
	subnet_init(bfo_subnet_table, user_plugin_folder);
#endif // BASICFLOWOUTPUT_SUBNET_TEST 
}


binary_value_t* printHeader() {
	binary_value_t *bfo_main_bv = NULL, *bfo_act_bv = NULL;
	
#if BASICFLOWOUTPUT_FLOWINDEX != 0
	bfo_act_bv = bv_new_bv("Flow Index", "flowInd", 0, 1, bt_int_32);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_FLOWINDEX

#if BASICFLOWOUTPUT_FLOW_STATUS != 0
	bfo_act_bv = bv_new_bv("Flow Status", "flowStat", 0, 1, bt_hex_32);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_FLOW_STATUS

#if BASICFLOWOUTPUT_DATE_TIME != 0
	bfo_act_bv = bv_new_bv("Date Time of first packet", "DateTimeFirst", 0, 1, bt_time);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);

	bfo_act_bv = bv_new_bv("Date Time of last packet", "DateTimeLast", 0, 1, bt_time);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#else // BASICFLOWOUTPUT_DATE_TIME == 0
	bfo_act_bv = bv_new_bv("System time of first packet", "UnixTimeFirst", 0, 1, bt_unix_time);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);

	bfo_act_bv = bv_new_bv("System time of last packet", "UnixTimeLast", 0, 1, bt_unix_time);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_DATE_TIME

	bfo_act_bv = bv_new_bv("Flow duration", "Duration", 0, 1, bt_unix_time);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#if (BASICFLOWOUTPUT_ETHERTYPE == 1 && BF_MAX_ETHTYPE_DEPTH > 0)
#if BASICFLOWOUTPUT_ETHERTYPE_HEX == 1
	bfo_act_bv = bv_new_bv("Ether Type: Vlan, ..", "ETHTypeVlan", 1, 1, bt_hex_32);	
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#else // BASICFLOWOUTPUT_ETHERTYPE_HEX == 0
	bfo_act_bv = bv_new_bv("Ether VlanID", "ETHVlanID", 1, 1, bt_int_16);	
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_ETHERTYPE_HEX
#endif // BASICFLOWOUTPUT_ETHERTYPE

#if (BASICFLOWOUTPUT_MPLS == 1 && MAX_MPLS_DEPTH > 0)
#if BASICFLOWOUTPUT_MPLS_DETAIL == 1
	bfo_act_bv = bv_new_bv("MPLS tags detail", "MPLSLabel_Tos_S_TTL", 1, 4, bt_uint_32, bt_uint_8, bt_uint_8, bt_uint_8);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#else // BASICFLOWOUTPUT_MPLS_DETAIL == 0
	bfo_act_bv = bv_new_bv("MPLS tag hex", "MplsTagsHex", 1, 1, bt_hex_32);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif
#endif // (BASICFLOWOUTPUT_MPLS == 1 && MAX_MPLS_DEPTH > 0) 

#if BASICFLOWOUTPUT_PPP == 1
	bfo_act_bv = bv_new_bv("PPP header", "PppHdr", 0, 1, bt_hex_32);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif


#if BASICFLOWOUTPUT_L2TP == 1
	bfo_act_bv = bv_new_bv("L2TP header", "L2tpHdr", 0, 1, bt_hex_16);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
	bfo_act_bv = bv_new_bv("L2TP Tunnel ID", "L2tpTid", 0, 1, bt_hex_16);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
	bfo_act_bv = bv_new_bv("L2TP Session ID", "L2tpSid", 0, 1, bt_hex_16);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	bfo_act_bv = bv_new_bv("L2TP source IP address hex", "L2TPsrcIPx", 0, 1, bt_hex_32);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	bfo_act_bv = bv_new_bv("L2TP source IP address", "L2TPsrcIP", 0, 1, bt_ip4_addr);
#endif // BASICFLOWOUTPUT_IP_HEX
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1
#if NETWORK_HEXLABEL == 1
	bfo_act_bv = bv_new_bv("Subnet number of L2TP source IP", "subnet_L2TPsrcIPx", 0, 1, bt_hex_32);
#else // NETWORK_HEXLABEL == 0 
	bfo_act_bv = bv_new_bv("Subnet number of L2TP source IP", "subnet_L2TPsrcIP", 0, 1, bt_uint_32);
#endif // NETWORK_HEXLABEL
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_SUBNET_TEST_L2TP 

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	bfo_act_bv = bv_new_bv("L2TP destination IP address hex", "L2TPdstIPx", 0, 1, bt_hex_32);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	bfo_act_bv = bv_new_bv("L2TP destination IP address hex", "L2TPdstIP", 0, 1, bt_ip4_addr);
#endif // BASICFLOWOUTPUT_IP_HEX
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1
#if NETWORK_HEXLABEL == 1
	bfo_act_bv = bv_new_bv("Subnet number of L2TP destination IP", "SubnetL2TPdstIP", 0, 1, bt_hex_32);
#else // BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 0
	bfo_act_bv = bv_new_bv("Subnet number of L2TP destination IP", "SubnetL2TPdstIP", 0, 1, bt_uint_32);
#endif // NETWORK_HEXLABEL
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_SUBNET_TEST_L2TP 

#endif // BASICFLOWOUTPUT_L2TP == 1


#if BASICFLOWOUTPUT_GRE == 1
	bfo_act_bv = bv_new_bv("GRE header", "GREHdr", 0, 1, bt_hex_32);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	bfo_act_bv = bv_new_bv("GRE source IP address hex", "GREsrcIPx", 0, 1, bt_hex_32);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	bfo_act_bv = bv_new_bv("GRE source IP address", "GREsrcIP", 0, 1, bt_ip4_addr);
#endif // BASICFLOWOUTPUT_IP_HEX
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1
#if NETWORK_HEXLABEL == 1
	bfo_act_bv = bv_new_bv("Subnet number of GRE source IP", "subnet_GREsrcIPx", 0, 1, bt_hex_32);
#else // NETWORK_HEXLABEL == 0 
	bfo_act_bv = bv_new_bv("Subnet number of GRE source IP", "subnet_GREsrcIP", 0, 1, bt_uint_32);
#endif // NETWORK_HEXLABEL
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	bfo_act_bv = bv_new_bv("GRE destination IP address hex", "GREdstIPx", 0, 1, bt_hex_32);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	bfo_act_bv = bv_new_bv("GRE destination IP address hex", "GREdstIP", 0, 1, bt_ip4_addr);
#endif // BASICFLOWOUTPUT_IP_HEX
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // (AGGREGATIONFLAG & SUBNET) == 0

#if BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1
#if NETWORK_HEXLABEL == 1
	bfo_act_bv = bv_new_bv("Subnet number of GRE destination IP", "SubnetGREdstIP", 0, 1, bt_hex_32);
#else // BASICFLOWOUTPUT_SUBNET_TEST_GRE == 0
	bfo_act_bv = bv_new_bv("Subnet number of GRE destination IP", "SubnetGREdstIP", 0, 1, bt_uint_32);
#endif // NETWORK_HEXLABEL
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1

#endif // BASICFLOWOUTPUT_GRE == 1


#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
#if IPV6_ACTIVATE == 1
	bfo_act_bv = bv_new_bv("Source IP6 address hex", "SrcIP6x", 0, 2, bt_hex_64, bt_hex_64);
#else
	bfo_act_bv = bv_new_bv("Source IP address hex", "SrcIP4x", 0, 1, bt_hex_32);
#endif
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#else // BASICFLOWOUTPUT_IP_HEX == 0
#if IPV6_ACTIVATE == 1
	bfo_act_bv = bv_new_bv("Source IPv6 address", "SrcIP6", 0, 1, bt_ip6_addr);
#else // IPV6_ACTIVATE == 0
	bfo_act_bv = bv_new_bv("Source IPv4 address", "SrcIP4", 0, 1, bt_ip4_addr);
#endif // IPV6_ACTIVATE
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_IP_HEX
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST == 1
#if NETWORK_HEXLABEL == 1
	bfo_act_bv = bv_new_bv("Subnet number of source IPv4", "SubnetSrcIP4x", 0, 1, bt_hex_32);
#else // NETWORK_HEXLABEL == 0
	bfo_act_bv = bv_new_bv("Subnet number of source IPv4", "SubnetSrcIP4", 0, 1, bt_uint_32);
#endif // NETWORK_HEXLABEL
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_SUBNET_TEST 

	bfo_act_bv = bv_new_bv("Source port", "srcPort", 0, 1, bt_uint_16);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
#if IPV6_ACTIVATE == 1
	bfo_act_bv = bv_new_bv("Destination IP6 address hex", "DstIP6x" 0, 2, bt_hex_64, bt_hex_64);
#else // IPV6_ACTIVATE == 0
	bfo_act_bv = bv_new_bv("Destination IP4 address hex", "DstIP4x", 0, 1, bt_hex_32);
#endif // IPV6_ACTIVATE
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#else // BASICFLOWOUTPUT_IP_HEX == 0
#if IPV6_ACTIVATE == 1
	bfo_act_bv = bv_new_bv("Destination IP6 address", "DstIP6", 0, 1, bt_ip6_addr);
#else // IPV6_ACTIVATE == 0
	bfo_act_bv = bv_new_bv("Destination IP4 address", "DstIP4", 0, 1, bt_ip4_addr);
#endif // IPV6_ACTIVATE
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_IP_HEX
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST == 1
#if NETWORK_HEXLABEL == 1
	bfo_act_bv = bv_new_bv("Subnet number of destination IP", "SubnetdstIPx", 0, 1, bt_hex_32);
#else
	bfo_act_bv = bv_new_bv("Subnet number of destination IP", "SubnetdstIP", 0, 1, bt_uint_32);
#endif
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);
#endif // BASICFLOWOUTPUT_SUBNET_TEST == 1

	bfo_act_bv = bv_new_bv("Destination port", "DstPort", 0, 1, bt_uint_16);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);

	bfo_act_bv = bv_new_bv("Layer 4 protocol", "Layer4Proto", 0, 1, bt_uint_8);
	bfo_main_bv = bv_append_bv(bfo_main_bv, bfo_act_bv);

	return bfo_main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {

#if ((BASICFLOWOUTPUT_MPLS == 1 && MAX_MPLS_DEPTH > 0) || BASICFLOWOUTPUT_PPP == 1 ||  BASICFLOWOUTPUT_L2TP == 1 ||  BASICFLOWOUTPUT_GRE == 1 || (BASICFLOWOUTPUT_ETHERTYPE == 1 && BF_MAX_ETHTYPE_DEPTH > 0) ) 
	encapsulateProtPerFlow_t *encapsulateProtPerFlowPtr = &encapsulateProtPerFlow[flowIndex];
	memset(encapsulateProtPerFlowPtr, '\0', sizeof(encapsulateProtPerFlow_t));
	uint32_t i = 0;	
	uint32_t *u;
#endif // BASICFLOWOUTPUT_MPLS ...

#if (BASICFLOWOUTPUT_ETHERTYPE == 1 && BF_MAX_ETHTYPE_DEPTH > 0)
	if ( packet->status & L2_VLAN && packet->vlans) {
		u = encapsulateProtPerFlowPtr->ETHType;
		do {
			u[i+1] = ntohl(packet->vlans[i++]);
		} while ( i < BF_MAX_ETHTYPE_DEPTH && (u[i] & ETHERTYPE32MASKn) == VLAN_8021Qn32 ); 
		u[0] = i;
	} 
#endif // BASICFLOWOUTPUT_ETHERTYPE ..

#if (BASICFLOWOUTPUT_MPLS == 1 && MAX_MPLS_DEPTH > 0)	
	if ( packet->status & L2_MPLS) {
		u = (uint32_t*)encapsulateProtPerFlowPtr->mplsh;
		i = 0;
		do {
			u[i+1] = ntohl(packet->mpls[i++]);
		} while ( i < MAX_MPLS_DEPTH && !(packet->mpls[i-1] & BTM_MPLS_STKn32) );
		u[0] = i;
	}
#endif // BASICFLOWOUTPUT_MPLS

#if BASICFLOWOUTPUT_PPP == 1
	if ( packet->status & L2_PPP && packet->pppHdr) encapsulateProtPerFlowPtr->pppHdr = (pppHu_t)ntohl(packet->pppHdr->pppHdrc);
#endif // BASICFLOWOUTPUT_PPP

#if BASICFLOWOUTPUT_L2TP == 1
	if ( packet->status & L2_L2TP && packet->l2tpLayer3Hdr) {
		encapsulateProtPerFlowPtr->l2tp_srcIP = packet->l2tpLayer3Hdr->ipHeader.ip_src;
		encapsulateProtPerFlowPtr->l2tp_dstIP = packet->l2tpLayer3Hdr->ipHeader.ip_dst;
		encapsulateProtPerFlowPtr->l2tpHdrBF = ntohs(*(uint16_t*)packet->l2TPHdr);
		if (encapsulateProtPerFlowPtr->l2tpHdrTID & 0x4000) i = 1; else i = 0;
		encapsulateProtPerFlowPtr->l2tpHdrTID = ntohs(*((uint16_t*)packet->l2TPHdr+i+1));
		encapsulateProtPerFlowPtr->l2tpHdrSID = ntohs(*((uint16_t*)packet->l2TPHdr+i+2));
	}
#endif // BASICFLOWOUTPUT_L2TP == 1

#if BASICFLOWOUTPUT_GRE == 1
	if ( packet->status & L2_GRE && packet->greLayer3Hdr) {
		encapsulateProtPerFlowPtr->greHdrBF = *(uint32_t*)packet->greHdr;
		encapsulateProtPerFlowPtr->gre_srcIP = packet->greLayer3Hdr->ipHeader.ip_src;
		encapsulateProtPerFlowPtr->gre_dstIP = packet->greLayer3Hdr->ipHeader.ip_dst;
	}
#endif // BASICFLOWOUTPUT_GRE == 1

}


void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	currFlow = &flows[flowIndex];
	encapsulateProtPerFlow_t *encapsulateProtPerFlowPtr = &encapsulateProtPerFlow[flowIndex];

	uint32_t bfo_offset = 0, subnetNr, i = 0;
	
	char bfo_buffer[BFO_BUFFER_SZ_ACT], *a;   
	memset(bfo_buffer, '\0', BFO_BUFFER_SZ_ACT);
		
#if BASICFLOWOUTPUT_FLOWINDEX != 0
	memcpy(&bfo_buffer[bfo_offset], &currFlow->findex, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_FLOWINDEX

#if BASICFLOWOUTPUT_FLOW_STATUS != 0
	memcpy(&bfo_buffer[bfo_offset], &currFlow->status, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_FLOW_STATUS

	memcpy(&bfo_buffer[bfo_offset], &currFlow->firstSeen.tv_sec, sizeof(time_t));
	bfo_offset += sizeof(uint64_t);

	uint32_t temp_tv_usec = currFlow->firstSeen.tv_usec * 1000;
	memcpy(&bfo_buffer[bfo_offset], &temp_tv_usec, sizeof(suseconds_t));
	bfo_offset += sizeof(uint32_t);

	memcpy(&bfo_buffer[bfo_offset], &currFlow->lastSeen.tv_sec, sizeof(time_t));
	bfo_offset += sizeof(uint64_t);

	temp_tv_usec = currFlow->lastSeen.tv_usec * 1000;
	memcpy(&bfo_buffer[bfo_offset], &temp_tv_usec, sizeof(suseconds_t));
	bfo_offset += sizeof(uint32_t);

	memcpy(&bfo_buffer[bfo_offset], &currFlow->duration.tv_sec, sizeof(time_t));
	bfo_offset += sizeof(uint64_t);

	temp_tv_usec = currFlow->duration.tv_usec * 1000;
	memcpy(&bfo_buffer[bfo_offset], &temp_tv_usec, sizeof(suseconds_t));
	bfo_offset += sizeof(uint32_t);
	
#if (BASICFLOWOUTPUT_ETHERTYPE == 1 && BF_MAX_ETHTYPE_DEPTH > 0)
	uint16_t j;
	uint32_t k ,*v = encapsulateProtPerFlowPtr->ETHType;
	if (v[0] > 0) k = v[0] - 1; else k = 1;
	memcpy(&bfo_buffer[bfo_offset], &k, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
	for (i = 1; i <= k; i++) {
#if BASICFLOWOUTPUT_ETHERTYPE_HEX == 1
		memcpy(&bfo_buffer[bfo_offset], (v+i), sizeof(uint32_t));
		bfo_offset += sizeof(uint32_t);
#else // BASICFLOWOUTPUT_ETHERTYPE_HEX == 0
		j = v[i] & 0x0fff;
		memcpy(&bfo_buffer[bfo_offset], &j, sizeof(uint16_t));
		bfo_offset += sizeof(uint16_t);
#endif // BASICFLOWOUTPUT_ETHERTYPE_HEX
	}		
#endif // BASICFLOWOUTPUT_ETHERTYPE

#if (BASICFLOWOUTPUT_MPLS != 0 && MAX_MPLS_DEPTH > 0)
	mplsh_t *u = (mplsh_t*)encapsulateProtPerFlowPtr->mplsh;
	if (u->mplshu == 0) u->mplshu = 1;
	memcpy(&bfo_buffer[bfo_offset], (uint32_t*)u, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
	for (i = 1; i <= u->mplshu; i++) {
#if BASICFLOWOUTPUT_MPLS_DETAIL == 1
		uint32_t u1 = (uint32_t)u[i].mplshs.label;
		memcpy(&bfo_buffer[bfo_offset], &u1, sizeof(uint32_t));
		bfo_offset += sizeof(uint32_t);
		u1 = u[i].mplshs.Exp_ToS;
		memcpy(&bfo_buffer[bfo_offset], (uint8_t*)&u1, sizeof(uint8_t));
		bfo_offset += sizeof(uint8_t);
		u1 = u[i].mplshs.S;
		memcpy(&bfo_buffer[bfo_offset], (uint8_t*)&u1, sizeof(uint8_t));
		bfo_offset += sizeof(uint8_t);
		u1 = u[i].mplshs.TTL;
		memcpy(&bfo_buffer[bfo_offset], (uint8_t*)&u1, sizeof(uint8_t));
		bfo_offset += sizeof(uint8_t);
#else // hex output
		memcpy(&bfo_buffer[bfo_offset], (uint32_t*)(u+i), sizeof(uint32_t));
		bfo_offset += sizeof(uint32_t);
#endif
	}
#endif // BASICFLOWOUTPUT_MPLS != 0


#if BASICFLOWOUTPUT_PPP == 1
	memcpy(&bfo_buffer[bfo_offset], &encapsulateProtPerFlowPtr->pppHdr.pppHdrc, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_PPP == 1


#if BASICFLOWOUTPUT_L2TP == 1
	memcpy(&bfo_buffer[bfo_offset], &encapsulateProtPerFlowPtr->l2tpHdrBF, sizeof(uint16_t));
	bfo_offset += sizeof(uint16_t);
	memcpy(&bfo_buffer[bfo_offset], &encapsulateProtPerFlowPtr->l2tpHdrTID, sizeof(uint16_t));
	bfo_offset += sizeof(uint16_t);
	memcpy(&bfo_buffer[bfo_offset], &encapsulateProtPerFlowPtr->l2tpHdrSID, sizeof(uint16_t));
	bfo_offset += sizeof(uint16_t);

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	i = ntohl(encapsulateProtPerFlowPtr->l2tp_srcIP.s_addr); 
#else // BASICFLOWOUTPUT_IP_HEX == 0
	i = encapsulateProtPerFlowPtr->l2tp_srcIP.s_addr; 
#endif // BASICFLOWOUTPUT_IP_HEX
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1
	subnetNr = subnet_test(bfo_subnet_table, encapsulateProtPerFlowPtr->l2tp_srcIP.s_addr); // subnet test source ip
	memcpy(&bfo_buffer[bfo_offset], &subnetNr, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1 

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	i = ntohl(encapsulateProtPerFlowPtr->l2tp_dstIP.s_addr);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	i = encapsulateProtPerFlowPtr->l2tp_dstIP.s_addr;
#endif // BASICFLOWOUTPUT_IP_HEX
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1
	subnetNr = subnet_test(bfo_subnet_table, encapsulateProtPerFlowPtr->l2tp_dstIP.s_addr); // subnet test source ip
	memcpy(&bfo_buffer[bfo_offset], &subnetNr, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1 

#endif // BASICFLOWOUTPUT_L2TP == 1


#if BASICFLOWOUTPUT_GRE == 1
	i = ntohl(encapsulateProtPerFlowPtr->greHdrBF); 
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	i = ntohl(encapsulateProtPerFlowPtr->gre_srcIP.s_addr); 
#else // BASICFLOWOUTPUT_IP_HEX == 0
	i = encapsulateProtPerFlowPtr->gre_srcIP.s_addr; 
#endif // BASICFLOWOUTPUT_IP_HEX
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1
	subnetNr = subnet_test(bfo_subnet_table, encapsulateProtPerFlowPtr->gre_srcIP.s_addr); // subnet test source ip
	memcpy(&bfo_buffer[bfo_offset], &subnetNr, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_SUBNET_TEST 

#if (AGGREGATIONFLAG & SUBNET) == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	i = ntohl(encapsulateProtPerFlowPtr->gre_dstIP.s_addr);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	i = encapsulateProtPerFlowPtr->gre_dstIP.s_addr;
#endif // BASICFLOWOUTPUT_IP_HEX
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1
	subnetNr = subnet_test(bfo_subnet_table, encapsulateProtPerFlowPtr->gre_dstIP.s_addr); // subnet test source ip
	memcpy(&bfo_buffer[bfo_offset], &subnetNr, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_SUBNET_TEST 
#endif // BASICFLOWOUTPUT_GRE == 1


#if (AGGREGATIONFLAG & SUBNET) == 0
#if IPV6_ACTIVATE == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	i = ntohl(currFlow->srcIP.s_addr);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	i = currFlow->srcIP.s_addr;
#endif // BASICFLOWOUTPUT_IP_HEX
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#else // IPV6_ACTIVATE == 1
	a = (char*)&currFlow->srcIP.IPv6.s6_addr[0];	
	//a = (char*)&currFlow->srcIP.IPv6.__in6_u.u6_addr32[0];	
	memcpy(&bfo_buffer[bfo_offset], a, 16);
	bfo_offset += 16;
#endif // IPV6_ACTIVATE 
#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST == 1
	subnetNr = subnet_test(bfo_subnet_table, currFlow->srcIP.s_addr); // subnet test source ip
	memcpy(&bfo_buffer[bfo_offset], &subnetNr, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_SUBNET_TEST

	memcpy(&bfo_buffer[bfo_offset], &currFlow->srcPort, sizeof(uint16_t));
	bfo_offset += sizeof(uint16_t);

#if (AGGREGATIONFLAG & SUBNET) == 0
#if IPV6_ACTIVATE == 0
#if BASICFLOWOUTPUT_IP_HEX == 1
	i = ntohl(currFlow->dstIP.s_addr);
#else // BASICFLOWOUTPUT_IP_HEX == 0
	i = currFlow->dstIP.s_addr;
#endif // BASICFLOWOUTPUT_IP_HEX
	memcpy(&bfo_buffer[bfo_offset], &i, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#else // IPV6_ACTIVATE == 1
	a = (char*)&currFlow->dstIP.IPv6.s6_addr[0];	
//	a = (char*)&currFlow->dstIP.IPv6.__in6_u.u6_addr32[0];	
	memcpy(&bfo_buffer[bfo_offset], a, 16);
	bfo_offset += 16;
#endif // IPV6_ACTIVATE

#endif // AGGREGATIONFLAG & SUBNET

#if BASICFLOWOUTPUT_SUBNET_TEST == 1
	subnetNr = subnet_test(bfo_subnet_table, currFlow->dstIP.s_addr); // subnet test source ip
	memcpy(&bfo_buffer[bfo_offset], &subnetNr, sizeof(uint32_t));
	bfo_offset += sizeof(uint32_t);
#endif // BASICFLOWOUTPUT_SUBNET_TEST

	memcpy(&bfo_buffer[bfo_offset], &currFlow->dstPort, sizeof(uint16_t));
	bfo_offset += sizeof(uint16_t);

	memcpy(&bfo_buffer[bfo_offset], &currFlow->layer4Protocol, sizeof(uint8_t));
	bfo_offset += sizeof(uint8_t);

	// append to main output buffer
	outputBuffer_append(main_output_buffer, bfo_buffer, bfo_offset);
}


void onApplicationTerminate() {

#if (BASICFLOWOUTPUT_SUBNET_TEST == 1 || BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1)
	if (bfo_subnet_table) free(bfo_subnet_table);
	bfo_subnet_table = NULL;
#endif

	if (encapsulateProtPerFlow) free(encapsulateProtPerFlow);
	encapsulateProtPerFlow = NULL;

}
