/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __BASICFLOWOUTPUT2_H__
#define __BASICFLOWOUTPUT2_H__

// User defined switches
#define BASICFLOWOUTPUT_FLOWINDEX 1 // 1: flow index on; 0: flow index off

#define BASICFLOWOUTPUT_DATE_TIME 0 // 1: Human readable Date time; 0: Unix time format for timestamps

#define BASICFLOWOUTPUT_FLOW_STATUS 1 // 1: Enables / 0: Disables FLOW STATUS output

#define BASICFLOWOUTPUT_ETHERTYPE 1 // 1: Enables / 0: Disables Ether Type header output
#define BASICFLOWOUTPUT_ETHERTYPE_HEX 0 // 1: Enables / 0: Disables Hex decoded decimal output

#define BASICFLOWOUTPUT_MPLS 0 // 1: Enables MPLS header output
#define BASICFLOWOUTPUT_MPLS_DETAIL 0 // 1: MPLS detail header output

#define BASICFLOWOUTPUT_L2TP 0 // 1: Enables / 0: Disables L2TP header output information
#define BASICFLOWOUTPUT_GRE 0 // 1: Enables / 0: Disables GRE header output information
#define BASICFLOWOUTPUT_PPP 0 // 1: Enables / 0: Disables PPP header output

#define BASICFLOWOUTPUT_IP_HEX 0 // 1: Enables / 0: Disables IP Hex output

#define BASICFLOWOUTPUT_SUBNET_TEST !IPV6_ACTIVATE * 1 // 1: Enables / 0: Disables Subnet Test
#define BASICFLOWOUTPUT_SUBNET_TEST_L2TP !IPV6_ACTIVATE * 0 // 1: Enables / 0: Disables Subnet Test on L2TP adresses
#define BASICFLOWOUTPUT_SUBNET_TEST_GRE !IPV6_ACTIVATE * 0 // 1: Enables / 0: Disables Subnet Test on GRE adresses

// buffer definitions
#define MAX_MPLS_DEPTH 3 // MPLS depth to be copied into flow structure
#define BF_MAX_MPLS_DEPTH (MAX_MPLS_DEPTH + 1) // MPLS depth to be copied into flow structure
#define	BFO_BUFFER_SZ	33351	// BFO print buffer min size (49), default 51
#define MAX_ETHTYPE_DEPTH 3

// calculated buffer defs
#define BF_MAX_ETHTYPE_DEPTH (MAX_ETHTYPE_DEPTH + 1) // ETHERType depth to be copied into flow structure
#define BFO_BUFFER_SZ_ACT ( BFO_BUFFER_SZ + BASICFLOWOUTPUT_FLOWINDEX*sizeof(uint32_t) + BASICFLOWOUTPUT_FLOW_STATUS*sizeof(uint32_t) + BASICFLOWOUTPUT_SUBNET_TEST*sizeof(uint32_t)*2 + BASICFLOWOUTPUT_SUBNET_TEST_GRE*sizeof(uint32_t)*2  + sizeof(uint32_t)*BF_MAX_ETHTYPE_DEPTH*BASICFLOWOUTPUT_ETHERTYPE + sizeof(mplsh_t)*BF_MAX_MPLS_DEPTH*BASICFLOWOUTPUT_MPLS + (sizeof(struct in_addr)*2 + sizeof(uint16_t)*3)*BASICFLOWOUTPUT_L2TP + (sizeof(struct in_addr)*2 + sizeof(uint32_t))*BASICFLOWOUTPUT_GRE  + sizeof(pppHu_t)*BASICFLOWOUTPUT_PPP  + 16*IPV6_ACTIVATE )

// includes
#include <stdio.h>
#include <string.h>
#include <time.h>

// local includes
#include "tranalyzer.h"
#include "main.h"
#include "networkHeaders.h"
#include "getter.h"

#if (BASICFLOWOUTPUT_SUBNET_TEST == 1 || BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1 || BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1)
#include "subnet.h"
#define SUBNETFILE "subnet.txt"
#endif // BASICFLOWOUTPUT_SUBNET_TEST 

#if BASICFLOWOUTPUT_MALSITE_TEST != 0
// #include "malsite.h"
#define MALSITEFILE "malsite.txt"
#endif // BASICFLOWOUTPUT_MALSITE_TEST  

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();

void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

// flow vars

typedef struct encapsulateProtPerFlow_s {
#if BASICFLOWOUTPUT_ETHERTYPE == 1
	uint32_t ETHType[BF_MAX_ETHTYPE_DEPTH];
#endif
#if BASICFLOWOUTPUT_MPLS == 1
	mplsh_t mplsh[BF_MAX_MPLS_DEPTH];
#endif // BASICFLOWOUTPUT_MPLS
#if BASICFLOWOUTPUT_PPP == 1 
	pppHu_t pppHdr;	
#endif // BASICFLOWOUTPUT_PPP
#if BASICFLOWOUTPUT_GRE == 1 
	struct in_addr gre_srcIP, gre_dstIP;
	uint32_t greHdrBF;
#endif // BASICFLOWOUTPUT_GRE == 1
#if BASICFLOWOUTPUT_L2TP == 1 
	struct in_addr l2tp_srcIP, l2tp_dstIP;
	uint16_t l2tpHdrBF;
	uint16_t l2tpHdrTID;
	uint16_t l2tpHdrSID;
#endif // BASICFLOWOUTPUT_L2TP == 1
} encapsulateProtPerFlow_t; 

// variables
flow_t *currFlow;
unsigned long pluginIndexBasicFlowOutput;
encapsulateProtPerFlow_t *encapsulateProtPerFlow;
#ifdef __APPLE__
// global variables
extern hashMap_t *mainHashMap;
extern flow_t *flows;
extern outputBuffer_t* main_output_buffer;
#endif // __APPLE__
#if BASICFLOWOUTPUT_ALTERNATIVE_TIME_PRINT != 0
char *bfo_time_string1, *bfo_time_string2;
struct tm *bfo_time1, *bfo_time2;
#endif


#if (BASICFLOWOUTPUT_SUBNET_TEST == 1 || BASICFLOWOUTPUT_SUBNET_TEST_L2TP == 1 || BASICFLOWOUTPUT_SUBNET_TEST_GRE == 1)
subnettable_t *bfo_subnet_table;
#endif // BASICFLOWOUTPUT_SUBNET_TEST 

#endif // __BASICFLOWOUTPUT_H2__ 

