/* This file belongs to tranalyzer2 and was taken from tranalyzer
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include "subnet.h"


void subnet_init(subnettable_t* table, char* bw_file) {
	struct in_addr ip;

	table->count = subnet_load(bw_file, NULL);
	if (table->count == 0) return;

	table->subnets = (subnet_t *)malloc(sizeof(subnet_t) * table->count);
	table->count = subnet_load(bw_file, table->subnets);
/*

	int i;
	for (i = 0; i< table->count; i++) {
		ip.s_addr = table->subnets[i].net;
		printf("   %d: %s",i,inet_ntoa(ip));
		ip.s_addr = table->subnets[i].mask;
		printf(" / %s",inet_ntoa(ip));
	//	printf("\t%d\n",table->subnets[i].memberNet);
	}
*/
}

/* returns the number of subnets and file table, if given */
int subnet_load(char* bw_file, subnet_t* subnets) {
	FILE* subnets_file;
	char  line[1024];
	int count = 0;
	int   mask;
	char* p, *q;
	struct in_addr ip;
	
	subnets_file =  fopen(bw_file,"r");

	if (! subnets_file) {
		printf("basicFlow_subnet: Cannot open subnet file %s\n", bw_file);
		return 0;
	}

	while ( fgets(line,1024,subnets_file) ) {

		/* Skip lines starting with a # */
		if (line[0] == '#') continue;

		/* Skip Whitespace */
		if (line[0] == '\n') continue;
		if (line[0] == ' ') continue;

		/* parse "ip / mask" */
		p = strtok(line,"/");
		q = strtok(NULL,"/	");
		if (p == NULL || q == NULL ) {
			if (subnets) printf("basicFlow_subnet: Ignored invalid CIDR notation @ %d: %s\n",count, line);
			continue;
		}

		/* convert network from ascii to binary networkorder */
		if (inet_aton(p,&ip) < 0){
			if (subnets) printf("basicFlow_subnet: Ignored invalid IP @ %d: %s\n",count, line);
			continue;
		}

		if (subnets) subnets[count].net=ip.s_addr;

		/* convert cidr subnet size from ascii to int */
		mask = atoi(q);
		if (mask==0) {
			if (subnets) printf("basicFlow_subnet: Ignored invalid mask @ %d: %s\n",count, line);
			continue;
		}
		if (subnets) subnets[count].mask = htonl(0xFFFFFFFF<< (32-mask));

		/* acquire network membership */

#if NETWORKLABELING == 1
		char *r;	
		if ( subnets){ 
			r = strtok(NULL,"\n");
			if (r != NULL) {
#if NETWORK_HEXLABEL == 1 			
				uint32_t val;
				sscanf(r, "%x", &val);
				subnets[count].memberNet = val; 
#else
				subnets[count].memberNet = atoi(r);
#endif
			} else subnets[count].memberNet = 0;
		}
#endif
		/* count number of subnets in file */
		count++;
	}
	
	fclose(subnets_file);
	return count;
}


/* this function tests whether a given IP is part of a subnet defined in the config file */
uint32_t subnet_test(subnettable_t* table, in_addr_t ip) {

	for (subnet_i = 0; subnet_i < table->count; subnet_i++) {
		if ((ip & table->subnets[subnet_i].mask ) == table->subnets[subnet_i].net) {
#if NETWORKLABELING == 1
			return table->subnets[subnet_i].memberNet;
#else
			return subnet_i+1;
#endif
		}
	}

	return 0;
}
