/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * connectionCounter.c
 */

#include "tranalyzer.h"

#include "connectionCounter.h"
#include "packetCapture.h"
#include "hashTable.h"
#include "main.h"

char* get_plugin_name() {
	return "connectionCounter";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	pluginConnectionCounterIndex = pPluginIndex;
	
	ccnt_oflow_prt = 0; // Flow print state variable

	// initialize the hashMaps
	ipPairHashMap = hashTable_initialize(1.0f, sizeof(flowID_t));
	ipHashMap = hashTable_initialize(1.0f, sizeof(flowID_t));
	
	// initialize the arrays
	ipPairConnections = (uint32_t*) malloc(sizeof(uint32_t) * ipPairHashMap->hashChainTableSize);
	ipConnections = (uint32_t*) malloc(sizeof(uint32_t) * ipHashMap->hashChainTableSize); 
	memset(ipPairConnections, '\0', ipPairHashMap->hashChainTableSize * sizeof(uint32_t));
	memset(ipConnections, '\0', ipHashMap->hashChainTableSize * sizeof(uint32_t));
}


binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;

	binary_value_t *act_bv = bv_new_bv("Number of connections from source IP to different hosts", "connSrc", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Number of connections from destination IP to different hosts", "connDst", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Number of connections between source IP and destination IP", "connSrcDst", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	return main_bv;
}


void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
//	flowID_t flowHelper[2];
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *ccnt_currFlow = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
	uint64_t flowHelper[4];
	uint64_t ip[4];
#else // IPV6_ACTIVATE == 0
	uint32_t flowHelper[2];
	uint32_t ip[2];
#endif // IPV6_ACTIVATE
	unsigned long ccntIPIndex1, ccntIPIndex2, ccntIPPairIndex;

	// if opposite flow exists then already counted
	if (ccnt_currFlow->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) return;

	// Check which of the IPs is the lower and which is the higher

#if IPV6_ACTIVATE == 0
	if (ccnt_currFlow->srcIP.s_addr > ccnt_currFlow->dstIP.s_addr) {
		ip[0] = ccnt_currFlow->dstIP.s_addr;
		ip[1] = ccnt_currFlow->srcIP.s_addr;
	} else {
		ip[0] = ccnt_currFlow->srcIP.s_addr;
		ip[1] = ccnt_currFlow->dstIP.s_addr;
#else // IPV6_ACTIVATE == 1
	if (ccnt_currFlow->srcIP.IPv6L[0] >= ccnt_currFlow->dstIP.IPv6L[0] ) {
		if (ccnt_currFlow->srcIP.IPv6L[0] == ccnt_currFlow->dstIP.IPv6L[0] && ccnt_currFlow->srcIP.IPv6L[1] <= ccnt_currFlow->dstIP.IPv6L[1]) goto low; 
		ip[0] = ccnt_currFlow->dstIP.IPv6L[0];
		ip[1] = ccnt_currFlow->dstIP.IPv6L[1];
		ip[2] = ccnt_currFlow->srcIP.IPv6L[0];
		ip[3] = ccnt_currFlow->srcIP.IPv6L[1];
	} else {
low:		ip[0] = ccnt_currFlow->srcIP.IPv6L[0];
		ip[1] = ccnt_currFlow->srcIP.IPv6L[1];
		ip[2] = ccnt_currFlow->dstIP.IPv6L[0];
		ip[3] = ccnt_currFlow->dstIP.IPv6L[1];
#endif // IPV6_ACTIVATE
	}

	// Look in the ipPair hashMap if there is already an entry

#if IPV6_ACTIVATE == 0
	flowHelper[0] = ip[0];
	flowHelper[1] = ip[1];
#else // IPV6_ACTIVATE == 1
	flowHelper[0] = ip[0];
	flowHelper[1] = ip[1];
	flowHelper[2] = ip[2];
	flowHelper[3] = ip[3];
#endif // IPV6_ACTIVATE
	if ((ccntIPPairIndex = hashTable_lookup(ipPairHashMap, (char*)&flowHelper)) == HASHTABLE_ENTRY_NOT_FOUND) {
		// iFirst connection between these two hosts -> new hashMap entry and store info in array
		ccntIPPairIndex = hashTable_insert(ipPairHashMap, (char*)&flowHelper);
		ipPairConnections[ccntIPPairIndex] = 1;

		// Number of distinct connections between both hosts has to be incremented -> check if an higher IP entry exists
#if IPV6_ACTIVATE == 0
		flowHelper[0] = ip[1];
		flowHelper[1] = 0;
#else // IPV6_ACTIVATE == 1
		flowHelper[0] = ip[2];
		flowHelper[1] = ip[3];
		flowHelper[2] = 0;
		flowHelper[3] = 0;
#endif // IPV6_ACTIVATE
		if ((ccntIPIndex1 = hashTable_lookup(ipHashMap, (char*)&flowHelper)) == HASHTABLE_ENTRY_NOT_FOUND) {
			// There is no entry, so generate one
			ccntIPIndex1 = hashTable_insert(ipHashMap, (char*)&flowHelper);
			ipConnections[ccntIPIndex1] = 1;
		}
		else {
			// There is already an index, so increment it;
			ipConnections[ccntIPIndex1]++;
		}

		// check if an entry for lower IP exists
#if IPV6_ACTIVATE == 0
		flowHelper[0] = ip[0];
		flowHelper[1] = 0;
#else // IPV6_ACTIVATE == 1
		flowHelper[0] = ip[0];
		flowHelper[1] = ip[1];
		flowHelper[2] = 0;
		flowHelper[3] = 0;
#endif // IPV6_ACTIVATE
		if ((ccntIPIndex1 = hashTable_lookup(ipHashMap, (char*)&flowHelper)) == HASHTABLE_ENTRY_NOT_FOUND) {
			// There is no entry, so generate one
			ccntIPIndex1 = hashTable_insert(ipHashMap, (char*)&flowHelper);
			ipConnections[ccntIPIndex1] = 1;
		}
		else {
			// Index exists, so increment it;
			ipConnections[ccntIPIndex1]++;
		}
	}
	else {
		// Index was found -> increment counter
		ipPairConnections[ccntIPPairIndex]++;
	}
}


void onFlowTerminate(unsigned long flowIndex) {
//	flowID_t flowHelper;
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	flow_t *ccnt_currFlow = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
	uint64_t flowHelper[4];
	uint64_t ip[4];
#else // IPV6_ACTIVATE == 0
	uint32_t flowHelper[2];
	uint32_t ip[2];
#endif // IPV6_ACTIVATE
	unsigned long *srcIP_i, *dstIP_i;
	unsigned long ccntIPIndex1, ccntIPIndex2, ccntIPPairIndex;

	// Check which of the IPs is the lower and which is the higher
#if IPV6_ACTIVATE == 0
	if (ccnt_currFlow->srcIP.s_addr > ccnt_currFlow->dstIP.s_addr) {
		ip[0] = ccnt_currFlow->dstIP.s_addr;
		ip[1] = ccnt_currFlow->srcIP.s_addr;
		srcIP_i = &ccntIPIndex1;
		dstIP_i = &ccntIPIndex2;
	} else {
		ip[0] = ccnt_currFlow->srcIP.s_addr;
		ip[1] = ccnt_currFlow->dstIP.s_addr;
		srcIP_i = &ccntIPIndex2;
		dstIP_i = &ccntIPIndex1;
#else // IPV6_ACTIVATE == 1
	if (ccnt_currFlow->srcIP.IPv6L[0] >= ccnt_currFlow->dstIP.IPv6L[0] ) {
		if (ccnt_currFlow->srcIP.IPv6L[0] == ccnt_currFlow->dstIP.IPv6L[0] && ccnt_currFlow->srcIP.IPv6L[1] <= ccnt_currFlow->dstIP.IPv6L[1]) goto low;
		ip[0] = ccnt_currFlow->dstIP.IPv6L[0];
		ip[1] = ccnt_currFlow->dstIP.IPv6L[1];
		ip[2] = ccnt_currFlow->srcIP.IPv6L[0];
		ip[3] = ccnt_currFlow->srcIP.IPv6L[1];
		srcIP_i = &ccntIPIndex1;
		dstIP_i = &ccntIPIndex2;
	} else {
low:		ip[0] = ccnt_currFlow->srcIP.IPv6L[0];
		ip[1] = ccnt_currFlow->srcIP.IPv6L[1];
		ip[2] = ccnt_currFlow->dstIP.IPv6L[0];
		ip[3] = ccnt_currFlow->dstIP.IPv6L[1];
		srcIP_i = &ccntIPIndex2;
		dstIP_i = &ccntIPIndex1;
#endif // IPV6_ACTIVATE
	}

	// Defensive programming, bug detection: If ipPair does not exist
#if IPV6_ACTIVATE == 0
	flowHelper[0] = ip[0];
	flowHelper[1] = ip[1];
#else // IPV6_ACTIVATE == 1
	flowHelper[0] = ip[0];
	flowHelper[1] = ip[1];
	flowHelper[2] = ip[2];
	flowHelper[3] = ip[3];
#endif // IPV6_ACTIVATE
	if ( (ccntIPPairIndex = hashTable_lookup(ipPairHashMap, (char*)&flowHelper) ) == HASHTABLE_ENTRY_NOT_FOUND) {
		printf("WARNING connection plugin: flowIndex: %lu, findex: %"PRIu32" has no ip pair number connections entry!\n", flowIndex, ccnt_currFlow->findex);
		return;
	}

	// Defensive programming, bug detection: Same as above
#if IPV6_ACTIVATE == 0
	flowHelper[0] = ip[1];
	flowHelper[1] = 0;
#else // IPV6_ACTIVATE == 1
	flowHelper[0] = ip[2];
	flowHelper[1] = ip[3];
	flowHelper[2] = 0;
	flowHelper[3] = 0;
#endif // IPV6_ACTIVATE
	if ( (ccntIPIndex1 = hashTable_lookup(ipHashMap, (char*)&flowHelper) ) == HASHTABLE_ENTRY_NOT_FOUND) {
		printf("WARNING connection plugin: flowIndex %lu, findex: %"PRIu32" has no ip_high number connections entry!\n", flowIndex, ccnt_currFlow->findex);
		return;
	}

#if IPV6_ACTIVATE == 0
	flowHelper[0] = ip[0];
	flowHelper[1] = 0;
#else // IPV6_ACTIVATE == 1
	flowHelper[0] = ip[0];
	flowHelper[1] = ip[1];
	flowHelper[2] = 0;
	flowHelper[3] = 0;
#endif // IPV6_ACTIVATE
	if ( (ccntIPIndex2 = hashTable_lookup(ipHashMap, (char*)&flowHelper)) == HASHTABLE_ENTRY_NOT_FOUND ) {
		printf("WARNING connection plugin: flowIndex %lu, findex: %"PRIu32" has no ip_low number connections entry!\n", flowIndex, ccnt_currFlow->findex);
		return;
	}

	outputBuffer_append(main_output_buffer, (char*) &ipConnections[*srcIP_i], sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &ipConnections[*dstIP_i], sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &ipPairConnections[ccntIPPairIndex], sizeof(uint32_t));

	// update counts only if opposite flow is printed
	if ( ccnt_currFlow->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) {
		if ( ccnt_oflow_prt ) ccnt_oflow_prt = 0;
		else {
			ccnt_oflow_prt = 1;
			return;
		}  
	}

	// There is no opposite flow, so decrement the ip pair connection counter
	ipPairConnections[ccntIPPairIndex]--;

	// Check if the counter is zero
	if (ipPairConnections[ccntIPPairIndex] == 0) {
		// Last connection between these two hosts, so remove entry in hashTable and decrement ip connection counters for
		// both hosts and delete them if they are zero.
#if IPV6_ACTIVATE == 0
		flowHelper[0] = ip[0];
		flowHelper[1] = ip[1];
#else // IPV6_ACTIVATE == 1
		flowHelper[0] = ip[0];
		flowHelper[1] = ip[1];
		flowHelper[2] = ip[2];
		flowHelper[3] = ip[3];
#endif // IPV6_ACTIVATE
		hashTable_remove(ipPairHashMap, (char*)&flowHelper);

		// Decrement the counter
		ipConnections[ccntIPIndex1]--;

		// Check if it is zero. If it is, remove the hashTable entry
#if IPV6_ACTIVATE == 0
		flowHelper[0] = ip[1];
		flowHelper[1] = 0;
#else // IPV6_ACTIVATE == 1
		flowHelper[0] = ip[2];
		flowHelper[1] = ip[3];
		flowHelper[2] = 0;
		flowHelper[3] = 0;
#endif // IPV6_ACTIVATE
		if (ipConnections[ccntIPIndex1] == 0) hashTable_remove(ipHashMap, (char*)&flowHelper);

		// Decrement the counter
		ipConnections[ccntIPIndex2]--;

		// Check if it is zero. If it is, remove the hashTable entry
#if IPV6_ACTIVATE == 0
		flowHelper[0] = ip[0];
		flowHelper[1] = 0;
#else // IPV6_ACTIVATE == 1
		flowHelper[0] = ip[0];
		flowHelper[1] = ip[1];
		flowHelper[2] = 0;
		flowHelper[3] = 0;
#endif // IPV6_ACTIVATE
		if (ipConnections[ccntIPIndex2] == 0) hashTable_remove(ipHashMap, (char*)&flowHelper);
	}
}


void onApplicationTerminate() {
	hashTable_destroy(ipPairHashMap); 
	ipPairHashMap = NULL;
	hashTable_destroy(ipHashMap); 
	ipHashMap = NULL;
	if (ipPairConnections) { 
		free(ipPairConnections); 
		ipPairConnections = NULL;
	}
	if (ipConnections) {
		free(ipConnections); 
		ipConnections = NULL;
	}
}
