/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * connectionCounter.h
 * counts the number of connections between two hosts and the number of distinct
 * connections of one host. Here, distinct means that only the number of
 * different hosts the actual host is connected to, are counted.
 * Please note that because of the nature of this program, not all connections
 * of a host might be observed. For example if the program is sniffing the
 * traffic between a gateway and a local intranet, it is not able to observe the
 * connections between two hosts inside the intranet. Therefore, these values
 * are to handle with care.
 */

#ifndef CONNECTIONCOUNTER_H_
#define CONNECTIONCOUNTER_H_

// includes
#include <stdio.h>
#include <string.h>

// local includes
#include "tranalyzer.h"
#include "hashTable.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "networkHeaders.h"
#include "getter.h"

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

// plugin definitions 

typedef struct flowID_s {
#if IPV6_ACTIVATE == 1
	uint64_t srcIP[2], dstIP[2];
#else // IPV6_ACTIVATE == 0
	uint32_t srcIP, dstIP;
#endif
} flowID_t;

// plugin globals 
// hashMaps for the number of connections
hashMap_t *ipPairHashMap, *ipHashMap;
uint32_t *ipPairConnections, *ipConnections;

// plugin worker globals
uint8_t ccnt_oflow_prt;
 
unsigned long pluginConnectionCounterIndex;
//global variables
#ifdef __APPLE__
extern outputBuffer_t *main_output_buffer;
extern flow_t *flows;
#endif // __APPLE__
#endif /* CONNECTIONCOUNTER_H_ */
