/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * descriptiveStatistics.c
 * These functions use the packetSizeInterArrivalTimeHisto structures to
 * calculate the statistics. Therefore the activation of the psiatHisto plugin
 * is necessary.
 */

#include "descriptiveStatistics.h"
#include "packetSizeInterArrivalTimeHisto.h"
#include "tranalyzer.h"
#include <math.h>

long dStats_getMinPacketLength(redBlackNode_t *tree);
long dStats_getMaxPacketLength(redBlackNode_t *tree);
long long dStats_getSumPacketLengths(redBlackNode_t *tree);
int dStats_getStatsFromTree(packetSizeInterArrivalTimeTrees_t *tree, uint32_t *numOfPackets, float *lower, float *median, float *upper, float *mode, float *variance, float* skewness, float *excess);
void dStats_getStatsFromTreeRek(redBlackNode_t *tree, unsigned long *actNumPackets, float *lower, float *median, float *upper, float *mode, float *variance, float* skewness, float *excess);

char* get_plugin_name() {
	return "descriptiveStatistics";
}

char* get_plugin_version() {
	return "0.5.1";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

char* get_dependencies() {
	return "packetSizeInterArrivalTimeHisto";
}

binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;
	binary_value_t *act_bv;

#if ENABLE_PACKETSIZE_CALCULATION != 0
	act_bv = bv_new_bv("Minimum packet length", "MinPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Maximum packet length", "MaxPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Mean packet length", "MeanPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Lower quartile of packet lengths", "LowQuartilePl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Median of packet lengths", "MedianPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Upper quartile of packet lengths", "UppQuartilePl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Inter quartile distance of packet lengths", "IqdPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Mode of packet lengths", "ModePl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Range of packet lengths", "RangePl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Standard deviation of packet lengths", "StdPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Robust standard deviation of packet lengths", "StdrobPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Skewness of packet lengths", "SkewPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Excess of packet lengths", "ExcPl", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif /* ENABLE_PACKETSIZE_CALCULATION != 0 */

#if ENABLE_IAT_CALCULATION != 0
	act_bv = bv_new_bv("Minimum inter arrival time", "MinIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Maximum inter arrival time", "MaxIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Mean inter arrival time", "MeanIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Lower quartile of inter arrival times", "LowQuartileIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Median inter arrival times", "MedianIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Upper quartile of inter arrival times", "UppQuartileIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Inter quartile distance of inter arrival times", "IqdIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Mode of inter arrival times", "ModeIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Range of inter arrival times", "RangeIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Standard deviation of inter arrival times", "StdIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Robust standard deviation of inter arrival times", "RobStdIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Skewness of inter arrival times", "SkewIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("Excess of inter arrival times", "ExcIat", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif /* ENABLE_IAT_CALCULATION != 0 */
	return main_bv;
}

void initialize(unsigned long pPluginIndex) {
	pluginDescriptiveStatisticsIndex = pPluginIndex;
	dStats_actNumPackets = malloc(sizeof(unsigned long));
}

void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	main_output_buffer = getMainOutputBuffer();
	packetSizeInterArrivalTime_trees = getPacketSizeInterArrivalTimeTrees();
#endif // __APPLE__
	dStats_currTree = &packetSizeInterArrivalTime_trees[flowIndex];
#if ENABLE_PACKETSIZE_CALCULATION != 0
	/* calculate and print packetlength statistics */
	memset(&dStats_packetLengths, '\0', sizeof(dStats_t));
	dStats_packetLengths.minimum = (float) dStats_getMinPacketLength(dStats_currTree->packetTree);
	dStats_packetLengths.maximum = (float) dStats_getMaxPacketLength(dStats_currTree->packetTree);
	dStats_packetLengths.range = (float) dStats_packetLengths.maximum - dStats_packetLengths.minimum;
	*dStats_actNumPackets = 0;
	dStats_packetLengths.mean = (float) dStats_getSumPacketLengths(dStats_currTree->packetTree) / (float) *dStats_actNumPackets;
	dStats_packetLengths.numOfPackets = *dStats_actNumPackets;
	dStats_getStatsFromTree(dStats_currTree,
			&(packetSizeInterArrivalTime_trees[flowIndex].numOfPackets),
			&(dStats_packetLengths.lowerQuartile),
			&(dStats_packetLengths.median),
			&(dStats_packetLengths.upperQuartile),
			&(dStats_packetLengths.mode),
			&(dStats_packetLengths.stddev),
			&(dStats_packetLengths.skewness),
			&(dStats_packetLengths.excess));
	dStats_packetLengths.stddev = sqrt(dStats_packetLengths.stddev);

	dStats_packetLengths.iqd = dStats_packetLengths.upperQuartile - dStats_packetLengths.lowerQuartile;

	if (dStats_packetLengths.stddev < dStats_packetLengths.iqd * 0.7413f) dStats_packetLengths.stdrob = dStats_packetLengths.stddev;
	else dStats_packetLengths.stdrob = dStats_packetLengths.iqd * 0.7413f;

	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.minimum, 13 * sizeof(float));

	/*
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.minimum, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.maximum, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.mean, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.lowerQuartile, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.median, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.upperQuartile, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.iqd, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.mode, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.range, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.stddev, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.stdrob, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.skewness, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_packetLengths.excess, sizeof(float));
	*/

#endif /* ENABLE_PACKETSIZE_CALCULATION != 0 */

#if ENABLE_IAT_CALCULATION != 0
	/*
	 * calculate and print IAT statistics
	 *
	 * A Note to the calculation: Because the inter arrival times are binned, ALL values
	 * a just approximations, more or less bad. Please be aware of it!
	 * */

	// We ignore the IAT of the first packet, because it is always zero
	// -> decrement value in bucket zero by one. If numOfPackets is one, print dummy output
	if (dStats_currTree->numOfPackets > 1) {
		dStats_currTree->numOfpacketsInTimeBin[0]--;
		dStats_currTree->numOfPackets--;
	}
	else {

		memset(&main_output_buffer->buffer[main_output_buffer->act_pos], '\0', 13 * sizeof(float));
		main_output_buffer->act_pos += 13 * sizeof(float);
		return;
	}

	// Now we have to solve the special case of one packet
	if (dStats_currTree->numOfPackets == 1) {
		// search for the bin of this packet
		dStats_actBin = 0;
		while(dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] == 0) dStats_actBin++;
		// now print stats
		float d_t = (float) bin2iat(dStats_actBin);
		int d_t_inc;

		for (d_t_inc = 0; d_t_inc < 6; d_t_inc++) {
			outputBuffer_append(main_output_buffer, (char*) &d_t, sizeof(float));
		}

		outputBuffer_append(main_output_buffer, "\0\0\0\0", sizeof(float));
		outputBuffer_append(main_output_buffer, (char*) &d_t, sizeof(float));
		memset(&main_output_buffer->buffer[main_output_buffer->act_pos], '\0', 5 * sizeof(float));
		main_output_buffer->act_pos += 5 * sizeof(float);
		return;
	}




	memset(&dStats_IATs, '\0', sizeof(dStats_t));

	// get minimum iat
	// TODO: IAT Bin 0 handling improvement: if bin[0] > 1 ? ...
	for (dStats_i = 0; dStats_i < IATBINUMAX; dStats_i++) {
		if (dStats_currTree->numOfpacketsInTimeBin[dStats_i] > 0) {
			dStats_IATs.minimum = (float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1)) / 2.0f;
			break;
		}
	}

	// get maximum iat
	for (dStats_i = IATBINUMAX - 1; dStats_i >= 0; dStats_i--) {
		if (dStats_currTree->numOfpacketsInTimeBin[dStats_i] > 0) {
			dStats_IATs.maximum = (float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1)) / 2.0f;
			break;
		}
	}

	// get range of iat's
	dStats_IATs.range = dStats_IATs.maximum - dStats_IATs.minimum;

	// get mean iat
	// because of this crazy binning we have to transform it
	// into the mean ms value of the bin and
	// later back into the bin of the mean ms, understood ?!? If not, don't care...
	// Next problem: Last bin is from 1000ms to infinity
	// -> take 1000ms as approx value -> stat gets falsified, be aware of it!!!
	for (dStats_i = 0; dStats_i < IATBINUMAX - 1; dStats_i++) {
		dStats_IATs.mean += (float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1)) / 2.0 * (float) dStats_currTree->numOfpacketsInTimeBin[dStats_i] / (float) dStats_currTree->numOfPackets;
	}
	dStats_IATs.mean += (float) bin2iat(IATBINUMAX - 1) * (float) dStats_currTree->numOfpacketsInTimeBin[IATBINUMAX - 1] / (float) dStats_currTree->numOfPackets;

	/*
	 * get iat quartiles
	 * we want to avoid going through the array several times, so this code is not very
	 * readable - sorry
	 */

	// Note: The calculation of the iat quantiles differs a bit from the calculation
	// of the packetsize quantiles because of the binning.
	quartilePosLower = (float) (dStats_currTree->numOfPackets - 1) * 0.25f + 1.0f;
	quartilePosMedian = (float) (dStats_currTree->numOfPackets - 1) * 0.5f + 1.0f;
	quartilePosUpper = (float) (dStats_currTree->numOfPackets - 1) * 0.75f + 1.0f;

	dStats_i = 0;
	dStats_actBin = 0;
	dStats_actNumPacketsIAT = 0;

	// get lower quartile
	carryOverLowerIAT = 0;
	// check if lower quartile pos is an integer
	if (quartilePosLower == roundf(quartilePosLower)) {
		// it is an integer
		// get bin of element at quartilePosLower
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < quartilePosLower) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				carryOverLowerIAT = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				break;
			}
		}
		// get bin of element at quartilePosLower + 1
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < quartilePosLower + 1) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
				if (dStats_actBin > IATBINUMAX-1) {
					fprintf(stdout, "BUG in descriptiveStatistics, calc of lower quartile:\n#packets: %u\n",dStats_currTree->numOfPackets - 1);
					fprintf(stdout, "l: %f, m: %f, u %f\n", quartilePosLower, quartilePosMedian, quartilePosUpper);
					int ii;
					for (ii = 0; ii < IATBINUMAX; ii++) {
						fprintf(stdout, "[%"PRId32"]", dStats_currTree->numOfpacketsInTimeBin[ii]);
					}
					exit(1);
				}
			} else {
				// we found the bin
				dStats_IATs.lowerQuartile = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				dStats_IATs.lowerQuartile = (carryOverLowerIAT + dStats_IATs.lowerQuartile) / 2.0f;
				break;
			}
		}

	} else {
	// it is not an integer
	// get bin of element at ceil(quartilePosLower)
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < ceilf(quartilePosLower)) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				dStats_IATs.lowerQuartile = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				break;
			}
		}
	}

	// get median
	carryOverMedianIAT = 0;
	// check if median quartile pos is an integer
	if (quartilePosMedian == roundf(quartilePosMedian)) {
		// it is an integer
		// get bin of element at quartilePosMedian
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < quartilePosMedian) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				carryOverMedianIAT = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				break;
			}
		}
		// get bin of element at quartilePosMedian + 1
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < quartilePosMedian + 1) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				dStats_IATs.median = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				dStats_IATs.median = (carryOverMedianIAT + dStats_IATs.median) / 2.0f;
				break;
			}
		}

	} else {
	// it is not an integer
	// get bin of element at ceil(quartilePosMedian)
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < ceilf(quartilePosMedian)) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				dStats_IATs.median = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				break;
			}
		}
	}

	// get upper quartile
	carryOverUpperIAT = 0;
	// check if upper quartile pos is an integer
	if (quartilePosUpper == roundf(quartilePosUpper)) {
		// it is an integer
		// get bin of element at quartilePosUpper
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < quartilePosUpper) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				carryOverUpperIAT = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				break;
			}
		}
		// get bin of element at quartilePosUpper + 1
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < quartilePosUpper + 1) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				dStats_IATs.upperQuartile = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				dStats_IATs.upperQuartile = (carryOverUpperIAT + dStats_IATs.upperQuartile) / 2.0f;
				break;
			}
		}

	} else {
	// it is not an integer
	// get bin of element at ceil(quartilePosUpper)
		while (1) {
			if (dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT < ceilf(quartilePosUpper)) {
				dStats_actNumPacketsIAT = dStats_currTree->numOfpacketsInTimeBin[dStats_actBin] + dStats_actNumPacketsIAT;
				dStats_actBin++;
			} else {
				// we found the bin
				dStats_IATs.upperQuartile = ((float) (bin2iat(dStats_actBin) + bin2iat(dStats_actBin + 1))) / 2.0f;
				break;
			}
		}
	}

	// get iqd
	dStats_IATs.iqd = dStats_IATs.upperQuartile - dStats_IATs.lowerQuartile;

	// get mode iat
	// TODO: Change to bin mean -> change var type of mode
	actNumPacketsForMode = dStats_currTree->numOfpacketsInTimeBin[0];
	dStats_IATs.mode = (float) (bin2iat(0) + bin2iat(1)) / 2.0f;
	for (dStats_i = 1; dStats_i < IATBINUMAX; dStats_i++) {
		if (dStats_currTree->numOfpacketsInTimeBin[dStats_i] > actNumPacketsForMode) {
			actNumPacketsForMode = dStats_currTree->numOfpacketsInTimeBin[dStats_i];
			dStats_IATs.mode = ((float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1))) / 2.0f;
		}
	}

	/* get stddev */
	dStats_IATs.stddev = 0.0;
	float dStats_temp = 0.0;

	// sum bins
	for (dStats_i = 0; dStats_i < IATBINUMAX; dStats_i++) {
		dStats_temp += dStats_currTree->numOfpacketsInTimeBin[dStats_i] * powf(((float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1)) / 2.0f) - dStats_IATs.mean, 2);
	}
	// divide it thru the numOfPackets and take the squareroot
	dStats_IATs.stddev=sqrtf((1.0 / (float) (dStats_currTree->numOfPackets - 1)) * dStats_temp);
	// check if NaN
	if (isnanf(dStats_IATs.stddev)) dStats_IATs.stddev = 0.0f;


	/* get stdrob */
	if (dStats_IATs.stddev < dStats_IATs.iqd * 0.7413f) dStats_IATs.stdrob = dStats_IATs.stddev;
		else dStats_IATs.stdrob = dStats_IATs.iqd * 0.7413f;

	/* get skewness */
	dStats_IATs.skewness = 0.0;
	// sum bins
	for (dStats_i = 0; dStats_i < IATBINUMAX; dStats_i++) {
		dStats_IATs.skewness += dStats_currTree->numOfpacketsInTimeBin[dStats_i] * powf(((float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1)) / 2.0f) - dStats_IATs.mean, 3);
	}

	// calc skewness, reuse value in dStats_temp
	dStats_IATs.skewness = ((1.0f / (float) dStats_currTree->numOfPackets) * dStats_IATs.skewness) / sqrtf(powf((1.0f / (float) dStats_currTree->numOfPackets) * dStats_temp, 3));
	// check if NaN
	if (isnanf(dStats_IATs.skewness)) dStats_IATs.skewness = 0.0f;

	/* get excess */
	dStats_IATs.excess = 0.0;
	// sum bins
	for (dStats_i = 0; dStats_i < IATBINUMAX; dStats_i++) {
		dStats_IATs.excess += dStats_currTree->numOfpacketsInTimeBin[dStats_i] * powf(((float) (bin2iat(dStats_i) + bin2iat(dStats_i + 1)) / 2.0f) - dStats_IATs.mean, 4);
	}
	// calc excess, reuse value in dStats_temp
	dStats_IATs.excess = ((1.0f / (float) dStats_currTree->numOfPackets) * dStats_IATs.excess) / powf((1.0f / (float) dStats_currTree->numOfPackets) * dStats_temp, 2) - 3;
	// check if NaN
	if (isnanf(dStats_IATs.excess)) dStats_IATs.excess = 0.0f;

	// print out
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.minimum, 13 * sizeof(float));
	/*
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.minimum, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.maximum, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.mean, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.lowerQuartile, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.median, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.upperQuartile, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.iqd, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.mode, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.range, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.stddev, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.stdrob, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.skewness, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &dStats_IATs.excess, sizeof(float));
	*/
	// increment first bucket to return to original values
	dStats_currTree->numOfpacketsInTimeBin[0]++;
	dStats_currTree->numOfPackets++;
# endif /* ENABLE_IAT_CALCULATION != 0 */
}

void onApplicationTerminate() {
	free(dStats_actNumPackets); dStats_actNumPackets = NULL;
}

long dStats_getMinPacketLength(redBlackNode_t *tree) {
	// search for the most left entry
	while (tree->leftChild) tree = tree->leftChild;
	return tree->value;
}

long dStats_getMaxPacketLength(redBlackNode_t *tree) {
	// search for the most right entry
	while (tree->rightChild) tree = tree->rightChild;
	return tree->value;
}

long long dStats_getSumPacketLengths(redBlackNode_t *tree) {
	long long sum = 0;
#ifdef __APPLE__
	psiat_vals = getPsiatVals();
	packetSizeInterArrivalTime_treeNodePool = getPacketSizeInterArrivalTimeTreeNodePool();
#endif // __APPLE__
	if (tree->leftChild) sum += dStats_getSumPacketLengths(tree->leftChild);
	sum += tree->value * psiat_vals[tree - &packetSizeInterArrivalTime_treeNodePool->nodePool[0]].numOfPackets;
	*dStats_actNumPackets += psiat_vals[tree - &packetSizeInterArrivalTime_treeNodePool->nodePool[0]].numOfPackets;
	if (tree->rightChild) sum += dStats_getSumPacketLengths(tree->rightChild);
	return sum;
}

int dStats_getStatsFromTree(packetSizeInterArrivalTimeTrees_t *tree, uint32_t *numOfPackets, float *lower, float *median, float *upper, float *mode, float *variance, float *skewness, float *excess) {
	// Calculate the absolute positions of the quantiles
	// The sorted packet lengths are taken as
	// the (0.5/N), (1.5/n),..., ((N-0.5)/N) quantiles,
	// N is the number of packets.
	// Therefore the position x of quantile q equals q * N - 0.5 (starting with index nr 0)
	quartilePosLower = (float) *numOfPackets * 0.25f + 0.5f;
	quartilePosMedian = (float) *numOfPackets * 0.5f + 0.5f;
	quartilePosUpper = (float) *numOfPackets * 0.75f + 0.5f;

	// special debug output, was needed for bugfixing
	// fprintf(stdout, "%f\t%f\t%f\n", quartilePosLower, quartilePosMedian, quartilePosUpper);

	// fill in max and min to be safe
	*lower = dStats_packetLengths.minimum;
	*upper = dStats_packetLengths.maximum;
	*median = (*lower + *upper) / 2.0f;
	*mode = (float) tree->packetTree->value;
	*variance = 0.0f;

	carryOverLower = 0;
	carryOverMedian = 0;
	carryOverUpper = 0;
	actNumPacketsForMode = 0;

	// start with the searching
	*dStats_actNumPackets = 0;

	// only process if there are at least 2 packets
	if (*numOfPackets > 1) {
		dStats_getStatsFromTreeRek(tree->packetTree, dStats_actNumPackets, lower, median, upper, mode, variance, skewness, excess);
		if (*variance != 0.0f) {
			*skewness = *skewness / powf(*variance, 3.0f / 2.0f);
			*excess = *excess / (*variance * *variance) - 3;// powf(*variance, 2.0f) - 3;
		}
	}
	return 0;
}

void dStats_getStatsFromTreeRek(redBlackNode_t *tree, unsigned long *actNumPackets, float *lower, float *median, float *upper, float *mode, float *variance, float *skewness, float *excess) {

	// PLEASE NOTE:
	// The quartiles are calculated either using mean or linear interpolation.
	// Codelines for both ways are already given, you just have to uncomment the
	// one you want.
	// current version: linear interpolation

	// check left child
	if (tree->leftChild) dStats_getStatsFromTreeRek(tree->leftChild, actNumPackets, lower, median, upper, mode, variance, skewness, excess);

	dStatsBuffer_numPackets =  psiat_vals[tree - &packetSizeInterArrivalTime_treeNodePool->nodePool[0]].numOfPackets;

	// lower quantile
	// quantile lies in this bucket
	if ((quartilePosLower >= *actNumPackets + 1) && (quartilePosLower <= *actNumPackets + dStatsBuffer_numPackets)) *lower = tree->value;
	// quantile lies in between this and the next bucket -> carry over needed!
	else if ((quartilePosLower > *actNumPackets + dStatsBuffer_numPackets) && (quartilePosLower < *actNumPackets + dStatsBuffer_numPackets + 1)) carryOverLower = tree->value;
	// quantile lies in between the last and this bucket
	else if ((quartilePosLower > *actNumPackets) && (quartilePosLower < *actNumPackets + 1)) {
		// *lower = (float) (carryOverLower + tree->value) / 2.0f; // mean
		*lower = (float) (tree->value - carryOverLower) * (quartilePosLower - floor(quartilePosLower)) + carryOverLower; // linear interpolation
	}

	// median
	// quantile lies in this bucket
	if ((quartilePosMedian >= *actNumPackets + 1) && (quartilePosMedian <= *actNumPackets + dStatsBuffer_numPackets)) *median = tree->value;
	// quantile lies in between this and the next bucket -> carry over needed!
	else if ((quartilePosMedian > *actNumPackets + dStatsBuffer_numPackets) && (quartilePosMedian < *actNumPackets + dStatsBuffer_numPackets + 1)) carryOverMedian = tree->value;
	// quantile lies in between the last and this bucket
	else if ((quartilePosMedian > *actNumPackets) && (quartilePosMedian < *actNumPackets + 1)) {
		// *median = (float) (carryOverMedian + tree->value) / 2.0f; // mean
		*median = (float) (tree->value - carryOverMedian) * (quartilePosMedian - floor(quartilePosMedian)) + carryOverMedian; // linear interpolation
	}

	// upper quantile
	// quantile lies in this bucket
	if ((quartilePosUpper >= *actNumPackets + 1) && (quartilePosUpper <= *actNumPackets + dStatsBuffer_numPackets)) *upper = tree->value;
	// quantile lies in between this and the next bucket -> carry over needed!
	else if ((quartilePosUpper > *actNumPackets + dStatsBuffer_numPackets) && (quartilePosUpper < *actNumPackets + dStatsBuffer_numPackets + 1)) carryOverUpper = tree->value;
	// quantile lies in between the last and this bucket
	else if ((quartilePosUpper > *actNumPackets) && (quartilePosUpper < *actNumPackets + 1)) {
		// *upper = (float) (carryOverUpper + tree->value) / 2.0f; // mean
		*upper = (float) (tree->value - carryOverUpper) * (quartilePosUpper - floor(quartilePosUpper)) + carryOverUpper; // linear interpolation

	}

	// mode
	if (dStatsBuffer_numPackets > actNumPacketsForMode) {
		*mode = (float) tree->value;
		actNumPacketsForMode = dStatsBuffer_numPackets;
	}

	float dStats_temp1 = (float) tree->value - dStats_packetLengths.mean;
	float dStats_temp2 = dStats_temp1 * dStats_temp1;
	// variance
	*variance += (1.0f / (float) (dStats_packetLengths.numOfPackets)) * dStatsBuffer_numPackets * dStats_temp2; // powf((float) tree->value - dStats_packetLengths.mean, 2.0f);

	// skewness
	dStats_temp2 = dStats_temp2 * dStats_temp1;
	*skewness += (1.0f / (float) (dStats_packetLengths.numOfPackets)) * dStatsBuffer_numPackets * dStats_temp2; // powf((float) tree->value - dStats_packetLengths.mean, 3.0f);

	// kurtosis
	dStats_temp2 = dStats_temp2 * dStats_temp1;
	*excess += (1.0f / (float) (dStats_packetLengths.numOfPackets)) * dStatsBuffer_numPackets * dStats_temp2; // powf((float) tree->value - dStats_packetLengths.mean, 4.0f);


	// update actual number of packets
	*actNumPackets += dStatsBuffer_numPackets;

	// proceed with right child
	if (tree->rightChild) dStats_getStatsFromTreeRek(tree->rightChild, actNumPackets, lower, median, upper, mode, variance, skewness, excess);
}

