/*
 * dhcpDecode.c
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 */


#include "dhcpDecode.h"

// Global variables 
#ifdef __APPLE__
extern hashMap_t* mainHashMap;
extern outputBuffer_t* main_output_buffer;
#endif // __APPLE__

// plugin variables
unsigned long plugin_dhcpDecode_Index;
dhcpFlow_t *dhcpFlow;

// Tranalyzer functions

char* get_plugin_name() {
	return "dhcpDecode";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

char* get_dependencies() {
	return "";
}

void initialize(unsigned long pPluginIndex) {
	plugin_dhcpDecode_Index = pPluginIndex;
#ifdef __APPLE__
        mainHashMap = getMainHashMap();
        baseFileName = getBaseFileName();
        specialOutput = getSpecialOutput();
#endif // __APPLE__

        dhcpFlow = malloc(sizeof(dhcpFlow_t) * mainHashMap->hashChainTableSize); // allocate struct for all flows
        memset(dhcpFlow, '\0', sizeof(dhcpFlow_t) * mainHashMap->hashChainTableSize); // set all to 0
}


binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;

	// 
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Status", "DHCPStat", 0, 1, bt_hex_8)); 
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Hardware Type", "DHCPHWTyp", 0, 1, bt_hex_32)); 
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Client Hardware Address", "DHCPCHWAdd", 0, 1, bt_mac_addr)); 
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Hop Count", "DHCPHopCnt", 0, 1, bt_uint_8)); 
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Option Count", "DHCPOpCnt", 0, 1, bt_uint_16)); 
#if DHCPBITFLD == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Option BitFiled", "DHCPOptiBF", 0, 3, bt_hex_64, bt_hex_64, bt_hex_64));
#else // DHCPBITFLD == 0
	main_bv = bv_append_bv(main_bv, bv_new_bv("DHCP Options", "DHCPOpt", 1, 1, bt_uint_8));
#endif // DHCPBITFLD

	return main_bv;

}


void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *flowP = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
	dhcp6Header_t *dhcpH = (dhcp6Header_t*)packet->layer7Header;
#else // IPV6_ACTIVATE == 0
	dhcpHeader_t *dhcpH = (dhcpHeader_t*)packet->layer7Header;
#endif // IPV6_ACTIVATE

	dhcpFlow_t *dhcpFlowP = &dhcpFlow[flowIndex];

	memset( dhcpFlowP, '\0', sizeof(dhcpFlow_t) ); // set everything to 0

}



void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	udpHeader_t *udpHdr = (udpHeader_t*)packet->layer4Header;
	if (packet->layer3Header->ipHeader.ip_p == L3_UDP) {
		if ( (udpHdr->source != DHCPUDPCPn || udpHdr->dest != DHCPUDPSPn) && (udpHdr->source != DHCPUDPSPn || udpHdr->dest != DHCPUDPCPn) ) return;
	} else return;

	int32_t i, dhcpOptLen = ntohs(udpHdr->len) - DHCPOPTUDPOFF;
	unsigned char *dhcpOpt;

	dhcpFlow_t *dhcpFlowP = &dhcpFlow[flowIndex];
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *flowP = &flows[flowIndex];

#if IPV6_ACTIVATE == 1
	dhcp6Header_t *dhcpHdr = (dhcp6Header_t*)packet->layer7Header;
#else // IPV6_ACTIVATE == 0
	dhcpHeader_t *dhcpHdr = (dhcpHeader_t*)packet->layer7Header;
#endif // IPV6_ACTIVATE

	dhcpFlowP->stat |= 1 << (dhcpHdr->opcode-1);
	dhcpFlowP->hwType |= 1 << dhcpHdr->hwType;
	dhcpFlowP->hopCnt |= dhcpHdr->hopCnt;
	dhcpFlowP->stat |= 8 >> (dhcpHdr->flags & DHCPBCST); 
	dhcpFlowP->clientIP = dhcpHdr->clientIP;
	dhcpFlowP->yourIP = dhcpHdr->yourIP;
	dhcpFlowP->servIP = dhcpHdr->servIP;
	dhcpFlowP->gwIP = dhcpHdr->gwIP;
	dhcpFlowP->clHWAdd[0] = dhcpHdr->clientHWaddr[0];
	dhcpFlowP->clHWAdd[1] = dhcpHdr->clientHWaddr[1];

	if (dhcpHdr->optMagNum != MAGICNUMBERn) {
		dhcpFlowP->stat |= DHCPMAGNUMERR;
		return;
	}

	dhcpOpt =  (unsigned char*)&dhcpHdr->optMagNum + 4;
	
	if (packet->snapL7Length - DHCPOPTOFF >= dhcpOptLen ) { // option field exists or crafted packet?
		for ( i = 0; i < dhcpOptLen, dhcpOpt[i], dhcpOpt[i] != DHCPOPTEND; i += dhcpOpt[i+1] + 2 ) {
#if DHCPBITFLD == 1
			if (dhcpOpt[i] < 64 ) dhcpFlowP->optT[2] |= (uint64_t)1 << (dhcpOpt[i] & DHCP64MSK);
			else if (dhcpOpt[i] < 128 ) dhcpFlowP->optT[1] |= (uint64_t)1 << ((dhcpOpt[i] - 64) & DHCP64MSK);
			else dhcpFlowP->optT[0] |= (uint64_t)1 << ((dhcpOpt[i] - 128) & DHCP64MSK);
#else // DHCPBITFLD == 0
			if (dhcpFlowP->optNum < DHCPMAXOPT) dhcpFlowP->opt[dhcpFlowP->optNum++] = dhcpOpt[i];
#endif // DHCPBITFLD
			dhcpFlowP->optCntT++;
		}
	} else dhcpFlowP->stat |= DHCPOPTCORRPT; // warning: crafted packet or option field not acquired
	

}



void onFlowTerminate(unsigned long flowIndex) {
	dhcpFlow_t *dhcpFlowP = &dhcpFlow[flowIndex];
	uint32_t i, j;

#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	flow_t *FlowP = &flows[flowIndex];

	outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->stat, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->hwType, sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->clHWAdd[0], 6);
	outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->hopCnt, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->optCntT, sizeof(uint16_t));
#if DHCPBITFLD == 1
	outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->optT, 3*sizeof(uint64_t));
#else // DHCPBITFLD == 0
	j = dhcpFlowP->optNum;
	outputBuffer_append(main_output_buffer, (char*)&j, sizeof(uint32_t));
	for (i = 0; i < j; i++) {
		outputBuffer_append(main_output_buffer, (char*)&dhcpFlowP->opt[i], sizeof(uint8_t));	
	}
#endif // DHCPBITFLD
}



void onApplicationTerminate() {

	if (dhcpFlow) {
		free(dhcpFlow);
		dhcpFlow = NULL;
	}

}


