/*
 * dhcpDecode.h
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 */

#ifndef DHCPDECODE_H_
#define DHCPDECODE_H_

// global includes

// local includes
#include "main.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "networkHeaders.h"

// user defines
#define DHCPBITFLD	0		// option Bitfield / option numbers in a row
#define DHCPMAXOPT	10		// if (DHCPBITFLD == 0) maximum stored options

// plugin defines
#define DHCP64MSK	0x3f		//
#define DHCPBCST	0x8000		// 
#define MAGICNUMBERn	0x63538263	// DHCP/BOOTP option magic number
#define DHCPOPTOFF	240		// dhcp header
#define DHCPOPTUDPOFF	(DHCPOPTOFF + 8)	// dhcp header + udp header
#define DHCPOPTEND	0xff		// dhcp opton end marker
#define DHCPUDPCPn	0x4400		// dhcp client port network order
#define DHCPUDPSPn	0x4300		// dhcp server port network order

// stat bit field
#define DHCPOPTCORRPT	0x20
#define DHCPMAGNUMERR	0x10

// sample plugin structures
typedef struct dhcpFlow_s { 
	uint64_t optT[3];
	uint32_t clHWAdd[2];
	uint32_t hwType;
        uint32_t clientIP;
        uint32_t yourIP;
        uint32_t servIP;
        uint32_t gwIP;
	uint16_t optCntT;
	uint16_t optNum;
        uint8_t hopCnt;
	uint8_t stat;
	uint8_t opt[DHCPMAXOPT];
} dhcpFlow_t;

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

#endif // DHCPDECODE

