/*
 * dnsDecode
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
*/
#include "tranalyzer.h"
#include "dnsDecode.h"

// Global variables
#if TF_MALSITE_TEST == 1
malsitetable_t *tf_malsite_table;
#endif // TF_MALSITE_TEST == 1

// tranalyzer routines
char* get_plugin_name() {
	return "dnsDecode";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	pluginIndexDNSDecodeIndex = pPluginIndex;
	
	dnsFlow = malloc(sizeof(dnsFlow_t) * mainHashMap->hashChainTableSize);
	memset(dnsFlow, '\0', sizeof(dnsFlow_t) * mainHashMap->hashChainTableSize);

#if TF_MALSITE_TEST == 1
	char user_plugin_folder_mal[256];
        memset(user_plugin_folder_mal,'\0', 256);
        strncpy(user_plugin_folder_mal, getenv("HOME"),255);
        strncat(user_plugin_folder_mal, "/", 1);
        strncat(user_plugin_folder_mal, PLUGIN_FOLDER, 255 - strlen(user_plugin_folder_mal));
        strncat(user_plugin_folder_mal, "/", 1);
        strncat(user_plugin_folder_mal, TMALFILE, 255 - strlen(user_plugin_folder_mal));
        tf_malsite_table = malloc(sizeof(malsitetable_t));
        malsite_init(tf_malsite_table, user_plugin_folder_mal);	
#endif // TF_MALSITE_TEST == 1
	
}


binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;

	binary_value_t *act_bv = bv_new_bv("DNS Status", "DNSStatus", 0, 1, bt_hex_8);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("DNS last Header Field", "DNSHdriOPField", 0, 1, bt_hex_16);
	main_bv = bv_append_bv(main_bv, act_bv);
	act_bv = bv_new_bv("DNS Aggregated Header Status_Opcode_ReturnCode Bit Field", "DnsStat_OpC_RetC", 0, 3, bt_hex_8, bt_hex_16, bt_hex_16);
	main_bv = bv_append_bv(main_bv, act_bv);

#if DNS_MODE  > 1
	act_bv = bv_new_bv("DNS Type Bit Fields", "DNSTypeBF", 0, 2, bt_hex_16, bt_hex_64);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // DNS_MODE > 1

#if DNS_MODE  > 0
	act_bv = bv_new_bv("DNS Question Answer Aux Additional cnt ", "DnsCntQu_Asw_Aux_Add", 0, 4, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // DNS_MODE > 0

#if DNS_MODE  > 2
	act_bv = bv_new_bv("DNS Query Name", "DNSQname", 1, 1, bt_string_class);
	main_bv = bv_append_bv(main_bv, act_bv);

#if (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 1)
	act_bv = bv_new_bv("DNS Domain Malware Code", "DmMalCode", 1, 1, bt_hex_32);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 1)

#if DNS_RECORD_MODE == 0
	act_bv = bv_new_bv("DNS Answer Name Record", "DNSAname", 1, 1, bt_string_class);
	main_bv = bv_append_bv(main_bv, act_bv);
#else // DNS_RECORD_MODE == 1
	act_bv = bv_new_bv("DNS Name CNAME entries", "DNSAPname", 1, 1, bt_string_class);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // DNS_RECORD_MODE

#if IPV6_ACTIVATE == 1
	act_bv = bv_new_bv("DNS Address entries IPv6", "DNSAaddress", 1, 1, bt_ip6_addr);
	main_bv = bv_append_bv(main_bv, act_bv);
#else // IPV6_ACTIVATE == 0
	act_bv = bv_new_bv("DNS Address entries IPv4", "DNSAaddress", 1, 1, bt_ip4_addr);
	main_bv = bv_append_bv(main_bv, act_bv);
#if (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 0)
	act_bv = bv_new_bv("DNS IP Malware Code", "IPMalCode", 1, 1, bt_hex_32);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 0)
#endif // IPV6_ACTIVATE

	act_bv = bv_new_bv("DNS Answer Record TTL entries", "DNSATTL", 1, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("DNS Answer Record Type entries", "DNSAType", 1, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // DNS_MODE > 2

	return main_bv;
}


void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	dnsFlow_t *dnsFlowP = &dnsFlow[flowIndex];
	
	memset(dnsFlowP, '\0', sizeof(dnsFlow_t));
}


void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
#if IPV6_ACTIVATE == 1
        if ( packet->ip6FragHdr && (packet->ip6FragHdr->frag_off & FRAG6ID_N) ) return; // only tcp and 1. frag packet will be processed
#else // IPV6_ACTIVATE == 0
	if ( (((ipHeader_t*)packet->layer3Header)->ip_off) & FRAGID_N ) return; // only icmp and 1. frag packet will be processed	
#endif
	register uint16_t sp, dp;
	
	sp = *(uint16_t*)packet->layer4Header;
	dp = *((int16_t*)packet->layer4Header+1);
	if ( sp != DNSPORTN && dp!= DNSPORTN && !(sp== DNSPORTMN && dp == DNSPORTMN) ) return; // only dns
	
	uint16_t *dnsPayload = (uint16_t*)packet->layer7Header;
 	if ( packet->snapL7Length  < 6 ) return; 	

	uint16_t dnsLen, u;
	
	if ( packet->layer4Type == L3_TCP ) { // tcp: length in dns header
		dnsLen = ntohs(*dnsPayload);
		if (dnsLen > packet->snapL7Length) dnsLen = packet->snapL7Length;
		dnsPayload++;
	} else if (packet->layer4Type == L3_UDP ) dnsLen = packet->snapL7Length;
	else return;
	
	if ( dnsLen < 4  ) return; // no dns payload
	
	dnsFlow_t *dnsFlowP = &dnsFlow[flowIndex];

//	numOfDNSPackets++; // count only unfragmented dns packets

	dnsPayload++;
	u = ntohs(*dnsPayload);
	
	dnsFlowP->dnsHdField = u;   
	dnsFlowP->dnsStatBfield |= ((u & 0x07F0)>>4) | ((u & 0x8000)>>8);
	dnsFlowP->dnsOpCodeBfield |= (1 << ((u & 0x7800)>>11));
	dnsFlowP->dnsRCodeBfield |= (1 << (u & 0x000F));

#if DNS_MODE > 0
	if ( dnsLen < 24 ) return; // no dns payload

#if DNS_MODE > 1
	uint16_t dnsQDBCnt, dnsANBCnt;
	dnsQDBCnt = dnsFlowP->dnsQDCnt;
	dnsANBCnt = dnsFlowP->dnsANCnt;
#endif // DNS_MODE > 1

	dnsFlowP->dnsQDCnt += ntohs(*(++dnsPayload));
	dnsFlowP->dnsANCnt += ntohs(*(++dnsPayload));
	dnsFlowP->dnsNSCnt += ntohs(*(++dnsPayload));
	dnsFlowP->dnsARCnt += ntohs(*(++dnsPayload));
#endif // DNS_MODE > 0
#if DNS_MODE > 1
	register uint32_t i, m, t = dnsFlowP->dnsQDECnt, j = dnsFlowP->dnsQDECnt;
	uint16_t k, l, len = 0;
	uint8_t *dnsPayloadB;
	dnsPayload -= 5; // reset uiint16_t ptr to beginning of DNS payload
	dnsPayloadB = (uint8_t*)dnsPayload; // set byte ptr

//	if  ( dnsFlowP->dnsQDCnt > DNS_QRECMAX ) dnsFlowP->dnsQDECnt = DNS_QRECMAX;
//	else dnsFlowP->dnsQDECnt = dnsFlowP->dnsQDCnt;
	if  ( dnsFlowP->dnsQDCnt > DNS_QRECMAX ) t = DNS_QRECMAX;
	else t = dnsFlowP->dnsQDCnt;

	for ( i = dnsQDBCnt, l = DNS_RSTART; i < t; i++ ) {
		k = 0;
		l = dns_parse( dnsFlowP->dnsQname[j], dnsLen, l, &k, dnsPayloadB );
		if ( l > DNS_MAXUDPLEN ) goto err;
		for ( m = 0; m < i; m++ ) {
			if ( !strncmp(dnsFlowP->dnsQname[m], dnsFlowP->dnsQname[j], DNS_HNLMAX) ) {
                		break;
			} 
		}

		dnsPayload = (uint16_t*)(dnsPayloadB + l);
		dnsFlowP->dnsQType[t] = ntohs(*dnsPayload++);
		dnsFlowP->dnsQClass[t] = ntohs(*dnsPayload++);

		l += 4; // advance byte ptr to unit16_t ptr
		if ( m == i ) j++;
	}
	dnsFlowP->dnsQDECnt = j;

	if  ( dnsFlowP->dnsANCnt > DNS_ARECMAX ) dnsFlowP->dnsANECnt = DNS_ARECMAX;
	else dnsFlowP->dnsANECnt = dnsFlowP->dnsANCnt;
	
	for ( i = dnsANBCnt ; i < dnsFlowP->dnsANECnt; i++ ) {
		k = 0;
		l = dns_parse( dnsFlowP->dnsAname[i], dnsLen, l, &k, dnsPayloadB );
		if ( l > DNS_MAXUDPLEN ) goto err;
	
		dnsPayload = (uint16_t*)(dnsPayloadB + l );	
		m = dnsFlowP->dnsAType[i] = ntohs(*dnsPayload++);
		if ( m < 64 ) dnsFlowP->dnsATypeLBF |= ((uint64_t)1 << m);
		if ( m > 248 && m < 258) dnsFlowP->dnsATypeUBF |= (1 << (m - 249));
		if ( m > 32767 && m < 32780 ) dnsFlowP->dnsATypeUBF |= (1 << (m - 32754));
		if ( dnsFlowP->dnsAType[i] == DNS_ZONE || dnsFlowP->dnsAType[i] == DNS_ZONEA ) dnsFlowP->dnsStatus |= DNS_ZTRANS;
#if DNS_MODE > 2
		dnsFlowP->dnsAClass[i] = ntohs(*dnsPayload++);
		dnsFlowP->dnsAttl[i] = ntohl(*(uint32_t*)dnsPayload++);
		dnsPayload++;
		len = ntohs(*dnsPayload++);

		if ( dnsFlowP->dnsAType[i] == DNS_MX ) {
			dnsFlowP->dnsMXPref[i] = ntohs(*dnsPayload++);
			l += 2;
		}

		l += 10; // advance byte ptr to unit16_t ptr

		if (dnsFlowP->dnsAType[i] == DNS_HOST ) { 
#if DNS_RECORD_MODE == 1
#if IPV6_ACTIVATE == 1
			dnsFlowP->dnsAadd[dnsFlowP->dnsAIndex++] = *(ipAddr_t*)dnsPayload++;
#else // IPV6_ACTIVATE == 0

			dnsFlowP->dnsAadd[dnsFlowP->dnsAIndex++] = *(uint32_t*)dnsPayload++;
#endif //IPV6_ACTIVATE
#else // DNS_RECORD_MODE == 0
#if IPV6_ACTIVATE == 1
			dnsFlowP->dnsAadd[i] = *(ipAddr_t*)dnsPayload++;
#else // IPV6_ACTIVATE == 0
			dnsFlowP->dnsAadd[i] = *(uint32_t*)dnsPayload++;
#endif //IPV6_ACTIVATE
#endif // DNS_RECORD_MODE
			l += 4;
		} else {
			k = 0;
#if DNS_RECORD_MODE == 1
			l = dns_parse( dnsFlowP->dnsPname[dnsFlowP->dnsPIndex++], dnsLen, l, &k, dnsPayloadB );
			if ( l > DNS_MAXUDPLEN ) goto err;
#else // DNS_RECORD_MODE == 0
			l = dns_parse( dnsFlowP->dnsPname[i], dnsLen, l, &k, dnsPayloadB );
			if ( l > DNS_MAXUDPLEN ) goto err;
#endif // DNS_RECORD_MODE
		}
#endif // DNS_MODE > 2
			
	}

	return;

// Error Handling
err:	if ( l == 65531 ) dnsFlowP->dnsStatus |= DNS_ERRLEN;
	if ( l == 65532 ) dnsFlowP->dnsStatus |= DNS_ERRPTR;
#endif // DNS_MODE > 1

}


void onFlowTerminate(unsigned long flowIndex) {
	uint32_t i,j,k;
	char* mal_sdomain=NULL;

	dnsFlow_t *dnsFlowP = &dnsFlow[flowIndex];
	
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsStatus, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsHdField, sizeof(uint16_t));

	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsStatBfield, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsOpCodeBfield, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsRCodeBfield, sizeof(uint16_t));

#if DNS_MODE > 1
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsATypeUBF, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsATypeLBF, sizeof(uint64_t));
#endif // DNS_MODE > 1


#if DNS_MODE > 0
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsQDCnt, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsANCnt, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsNSCnt, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsARCnt, sizeof(uint16_t));
#endif // DNS_MODE > 0


#if DNS_MODE > 2
	j = dnsFlowP->dnsQDECnt;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0 ; i < j; i++ ) {
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsQname[i][0], strlen(dnsFlowP->dnsQname[i])+1);
	}

#if (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 1)
	j = dnsFlowP->dnsQDECnt;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
		k = maldomain_test(tf_malsite_table, &dnsFlowP->dnsQname[i][0], strlen(dnsFlowP->dnsQname[i])); 
		if (k) numAlarms++;
		outputBuffer_append(main_output_buffer, (char*) &k, sizeof(uint32_t));
	}
#endif // (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 1)


#if DNS_RECORD_MODE == 0
	j = dnsFlowP->dnsANECnt;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsAname[i][0], strlen(dnsFlowP->dnsAname[i])+1);
	}
#else // DNS_RECORD_MODE == 1
	j = dnsFlowP->dnsPIndex;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsPname[i][0], strlen(dnsFlowP->dnsPname[i])+1);
	}
#endif // DNS_RECORD_MODE

#if DNS_RECORD_MODE == 1
	j = dnsFlowP->dnsAIndex;
#else // DNS_RECORD_MODE == 0
	j = dnsFlowP->dnsANECnt;
#endif // DNS_RECORD_MODE
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
#if IPV6_ACTIVATE == 1
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsAadd[i], sizeof(ipAddr_t));
#else // IPV6_ACTIVATE == 0
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsAadd[i], sizeof(uint32_t));
#endif // IPV6_ACTIVATE
	}

#if (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 0)
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
		k = malip_test(tf_malsite_table, dnsFlowP->dnsAadd[i]); 
		if (k) numAlarms++;
		outputBuffer_append(main_output_buffer, (char*) &k, sizeof(uint32_t));
	}
#endif // (TF_MALSITE_TEST == 1 && TF_MALSITE_DOMAIN == 0)

	j = dnsFlowP->dnsANECnt;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsAttl[i], sizeof(uint32_t));
	}

	j = dnsFlowP->dnsANECnt;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for ( i = 0; i < j; i++ ) {
		outputBuffer_append(main_output_buffer, (char*) &dnsFlowP->dnsAType[i], sizeof(uint16_t));
	}
#endif // DNS_MODE > 2
}



void onApplicationTerminate(unsigned long flowIndex) {
	
	if (dnsStatsFile) fclose(dnsStatsFile);
	if (dnsFlow) {
		free(dnsFlow);
		dnsFlow = NULL;
	}
}


uint16_t dns_parse( unsigned char *dnsName, uint16_t len, uint16_t l, uint16_t *kp, uint8_t *dnsPayloadB ) {
	uint16_t j, k;
	uint8_t sw = 1;
	k = *kp;

	while ( 1 ) {
		if ( l + 1 >= len || l < DNS_RSTART) {
			l = 65530; // ++l -> 65531
			break;
		}
		j = *(uint16_t*)(dnsPayloadB+l);
		if ( j & DNS_PTRN) {
			if ( (j & DNS_PTRN) == DNS_PTRN ) {
				dns_parse( dnsName, len, ntohs(j & DNS_PTRVN), &k, dnsPayloadB );
				return l + 2;
			} else {
				l = 65531; // ++l -> 65532
				break;
			}
		} else {
			if ( sw && (k + dnsPayloadB[l] < DNS_HNLMAX) ) { 
				for ( j = l + 1; j <= l + dnsPayloadB[l]; j++ ) { 
					dnsName[k++] = dnsPayloadB[j];
				//	if ( j <= len) dnsName[k++] = dnsPayloadB[j];
				}
				l = j;
			} else { 
				l += dnsPayloadB[l] + 1;
				sw = 0;
			}
			
			if ( dnsPayloadB[l] == '\0' ) break;
			if (sw) dnsName[k++] = '.';
		}
	}
	dnsName[k] = '\0';
	*kp = k;
	return ++l;
}

