 /*
 * dnsDecode
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
*/

#ifndef __DNSDECODE_H__
#define __DNSDECODE_H__

// local includes
#include "networkHeaders.h"
#include "main.h"
#include "binaryValue.h"

// user config: compile switches
#define TF_MALSITE_TEST !IPV6_ACTIVATE * 1	// activate trueflow mode, currently only ipv4
#define	DNS_MODE	2	// 0: Only aggregated header info, 1: +Record counts, 2: +Type bit field, 3: +Extract Content
#define DNS_RECORD_MODE 0	// if (DNS\_MODE > 2) 0: Answer Name Records, 1: CNAME entries
#define	DNS_QRECMAX 	10	// Max # of query records / flow
#define	DNS_ARECMAX 	10	// Max # of answer records / flow
#define	DNS_HNLMAX 	40 	// Max name length in flow structure

#if TF_MALSITE_TEST == 1 
#include "malsite.h"
#if TF_MALSITE_DOMAIN == 1
#define TMALFILE "tf_dm_malsite.txt"
#else // TF_MALSITE_DOMAIN == 0
#define TMALFILE "tf_ip_malsite.txt"
#endif // TF_MALSITE_DOMAIN
#endif // TF_MALSITE_TEST == 1

// Local defines

// DNS Status
#define	DNS_ERRLEN	0x01	// Error: DNS record length error
#define	DNS_ERRPTR	0x02	// Error: Wrong DNS PTR detected
#define	DNS_ZTRANS	0x80	// Warning: DNS zone transfer detected

// DNS boundary conditions
#define	DNS_MAXUDPLEN	65445	// Max DNS udp payload length
#define	DNS_MAXTCPLEN	65443	// Max DNS tcp payload length
#define DNS_RSTART	12	// DNS record start in payload

// DNS Ports network order
#define DNSPORTN	0x3500	// 53
#define DNSPORTMN	0xE914	// 5353

// DNS types
#define DNS_QR	0x8000
#define DNS_AA	0x0400
#define DNS_TC	0x0200
#define DNS_RD	0x0100
#define DNS_RA	0x0080
#define DNS_BF	0x0780

#define	DNS_OPC_MASKn	0x7000
#define	DNS_RC_MASKn	0x000F

#define DNS_PTRN	0x00C0
#define DNS_PTRVN	~DNS_PTRN	

// Type codes
#define DNS_HOST	0x0001
#define DNS_CNAME	0x0005
#define DNS_SOA		0x0006
#define DNS_PTR		0x000C
#define DNS_MX		0x000F
#define DNS_TXT		0x0010
#define DNS_ZONE	0x00FC
#define DNS_ZONEA	0x00FF



// functions
extern void terminate();

char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);
void onApplicationTerminate(unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
binary_value_t* printHeader();


//structures
typedef struct dnsHCode_s {
	uint16_t rCode:4;
	uint16_t z:3;
	uint16_t aCode:4;
	uint16_t opCode:4;
	uint16_t qr:1;
} dnsHCode_t;

typedef union dnsHC_u {
        dnsHCode_t dnsHCs;
        uint16_t dnsHCu;
} dnsHC_t;

typedef struct dnsHeader_s {
//	uint16_t len; // only TCP
	uint16_t id;
	dnsHC_t dnsHCode;	
	uint16_t qdCount;	
	uint16_t anCount;	
	uint16_t nsCount;	
	uint16_t arCount;	
} dnsHeader_t;


typedef struct dnsFlow_s {
	uint64_t dnsATypeLBF;
#if DNS_MODE > 2
#if IPV6_ACTIVATE == 1
	ipAddr_t dnsAadd[DNS_ARECMAX];
#else // IPV6_ACTIVATE == 0
	uint32_t dnsAadd[DNS_ARECMAX];
#endif // IPV6_ACTIVATE
	uint32_t dnsAttl[DNS_ARECMAX];
	uint16_t dnsAClass[DNS_ARECMAX];
	uint16_t dnsMXPref[DNS_ARECMAX];
	uint16_t dnsATypeUBF;
#endif // DNS_MODE > 2
#if DNS_RECORD_MODE == 1
	uint16_t dnsAIndex;
	uint16_t dnsPIndex;
#endif // DNS_RECORD_MODE == 1
#if DNS_MODE > 0
	uint16_t dnsQDCnt;
	uint16_t dnsANCnt;
	uint16_t dnsNSCnt;
	uint16_t dnsARCnt;
	uint16_t dnsAType[DNS_ARECMAX];
#endif // DNS_MODE > 0
	uint16_t dnsHdField;
	uint16_t dnsRCodeBfield;
	uint16_t dnsOpCodeBfield;
#if DNS_MODE > 1
	uint16_t dnsQDECnt;
	uint16_t dnsANECnt;
	uint16_t dnsQType[DNS_QRECMAX];
	uint16_t dnsQClass[DNS_QRECMAX];
	unsigned char dnsQname[DNS_QRECMAX][DNS_HNLMAX];
	unsigned char dnsAname[DNS_ARECMAX][DNS_HNLMAX];
	unsigned char dnsPname[DNS_ARECMAX][DNS_HNLMAX];
#endif // DNS_MODE > 1
	uint8_t dnsStatBfield;
	uint8_t dnsStatus;
} dnsFlow_t;

// variables
FILE *dnsStatsFile = NULL;
unsigned long pluginIndexDNSDecodeIndex;
dnsFlow_t *dnsFlow;

// 
#if DNS_MODE > 1
uint16_t dns_parse( unsigned char *dnsName, uint16_t len, uint16_t l, uint16_t *kp, uint8_t *dnsPayloadB );
#endif // DNS_MODE > 1

#endif // __DNSDECODE_H__

