/*
 * malsite.c
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Noe Thalheim, RDNIZ
 * Co-Author: Stefan Burschka, RDNIZ
 *
*/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <arpa/inet.h>
#include "malsite.h"

void malsite_init(malsitetable_t* table, char* bw_file) {

	table->count = malsite_load(bw_file, NULL); //return the numbers of lines in the subnet file. 
	if (table->count == 0) return;
	table->malsites = (malsite_t *)malloc(sizeof(malsite_t) * table->count); //allocate memory dependent of the numbers of lines in the subnet file.
	table->count = malsite_load(bw_file, table->malsites);  
}

/* returns the number of subnets and file table, if given */
int malsite_load(char* bw_file, malsite_t* malsites) {
	FILE* malsite_file;
	char  line[281], domain[MAXLENDMN];
	uint32_t count = 0, id, ip;
	malsite_file =  fopen(bw_file,"r");

	if (! malsite_file) {
		printf("Cannot open malsite file %s\n", bw_file);
		return 0;
	}
	while ( fgets(line,280,malsite_file) ) {
		line[280]='\0'; //Set the last character as '\0'
		if (line[0] == '\n' || line[0] == '#' || line[0] == ' ' || line[0] == '\t') continue;
		if (malsites) {
#if TF_MALSITE_DOMAIN == 1 
			sscanf(line,"%[^\t\n]\t%x\n", domain, &id);
			ip = strlen(domain);
			malsites[count].len = ip;
			memcpy(malsites[count].malDomain, domain, ip);
#else // TF_MALSITE_DOMAIN == 0
			sscanf(line,"%x\t%x\n", &ip, &id);
			malsites[count].malIp = ip;
#endif // TF_MALSITE_DOMAIN
		malsites[count].malId = id;
		}
		count++;
	}	
#if TF_MALSITE_DOMAIN == 1 
	if (malsites) printf("# mal domains: %d\n", count);
#else // TF_MALSITE_DOMAIN == 0
	if (malsites) printf("# mal ips: %d\n", count);
#endif // TF_MALSITE_DOMAIN
	fclose(malsite_file);

	return count;
}



#if TF_MALSITE_DOMAIN == 1 

/* this function tests whether a given domain name is a malware host defined in the config file */
inline uint32_t maldomain_test(malsitetable_t* table, char* dname, uint8_t len) {
	register int malsite_end, malsite_start, malsite_i, i;

	if (*dname != '\0') {
		while (1) {
			malsite_end = table->count - 1;
			malsite_start = 0;
			while (malsite_start <= malsite_end) {
				malsite_i = (malsite_end + malsite_start) / 2; //define malsite_i as middle between malsite_start and malsite_end.
				i = strncmp(dname, table->malsites[malsite_i].malDomain, len);
				if ( !i ) 
					return table->malsites[malsite_i].malId; //return the located malsite codes. 
				else if ( i < 0 ) 
					malsite_end = malsite_i - 1; //set the endpoint one under the currently middle.
				else 
					malsite_start = malsite_i + 1; //set the startpoint one over the currently middle. 
			}
			return 0; // in case the ip isn't in the file, return 0.
		}
	} else return 0;
}

#else // TF_MALSITE_DOMAIN == 0

/* this function tests whether a given IP is a malware IP defined in the config file */
inline uint32_t malip_test(malsitetable_t* table, in_addr_t ip) {
	register int malsite_end, malsite_start, malsite_i;
	register uint32_t i;

	if (ip != 0) {
		while (1) {
			ip = ntohl(ip);
			malsite_end = table->count - 1;
			malsite_start = 0;
			while (malsite_start <= malsite_end) {
				malsite_i = (malsite_end + malsite_start) / 2; //define malsite_i as middle between malsite_start and malsite_end.
				i = table->malsites[malsite_i].malIp;
				if (ip  == i) 
					return table->malsites[malsite_i].malId; //return the located malsite codes. 
				else if (ip < i) 
					malsite_end = malsite_i - 1; //set the endpoint one under the currently middle.
				else 
					malsite_start = malsite_i + 1; //set the startpoint one over the currently middle. 
			}
			return 0; // in case the ip isn't in the file, return 0.
		}
	} else return 0;
}

#endif // TF_MALSITE_DOMAIN
