/*
 * entropy.h
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 * Co-Author: Mischa Lehmann, RDNIZ
 */

#ifndef entropy_H_
#define entropy_H_

#include "main.h"
#include "outputBuffer.h"
#include "networkHeaders.h"
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>
#include <math.h>

//Constants
#define	ENT_MAXPBIN	1 << 8	// N = 8 Bit Word, vocabulary: 256	
#define ENT_ALPHA_D 	0	// 1: print Alphabet distribution in flow file

#define ENT_MK_FRAGS 	0	// 1: entropy time series 

#define ENT_MAX_DATA_SIZE	100000000	// Maximum data-size before crash
#define ENT_MAX_FRAG_COUNT	10 		// Maximum amount of different fragments ( WHEN YOU CHANGE THIS ALSO CHANGE the ENT_AGS 
#define ENT_ARGS 	bt_float, bt_float, bt_float, bt_float, bt_float, bt_float, bt_float, bt_float, bt_float, bt_float // 
#define ENT_FRAG_SIZE 	100 		// size in bytes of each fragment !!!!!! IMPORTANT :: IF THIS DEFINE IS > 255 you have to change the ENT_DATATYPE !!!!!!
#define ENT_DATATYPE 	uint8_t 	// the used datatype for counting occurances


// everything that is calculated per flow. ~1000 Bytes allocated
typedef struct collectedPayloadPerFlow_s {
        uint32_t numBytes; 			// Number of bytes collected 
        uint32_t binCount[ENT_MAXPBIN]; 	// Count of each bin value

#if ENT_MK_FRAGS > 0
	ENT_DATATYPE frag_count;		// The count of time series ragments
	ENT_DATATYPE frag_size;			// The size of the current fragment
	ENT_DATATYPE frag_binCount[ENT_MAXPBIN];	// The current time series fragment
	float frag_entropy[ENT_MAX_FRAG_COUNT];	// Calculated entropies
#endif
} collectedPayloadPerFlow_t;

// variables
unsigned long pluginEntropyIndex;
#ifdef __APPLE__
//gloabl variables 
extern hashMap_t* mainHashMap;
extern outputBuffer_t* main_output_buffer;
#endif // __APPLE__
// implemented functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();

void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
void onApplicationTerminate();

#endif /* entropy_H_ */
