/*
 * global_statistics.h
 *
 * Autor: Torben Ruehl, RUAG (SWITZERLAND) Ltd., RDNIZ
 *
 * Copyright: 2013 by RUAG (Switzerland) Ltd.
 */
#ifndef GLOBAL_STATISTICS_H_
#define GLOBAL_STATISTICS_H_

// #define GS_DEBUG

// global includes
#include <stdio.h>
#include <stdarg.h>
#include <stdint.h>
#include <stdlib.h>
#include <stddef.h>

// local includes
#include "main.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "networkHeaders.h"

#define GS_SAMPLE_TIME 10000000 // sample time of statistics, in microseconds
#define GS_NUMBER_OF_HISTO_SECONDS 600 // Number of seconds represented in the Histogram
#define GS_BIN_WIDTH 1000000 // Histogram bin width, in microseconds

typedef struct gs_stat_s {
    uint64_t packets_sec;
    uint64_t bytes_sec;
    uint64_t flows_started;
    uint64_t flows_stopped;
    uint64_t current_flows_mean;
    uint64_t current_flows_max;
    uint64_t current_flows_min;
    double_t current_flows_stddev;
    uint64_t timer_changes;
    uint64_t flow_age_mean;
    uint64_t flow_age_max;
    uint64_t flow_age_stddev;
    uint64_t flow_age_histo[GS_NUMBER_OF_HISTO_SECONDS * 1000000 / GS_BIN_WIDTH];

} gs_stat_t;

gs_stat_t gs_stats;

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer2Information(packet_t* packet);
void claimLayer3Information(packet_t* packet);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

#endif // GLOBAL_STATISTICS_H_

