/*
 * httpSniffer
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Authors: Mischa Lehmann, RDNIZ, Stefan Burschka, RDNIZ 
 */


// local includes    

#include "httpSniffer.h"


// flow worker globals    

unsigned long plugin_httpSniffer_Index;
static http_flow_t *http_flow;
static int32_t httpSniffer_fd_cnt, httpSniffer_fd_cnt_max;

//    local function prototype declarations

// Arguments:
//	- data : the packetdata
//	- http_data_len : length of packet (data-field)

static char* http_get_linesize(uint8_t *data, int16_t http_data_len); // return size of http-header line

/*
*	static int http_read_method(char *data,http_method *retmethod): analyze http method
*	Arguments:
*		- data : the packetdata
*		- retmethod : the found method
*/
static int http_read_method(char *data, http_method *retmethod); // 

/*
*	static int http_read_mimetype(char *data,http_mimetype *retmime):  analyze mimetype
*	Arguments:
*		- data : the packetdata
*		- retmime : the found mimetype
*/
static int http_read_mimetype(char *data, http_mimetype *retmime);

/*
*	static int http_read_header_data(char* data,const char *header,char **retdata): read header-field
*	Arguments:
*		- data : the packetdata
*		- header : the lookup-header field
*		- retdata : return data of the specified header field in header
*/

static char* http_read_header_data(char* data, uint16_t data_len, const char *header, uint16_t header_len);


// static void http_fname( httpFlow_t* httpFlowPO, flow_t* flowP, uint8_t* name_p, uint16_t cnt ); 


//     Tranalyzer functions

char* get_plugin_name() {
	return "httpSniffer";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}


void initialize(unsigned long pPluginIndex) {

	plugin_httpSniffer_Index = pPluginIndex;

	if ((http_flow = malloc(mainHashMap->hashChainTableSize * sizeof(http_flow_t))) == NULL) {
		fprintf( stderr, "http_flows: Couldn't allocate memory for flow statistics. Exit program");
		terminate();
	}
	memset(http_flow, 0x0,mainHashMap->hashChainTableSize * sizeof(http_flow_t));

#if HTTP_RM_PICDIR == 1
	int i;
#if HTTP_SAVE_IMAGE == 1
	i  = system("rm -r "HTTP_IMAGE_PATH);
#endif // HTTP_SAVE_IMAGE == 1 
#if HTTP_SAVE_VIDEO == 1
	i  = system("rm -r "HTTP_VIDEO_PATH);
#endif // HTTP_SAVE_VIDEO == 1 
#if HTTP_SAVE_AUDIO == 1
	i  = system("rm -r "HTTP_AUDIO_PATH);
#endif // HTTP_SAVE_AUDIO == 1 
#if HTTP_SAVE_MSG == 1
	i  = system("rm -r "HTTP_MSG_PATH);
#endif // HTTP_SAVE_MSG == 1 
#if HTTP_SAVE_TEXT == 1
	i  = system("rm -r "HTTP_TEXT_PATH);
#endif // HTTP_SAVE_TEXT == 1 
#if HTTP_SAVE_APPL == 1
	i  = system("rm -r "HTTP_APPL_PATH);
#endif // HTTP_SAVE_APPL == 1 
#endif //  HTTP_RM_PICDIR == 1

#if HTTP_SAVE_IMAGE == 1
	mkdir(HTTP_IMAGE_PATH, S_IRWXU);
#endif // HTTP_SAVE_IMAGE == 1 
#if HTTP_SAVE_VIDEO == 1
	mkdir(HTTP_VIDEO_PATH, S_IRWXU);
#endif // HTTP_SAVE_VIDEO == 1 
#if HTTP_SAVE_AUDIO == 1
	mkdir(HTTP_AUDIO_PATH, S_IRWXU);
#endif // HTTP_SAVE_AUDIO == 1 
#if HTTP_SAVE_MSG == 1
	mkdir(HTTP_MSG_PATH, S_IRWXU);
#endif // HTTP_SAVE_MSG == 1 
#if HTTP_SAVE_TEXT == 1
	mkdir(HTTP_TEXT_PATH, S_IRWXU);
#endif // HTTP_SAVE_TEXT == 1 
#if HTTP_SAVE_APPL == 1
	mkdir(HTTP_APPL_PATH, S_IRWXU);
#endif // HTTP_SAVE_APPL == 1 

	httpSniffer_fd_cnt = 0;
	httpSniffer_fd_cnt_max = 0;

}


binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;

	main_bv = bv_append_bv(main_bv, bv_new_bv("Flags of Http flow", "httpFlags", 0, 1, bt_hex_16));
	main_bv = bv_append_bv(main_bv, bv_new_bv("Http Methods in flow", "httpMethods", 0, 1, bt_hex_8));
	main_bv = bv_append_bv(main_bv, bv_new_bv("HEADMIME-TYPES in flow", "httpHeadMimes", 0, 1, bt_hex_16));
#if HTTP_BODY == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Http Body anomalies in flow", "httpBodyFlgs", 0, 1, bt_hex_8));
#endif // HTTP_BODY
	main_bv = bv_append_bv(main_bv, bv_new_bv("Mime-Type count in Http flow", "httpMimesCnt", 0, 1, bt_uint_16));
	main_bv = bv_append_bv(main_bv, bv_new_bv("Set Cookie count in Http flow", "httpCookiesCnt", 0, 1, bt_uint_16));
	main_bv = bv_append_bv(main_bv, bv_new_bv("User Agent count in Http flow", "httpAgentsCnt", 0, 1, bt_uint_16));

	main_bv = bv_append_bv(main_bv, bv_new_bv("Number of Images;Videos;Audios;messsges;texts;applications in Http-Flow", "Img_Vid_Aud_Msg_App_Txt", 0, 6, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_16));

	main_bv = bv_append_bv(main_bv, bv_new_bv("Status count in Http flow", "httpStatCnt", 0, 1, bt_uint_16));

#if HTTP_STAT == 1	
	main_bv = bv_append_bv(main_bv, bv_new_bv("Http Status messages in flow", "httpStat", 1, 1, bt_uint_16));
#endif // HTTP_STAT

#if HTTP_MIME == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("MIME-types in Http-Flow", "httpMimes", 1, 1, bt_string));
#endif // HTTP_MIME

#if HTTP_COOKIE == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Cookies in Http-Flow", "httpCookies", 1, 1, bt_string));
#endif // HTTP_COOKIE

#if HTTP_IMAGE == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Images in Http-Flow", "httpImages", 1, 1, bt_string));
#endif // HTTP_IMAGE

#if HTTP_VIDEO == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Videos in Http-Flow", "httpVideos", 1, 1, bt_string));
#endif // HTTP_VIDEO

#if HTTP_AUDIO == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Audios in Http-Flow", "httpAudios", 1, 1, bt_string));
#endif // HTTP_AUDIO

#if HTTP_MSG == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Messages in Http-Flow", "httpMsgs", 1, 1, bt_string));
#endif // HTTP_MSG

#if HTTP_APPL == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Applications in Http-Flow", "httpAppl", 1, 1, bt_string));
#endif // HTTP_APPL

#if HTTP_TEXT == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Texts in Http-Flow", "httpText", 1, 1, bt_string));
#endif // HTTP_TEXT

#if (HTTP_BODY == 1 && HTTP_BDURL == 1)
	main_bv = bv_append_bv(main_bv, bv_new_bv("Body: Refresh, set_Cookie URL in Http-Flow", "httpBdyURL", 1, 1, bt_string));
#endif // (HTTP_BODY == 1 && HTTP_BDURL == 1)

#if HTTP_USRAG == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("User agent info in Http-Flow", "httpUsrAg", 1, 1, bt_string));
#endif // HTTP_USRAG

	return main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex)	{
	http_flow_t *httpFlowP = &(http_flow[flowIndex]);

	memset(httpFlowP, '\0', sizeof(http_flow_t));
	httpFlowP->flags |= HTTP_F_HTTP_HDR;
}

void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	flow_t *flowP = &(flows[flowIndex]); // get flow
	http_flow_t *httpFlowP = &(http_flow[flowIndex]); // get httpFlowP
	http_flow_t *httpFlowPO; // get opposite httpFlowP

	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header; // the ip-header
	register uint32_t i;
	int32_t j;
	uint16_t http_data_len = packet->snapL7Length; // length of http_data (if there is)
	http_mimetype mimetype;

	// check whether opposite exists
	if( flowP->oppositeFlowIndex == HASHTABLE_ENTRY_NOT_FOUND ) httpFlowPO = NULL;
	else httpFlowPO = &http_flow[flowP->oppositeFlowIndex];
	

	if(packet->layer4Type != 6 || http_data_len == 0) return;
 
	char *data_ptr = (char*)packet->layer7Header; // goto data 
	char* line_eptr = http_get_linesize(data_ptr, http_data_len); // size of current line (if 0 header is completely read)
	httpFlowP->pktcnt++;

	if ( !(httpFlowP->flags & HTTP_F_PLD_S) && line_eptr == NULL) return;
	
	tcpHeader_t *tcpHeader = (tcpHeader_t*)packet->layer4Header; // the tcp-header
	uint32_t tcpSeq = ntohl(tcpHeader->seq); // absolute / relative tcp sequence number
	uint32_t tcpSeqDiff = 0; // 
	uint32_t line_count = 0; // number of header lines
	uint32_t namelen = 0; // file or image name length
	uint16_t line_size = (uint16_t)(line_eptr - data_ptr); // size of current line (if 0 header is completely read)
	uint16_t linesz;
	uint16_t http_data_len_chkd;
	char *http_header_data, *name_p="A", *dp, *dp1, h[5];
	char imfilename[HTTP_MXIMNM_LEN];
	

	// HTTP-HEADER-Parsing
	while( line_size > 0 &&	httpFlowP->flags & HTTP_F_HTTP_HDR ) {

		// firstline: identify http-flow.
		if ( line_count == 0 ) {
			if (line_eptr == NULL) return; 
			*line_eptr = 0x00; 
			http_header_data = strstr(data_ptr, HTTP_IDENTIFICATION);  // HTTP FOUND!
			*line_eptr = 0x0d;
			if ( http_header_data && http_header_data - data_ptr < http_data_len ) { // HTTP FOUND!
				http_method method;
				httpFlowP->flags &= ~HTTP_F_S;
				httpFlowP->flags |= (HTTP_F_DETECT | HTTP_F_HTTP_HDR);
#if HTTP_STAT == 1	
				if (http_header_data[8] == ' ' && httpFlowP->stat_c < HTTP_DATA_C_MAX) {
					memcpy(h, http_header_data + 9, 3);
					h[3]=0x00;
					httpFlowP->stat[httpFlowP->stat_c] = (uint16_t)atoi(h);
#endif // HTTP_STAT == 1
					httpFlowP->stat_c++; // inc
#if HTTP_STAT == 1	
				}
#endif // HTTP_STAT == 1
				// look for http-methods
				if ( http_read_method(data_ptr, &method) > 0 ) {
					httpFlowP->httpMethods |= method; // method seen = true
					
					if (method == GET){ // is get, pare filename
						namelen = strcspn((data_ptr + 4)," "); // look for namelen (skip 4 bytes -> GET + SPACE)

						if (namelen > 0) {

							if ( namelen >= HTTP_MXFILE_LEN ) {
								namelen = HTTP_MXFILE_LEN; 
								httpFlowP->flags |= HTTP_F_FNM_LN;
							}

							// copy getfile
							memcpy(httpFlowP->getFile,(data_ptr + 4), namelen);
							httpFlowP->getFile[namelen] = '\0';

							httpFlowP->flags |= HTTP_F_GET; // finaly set hasget-flag to true
						}
					}
				} else {
	/*				*line_eptr = 0x00; 
					http_header_data = strstr(http_header_data + sizeof(HTTP_IDENTIFICATION), "OK");
					*line_eptr = 0x0d;
					if ( http_header_data ) httpFlowP->flags = 0x00;*/
				}
			} else {
				// no http header present seen
				if (!(httpFlowP->flags & HTTP_F_DETECT)) { // ignore if not a http flow 
					httpFlowP->flags = 0x00;
				}
			}
		} else {
			// is this a http flow?
			if (!(httpFlowP->flags & HTTP_F_DETECT)) { // ignore if not a http flow 
				httpFlowP->flags = 0x00;
			}
		}
		
		// continue parsing http-header
		if ( httpFlowP->flags & HTTP_F_HTTP_HDR ) { 

			//Mime-Type-sniffing
			http_header_data = http_read_header_data(data_ptr, http_data_len, CONTENT_TYPE, sizeof(CONTENT_TYPE)-1);

			if ( http_header_data ) {
				// save mime type

#if HTTP_MIME == 1
				if (httpFlowP->mime_c < HTTP_DATA_C_MAX) {
					linesz = (uint16_t)(line_eptr - http_header_data);
					
					httpFlowP->mime[httpFlowP->mime_c] = malloc(linesz + 1); // malloc space for mime-type
					memcpy(httpFlowP->mime[httpFlowP->mime_c], http_header_data, linesz); // copy mime type..
					httpFlowP->mime[httpFlowP->mime_c][linesz] = 0x00;
				} else {
					httpFlowP->flags |= HTTP_F_OVRFLW;
				}
#endif // HTTP_MIME
				httpFlowP->mime_c++; // inc

				// do mimetype based actions
				if ( http_read_mimetype(http_header_data, &mimetype) > 0 ) {
					httpFlowP->mimeTypes |= mimetype; // add to seen mimetypes
					i = flowP->status & 0x0001;
					
					switch (mimetype) {
						case image:
#if (HTTP_IMAGE == 1 || HTTP_SAVE_IMAGE == 1)
							name_p = httpFlowP->image[httpFlowP->images_c];
							if(httpFlowP->images_c < HTTP_DATA_C_MAX) {
								// get image name
								if( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // saw a get request
									j = strlen(httpFlowPO->getFile);
									namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

									name_p = malloc(namelen + 1);
									snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, i,flowP->findex, httpFlowP->pktcnt, httpFlowP->images_c);

									// replace all '/' with '_'
									for( i = 0; i < j; i++ ) if(name_p[i] == '/') name_p[i] = '_';
								} else {
									name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
									snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, i,flowP->findex, httpFlowP->pktcnt, httpFlowP->images_c);
								}
								// save image
								httpFlowP->image[httpFlowP->images_c] = name_p;
#if HTTP_SAVE_IMAGE == 1
								remove(name_p);
								httpFlowP->flags |= HTTP_F_PLD_IMG_S; // start sniffing content..
#endif	// HTTP_SAVE_IMAGE == 1

							} else httpFlowP->flags |= HTTP_F_OVRFLW;
#endif	// (HTTP_IMAGE == 1 || HTTP_SAVE_IMAGE == 1)
							httpFlowP->images_c++;
						break;

						case message:
#if (HTTP_MSG == 1 || HTTP_SAVE_MSG == 1)
							name_p = httpFlowP->msg[httpFlowP->msg_c];
							if(httpFlowP->images_c < HTTP_DATA_C_MAX) {
								// get image name
								if( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // saw a get request
									j = strlen(httpFlowPO->getFile);
									namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

									name_p = malloc(namelen + 1);
									snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->msg_c);
									// replace all '/' with '_'
									for( i = 0; i < namelen; i++ ) if (name_p[i] == '/') name_p[i] = '_';
								} else {
									name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
									snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->msg_c);
								}
								// save message 
								httpFlowP->msg[httpFlowP->msg_c] = name_p;
#if HTTP_SAVE_MSG == 1
								remove(name_p);
								httpFlowP->flags |= HTTP_F_PLD_MSG_S; // start sniffing content..
#endif	// HTTP_SAVE_MSG == 1

							} else httpFlowP->flags |= HTTP_F_OVRFLW;
#endif	// (HTTP_MSG == 1 || HTTP_SAVE_MSG == 1)
							httpFlowP->msg_c++;

						break;
							
						case text:
#if (HTTP_TEXT == 1 || HTTP_SAVE_TEXT == 1)
							name_p = httpFlowP->text[httpFlowP->text_c];
							if(httpFlowP->text_c < HTTP_DATA_C_MAX) {
								// get image name
								if( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // saw a get request
									j = strlen(httpFlowPO->getFile);
									namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

									name_p = malloc(namelen + 1);
									snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->text_c);
									// replace all '/' with '_'
									for( i = 0; i < namelen; i++ ) if (name_p[i] == '/') name_p[i] = '_';
								} else {
									name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
									snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->text_c);
								}
								// save text
								httpFlowP->text[httpFlowP->text_c] = name_p;
#if HTTP_SAVE_TEXT == 1
								remove(name_p);
								httpFlowP->flags |= HTTP_F_PLD_TXT_S; // start sniffing content..
#endif	// HTTP_SAVE_TEXT == 1

							} else httpFlowP->flags |= HTTP_F_OVRFLW;
#endif	// (HTTP_TEXT == 1 || HTTP_SAVE_TEXT == 1)
							httpFlowP->text_c++;
						break;
							
						case video:
#if (HTTP_VIDEO == 1 || HTTP_SAVE_VIDEO == 1)
							name_p = httpFlowP->video[httpFlowP->video_c];
							if(httpFlowP->video_c < HTTP_DATA_C_MAX) {
								// get image name
								if( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // saw a get request
									j = strlen(httpFlowPO->getFile);
									namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

									name_p = malloc(namelen + 1);
									snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->video_c);
									// replace all '/' with '_'
									for( i = 0; i < namelen; i++ ) if (name_p[i] == '/') name_p[i] = '_';
								} else {
									name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
									snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->video_c);
								}
								// save video 
								httpFlowP->video[httpFlowP->video_c] = name_p;
#if HTTP_SAVE_VIDEO == 1
								remove(name_p);
								httpFlowP->flags |= HTTP_F_PLD_VID_S; // start sniffing content..
#endif	// HTTP_SAVE_VIDEO == 1

							} else httpFlowP->flags |= HTTP_F_OVRFLW;
#endif	// (HTTP_VIDEO == 1 || HTTP_SAVE_VIDEO == 1)
							httpFlowP->video_c++;
						break;
						
						case audio:
#if (HTTP_AUDIO == 1 || HTTP_SAVE_AUDIO == 1)
							name_p = httpFlowP->audio[httpFlowP->audio_c];
							if (httpFlowP->audio_c < HTTP_DATA_C_MAX) {
								// get image name
								if( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // saw a get request
									j = strlen(httpFlowPO->getFile);
									namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

									name_p = malloc(namelen + 1);
									snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->audio_c);
									// replace all '/' with '_'
									for( i = 0; i < namelen; i++ ) if (name_p[i] == '/') name_p[i] = '_';
								} else {
									name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
									snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->audio_c);
								}
								// save audio
								httpFlowP->audio[httpFlowP->audio_c] = name_p;
#if HTTP_SAVE_AUDIO == 1
								remove(name_p);
								httpFlowP->flags |= HTTP_F_PLD_AUD_S; // start sniffing content..
#endif	// HTTP_SAVE_AUDIO == 1

							} else httpFlowP->flags |= HTTP_F_OVRFLW;
#endif	// (HTTP_AUDIO == 1 || HTTP_SAVE_AUDIO == 1)
							httpFlowP->audio_c++;
						break;

						case application:
#if (HTTP_APPL == 1 || HTTP_SAVE_APPL == 1)
							name_p = httpFlowP->appl[httpFlowP->appl_c];
							if (httpFlowP->appl_c < HTTP_DATA_C_MAX) {
								// get image name
								if( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // saw a get request
									j = strlen(httpFlowPO->getFile);
									namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

									name_p = malloc(namelen + 1);
									snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->appl_c);
									// replace all '/' with '_'
									for( i = 0; i < namelen; i++ ) if (name_p[i] == '/') name_p[i] = '_';
								} else {
									name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
									snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, i, flowP->findex, httpFlowP->pktcnt, httpFlowP->appl_c);
								}
								// save audio
								httpFlowP->appl[httpFlowP->appl_c] = name_p;
#if HTTP_SAVE_APPL == 1
								remove(name_p);
								httpFlowP->flags |= HTTP_F_PLD_APP_S; // start sniffing content..
#endif	// HTTP_SAVE_APPL == 1

							} else httpFlowP->flags |= HTTP_F_OVRFLW;
#endif	// (HTTP_APPL == 1 || HTTP_SAVE_APPL == 1)
							httpFlowP->appl_c++;
						break;
					}
				}
			}

			//Content-Length extraction
			http_header_data = http_read_header_data(data_ptr, http_data_len, CONTENT_LENGTH, sizeof(CONTENT_LENGTH)-1);
			if ( http_header_data ) {
				*line_eptr = 0x00; 
				httpFlowP->contentLength = atoi(http_header_data);
				*line_eptr = 0x0d;
			}
			
			//Transfer-Encoding extraction
			http_header_data = http_read_header_data(data_ptr, http_data_len, TRANS_ENCODING, sizeof(TRANS_ENCODING)-1);
			if ( http_header_data ) {
				*line_eptr = 0x00; 
				 if ( strstr(http_header_data, "chunked") ) httpFlowP->flags |= HTTP_F_CHKD;
				*line_eptr = 0x0d;
			}
		}

		// Cookie sniffing
		http_header_data = http_read_header_data(data_ptr, http_data_len, SET_COOKIE, sizeof(SET_COOKIE)-1);
		if( http_header_data ) {
#if HTTP_COOKIE == 1
			// save cookie 
			if( httpFlowP->cookie_c < HTTP_DATA_C_MAX ) { // cookie limit?
				linesz = (uint16_t)(line_eptr - http_header_data);
				httpFlowP->cookie[httpFlowP->cookie_c] = malloc(linesz + 1);
				memcpy(httpFlowP->cookie[httpFlowP->cookie_c], http_header_data, linesz); 
				httpFlowP->cookie[httpFlowP->cookie_c][linesz] = 0x00;
#endif // HTTP_COOKIE = 1
				httpFlowP->cookie_c++; // inc
#if HTTP_COOKIE == 1
			} else {
				httpFlowP->flags |= HTTP_F_OVRFLW;
			}
#endif // HTTP_COOKIE = 1
		}

		// User Agent
		http_header_data = http_read_header_data(data_ptr, http_data_len, USER_AGENT, sizeof(USER_AGENT)-1);
		if( http_header_data ) {
#if HTTP_USRAG == 1
			// save cookie 
			if( httpFlowP->userAg_c < HTTP_DATA_C_MAX ) { // ?
				linesz = (uint16_t)(line_eptr - http_header_data);
				if ( linesz >= HTTP_MXAG_LEN ) {
					linesz = HTTP_MXAG_LEN; 
			//		httpFlowP->flags |= HTTP_F_FNM_LN;
				}
				httpFlowP->usrAg[httpFlowP->userAg_c] = malloc(linesz + 1);
				memcpy(httpFlowP->usrAg[httpFlowP->userAg_c], http_header_data, linesz); 
				httpFlowP->usrAg[httpFlowP->userAg_c][linesz] = 0x00;
#endif // HTTP_USRAG = 1
				httpFlowP->userAg_c++; // inc
#if HTTP_USRAG == 1
			} else {
				httpFlowP->flags |= HTTP_F_OVRFLW;
			}
#endif // HTTP_USRAG = 1
		}

		// cleanup after line..
		line_count++;

		// go to next line 
		if( httpFlowP->flags & HTTP_F_HTTP_HDR ){
			data_ptr += (line_size + 2); // +2 skip line ending \r\n
			http_data_len -= (line_size + 2);

			// Get new linesize
			line_eptr = http_get_linesize(data_ptr, http_data_len); // size of current line (if 0 header is completely read)
			if ( line_eptr ) line_size = (uint16_t)(line_eptr - data_ptr); // size of current line (if 0 header is completely read)
			else line_size = 0;
		}

	}

	// skip \r\n at end of http-header
	if( httpFlowP->flags & HTTP_F_HTTP_HDR ) { // was in HTTP-Header
		if ( http_data_len == 0 || line_eptr == NULL ) return;
		if ( http_data_len >= 2 ) {
			data_ptr += 2; // skip \r\n 
			http_data_len -= 2; // length of all body payload to come
		} else httpFlowP->flags |= HTTP_F_PRS_ERR; // Parse Error

#if HTTP_SAVE == 1
		if (httpFlowP->fd) { 
			fclose(httpFlowP->fd);
			httpFlowP->fd = NULL;
			httpSniffer_fd_cnt--;
		}
		if ( httpFlowP->flags & HTTP_F_CHKD ) {
			line_eptr = http_get_linesize(data_ptr, http_data_len); // size of current line (if 0 header is completely read)
			if ( line_eptr ) line_size = (uint16_t)(line_eptr - data_ptr); // size of current line (if 0 header is completely read)
			else line_size = 0;
			if ( line_eptr ) {
				*line_eptr = 0x00; 
				sscanf(data_ptr, "%x", &(httpFlowP->contentLength));
				*line_eptr = 0x0d;
				data_ptr += (line_size + 2); // +2 skip line ending \r\n
				http_data_len -= (line_size + 2); // +2 skip line ending \r\n
		// printf("wurstB: %d %x %X - %d %d %d\n", flowP->findex, httpFlowP->fd, httpFlowP->flags, http_data_len, httpFlowP->contentLength, httpFlowP->sniffedContent);
		// fflush(stdout);
			}
		}

		// printf("wurstA: %d %d %x %X - %d %d %d\n", flowP->findex, httpFlowP->pktcnt, httpFlowP->fd, httpFlowP->flags, http_data_len, httpFlowP->contentLength, httpFlowP->sniffedContent);
		// fflush(stdout);
		httpFlowP->sniffedContent = 0;

	//	if ( httpFlowP->flags & HTTP_F_PLD_S && http_data_len <= (httpFlowP->contentLength - httpFlowP->sniffedContent)) {
		if ( httpFlowP->flags & HTTP_F_PLD_S ) {
			switch (httpFlowP->flags & HTTP_F_PLD_S) {
#if HTTP_SAVE_IMAGE == 1
				case HTTP_F_PLD_IMG_S:
					name_p = httpFlowP->image[(httpFlowP->images_c-1)]; //
					memcpy(imfilename, HTTP_IMAGE_PATH, sizeof(HTTP_IMAGE_PATH));
					memcpy( imfilename + sizeof(HTTP_IMAGE_PATH) - 1, name_p, strlen(name_p)+1 ); // prepend path to filename
					 
				break;
#endif // HTTP_SAVE_IMAGE == 1
#if HTTP_SAVE_VIDEO == 1
				case HTTP_F_PLD_VID_S:
					name_p = httpFlowP->video[(httpFlowP->video_c-1)]; // 
					memcpy(imfilename, HTTP_VIDEO_PATH, sizeof(HTTP_VIDEO_PATH));
					memcpy( imfilename + sizeof(HTTP_VIDEO_PATH) - 1, name_p, strlen(name_p)+1 ); // prepend path to filename
				break;
#endif // HTTP_SAVE_VIDEO == 1
#if HTTP_SAVE_AUDIO == 1
				case HTTP_F_PLD_AUD_S:
					name_p = httpFlowP->audio[(httpFlowP->audio_c-1)]; // 
					memcpy(imfilename, HTTP_AUDIO_PATH, sizeof(HTTP_AUDIO_PATH));
					memcpy( imfilename + sizeof(HTTP_AUDIO_PATH) - 1, name_p, strlen(name_p)+1 ); // prepend path to filename
				break;
#endif // HTTP_SAVE_AUDIO == 1
#if HTTP_SAVE_MSG == 1
				case HTTP_F_PLD_MSG_S:
					name_p = httpFlowP->msg[(httpFlowP->msg_c-1)]; // 
					memcpy(imfilename, HTTP_MSG_PATH, sizeof(HTTP_MSG_PATH));
					memcpy( imfilename + sizeof(HTTP_MSG_PATH) - 1, name_p, strlen(name_p)+1 ); // prepend path to filename
				break;
#endif // HTTP_SAVE_MSG == 1
#if HTTP_SAVE_TEXT == 1
				case HTTP_F_PLD_TXT_S:
					name_p = httpFlowP->text[(httpFlowP->text_c-1)]; // 
					memcpy(imfilename, HTTP_TEXT_PATH, sizeof(HTTP_TEXT_PATH));
					memcpy( imfilename + sizeof(HTTP_TEXT_PATH) - 1, name_p, strlen(name_p)+1 ); // prepend path to filename
				break;
#endif // HTTP_SAVE_TEXT == 1
#if HTTP_SAVE_APPL == 1
				case HTTP_F_PLD_APP_S:
					name_p = httpFlowP->appl[(httpFlowP->appl_c-1)]; // 
					memcpy(imfilename, HTTP_APPL_PATH, sizeof(HTTP_APPL_PATH));
					memcpy( imfilename + sizeof(HTTP_APPL_PATH) - 1, name_p, strlen(name_p)+1 ); // prepend path to filename
				break;
#endif // HTTP_SAVE_APPL == 1
				default:
					httpFlowP->flags &= ~HTTP_F_PLD_S;
					return;
			}

			imfilename[strlen(imfilename)] = '\0';
			httpFlowP->fd = fopen(imfilename, "w+b"); 
			if ( httpFlowP->fd == NULL ) {
				printf("ERROR http_Sniffer: fopen %s %s \n", imfilename, strerror(errno));
				terminate();
			}
			httpFlowP->flags |= HTTP_F_SEQ_INIT;
			if ( http_data_len ) httpFlowP->flags |= HTTP_F_SHFT;
			httpFlowP->flags &= ~HTTP_F_HTTP_HDR;

			if ( ++httpSniffer_fd_cnt > httpSniffer_fd_cnt_max ) httpSniffer_fd_cnt_max = httpSniffer_fd_cnt;
		}
#endif // HTTP_SAVE == 1 
		
	}

#if (HTTP_BODY == 1 && HTTP_TEXT == 1)
		if (httpFlowP->mimeTypes & HTTP_C_TEXT) { // text
			if ( (dp = strstr(data_ptr, "\"Refresh\"")) ) { 
				for (i = 2; i < http_data_len; i++) { 
					if (dp[i] == '>') { 
						*h = dp[i-3];
						dp[i-3] = 0x00;
						if ( (dp1 = strstr(dp+9, "URL")) ) {
							httpFlowP->httpBdyFlgs |= HTTP_REFRESH;
#if HTTP_BDURL == 1
							j = strlen(dp1);
							name_p = malloc(j);
							memcpy(name_p, dp1 + 4, j-4);
							httpFlowP->refURL[httpFlowP->refURL_c++] = name_p;
#endif // HTTP_BDURL == 1
						}
						dp[i-3] = *h;
						break;
					}
				}
			} else if ( (dp = strstr(data_ptr, "\"Set-Cookie\"")) ) {
				for (i = 2; i < http_data_len; i++) { 
					if (dp[i] == '>') { 
						*h = dp[i-3];
						dp[i-3] = 0x00;
						if ( (dp1 = strstr(dp+12, "URL")) ) {
							httpFlowP->httpBdyFlgs |= HTTP_STCOOKIE;
#if HTTP_BDURL == 1
							j = strlen(dp1);
							name_p = malloc(j);
							memcpy(name_p, dp1 + 4, j-4);
							httpFlowP->refURL[httpFlowP->refURL_c++] = name_p;
#endif // HTTP_BDURL == 1
						}
						dp[i-3] = *h;
						break;
					}
				}
			}
		}
#endif // (HTTP_BODY == 1 && HTTP_TEXT == 1)


#if HTTP_SAVE == 1
	// if Http-Flow do analyze data 
	if( (httpFlowP->flags & HTTP_F_PLD_S) && http_data_len > 0 ) {
		
		if ( http_data_len > (httpFlowP->contentLength - httpFlowP->sniffedContent) ) http_data_len_chkd = httpFlowP->contentLength - httpFlowP->sniffedContent;
		else http_data_len_chkd = http_data_len;

		if ( httpFlowP->flags & HTTP_F_SEQ_INIT ) {
			httpFlowP->tcpSeqInit = tcpSeq;
			httpFlowP->hdr_len = packet->snapL7Length - http_data_len_chkd;
			httpFlowP->flags &= ~HTTP_F_SEQ_INIT;
			if ( httpFlowP->flags & HTTP_F_CHKD ) {
				// printf("wurstC: %d %x %X - %d %d %d %d %s\n", flowP->findex, httpFlowP->fd, httpFlowP->flags, http_data_len, http_data_len_chkd, httpFlowP->contentLength, httpFlowP->sniffedContent, name_p );
		 		//fflush(stdout);
			}
		} else if ( httpFlowP->flags & HTTP_F_CHKD ) {
			line_eptr = http_get_linesize(data_ptr, http_data_len_chkd); // size of current line (if 0 header is completely read)
			if ( line_eptr ) {
				line_size = (uint16_t)(line_eptr - data_ptr); // size of current line (if 0 header is completely read)
				*line_eptr = 0x00; 
				sscanf(data_ptr, "%x", &(httpFlowP->contentLength));
				*line_eptr = 0x0d;
				data_ptr += (line_size + 2); // +2 skip line ending \r\n
				http_data_len_chkd -= (line_size + 2); // +2 skip line ending \r\n
		 //printf("wurstD: %d %d %x %X - %d %d %d %d %s\n", flowP->findex, httpFlowP->pktcnt, httpFlowP->fd, httpFlowP->flags, http_data_len, http_data_len_chkd, httpFlowP->contentLength, httpFlowP->sniffedContent, httpFlowP->image[(httpFlowP->images_c-1)]);
		 //fflush(stdout);
			}
		}

		tcpSeqDiff = tcpSeq - httpFlowP->tcpSeqInit; 
			
		j = tcpSeqDiff - httpFlowP->sniffedContent; 
		if ( tcpSeqDiff && httpFlowP->flags & HTTP_F_SHFT ) j -= httpFlowP->hdr_len; // remove header part
				
		if ( j  && !(httpFlowP->flags & HTTP_F_CHKD)) { // appropriate transmission j == 0				
			httpFlowP->flags |= HTTP_F_SQ_NM;
			if ( j > 0 ) { // time jump forward in flow
				if (httpFlowP->flags & HTTP_F_PLD_VID_S) fwrite(data_ptr, 1, j, httpFlowP->fd); // fill the gap 
				httpFlowP->sniffedContent += j; 						// and advance sniffed len
			} else if (tcpSeq < httpFlowP->seq) { // time jump backwards in flow
				fseek(httpFlowP->fd, tcpSeqDiff - httpFlowP->hdr_len, SEEK_SET);
				fwrite(data_ptr, 1, http_data_len_chkd, httpFlowP->fd);
				if ( !(httpFlowP->flags & HTTP_F_CHKD) ) {
					fseek(httpFlowP->fd, 0, SEEK_END);
					return;
				}
			}
		}

		// write normal sequence to file
		if (j >= 0) {
			if ( http_data_len_chkd != fwrite(data_ptr, 1, http_data_len_chkd, httpFlowP->fd) ) { 
				printf("ERROR http_Sniffer: fwrite %s %s \n", imfilename, strerror(errno));
				fclose(httpFlowP->fd);
				terminate();
			}
			
			httpFlowP->sniffedContent += http_data_len_chkd; // more sniffed content
		}	
		httpFlowP->seq = tcpSeq;	
		 
		
		if (httpFlowP->contentLength <= httpFlowP->sniffedContent) {
			if ( httpFlowP->flags & HTTP_F_CHKD ) {
				do {
					i = http_data_len_chkd + 2;
					data_ptr += i;
					line_eptr = http_get_linesize(data_ptr, http_data_len - i); // size of current line (if 0 header is completely read)
					if (!line_eptr) {
						httpFlowP->contentLength = 0;
						return;
					}
					line_size = (uint16_t)(line_eptr - data_ptr); //
					 
					*line_eptr = 0x00; 
					sscanf( data_ptr, "%x", &(httpFlowP->contentLength) );
					*line_eptr = 0x0d;
					if (httpFlowP->contentLength == 0) break;
					i = line_size + 2;
					data_ptr += i; // +2 skip line ending \r\n
					http_data_len -= (http_data_len_chkd + i + 2); // +2 skip line ending \r\n
					fwrite( data_ptr, 1, http_data_len, httpFlowP->fd );
					httpFlowP->sniffedContent += http_data_len_chkd; // more sniffed content
		 //printf("wurstE: %d %d %x %X - %d %d %d %d %s\n", flowP->findex, httpFlowP->pktcnt, httpFlowP->fd, httpFlowP->flags, http_data_len, http_data_len_chkd, httpFlowP->contentLength, httpFlowP->sniffedContent, httpFlowP->image[(httpFlowP->images_c-1)]);
		 //fflush(stdout);
				} while ( httpFlowP->contentLength < httpFlowP->sniffedContent ); 
	
				if ( httpFlowP->contentLength ) {
					httpFlowP->flags &= ~HTTP_F_SHFT;	
					httpFlowP->sniffedContent = http_data_len;
					httpFlowP->tcpSeqInit = tcpSeq;
					return;
				}
			}
 
			if (httpFlowP->fd) { 
				fclose(httpFlowP->fd);
				httpFlowP->fd = NULL;
				httpSniffer_fd_cnt--;
			}
			httpFlowP->flags &= ~HTTP_F_S; // content processing finished
			httpFlowP->flags |= HTTP_F_HTTP_HDR;
			httpFlowP->sniffedContent = 0;
			httpFlowP->contentLength = 0;
			httpFlowP->tcpSeqInit = 0;
			httpFlowP->hdr_len = 0;
			http_data_len = 0;
		}

	}

#endif // HTTP_SAVE == 1 

}

static inline char* http_get_linesize(uint8_t *data, int16_t http_data_len){
	unsigned int i;

	if(http_data_len > 1){ // data length should at least be 1
		for (i = 0; i < http_data_len -1; i++) { 
			if ( data[i] == '\r' ) {
				if (data[i+1] == '\n') return &data[i];
			}
		}
	}

	return NULL; // no \r\n found
}

static inline int http_strncmp(char *s1,char *s2,size_t maxcmp){
	uint32_t i;
	uint32_t s1_len = strlen(s1);
	uint32_t s2_len = strlen(s2);
	uint32_t cmpwidth = (s1_len < s2_len) ? s1_len : s2_len; // get smaller string

	for( i = 0; i < cmpwidth && i < maxcmp; i++) {
		if(s1[i] != s2[i]) return -1; // strings not equal
	}

	return 0;
}

static inline int http_read_method(char *data, http_method *retmethod ) {
	uint32_t method_size = strcspn(data," "); // search first space

	if (strlen(data) < method_size) return -1;

	if ( http_strncmp(data, HTTP_IDENTIFICATION, method_size) == 0 ) (*retmethod)=RESPONSE;
	else if ( http_strncmp(data, "GET", method_size) == 0 ) (*retmethod)=GET;
	else if ( http_strncmp(data, "POST", method_size) == 0 ) (*retmethod)=POST;
	else if ( http_strncmp(data, "OPTIONS", method_size) == 0 ) (*retmethod)=OPTIONS;
	else if ( http_strncmp(data, "HEAD", method_size) == 0 ) (*retmethod)=HEAD;
	else if ( http_strncmp(data, "PUT", method_size) == 0 ) (*retmethod)=PUT;
	else if ( http_strncmp(data, "DELETE", method_size) == 0 ) (*retmethod)=DELETE;
	else if ( http_strncmp(data, "TRACE", method_size) == 0 ) (*retmethod)=TRACE;
	else if ( http_strncmp(data, "CONNECT", method_size) == 0 ) (*retmethod)=CONNECT;
	else return -1;
	

	return 1;
}

static inline int http_read_mimetype(char *data,http_mimetype *retmime){
	size_t mime_size = strcspn(data,"/");
	if(strlen(data) < mime_size) return -1;

	if (http_strncmp(data,"application",mime_size)==0 ) (*retmime)=application;
	else if (http_strncmp(data,"audio",mime_size)==0) (*retmime)=audio;
	else if (http_strncmp(data,"image",mime_size)==0) (*retmime)=image;
	else if (http_strncmp(data,"message",mime_size)==0) (*retmime)=message;
	else if (http_strncmp(data,"model",mime_size)==0) (*retmime)=model;
	else if (http_strncmp(data,"multipart",mime_size)==0) (*retmime)=multipart;
	else if (http_strncmp(data,"text",mime_size)==0) (*retmime)=text;
	else if (http_strncmp(data,"video",mime_size)==0) (*retmime)=video;
	else if (http_strncmp(data,"vnd",mime_size)==0) (*retmime)=vnd;
	else if (http_strncmp(data,"x",mime_size)==0) (*retmime)=x;
	else if (http_strncmp(data,"x-pkcs",mime_size)==0) (*retmime)=xpkcs;
	else return -1;

	return 1;
}

static inline char* http_read_header_data(char* data, uint16_t data_len, const char *header, uint16_t header_len){

	if ( memcmp((void*)data, (void*)header, header_len) !=0 ) return 0; // data not found..
	
	while(data[header_len] == ' ' && header_len < data_len) header_len++;

	return &data[header_len];
}

/*
static inline uint16_t http_read_mimetype(char *data){
	size_t mime_size = strcspn(data,"/");
	if(strlen(data) < mime_size) return 0;

	if (http_strncmp(data,"application",mime_size)==0 ) return HTTP_C_APPL;
	else if (http_strncmp(data,"audio",mime_size)==0) return HTTP_C_AUDIO;
	else if (http_strncmp(data,"image",mime_size)==0) return HTTP_C_IMAGE;
	else if (http_strncmp(data,"message",mime_size)==0) return HTTP_C_MSG;
	else if (http_strncmp(data,"model",mime_size)==0) return HTTP_C_MODEL;
	else if (http_strncmp(data,"multipart",mime_size)==0) return HTTP_C_MLTPRT;
	else if (http_strncmp(data,"text",mime_size)==0) return HTTP_C_TEXT;
	else if (http_strncmp(data,"video",mime_size)==0) return HTTP_C_VIDEO;
	else if (http_strncmp(data,"vnd",mime_size)==0) return HTTP_C_VND;
	else if (http_strncmp(data,"x",mime_size)==0) return HTTP_C_X;
	else if (http_strncmp(data,"x-pkcs",mime_size)==0) return HTT_C_XPKCS;
	return 0;
}*/


/*
static inline void http_fname( httpFlow_t* httpFlowP, httpFlow_t* httpFlowPO, flow_t* flowP, char* name_pi, uint16_t cnt ) {
	unsigned int namelen;

	if (cnt < HTTP_DATA_C_MAX) {
		// get image name
		if ( httpFlowPO && (httpFlowPO->flags & HTTP_F_GET) ) { // get request
			j = strlen(httpFlowPO->getFile);
			namelen = j + HTTP_CNT_LEN + HTTP_findex_LEN;

			name_p = malloc(namelen + 1);
			snprintf(name_p, namelen, "%s_%d_%d_%d_%d", httpFlowPO->getFile, (flowP->status & 0x0001), flowP->findex, httpFlowP->pktcnt, httpFlowP->audio_c);

			// replace all '/' with '_'
			for ( i = 0; i < j; i++ ) if (name_p[i] == '/') name_p[i] = '_';
		} else {
			name_p = malloc(HTTP_NONAME_IMAGE_LEN + 1);
			snprintf(name_p, HTTP_NONAME_IMAGE_LEN, "%s_%d_%d_%d_%d", HTTP_NONAME_IMAGE, (flowP->status & 0x0001), flowP->findex, httpFlowP->pktcnt, httpFlowP->audio_c);
		}
	} else httpFlowP->flags |= HTTP_F_OVRFLW;
}
*/


void onFlowTerminate(unsigned long flowIndex) {
	uint32_t i, j;
	char *cookie;
	flow_t *flowP = &(flows[flowIndex]); // get flow
	http_flow_t *httpFlowP = &(http_flow[flowIndex]); // get flow

#if HTTP_SAVE == 1 
	if (httpFlowP->fd) { 
		fseek(httpFlowP->fd, 0, SEEK_END);	
		fclose(httpFlowP->fd);
		httpFlowP->fd = NULL;
		httpSniffer_fd_cnt--;
	}
#endif // HTTP_SAVE == 1 

	outputBuffer_append(main_output_buffer, (char*) &(httpFlowP->flags), sizeof(uint16_t)); 	// Flags
	outputBuffer_append(main_output_buffer, (char*) &(httpFlowP->httpMethods), sizeof(uint8_t)); 	// Methods
	outputBuffer_append(main_output_buffer, (char*) &(httpFlowP->mimeTypes), sizeof(uint16_t)); 	// Mime Types
#if HTTP_BODY == 1
	outputBuffer_append(main_output_buffer, (char*) &(httpFlowP->httpBdyFlgs), sizeof(uint8_t)); 	// Content flags
#endif // HTTP_BODY	
	outputBuffer_append(main_output_buffer, (char*) &(httpFlowP->mime_c), 10*sizeof(uint16_t)); 	// ++ counts

#if HTTP_STAT == 1	
	j = httpFlowP->stat_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)(&httpFlowP->stat[i]), sizeof(uint16_t));
	}
#endif // HTTP_STAT == 1

	// print mime types in flow file
#if HTTP_MIME == 1
	j = httpFlowP->mime_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->mime[i], strlen(httpFlowP->mime[i])+1);
		free(httpFlowP->mime[i]); // and free buffer..
	}
#endif // HTTP_MINE

	// print cookie names in flow file
#if HTTP_COOKIE == 1
	j = httpFlowP->cookie_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t)); // cookie count
	// seen cookies
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->cookie[i] , strlen(httpFlowP->cookie[i])+1);
		free(httpFlowP->cookie[i]);
	}
#endif // HTTP_COOKIE 

	// print image names in flow file
#if HTTP_IMAGE == 1
	j = httpFlowP->images_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  images
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->image[i], strlen(httpFlowP->image[i])+1);
		free(httpFlowP->image[i]); // and free buffer..
	}
#endif // HTTP_IMAGE
	
	// print video names in flow file
#if HTTP_VIDEO == 1
	j = httpFlowP->video_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  videos 
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->video[i], strlen(httpFlowP->video[i])+1);
		free(httpFlowP->video[i]); // and free buffer..
	}
#endif // HTTP_VIDEO

	// print audio names in flow file
#if HTTP_AUDIO == 1
	j = httpFlowP->audio_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  audios
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->audio[i], strlen(httpFlowP->audio[i])+1);
		free(httpFlowP->audio[i]); // and free buffer..
	}
#endif // HTTP_AUDIO

	// print message names in flow file
#if HTTP_MSG == 1
	j = httpFlowP->msg_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  messages
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->msg[i], strlen(httpFlowP->msg[i])+1);
		free(httpFlowP->msg[i]); // and free buffer..
	}
#endif // HTTP_MSG

	// print application names in flow file
#if HTTP_APPL == 1
	j = httpFlowP->appl_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  applications
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->appl[i], strlen(httpFlowP->appl[i])+1);
		free(httpFlowP->appl[i]); // and free buffer..
	}
#endif // HTTP_APPL

	// print text names in flow file
#if HTTP_TEXT == 1
	j = httpFlowP->text_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  texts
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->text[i], strlen(httpFlowP->text[i])+1);
		free(httpFlowP->text[i]); // and free buffer..
	}
#endif // HTTP_TEXT

	// print body set-cookie, refresh URL in flow file
#if ( HTTP_BODY == 1 && HTTP_BDURL == 1 )
	j = httpFlowP->refURL_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  body content
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->refURL[i], strlen(httpFlowP->refURL[i])+1);
		free(httpFlowP->refURL[i]); // and free buffer..
	}
#endif // (HTTP_BODY == 1 && HTTP_BDURL == 1)

	// print user agents in flow file
#if HTTP_USRAG == 1
	j = httpFlowP->userAg_c;
	if ( j > HTTP_DATA_C_MAX) j = HTTP_DATA_C_MAX;
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	//  user agent content
	for (i = 0; i < j; i++){
		outputBuffer_append(main_output_buffer, (char*)httpFlowP->usrAg[i], strlen(httpFlowP->usrAg[i])+1);
		free(httpFlowP->usrAg[i]); // and free buffer..
	}
#endif // HTTP_USRAG

}


void onApplicationTerminate() {
	printf("httpSniffer max fd count: %"PRId32"\n", httpSniffer_fd_cnt_max);
	free(http_flow); // free BIG field
}


