/*
 * httpSniffer
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Mischa Lehmann, RDNIZ, Stefan Burschka, RDNIZ
 */

// Includes       

#ifndef _HTTPSNIFFER_H
#define _HTTPSNIFFER_H 1

/*all Libraries*/
#include <stdbool.h>

// local includes
#include "main.h"


// User defined output switches
#define HTTP_MIME 1 			// 1: print mime type in flow file; 0: print # of mime types only
#define HTTP_STAT 1 			// 1: print status code in flow file; 0: print # of status codes only
#define HTTP_COOKIE 1			// 1: print cookies in flow file; 0: print # of cookies only
#define HTTP_IMAGE 1			// 1: print image name in flow file; 0: print # of images only
#define HTTP_VIDEO 1			// 1: print video name in flow file; 0: print # of videos only
#define HTTP_AUDIO 1			// 1: print audio name in flow file; 0: print # of audios only
#define HTTP_MSG 1			// 1: print pdf name in flow file; 0: print # of pdf only
#define HTTP_APPL 1			// 1: print application name in flow file; 0: print # of applications only
#define HTTP_TEXT 1			// 1: print text name in flow file; 0: print # of texts only
#define HTTP_BODY 1			// 1: body exam, print anomaly bits in flow file; 0: none
#define HTTP_BDURL 1			// 1: print url name in flow file; 0: none
#define HTTP_USRAG 1			// 1: print user agents in flow file; 0: none

#define HTTP_SAVE_IMAGE 0		// 1: Save images in files under HTTP_IMAGE_PATH; 0: Dont save images
#define HTTP_SAVE_VIDEO 0		// 1: Save videos in files under HTTP_VIDEO_PATH; 0: Dont save videos
#define HTTP_SAVE_AUDIO 0 		// 1: Save audios in files under HTTP_TEXT_PATH; 0: Dont save audios
#define HTTP_SAVE_MSG 0			// 1: Save messages in files under HTTP_MSG_PATH; 0: Dont save pdfs
#define HTTP_SAVE_TEXT 0		// 1: Save texts in files under HTTP_TEXT_PATH; 0: Dont save texts
#define HTTP_SAVE_APPL 0		// 1: Save applications in files under HTTP_TEXT_PATH; 0: Dont save applications

#define HTTP_SAVE ( HTTP_SAVE_IMAGE | HTTP_SAVE_VIDEO | HTTP_SAVE_AUDIO | HTTP_SAVE_MSG | HTTP_SAVE_TEXT | HTTP_SAVE_APPL )

#define HTTP_RM_PICDIR 1 		// if HTTP_SAVE == 1 then remove pic dir in order to prevent appending to existing files

// User defined storage boundary conditions
#define HTTP_PATH "/home/stbf104g/"		// root path 
#define HTTP_IMAGE_PATH HTTP_PATH"TranPicture2/"	// Path for pictures
#define HTTP_VIDEO_PATH HTTP_PATH"TranVideo2/"		// Path for videos
#define HTTP_AUDIO_PATH HTTP_PATH"TranAudio2/"		// Path for audios
#define HTTP_MSG_PATH 	HTTP_PATH"TranMSG2/"		// Path for messages 
#define HTTP_TEXT_PATH 	HTTP_PATH"TranText2/"		// Path for texts
#define HTTP_APPL_PATH 	HTTP_PATH"TranAppl2/"		// Path for applications

#define HTTP_NONAME_IMAGE "nudel" 	// name of pictures without name

#define HTTP_DATA_C_MAX	40		// maximum amount of information to store (per flow)
#define HTTP_CNT_LEN	8 		// max # of cnt digits attached to file name. String length depends on HTTP_DATA_C_MAX !
#define HTTP_MXFILE_LEN	15		// maximum storage name length
#define HTTP_MXAG_LEN	49		// maximum user agent name length
#define HTTP_findex_LEN	10		// length of findex in decimal format

// Calculate Name lengths
#define HTTP_NONAME_IMAGE_LEN sizeof(HTTP_NONAME_IMAGE) + HTTP_CNT_LEN + HTTP_findex_LEN	// Standard name of pictures without name: name_findex_pkt_num
#define HTTP_MXIMNM_LEN	sizeof(HTTP_IMAGE_PATH) + HTTP_NONAME_IMAGE_LEN + HTTP_MXFILE_LEN + 1	// maximum Image name length

// http_status_flags
#define HTTP_F_OVRFLW	 0x0001		// More Data in Flow than HTTP_DATA_C_MAX can save
#define HTTP_F_FNM_LN 	 0x0002		// Filename larger than HTTP_MXIMNM_LEN or Linux def 
#define HTTP_F_SQ_NM 	 0x0004		// Sequence number violation
#define HTTP_F_PRS_ERR 	 0x0008		// Parse Error
#define HTTP_F_CHKD	 0x0010		// Internal State: Chunked transfer
#define HTTP_F_DETECT	 0x0020		// Internal State: HTTP Flow detected
#define HTTP_F_HTTP_HDR	 0x0040		// Internal State: http hdr in process
#define HTTP_F_GET	 0x0080		// Internal State: Pending GET-REQUEST 
#define HTTP_F_SEQ_INIT	 0x0100		// Internal State: sequence number init 
#define HTTP_F_SHFT	 0x0200		// Internal State: header shift
#define HTTP_F_PLD_IMG_S 0x0400		// Internal State: Image payload sniffing
#define HTTP_F_PLD_VID_S 0x0800		// Internal State: video payload sniffing
#define HTTP_F_PLD_AUD_S 0x1000		// Internal State: audio payload sniffing
#define HTTP_F_PLD_MSG_S 0x2000		// Internal State: message payload sniffing
#define HTTP_F_PLD_TXT_S 0x4000		// Internal State: text payload sniffing
#define HTTP_F_PLD_APP_S 0x8000		// Internal State: application payload sniffing

#define HTTP_F_PLD_S 	 0xfc00		// Internal states mask: Payload sniffing 
#define HTTP_F_S 	 0xfff0		// Internal states mask: Internal State 

// http content flags
#define HTTP_STCOOKIE	0x01		//
#define HTTP_REFRESH	0x02		//

// content Types
#define	HTTP_C_APPL	0x0001		//
#define	HTTP_C_AUDIO	0x0002		//
#define	HTTP_C_IMAGE	0x0004		//
#define	HTTP_C_MSG	0x0008		//
#define	HTTP_C_MODEL	0x0010		//
#define	HTTP_C_MLTPRT	0x0020		//
#define	HTTP_C_TEXT	0x0040		//
#define	HTTP_C_VIDEO	0x0080		//
#define	HTTP_C_VND	0x0100		//
#define	HTTP_C_X	0x0200		//
#define	HTTP_C_XPKCS	0x0400		//

//

#define HTTP_IDENTIFICATION "HTTP/1." 	// detect all HTTP 1.* - protocols

#define HTTP_HEADER_LINEEND {'\r','\n'}
#define HTTP_HEADER_CRLF "\r\n"
#define HTTP_COOKIE_VALSEPARATOR_C 2
#define HTTP_COOKIE_VALSEPARATOR ";"
#define HTTP_COOKIE_SEPARATOR '='

// definition of content types
#define CONTENT_TYPE "Content-Type:"
#define CONTENT_LENGTH "Content-Length:"
#define CONTENT_ENCODING "Content-Encoding:"
#define TRANS_ENCODING "Transfer-Encoding:"
#define SET_COOKIE "Set-Cookie:"
#define USER_AGENT "User-Agent:"


typedef struct http_flow_s {
#if HTTP_SAVE == 1
	FILE *fd;			// file descriptor per flow
#endif // HTTP_SAVE == 1
	uint32_t pktcnt;		// packet count for stored info
	uint32_t tcpSeqInit;		// last Tcp Sequence number if in sniff-content
	uint32_t seq;		// last Tcp Sequence number if in sniff-content
	uint32_t contentLength;		// The last Http-Content-Length field
	uint32_t sniffedContent;	// Amount of sniffed content
	uint32_t hdr_len;		// header length
	uint16_t mimeTypes;		// mime types in flow
	uint16_t mime_c;		// Amount of mime types in Flow
	uint16_t cookie_c;		// Amount of cookies in flow
	uint16_t userAg_c;		// Amount of user agent info in flow
	uint16_t images_c;		// Amount of images in flow
	uint16_t video_c;		// Amount of videos in flow
	uint16_t audio_c;		// Amount of audios in flow
	uint16_t msg_c;			// Amount of msgs in flow
	uint16_t text_c;		// Amount of texts in flow
	uint16_t appl_c;		// Amount of applications in flow
	uint16_t stat_c;		// Amount of status codes in Flow
#if ( HTTP_BDURL == 1 )
	uint16_t refURL_c;		// 
#endif // () 
	uint16_t flags;			// Http-Flags (see above)
#if HTTP_STAT == 1
	uint16_t stat[HTTP_DATA_C_MAX];	// http status
#endif // HTTP_STAT == 1
#if HTTP_MIME == 1
	char *mime[HTTP_DATA_C_MAX];	//  http Mimetypes
#endif // HTTP_MIME == 1
#if HTTP_COOKIE == 1
	char *cookie[HTTP_DATA_C_MAX];	// http cookie names
#endif // HTTP_COOKIES == 1
#if (HTTP_IMAGE == 1 || HTTP_SAVE_IMAGE == 1)
	char *image[HTTP_DATA_C_MAX];	// http image names
#endif // (HTTP_IMAGES == 1 || HTTP_SAVE_IMAGES == 1) 
#if (HTTP_VIDEO == 1 || HTTP_SAVE_VIDEO == 1)
	char *video[HTTP_DATA_C_MAX];	// http video names 
#endif // (HTTP_VIDEO == 1 || HTTP_SAVE_VIDEO == 1) 
#if (HTTP_AUDIO == 1 || HTTP_SAVE_AUDIO == 1)
	char *audio[HTTP_DATA_C_MAX];	// http audio names
#endif // (HTTP_AUDIO == 1 || HTTP_SAVE_AUDIO == 1) 
#if (HTTP_MSG == 1 || HTTP_SAVE_MSG == 1)
	char *msg[HTTP_DATA_C_MAX];	// http message names
#endif // (HTTP_MSG == 1 || HTTP_SAVE_MSG == 1) 
#if (HTTP_TEXT == 1 || HTTP_SAVE_TEXT == 1)
	char *text[HTTP_DATA_C_MAX];	// http text names
#endif // (HTTP_TEXT == 1 || HTTP_SAVE_TEXT == 1) 
#if (HTTP_APPL == 1 || HTTP_SAVE_APPL == 1)
	char *appl[HTTP_DATA_C_MAX];	// http text names
#endif // (HTTP_APPL == 1 || HTTP_SAVE_APPL == 1) 
#if ( HTTP_BDURL == 1 )
	char *refURL[HTTP_DATA_C_MAX];	// http text names
#endif // () 
#if HTTP_USRAG == 1
	char *usrAg[HTTP_DATA_C_MAX];	// http text names
#endif // () 
	char getFile[HTTP_MXIMNM_LEN];	// File requested by http-GET
	uint8_t httpBdyFlgs;		// http Content Anomalies
	uint8_t httpMethods;		// Seen Http-Methods in that Flow
} http_flow_t;

typedef enum{
	RESPONSE=0x00,
	OPTIONS=0x01,
	GET=0x02,
	HEAD=0x04,
	POST=0x08,
	PUT=0x10,
	DELETE=0x20,
	TRACE=0x40,
	CONNECT=0x80
}http_method;

typedef enum {
	application=0x0001,
	audio=0x0002,
	image=0x0004,
	message=0x0008,
	model=0x0010,
	multipart=0x0020,
	text=0x0040,
	video=0x0080,
	vnd=0x0100,
	x=0x0200,
	xpkcs=0x0400,
} http_mimetype;



// Tranalyzer-PluginFunctions

char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

#endif

