/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __ICMPDECODE_H__
#define __ICMPDECODE_H__

#define ICMP_SUFFIX "_icmpStats.txt"

// includes

// local includes
#include "main.h"
#include "binaryValue.h"
#include "networkHeaders.h"
#include "getter.h"

// ICMP types
#define ICMP4_ECHOREPLY          0       // Echo Reply
#define ICMP4_DEST_UNREACH       3       // Destination Unreachable
#define ICMP4_SOURCE_QUENCH      4       // Source Quench
#define ICMP4_REDIRECT           5       // Redirect (change route)
#define ICMP4_ECHO               8       // Echo Request
#define ICMP4_TIME_EXCEEDED      11      // Time Exceeded
#define ICMP4_PARAMETERPROB      12      // Parameter Problem
#define ICMP4_TIMESTAMP          13      // Timestamp Request
#define ICMP4_TIMESTAMPREPLY     14      // Timestamp Reply
#define ICMP4_INFO_REQUEST       15      // Information Request
#define ICMP4_INFO_REPLY         16      // Information Reply
#define ICMP4_ADDRESS            17      // Address Mask Request
#define ICMP4_ADDRESSREPLY       18      // Address Mask Reply
#define ICMP4_TRACEROUTE 	 30

/* Codes for UNREACH. */
#define ICMP4_NET_UNREACH        0       // Network Unreachable
#define ICMP4_HOST_UNREACH       1       // Host Unreachable
#define ICMP4_PROT_UNREACH       2       // Protocol Unreachable
#define ICMP4_PORT_UNREACH       3       // Port Unreachable
#define ICMP4_FRAG_NEEDED        4       // Fragmentation Needed/DF set
#define ICMP4_SR_FAILED          5       // Source Route failed
#define ICMP4_NET_UNKNOWN        6
#define ICMP4_HOST_UNKNOWN       7
#define ICMP4_HOST_ISOLATED      8
#define ICMP4_NET_ANO            9
#define ICMP4_HOST_ANO           10
#define ICMP4_NET_UNR_TOS        11
#define ICMP4_HOST_UNR_TOS       12
#define ICMP4_PKT_FILTERED       13      /* Packet filtered */
#define ICMP4_PREC_VIOLATION     14      /* Precedence violation */
#define ICMP4_PREC_CUTOFF        15      /* Precedence cut off */

/* Codes for REDIRECT. */
#define ICMP4_REDIR_NET          0       /* Redirect Net                 */
#define ICMP4_REDIR_HOST         1       /* Redirect Host                */
#define ICMP4_REDIR_NETTOS       2       /* Redirect Net for TOS         */
#define ICMP4_REDIR_HOSTTOS      3       /* Redirect Host for TOS        */

// Codes for TIME_EXCEEDED
#define ICMP4_EXC_TTL            0       /* TTL count exceeded           */
#define ICMP4_EXC_FRAGTIME       1       /* Fragment Reass time exceeded */

// icmp4 bit field length
#define ICMP4_TYPEFIELD  32      // if you change this value: %2 & change the datatype of icmptype if necessary 
#define ICMP4_CODEFIELD  16	// if you change this value: %2 & change the datatype of icmpcode if necessary

// icmp6 bit field length
#define ICMP6_TYPEFIELD  32      // if you change this value: %2 & change the datatype of icmptype if necessary 
#define ICMP6_CODEFIELD  16	// if you change this value: %2 & change the datatype of icmpcode if necessary

// functions
//extern void terminate();

char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);
void onApplicationTerminate(unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
binary_value_t* printHeader();

// variables
unsigned long numEchoRequests;
unsigned long numEchoReplies;
unsigned long numSourceQuench;
unsigned long numTraceroutes;
unsigned long numOfICMPPackets;
unsigned long numDestUnreach[16];
unsigned long numTimeExceeded[2];
unsigned long numRedirect[4];
/*
typedef struct icmpDecodeGVar_s {
	unsigned long numEchoRequests;
	unsigned long numEchoReplies;
	unsigned long numSourceQuench;
	unsigned long numTraceroutes;
	unsigned long numOfICMPPackets;
	unsigned long numDestUnreach[16];
	unsigned long numTimeExceeded[2];
	unsigned long numRedirect[4];
} icmpDecodeGVar_t;
*/

uint32_t num_icmp6[255][8];

typedef struct icmpDecodeFlow_s {
	uint32_t echoRequestsSent;
	uint32_t echoRepliesSent;
	uint32_t icmp_type_bfieldH;
	uint32_t icmp_type_bfieldL;
	uint16_t icmp_code_bfield;
} icmpDecodeFlow_t;

unsigned long pluginICMPDecodeIndex;
icmpDecodeFlow_t *icmpDecodeFlow;
//icmpDecodeGVar_t *icmpDecodeGVarP;
#ifdef __APPLE__
// global variables
extern hashMap_t *mainHashMap;
extern outputBuffer_t *main_output_buffer;
extern flow_t *flows;
extern char *baseFileName;
extern unsigned long numOfPackets;
#endif //__APPLE__
#endif

