/*
 * ipStats 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: ?, RDNIZ
 * Troubleshooter: Stefan BUrschka, RUAG AG, RUAG Defence AG, Switzerland
 */

#include "ipStats.h"

char* get_plugin_name() {
	return "ipStats";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

char* get_dependencies() {
	return "basicLayer4CalcStatistics";
}

void initialize(unsigned long pPluginIndex) {
	plugin_PLUGINNAME_Index = pPluginIndex;

        // File to write in
        memset(ipStatsFileName, '\0', MAX_FILENAME_LEN+1);
        strncpy(ipStatsFileName, baseFileName, MAX_FILENAME_LEN);
        strncat(ipStatsFileName, IPSTATS_SUFFIX, MAX_FILENAME_LEN-strlen(baseFileName));
        // Open File
        if ((ipStatsFile = fopen(ipStatsFileName, "w")) == NULL) {
                fprintf (stderr, "ipStats Error: Can't open file %s in write mode\n", ipStatsFileName);
                terminate();
        }
	// Allocate stuff
#if IPV6_ACTIVATE == 1
	ipStatsIpHashMap = hashTable_initialize(1.0f, sizeof(ipAddr_t));
#else // IPV6_ACTIVATE == 0
	ipStatsIpHashMap = hashTable_initialize(1.0f, sizeof(in_addr_t));
#endif // IPV6_ACTIVATE

	ipStats = (ipStats_t*)malloc(sizeof(ipStats_t) * ipStatsIpHashMap->hashChainTableSize);
	memset(ipStats, '\0', sizeof(ipStats_t) * ipStatsIpHashMap->hashChainTableSize);
}


void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {

	flow_t *currFlow = &flows[flowIndex];
	basicStatsPerFlow_t* basicStatsFlowPtr = &basicStatsPerFlow[flowIndex]; // Information from basicLayer4CalcStatistics

	//printf("0x%04x\t", currFlow->status);
	unsigned int i; // Counter
	unsigned long IpID; // Unique ID for every IP
	ipStats_t *ipStatsP;

	// srcIP Hashtable lookup
	if ((IpID = hashTable_lookup(ipStatsIpHashMap, (char*)&currFlow->srcIP)) == HASHTABLE_ENTRY_NOT_FOUND) {
		IpID = hashTable_insert(ipStatsIpHashMap, (char*)&currFlow->srcIP); // Add Value to IPHashMap
		ipStatsP = &ipStats[IpID];		
		ipStatsP->IP = currFlow->srcIP; // Add IP Address
		ipStatsP->inConnections = 0; // Connections to this IP
		ipStatsP->outConnections = 1; // Connections from this IP
		ipStatsP->numTransBytes = 0; // Number of Transmitted Bytes
		ipStatsP->numTransPackets = 0; // Number of transmitted Packets
		ipStatsP->numRecBytes = 0; // Number of received Bytes
		ipStatsP->numRecPackets = 0; // Number of received Packets
		ipStatsP->numPorts = 1; // Counter (how many Port-entries are there for this IP)
		ipStatsP->numfindex = 1; // Counter (how many Port-entries are there for this IP)
		ipStatsP->portList[0] = currFlow->dstPort; // first Port used by this IP
		ipStatsP->l4protList[0] = currFlow->layer4Protocol; // first L4 Protocol used by this IP
		ipStatsP->findex[0] = currFlow->findex; // first L4 Protocol used by this IP
	} else {
		ipStatsP = &ipStats[IpID];		
		ipStatsP->outConnections++; // Increase outgoing Connections
		// port stuff (viel zu kompliziert gelöst)
		if (ipStatsP->numPorts < IPSTATS_MAX_PORTS) { // Check if there are more Ports requested (in header-file)
			for( i = 0; i < ipStatsP->numPorts; i++) { // For every entry in ipStats[].portList
				if(ipStatsP->portList[i] == currFlow->dstPort && ipStatsP->l4protList[i] == currFlow->layer4Protocol) goto chkFIx; 
			}
			 // If the Port never appeared before
			ipStatsP->portList[ipStatsP->numPorts] = currFlow->dstPort; // Add the Port to the array
			ipStatsP->l4protList[ipStatsP->numPorts] = currFlow->layer4Protocol; // Add the Protocol to the array
			ipStatsP->numPorts++;
		}
	
chkFIx:		if (ipStatsP->numfindex < IPSTATS_MAX_FX) {	
			for( i = 0; i < ipStatsP->numfindex; i++) { 
				if ( ipStatsP->findex[i] == currFlow->findex  ) goto ppExist;
			}
			ipStatsP->findex[ipStatsP->numfindex] = currFlow->findex; 
			ipStatsP->numfindex++;
		}

	}

ppExist:
	// Same stuff as above for Destination host
	if ((IpID = hashTable_lookup(ipStatsIpHashMap, (char*)&currFlow->dstIP)) == HASHTABLE_ENTRY_NOT_FOUND) {
		IpID = hashTable_insert(ipStatsIpHashMap, (char*)&currFlow->dstIP);
		ipStatsP = &ipStats[IpID];		
		ipStatsP->IP = currFlow->dstIP;
		ipStatsP->inConnections = 1; // Destination host gets an inConnection instead of an outConnection
		ipStatsP->outConnections = 0;
		ipStatsP->numTransBytes = 0;
		ipStatsP->numTransPackets = 0;
		ipStatsP->numRecBytes = 0;
		ipStatsP->numRecPackets = 0;
		ipStatsP->numPorts = 0;
		ipStatsP->numfindex = 1; // Counter (how many Port-entries are there for this IP)
		ipStatsP->findex[0] = currFlow->findex; // first L4 Protocol used by this IP
	} else {
		// No Ports are added, as he is only the receiver
		ipStats[IpID].inConnections++;
		if (ipStatsP->numfindex < IPSTATS_MAX_FX) {	
			for( i = 0; i < ipStatsP->numfindex; i++) { 
				if ( ipStatsP->findex[i] == currFlow->findex  ) return; 
			}
			ipStatsP->findex[ipStatsP->numfindex] = currFlow->findex; 
			ipStatsP->numfindex++;
		}

	}


}

void onFlowTerminate(unsigned long flowIndex) {
	
	flow_t *currFlow = &flows[flowIndex];
	basicStatsPerFlow_t* basicStatsFlowPtr = &basicStatsPerFlow[flowIndex]; // Information from basicLayer4Calcstatistics

	unsigned long IpID; // Unique ID for every IP
	ipStats_t *ipStatsP;
	// Check if srcIP has already appeared (It really should, else we have a bug)
	if ((IpID = hashTable_lookup(ipStatsIpHashMap, (char*)&currFlow->srcIP)) == HASHTABLE_ENTRY_NOT_FOUND) {
                fprintf(stdout, "ipStats ERROR: hashTable_lookup failed unexpectedly \n");
	} else {
		ipStatsP = &ipStats[IpID];		
		ipStatsP->numTransBytes += basicStatsFlowPtr->numTransBytes; // sum up sent Bytes
		ipStatsP->numTransPackets += basicStatsFlowPtr->numTransPackets; // sum up sent Packets
		if (currFlow->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) { // Check opposite flow
			ipStatsP->numRecBytes += basicStatsPerFlow[currFlow->oppositeFlowIndex].numTransBytes; // sum up received Bytes
			ipStatsP->numRecPackets += basicStatsPerFlow[currFlow->oppositeFlowIndex].numTransPackets; // sum up received Packets
		}
		ipStatsP->status = currFlow->status; // first L4 Protocol used by this IP
	}
	if (IpID > maxIpID) maxIpID = IpID; // Check maxIpID
	// Do the same as above for destination host
	if ((IpID = hashTable_lookup(ipStatsIpHashMap, (char*)&currFlow->dstIP)) == HASHTABLE_ENTRY_NOT_FOUND) {
                fprintf(stdout, "ipStats ERROR: hashTable_lookup failed unexpectedly \n");
	} else {
		ipStatsP = &ipStats[IpID];		
		ipStatsP->numRecBytes += basicStatsFlowPtr->numTransBytes;
		ipStatsP->numRecPackets += basicStatsFlowPtr->numTransPackets;
		if (currFlow->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
			ipStatsP->numTransBytes += basicStatsPerFlow[currFlow->oppositeFlowIndex].numTransBytes;
			ipStatsP->numTransPackets += basicStatsPerFlow[currFlow->oppositeFlowIndex].numTransPackets;
		}	
		ipStatsP->status = currFlow->status; // first L4 Protocol used by this IP
	}
	if (IpID > maxIpID) maxIpID = IpID;
}

void onApplicationTerminate() {

	unsigned int i, k = 0; // counters
	double packet_asym, byte_asym; // Packet asymmetry and byte asymmetry
	ipStats_t *ipStatsP;

	/* OUTPUT */
	fprintf(ipStatsFile, "# IPID\tIP\tstatus\toutConns\tinConns\tnumTransPackets\tnumRecPackets\tPacketAsym\tnumTransBytes\tnumRecBytes\tByteAsym\txL4Protocols\tTCPPorts\tUDPPorts\tfindex\n");

	for( i = 0; i <= maxIpID; i++ ) {
		ipStatsP = &ipStats[i];		
		
		// Calculate packets asymmetry
		if ((ipStatsP->numTransPackets + ipStatsP->numRecPackets) != 0) {	
			packet_asym = ((double)ipStatsP->numTransPackets - (double)ipStatsP->numRecPackets) / ((double)ipStatsP->numTransPackets + (double)ipStatsP->numRecPackets);
		} else {
			packet_asym = 0;
		}
		// Calculate byte asymmetry
		if ((ipStatsP->numTransBytes + ipStatsP->numRecBytes) != 0) {
			byte_asym = ((double)ipStatsP->numTransBytes - (double)ipStatsP->numRecBytes) / ((double)ipStatsP->numTransBytes + (double)ipStatsP->numRecBytes);
		} else {
			byte_asym = 0;
		}


		fprintf(ipStatsFile, "%d\t", i+1);

		// Print IP
#if IPV6_ACTIVATE == 0

#if IPSTATS_IP_HEX == 0
		fprintf(ipStatsFile, "%d.%d.%d.%d\t", ipStatsP->IP.s_addr & 0xff, (ipStatsP->IP.s_addr >> 8) & 0xff, (ipStatsP->IP.s_addr >> 16) & 0xff, ipStatsP->IP.s_addr >> 24 );
#else //IPSTATS_IP_HEX == 1
		fprintf(ipStatsFile, "0x%08"PRIX32"\t", ntohl(ipStatsP->IP.s_addr));
#endif //IPSTATS_IP_HEX

#else // IPV6_ACTIVATE == 1
		uint16_t *c;
#if IPSTATS_IP_HEX == 0
		c = (uint16_t*)&ipStatsP->IP.IPv6.__in6_u.__u6_addr16[0];
#if IPSTATS_CMPRSS6 == 0
		fprintf(ipStatsFile, "%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16"\t", 
					ntohs(c[0]), ntohs(c[1]), ntohs(c[2]), ntohs(c[3]), ntohs(c[4]), ntohs(c[5]), ntohs(c[6]), ntohs(c[7]));
#else // IPSTATS_CMPRSS6 == 1
		uint16_t sw = 0;

		for ( k = 0; k < 8; k++ ) {
			if ( c[k] ) {
				if ( sw ) fprintf(ipStatsFile, ":");
				fprintf(ipStatsFile, "%x", ntohs(c[k]));
				sw = 1;
			} else if ( sw ) {
				if ( !c[k+1] && sw == 1) {
					fprintf(ipStatsFile, "::");
					sw = 0;
				} else fprintf(ipStatsFile, ":0");
			}
		}
		if ( !sw ) fprintf(ipStatsFile, "::");
		fprintf(ipStatsFile, "\t");
#endif // IPSTATS_CMPRSS6

#else //IPSTATS_IP_HEX == 1
		uint32_t *c32 = (uint32_t*)&ipStatsP->IP.IPv6.__in6_u.__u6_addr32[0];
		fprintf(ipStatsFile, "0x%08"PRIX32"%08"PRIX32"%08"PRIX32"%08"PRIX32"\t", ntohl(c32[0]), ntohl(c32[1]), ntohl(c32[2]), ntohl(c32[3]) );
#endif //IPSTATS_IP_HEX

#endif // IPV6_ACTIVATE

		fprintf(ipStatsFile, "0x%04x\t", ipStatsP->status);
 
		// Print outConns, inConns, TransPackets, RecPackets, packet_asym, TransBytes, RecBytes, byte_asym
		fprintf(ipStatsFile, "%d\t%d\t%"PRIu64"\t%"PRIu64"\t%f\t%"PRIu64"\t%"PRIu64"\t%f\t", 
			ipStatsP->outConnections, ipStatsP->inConnections, ipStatsP->numTransPackets, ipStatsP->numRecPackets, packet_asym,
			ipStatsP->numTransBytes, ipStatsP->numRecBytes, byte_asym);
		// Print Other Protocols (as ICMP, OSPF and stuff)

		for( k = 0; k < ipStatsP->numPorts; k++ ) {
			if (ipStatsP->portList[k] == 0) {
				fprintf(ipStatsFile, "%d ", ipStatsP->l4protList[k]);
			}
		}
		
		// Print TCP Ports
		fprintf(ipStatsFile, "\t");
		for( k = 0; k < ipStatsP->numPorts; k++ ) {
			if(ipStatsP->l4protList[k] == 6){
				fprintf(ipStatsFile, "%d ", ipStatsP->portList[k]);
			}
		}
		
		// Print UDP Ports
		fprintf(ipStatsFile, "\t");
		for( k = 0; k < ipStatsP->numPorts; k++ ) {
			if(ipStatsP->l4protList[k] == 17){
				fprintf(ipStatsFile, "%d ", ipStatsP->portList[k]);
			}
		}
		
		fprintf(ipStatsFile, "\t");
		for( k = 0; k < ipStatsP->numfindex; k++ ) {
			fprintf(ipStatsFile, "%ld ", ipStatsP->findex[k]);
		}
		

		fprintf(ipStatsFile, "\n");
	}

//	if (a) printf("ipStats WARNING: Ports list truncated, increase IPSTATS_MAX_PORTS in ipstats.h\n"); // Print a warning if storage length exeeded

	fclose(ipStatsFile);

	// Free Memory
	hashTable_destroy(ipStatsIpHashMap); ipStatsIpHashMap = NULL;
	if (ipStats) free(ipStats); 
	ipStatsP = NULL;
}
