/*
 * ipStats 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: ?, RDNIZ
 */

#ifndef IPSTATS_H_
#define IPSTATS_H_

// global includes
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

// local includes
#include "networkHeaders.h"
#include "main.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "../basicLayer4CalcStatistics/src/basicLayer4CalcStatistics.h"


// user defines
#define IPSTATS_CMPRSS6		1	// if (IPV6_ACTIVATE) then 0: full length :hex: , 1: Compressed output
#define IPSTATS_IP_HEX		0	// 0: IP in dotted dec format; 1: IP in Hex
#define IPSTATS_MAX_FX		10	// Maximum findex
#define IPSTATS_MAX_PORTS	10	// Maximum Ports to save (UDP, TCP)
#define IPSTATS_SUFFIX "_ipStats.txt"	// 

// HashMap
hashMap_t * ipStatsIpHashMap;

// Struct
typedef struct ipStats_s {
#if IPV6_ACTIVATE == 1 
		ipAddr_t IP;
#else // IPV6_ACTIVATE == 0
		struct in_addr IP;
#endif
		uint16_t inConnections; // Incoming connections
		uint16_t outConnections; // Outgoing connections
		uint64_t numTransPackets; // Transmitted Packets
		uint64_t numRecPackets; // Received Packets
		uint64_t numTransBytes; // Transmitted bytes
		uint64_t numRecBytes; // Received bytes
		uint64_t findex[IPSTATS_MAX_FX]; // 
		uint16_t numfindex;
		uint16_t numPorts;
		uint16_t status; // 
		uint16_t portList[IPSTATS_MAX_PORTS]; // Ports used
		uint8_t l4protList[IPSTATS_MAX_PORTS]; // L4 protos used
	} ipStats_t;

ipStats_t *ipStats;

int maxIpID;

char ipStatsFileName[MAX_FILENAME_LEN+1];
FILE *ipStatsFile;
extern char *baseFileName;
extern void terminate();

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
char* get_dependencies();
void initialize(unsigned long pPluginIndex);
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

// variables
unsigned long plugin_PLUGINNAME_Index;

#endif /* IPSTATS_H_ */

