/*
 * macRecorder.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "macRecorder.h"
#include <stdlib.h>
#include <stdio.h>
#include "hashTable.h"

char* get_plugin_name() {
	return "macRecorder";
}

char* get_plugin_version() {
	return "0.5.1";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
#ifdef __APPLE__
	mainHashMap = getMainHashMap();
#endif // __APPLE__
	pluginMacRecorderIndex = pPluginIndex;
	unsigned int i;

	// malloc memory
	macRecorder_macArray = malloc(mainHashMap->hashChainTableSize * sizeof(macRecorder_t));
	macRecorder_macList = malloc(2 * mainHashMap->hashChainTableSize * sizeof(macList_t));

	memset(macRecorder_macArray, '\0', mainHashMap->hashChainTableSize * sizeof(macRecorder_t));
	memset(macRecorder_macList, '\0', 2 * mainHashMap->hashChainTableSize * sizeof(macList_t));

	// connect free entrys with each other

	for (i = 0; i < 2 * mainHashMap->hashChainTableSize - 1; i++) {
		macRecorder_macList[i].next = &(macRecorder_macList[i + 1]);
	}

	// set freeList pointer on first entry
	macRecorder_freeList = &(macRecorder_macList[0]);
}

binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;

	binary_value_t *act_bv = bv_new_bv( "Source MAC address, destination MAC address, number of packets seen with this MAC address combination", "SrcMac_DstMac_NumP", 1, 3, bt_mac_addr, bt_mac_addr, bt_uint_64);
	main_bv = bv_append_bv(main_bv, act_bv);

	return main_bv;
}


void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	// first put all attached macList entries back into the free list
	macList_t *macRecorder_tempListPointer1 = macRecorder_macArray[flowIndex].macList;
	macList_t *macRecorder_tempListPointer2;

	while (macRecorder_tempListPointer1 != NULL) {
		macRecorder_tempListPointer2 = macRecorder_tempListPointer1->next;
		macRecorder_tempListPointer1->next = macRecorder_freeList;
		macRecorder_freeList = macRecorder_tempListPointer1;
		macRecorder_tempListPointer1 = macRecorder_tempListPointer2;
	}

	// reset array entry
	memset(&(macRecorder_macArray[flowIndex]), '\0', sizeof(macRecorder_t));
}


void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {

	macList_t *macRecorder_tempListPointer1 = macRecorder_macArray[flowIndex].macList;
	macList_t *macRecorder_tempListPointer2 = macRecorder_tempListPointer1;

	while(macRecorder_tempListPointer1 != NULL) {
		if ((memcmp(macRecorder_tempListPointer1->ethernetHeader.ether_dhost, packet->layer2Header->ethernetHeader.ether_dhost, 6) == 0) &&
			(memcmp(macRecorder_tempListPointer1->ethernetHeader.ether_shost, packet->layer2Header->ethernetHeader.ether_shost, 6) == 0)) {
			macRecorder_tempListPointer1->numOfPackets++;
			return;
		}
		macRecorder_tempListPointer2 = macRecorder_tempListPointer1;
		macRecorder_tempListPointer1 = macRecorder_tempListPointer1->next;
	}

	// the macList entry wasn't found
	// take a list entry out of the free list
	if (macRecorder_macArray[flowIndex].macList == NULL) {
		macRecorder_macArray[flowIndex].macList = macRecorder_freeList; // point to first entry in free list
		macRecorder_tempListPointer2 = macRecorder_macArray[flowIndex].macList; // move temp pointer to entry
	} else {
		macRecorder_tempListPointer2->next = macRecorder_freeList; // point to first entry in free list
		macRecorder_tempListPointer2 = macRecorder_tempListPointer2->next; // move temp pointer to entry
	}

	macRecorder_freeList = macRecorder_freeList->next; // move free list pointer
	macRecorder_tempListPointer2->next = NULL; // disconnect entry from free list

	// fill with the right values
	memcpy(&(macRecorder_tempListPointer2->ethernetHeader), &(packet->layer2Header->ethernetHeader), sizeof(ethernetHeader_t));
	macRecorder_tempListPointer2->numOfPackets = 1;

	// increment the number of mac combos
	macRecorder_macArray[flowIndex].num_of_entries++;
}


void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	macList_t *macRecorder_tempListPointer1 = macRecorder_macArray[flowIndex].macList;
	
	// write out number of entries, cause output is repeatable
	outputBuffer_append(main_output_buffer, (char*) &macRecorder_macArray[flowIndex].num_of_entries, sizeof(uint32_t));

	// point to actual entry
	while (macRecorder_tempListPointer1 != NULL) {
		// print source  and dest mac
		outputBuffer_append(main_output_buffer, (char*) &macRecorder_tempListPointer1->ethernetHeader.ether_shost, l_bt_mac_addr);
		outputBuffer_append(main_output_buffer, (char*) &macRecorder_tempListPointer1->ethernetHeader.ether_dhost, l_bt_mac_addr);

		// print number of packets with this src/dst combo
		outputBuffer_append(main_output_buffer, (char*) &macRecorder_tempListPointer1->numOfPackets, sizeof(uint64_t));

		// goto next entry
		macRecorder_tempListPointer1 = macRecorder_tempListPointer1->next;
	}
}


void onApplicationTerminate() {
	free(macRecorder_macArray);
	free(macRecorder_macList);
	 macRecorder_macArray = NULL;
	 macRecorder_macList = NULL;
}

