/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * nFirstPacketsStats.c
 */

#include "nFirstPacketsStats.h"

char* get_plugin_name() {
	return "nFirstPacketsStats";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	pluginNFirstPacketsStatsIndex = pPluginIndex;
#ifdef __APPLE__
	mainHashMap = getMainHashMap();
#endif // __APPLE__
	if ( ( nFirstPacketsStats = malloc(mainHashMap->hashChainTableSize * sizeof(nFirstPacketsStats_t) ) ) == NULL) {
		fprintf(stderr, "nFirstPacketStats: Couldn't allocate memory for packet statistics. Exit program");
		terminate();
	}
	memset(nFirstPacketsStats, '\0', mainHashMap->hashChainTableSize * sizeof(nFirstPacketsStats_t));
}

binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;

#if NFIRSTP_HDRINFO == 1
	binary_value_t *act_bv = bv_new_bv("L3Hdr, L4 Hdr, L2/L3/L4/Payload( s. PACKETLENGTH in packetCapture.h) length and inter-arrival times for the N first packets", "HSig:HD3l_HD4l_L2L3L4Pl_Iat", 1, 4, bt_uint_8, bt_uint_8, bt_uint_16, bt_unix_time);
#else // NFIRSTP_HDRINFO == 0 
	binary_value_t *act_bv = bv_new_bv("L2/L3/L4/Payload( s. PACKETLENGTH in packetCapture.h) length and inter-arrival times for the N first packets", "Sig:L2L3L4Pl_Iat", 1, 2, bt_uint_16, bt_unix_time);
#endif // NFIRSTP_HDRINFO

	main_bv = bv_append_bv(main_bv, act_bv);

	return main_bv;

}

void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
	nFirstPacketsStats_t *nFpStatsP = &nFirstPacketsStats[flowIndex];
	nFpStatsP->packetsAnalyzed = 0;
	nFpStatsP->lastPacketTime.tv_sec = packet->pcapHeader->ts.tv_sec;
	nFpStatsP->lastPacketTime.tv_usec = packet->pcapHeader->ts.tv_usec;
}

void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
	nFirstPacketsStats_t *nFpStatsP = &nFirstPacketsStats[flowIndex];
	packetStats_t *pP;

	uint32_t ipC = nFpStatsP->packetsAnalyzed;

	if (ipC < NFIRSTP_PKTCNT) {
		pP = &nFpStatsP->packets[ipC];
#if NFIRSTP_HDRINFO == 1
		pP->l3HDLen = packet->l2HdrLen; 
		pP->l4HDLen = packet->l4HdrLen;
#endif // NFIRSTP_HDRINFO 

		pP->packetSize = packet->packetLength;
		timersub(&packet->pcapHeader->ts, &nFpStatsP->lastPacketTime, &pP->interArrivalTime);

		// update the last seen packet time
		nFpStatsP->lastPacketTime.tv_sec = packet->pcapHeader->ts.tv_sec;
		nFpStatsP->lastPacketTime.tv_usec = packet->pcapHeader->ts.tv_usec;

		// increment number of analyzed packets
		nFpStatsP->packetsAnalyzed++;
	}
		
}

void onFlowTerminate(unsigned long flowIndex) {
	main_output_buffer = getMainOutputBuffer();
	uint32_t i, temp_tv_usec;
	uint64_t temp_tv_sec;
	nFirstPacketsStats_t *nFpStatsP = &nFirstPacketsStats[flowIndex];

	// number of entries, because output is repeatable
	outputBuffer_append(main_output_buffer, (char*) &nFpStatsP->packetsAnalyzed, sizeof(uint32_t));

	for (i = 0; i < nFpStatsP->packetsAnalyzed; i++) {
#if NFIRSTP_HDRINFO == 1
		outputBuffer_append(main_output_buffer, (char*) &nFpStatsP->packets[i].l3HDLen, sizeof(uint8_t));
		outputBuffer_append(main_output_buffer, (char*) &nFpStatsP->packets[i].l4HDLen, sizeof(uint8_t));
#endif // NFIRSTP_HDRINFO 

		outputBuffer_append(main_output_buffer, (char*) &nFpStatsP->packets[i].packetSize, sizeof(uint16_t));
		temp_tv_sec = (uint64_t)nFpStatsP->packets[i].interArrivalTime.tv_sec;
		outputBuffer_append(main_output_buffer, (char*) &temp_tv_sec, sizeof(uint64_t));

		temp_tv_usec = nFpStatsP->packets[i].interArrivalTime.tv_usec * 1000;
		outputBuffer_append(main_output_buffer, (char*) &temp_tv_usec, sizeof(uint32_t));

	}

}

void onApplicationTerminate() {
	if (nFirstPacketsStats) { 
		free(nFirstPacketsStats);
		nFirstPacketsStats = NULL;
	}
}

