/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * nFirstPacketsStats.h
 */

#ifndef NFIRSTPACKETSSTATS_H_
#define NFIRSTPACKETSSTATS_H_

/* includes */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <inttypes.h>

/* local includes */
#include "tranalyzer.h"
#include "main.h"
#include "networkHeaders.h"
#include "binaryValue.h"
#include "getter.h"

// User config parameters
#define NFIRSTP_PKTCNT 20 // defines how many first packets are recorded
#define NFIRSTP_HDRINFO 0 // extended 

/* struct to save basic statistic of a single packet */
typedef struct packetStats_s {
	struct timeval interArrivalTime; // 8 bytes
	uint16_t packetSize; // 2 bytes
#if NFIRSTP_HDRINFO == 1
	uint8_t l3HDLen; // 1 bytes
	uint8_t l4HDLen; // 1 bytes
#endif // NFIRSTP_HDRINFO == 1
} packetStats_t; // 12 bytes -> 14 bytes allocated

/* struct to collect the stats of the first n packets of a flow */
typedef struct nFirstPacketsStats_s {
	struct timeval lastPacketTime; // 8 bytes
	uint32_t packetsAnalyzed; // 4 bytes
	packetStats_t packets[NFIRSTP_PKTCNT]; // n times 14 bytes
} nFirstPacketsStats_t;

/* variables */
nFirstPacketsStats_t *nFirstPacketsStats;
unsigned long pluginNFirstPacketsStatsIndex;

#ifdef __APPLE__
// global variables
hashMap_t *mainHashMap;
outputBuffer_t *main_output_buffer;
#endif // __APPLE__
/* functions */

char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

#endif /* NFIRSTPACKETSSTATS_H_ */

