/*
 * ospfDecode.c
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 */


#include "ospfDecode.h"

// Global variables 
#ifdef __APPLE__
extern hashMap_t* mainHashMap;
extern outputBuffer_t* main_output_buffer;
#endif // __APPLE__

// plugin variables
unsigned long plugin_ospf_Index;
ospfFlow_t *ospfFlow;


// Tranalyzer functions

char* get_plugin_name() {
	return "ospfDecode";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

char* get_dependencies() {
	return "";
}

void initialize(unsigned long pPluginIndex) {
	plugin_ospf_Index = pPluginIndex;
#ifdef __APPLE__
        mainHashMap = getMainHashMap();
        baseFileName = getBaseFileName();
        specialOutput = getSpecialOutput();
#endif // __APPLE__

        ospfFlow = calloc(sizeof(ospfFlow_t) * mainHashMap->hashChainTableSize, sizeof(ospfFlow_t) * mainHashMap->hashChainTableSize); // allocate struct for all flows and reset
//        memset(ospfFlow, '\0', sizeof(ospfFlow_t) * mainHashMap->hashChainTableSize); // set all to 0
}


binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;

	// 
	main_bv = bv_append_bv(main_bv, bv_new_bv("OSPF status", "ospfStat", 0, 1, bt_hex_8)); 
	main_bv = bv_append_bv(main_bv, bv_new_bv("OSPF type", "ospftype", 0, 1, bt_hex_8)); 

	return main_bv;

}


void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *flowP = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
	ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

	ospfFlow_t *ospfFlow_P = &ospfFlow[flowIndex];

// your code how you want to init
	memset( ospfFlow_P, '\0', sizeof(ospfFlow_t) ); // set everything to 0

}



void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {

#if IPV6_ACTIVATE == 1
	ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

	if (ipHeader->ip_p != L3_OSPF) return;

#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__

	flow_t *flowP = &flows[flowIndex];

	ospfFlow_t *ospfFlowP = &ospfFlow[flowIndex];
	
	ospfHeader_t *ospfHdrP = (ospfHeader_t*)packet->layer4Header;

	ospfFlowP->type |= 1 << ospfHdrP->type;
	uint16_t dlen = ospfFlowP->len - 24;
	ospfFlowP->routerID = ospfHdrP->routerID;
	ospfFlowP->areaID = ospfHdrP->areaID;
	ospfFlowP->autType |= (uint8_t)(1 << ospfHdrP->autType_res);

	ospfType1_t *ospfT1P = (ospfType1_t*)&ospfHdrP->data;	
	switch (ospfHdrP->type) {
		case 1:
			ospfFlowP->netMask1 = ospfT1P->netMask;
			ospfFlowP->helloInt1 = ospfT1P->helloInt;
			ospfFlowP->options1 = ospfT1P->options;
			break;
		case 2:
			break;
		case 3:
			break;
		case 4:
			break;
		case 5:
			break;
		default:
			break;
	}
}

void onFlowTerminate(unsigned long flowIndex) {
	ospfFlow_t *ospfFlowP = &ospfFlow[flowIndex];

#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	flow_t *FlowP = &flows[flowIndex];

// Sample code

	outputBuffer_append(main_output_buffer, (char*)&ospfFlowP->stat, 2*sizeof(uint8_t));
//	outputBuffer_append(main_output_buffer, (char*)&ospfFlowP->type, sizeof(uint8_t));


}

void onApplicationTerminate() {

	if (ospfFlow) {
		free(ospfFlow);
		ospfFlow = NULL;
	}

}


