/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * packetSizeInterArrivalTimeHisto.h
 */

// includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

/* local includes */
#include "tranalyzer.h"
#include "redBlackTree.h"
#include "networkHeaders.h"
#include "hashTable.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "getter.h"

#ifndef PACKETSIZEINTERARRIVALTIMEHISTO_H_
#define PACKETSIZEINTERARRIVALTIMEHISTO_H_

// User defines
#define PRINT_HISTO_IN_SEPARATE_FILE 0 // 1: print histo into separate histo file for Traviz3 and IPSOM
#define HISTO_NODEPOOL_FACTOR 15 // multiplication factor redblack tree nodepool: sizeof(nodepool) = HISTO_NODEPOOL_FACTOR * mainHashMap->hashChainTableSize
#define PRINT_HISTO 1 // 1: print histo into flow file
#define HISTO_PRINT_BIN 0 // 1: Bin number; 0: Minimum of assigned inter arrival time. (Example: Bin = 10 -> iat = [50:55) -> min(iat) = 50ms)
#define HISTO_PRINT_PROJECTION 1 // enables printing of the axis projections
#define HISTO_EARLY_CLEANUP 0 // enables the destroying of the tree and the cleaning of the struct directly after printout
#define HISTO_FILE_SUFFIX "_histo.txt"
#define HISTO_DEBUG 0 // enables debug output

// User statistics definitions
// integer flexible bin

#define IATNORM	1000	// select ms as basic unit

#define IATBINBu1   200 	// bin boundary of section one: [0, 200)ms
#define IATBINBu2   400
#define IATBINBu3   1000
#define IATBINBu4   10000
#define IATBINBu5   100000

#define IATBINWu1   5		// bin width 5ms
#define IATBINWu2   10
#define IATBINWu3   20
#define IATBINWu4   50
#define IATBINWu5   100

#define IATBINNu1   IATBINBu1 / IATBINWu1 	// # of bins in section one
#define IATBINNu2   (IATBINBu2 - IATBINBu1) / IATBINWu2 + IATBINNu1
#define IATBINNu3   (IATBINBu3 - IATBINBu2) / IATBINWu3 + IATBINNu2
#define IATBINNu4   (IATBINBu4 - IATBINBu3) / IATBINWu4 + IATBINNu3
#define IATBINNu5   (IATBINBu5 - IATBINBu4) / IATBINWu5 + IATBINNu4

#define IATSECMAX 3 // max # of section in statistics, last section comprises all elements > IATBINBu2
//#define IATBINUMAX IATBINNu3 + 1
#define TOKENHELPER(x, y) x ## y
#define TOKENPASTE(x ,y) TOKENHELPER(x, y)
#define IATBINUMAX TOKENPASTE(IATBINNu, IATSECMAX) + 1 // automated setting of timebin dimension

// definition of bin count fields
const uint32_t IATBinBu[IATSECMAX+1] = { 0, IATBINBu1, IATBINBu2, IATBINBu3 };
const uint32_t IATBinWu[IATSECMAX] = { IATBINWu1, IATBINWu2, IATBINWu3 };
const uint32_t IATBinNu[IATSECMAX+1] = { 0, IATBINNu1, IATBINNu2, IATBINNu3 };
//uint32_t IATBinNu[IATSECMAX+1];


// float flexible bin
/*
#define IATBINBF1   0.2f
#define IATBINBF2   0.4f
#define IATBINBF3   1.0f
#define IATBINBF4   10.0f
#define IATBINBF5   100.0f

#define IATBINWF1   0.005f
#define IATBINWF2   0.01f
#define IATBINWF3   0.02f
#define IATBINWF4   0.1f
#define IATBINWF5   1f

#define IATBINWIF1  1.0f / IATBINWF1 
#define IATBINWIF2  1.0f / IATBINWF2 
#define IATBINWIF3  1.0f / IATBINWF3 
#define IATBINWIF4  1.0f / IATBINWF4 
#define IATBINWIF5  1.0f / IATBINWF5 

#define IATBINNfu1   IATBINBf1 / IATBINWf1
#define IATBINNfu2   (IATBINBf2 - IATBINBf1) / IATBINWf2 + IATBINNfu1
#define IATBINNfu3   (IATBINBf3 - IATBINBf2) / IATBINWf3 + IATBINNfu2
#define IATBINNfu4   (IATBINBf4 - IATBINBf3) / IATBINWf4 + IATBINNfu3
#define IATBINNfu5   (IATBINBf5 - IATBINBf4) / IATBINWf5 + IATBINNfu4

#define IATBINFMAX IATBIINF3 + 1

const float IATBinBf[] = { 0.0f, IATBINBF1, IATBINBF2, IATBINBF3 };
const float IATBinWf[] = { IATBINWF1, IATBINWF2, IATBINWF3 };
const float IATBinWif[] = { IATBINWIF1, IATBINWIF2, IATBINWIF3 };
const uint32_t IATBinNfu[] = { 0, IATBINF1, IATBINF2, IATBINF3 };
*/


/* structs */
typedef struct psiat_val_s {
	uint32_t numOfPackets; // 4
	redBlackNode_t *iat_tree; // 4
} psiat_val_t;

typedef struct packetSizeInterArrivalTimeTrees_s {
	uint32_t numOfPackets; // the total amount of packets inside the tree. 4 bytes
	redBlackNode_t *packetTree; // the root of the stored tree. 4 bytes
	struct timeval lastPacketTime; // 8 bytes
	uint32_t numOfpacketsInTimeBin[IATBINUMAX];	// stores the packets with a specific IAT,
																					// so we don't need to traverse the tree several times. 372 bytes
} packetSizeInterArrivalTimeTrees_t; 												// 388 bytes

/* variables */
#if PRINT_HISTO_IN_SEPARATE_FILE != 0
char histoFileName[MAX_FILENAME_LEN+1]; // filename of the histo file
FILE *histoFile;
#endif /* PRINT_HISTO_IN_SEPARATE_FILE != 0 */

#if PRINT_HISTO != 0 || PRINT_HISTO_IN_SEPARATE_FILE != 0
outputBuffer_t* psiat_output_buffer;
#endif /* PRINT_HISTO != 0 || PRINT_HISTO_IN_SEPARATE_FILE != 0 */

redBlackTreeNodePool_t *packetSizeInterArrivalTime_treeNodePool;
packetSizeInterArrivalTimeTrees_t *packetSizeInterArrivalTime_trees;
psiat_val_t *psiat_vals;
unsigned long currPacketTreeBucket, currIATTreeBucket;
uint32_t psiat_counter;
uint8_t psiat_is_first; // a flag we need to determine if it is the first bin we print out. Important for printing the delimeter (";")
unsigned long packetSizeInterArrivalTimeHistoIndex;

#ifdef __APPLE__
//global variables 
extern hashMap_t *mainHashMap;
extern outputBuffer_t *main_output_buffer;
#endif // __APPLE__

/* functions */
uint32_t iat2bin(struct timeval iat);
int32_t bin2iat(uint32_t bin);

char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
#if PRINT_HISTO != 0
binary_value_t* printHeader();
#endif /* PRINT_HISTO != 0 */
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();
#ifdef __APPLE__
//global function
redBlackTreeNodePool_t* getPacketSizeInterArrivalTimeTreeNodePool();
packetSizeInterArrivalTimeTrees_t* getPacketSizeInterArrivalTimeTrees();
psiat_val_t* getPsiatVals();
#endif // __APPLE__
#endif /* PACKETSIZEINTERARRIVALTIMEHISTO_H_ */

