/*
 * pcapd.c
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 */

#include "pcapd.h"

// global defines
uint32_t *findexP, line_cnt;
pcap_t *fdPcap;
pcap_dumper_t *pd;
FILE *fifp;

// plugin functions

char* get_plugin_name() {
	return "pcapd";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}


void initialize(unsigned long pPluginIndex) {
	unsigned int isw = 1, i = 0, z;
	char *s, regline[MAXPCAPR], filename[256];

	memset(filename,'\0', 256);
	if ( esomFileName) {
		strncpy(filename, esomFileName, 255);
	} else {
		strncpy(filename, baseFileName, 255);
		strncat(filename, ESOMFILEN, 255 - strlen(filename));
	}

	if ( (fifp = fopen(filename, "r")) == NULL ) {
		printf("%s Error: Can't open file %s in read mode..\n", ERROR_PREFIX, filename);
		terminate();
	}
	
	while ( fgets(regline, MAXPCAPR, fifp) ) {

                if ( regline[0] == '#' || regline[0] == '\n' || regline[0] == ' ' ) continue;

		if ( isw && regline[0] == '%' ) {
			z = sscanf(regline, "%%%d", &line_cnt);
                	if (z == 0) {
                        	printf("pcapd error: %d - %d\n", z, line_cnt );
                        	exit(-1);
                	}
			findexP = (uint32_t*)malloc(line_cnt * sizeof(uint32_t));
			isw = 0; 
		} else {
			if ( regline[0] != '%') 
				//z = sscanf(regline, "%[^\n\t]%d\t%s", s, &findexP[i++],s);
				z = sscanf(regline, "%*1c\t%d\t%*c", &findexP[i++]);
			if ( i >= line_cnt) break;
			
		}
	}

	fdPcap = pcap_open_dead(DLT_EN10MB, 65535);

	memset(filename,'\0', 256);
	
	if ( esomFileName == NULL ) strncpy(filename, baseFileName, 255);
	else strncpy(filename, esomFileName, 255);
	
	strncat(filename, PCAPFILEN, 255 - strlen(filename));

	if ( (pd = pcap_dump_open(fdPcap, filename)) == NULL ) {
		printf("%s Error: Can't open pcap file %s in  write mode..\n", ERROR_PREFIX, filename);
		terminate();
	}

}


void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	flow_t *flowP = &flows[flowIndex];
	unsigned int i;	

	for ( i = 0; i < line_cnt; i++ ) {
		if ( flowP->findex == findexP[i])
			 pcap_dump((u_char *)pd, packet->pcapHeader, packet->raw_packet);
	}

}


void onApplicationTerminate() {
	if (pd) pcap_dump_close(pd);
}


