#!/bin/bash


# ------------------------------------------
#-------------- CONFIG PART ----------------
# ------------------------------------------

# Plugin name
PLUGINNAME=portBasedClassifier

# Plugin execution order, as 3-digit decimal
PLUGINORDER=014_





# ------------------------------------------
#-------------- SCRIPT PART ----------------
# ------------------------------------------

DEBUG_MODE=0
NOCONFIG=0
TRANALYZER_INSTALL_DIR=
TRANALYZER_DEFAULT_DIR=$HOME"/.tranalyzer/plugins"

function USAGE ()
{
    echo ""
    echo "USAGE:"
    echo "    autogen.sh [-dht]"
    echo ""
    echo "OPTIONS:"
    echo "    -n  only execute make and copy plugins"
    echo "    -d  compile in debug and profiling mode"
    echo "    -h  show this usage information"
    echo "    -t  tranalyzer install directory"
    echo ""
    exit $E_OPTERROR    # Exit and explain usage, if no argument(s) given.
}


#PROCESS ARGS
while getopts "ndht:" Option
do
    case $Option in
        n    ) NOCONFIG=1;;
        d    ) DEBUG_MODE=1;;
        t    ) TRANALYZER_INSTALL_DIR=$OPTARG;;
        h    ) USAGE
               exit 0;;
        *    ) echo ""
               echo "Unimplemented or incomplete option chosen!"
               USAGE   # DEFAULT
    esac
done


shift $(($OPTIND - 1))
#  Decrements the argument pointer so it points to next argument.
#  $1 now references the first non option item supplied on the command line
#+ if one exists.


if [ "$NOCONFIG" != 1 ]; then
	if [ ! -d "m4" ]; then
		mkdir m4
	fi

	autoreconf --install

	if [ "$DEBUG_MODE" == 1 ]; then
		echo "Compiling in debug and profiling mode..."
		CFLAGS="-O0 -p -pg -g" ./configure
	else
		./configure
	fi
fi

make clean
make
if [ -z "$TRANALYZER_INSTALL_DIR" ]; then
	TRANALYZER_INSTALL_DIR=$TRANALYZER_DEFAULT_DIR
	mkdir -p $TRANALYZER_INSTALL_DIR
fi

case $OSTYPE in 
	darwin*) 
		cp src/.libs/lib"$PLUGINNAME".0.dylib $HOME/.tranalyzer/plugins/$PLUGINORDER$PLUGINNAME.so
		;;
	linux*)
		cp src/.libs/lib"$PLUGINNAME".so.0.0.0 $TRANALYZER_INSTALL_DIR/$PLUGINORDER$PLUGINNAME.so
		;;
esac
echo "plugin copied into "$TRANALYZER_INSTALL_DIR

if [ -e "$TRANALYZER_INSTALL_DIR/portmap.txt" ]; then
	echo "portmap file already exists in "$TRANALYZER_INSTALL_DIR
else
	cp portmap.txt $TRANALYZER_INSTALL_DIR
	echo "portmap file copied into "$TRANALYZER_INSTALL_DIR
fi

