/*
 * portBasedClassifier.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "portBasedClassifier.h"

char* get_plugin_name() {
	return "portBasedClassifier";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
#ifdef __APPLE__
	pluginFolder = getPluginFolder();
#endif // __APPLE__
	pluginPortBasedClassifierIndex = pPluginIndex;
	FILE *port_file = NULL;
	char port_file_name[MAX_FILENAME_LEN + 1];
	char line[MAX_LINE_LEN];
	char appl_name[MAX_LINE_LEN];
	char appl_proto[MAX_LINE_LEN];
	uint32_t appl_port;
	
	//reset the memory
	memset(portClassificationArray,0,sizeof(t_portToAppl)*65536);

	/*Open the ports file*/

	if (pluginFolder) {
		strncpy(port_file_name, pluginFolder, MAX_FILENAME_LEN);
		if (port_file_name[strlen(port_file_name) - 1] != '/')
			strncat(port_file_name, "/", MAX_FILENAME_LEN - strlen(port_file_name));
	}
	else {
		strncpy(port_file_name, getenv("HOME"), MAX_FILENAME_LEN);
		if (port_file_name[strlen(port_file_name) - 1] != '/')
			strncat(port_file_name, "/", MAX_FILENAME_LEN - strlen(port_file_name));
		strncat(port_file_name, PLUGIN_FOLDER, MAX_FILENAME_LEN- strlen(port_file_name));
		if (port_file_name[strlen(port_file_name) - 1] != '/')
			strncat(port_file_name, "/", MAX_FILENAME_LEN - strlen(port_file_name));
	}
	strncat(port_file_name, PORT_BASED_CLASSIFICATION_FILE, MAX_FILENAME_LEN - strlen(port_file_name));

	if ((port_file = fopen(port_file_name, "r")) == NULL) {
		fprintf(stderr, "%s Error: Can't open file %s in read mode..\n",ERROR_PREFIX, port_file_name);
		go = 0;
		return;
	}

	/*Parse the Input*/

	while (fgets(line,MAX_LINE_LEN-1,port_file)!=NULL){
		line[MAX_LINE_LEN-1]='\0';

		//Discards line with comments or white lines
		if(line[0]!='#' && line[0]!=' ' && line[0]!='\n' && line[0]!='\t'){

			//scan the line
			sscanf(line, "%d\t%s\t%s\t", &appl_port, appl_proto, appl_name); //read corrected port key files
//			sscanf(line, "%s\t%d/%s\t", appl_name, &appl_port, appl_proto); // read original IANA files

 			if (appl_port >= 0 && appl_port < 65536) {
				//tcp
				if (strncmp("tcp", appl_proto, 3) == 0) {
					if (portClassificationArray[appl_port].tcp_appl_name[0]== '\0')
						strncpy(portClassificationArray[appl_port].tcp_appl_name,appl_name, 63);
					
				//udp
				} else if (strncmp("udp", appl_proto, 3) == 0) {
					if (portClassificationArray[appl_port].udp_appl_name[0]== '\0')
					strncpy(portClassificationArray[appl_port].udp_appl_name,appl_name, 63);
				//other:error. ignore it
				} else {
					fprintf(stderr, "%sError: proto invalid: %s",ERROR_PREFIX, appl_proto);
				}
			} else {
				fprintf(stderr, "%sError: port invalid: %d", ERROR_PREFIX,appl_port);
				go = 0;
				break; 
			} 
		}
	}
	fclose(port_file);
}

binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;
#if PBC_STRING == 1
	binary_value_t *act_bv = bv_new_bv("Port based classification of the destination port", "dstPort_class", 0, 1, bt_string_class);
#else // PBC_STRING == 0
	binary_value_t *act_bv = bv_new_bv("Port based classification of the destination port", "dstPort_class", 0, 1, bt_uint_16);
#endif // PBC_STRING
	main_bv = bv_append_bv(main_bv, act_bv);
	return main_bv;
}

void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	//Exit if the initialize function failed and we don't want to do any classification
	if (go == 0){
		outputBuffer_append(main_output_buffer, "-\0", 2);
		return;
	}

	flow_t *FlowP = &flows[flowIndex];
	char *protocol;
	uint16_t dstPort;

	if ( FlowP->status & L3FLOWINVERT ) dstPort = FlowP->srcPort;
	else dstPort = FlowP->dstPort;

	// assign text to ports

#if PBC_STRING == 1
	if (flows[flowIndex].layer4Protocol == L3_TCP) protocol = portClassificationArray[dstPort].tcp_appl_name;
	else if (flows[flowIndex].layer4Protocol == L3_UDP) protocol = portClassificationArray[dstPort].udp_appl_name;
	else protocol = NULL;
#endif // PBC_STRING

	// produce output for UDP and TCP
#if PBC_STRING == 1
	if (protocol != NULL && protocol[0] != 0) {
		outputBuffer_append(main_output_buffer, protocol, strlen(protocol));
		outputBuffer_append(main_output_buffer, "\0", 1);
	} else {
		outputBuffer_append(main_output_buffer, "unknown\0", 8);
	}
#else // PBC_STRING == 0
	outputBuffer_append(main_output_buffer, &dstPort, sizeof(uint16_t));
#endif // PBC_STRING
}

