/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "protocolStatistics.h"

char* get_plugin_name() {
	return "protocolStatistics";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
#ifdef __APPLE__
	baseFileName = getBaseFileName();
#endif // __APPLE
	char *temp;
	pluginProtocolStatistics = pPluginIndex;
	
	memset(numPackets, '\0', sizeof(numPackets));
	memset(tcpPortUsage, '\0', sizeof(tcpPortUsage));
	memset(udpPortUsage, '\0', sizeof(udpPortUsage));
	memset(scpPortUsage, '\0', sizeof(scpPortUsage));

	if (baseFileName) {
		temp = baseFileName;
	} else if (captureFileName) {
		temp = captureFileName;
	} else if (captureListFileName) {
		temp = captureListFileName;
	} else if (captureInterface) {
		temp = captureInterface;
	} else terminate();

	strncpy(protocolStatsFileName, temp, strlen(temp)+1);
	strncat(protocolStatsFileName, PROTOCOLS_SUFFIX, sizeof(PROTOCOLS_SUFFIX));

}

void claimLayer3Information(packet_t *packet) {
	numPackets[packet->layer4Type]++;
}

void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
#ifdef __APPLE__
    flows = getFlows();
#endif // __APPLE
	flow_t *flow = &flows[flowIndex];
	uint16_t dport;

	if (flow->status & L3FLOWINVERT) dport = flow->srcPort;
	else dport = flow->dstPort;

	// check for encapsulated packet
	switch (packet->layer4Type) {

		case 6: // TCP
			tcpPortUsage[dport]++;
			break;

		case 17: // UDP
			udpPortUsage[dport]++;
			break;

		case 132: //SCTP 
			scpPortUsage[dport]++;
			break;

		default:
			return; // no ports
	}
}

void onApplicationTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
    numOfV6Packets = getNumOfV6Packets();
    numOfV4Packets = getNumOfV4Packets();
#endif    
	FILE *ip_protfile, *l4_portfile;
	unsigned int i;
	unsigned long num;
	int z;
	char ip_protfilename[MAX_FILENAME_LEN+1], ip_protname[IPPROTMAX][IPPROTMAXLEN];
	char l4_portfilename[MAX_FILENAME_LEN+1], l4P[4], s[L4PORTMAXLEN];

        ip_protfilename[0] = '\0';
        strncpy(ip_protfilename, getenv("HOME"), MAX_FILENAME_LEN);
        strncat(ip_protfilename, "/", 1);
        strncat(ip_protfilename, PLUGIN_FOLDER, MAX_FILENAME_LEN - strlen(ip_protfilename));
        strncat(ip_protfilename, PROTOFILE, MAX_FILENAME_LEN - strlen(ip_protfilename));

	// Read Proto decoder file
	if ( (ip_protfile = fopen(ip_protfilename, "r")) == NULL ) {
		printf( "Protstat: Can't open file %s in read mode..\n", ip_protfilename );
        } else {
		for ( i = 0; i < IPPROTMAX; i++ ) z = fscanf(ip_protfile, "%lu\t%[^\n\t]\t%[^\n\t]", &num, s, ip_protname[i]);
		fclose(ip_protfile);	
	}

        l4_portfilename[0] = '\0';
        strncpy(l4_portfilename, getenv("HOME"), MAX_FILENAME_LEN);
        strncat(l4_portfilename, "/", 1);
        strncat(l4_portfilename, PLUGIN_FOLDER, MAX_FILENAME_LEN - strlen(l4_portfilename));
        strncat(l4_portfilename, PORTFILE, MAX_FILENAME_LEN - strlen(l4_portfilename));

	// Read Proto decoder file
	if ( (l4_portfile = fopen(l4_portfilename, "r")) == NULL ) {
		printf( "Protstat: Can't open file %s in read mode..\n", l4_portfilename );
        } else {
		do {  
			z = fscanf(l4_portfile, "%u\t%[^\n\t]\t%[^\n\t]\t%[^\n\t]", &i, l4P, s, s);
			if (strncmp("udp", l4P, 3) == 0) strncpy(l4_portname[0][i], s, L4PORTMAXLEN-1);
			else if (strncmp("tcp", l4P, 3) == 0) strncpy(l4_portname[1][i], s, L4PORTMAXLEN-1);
		} while (z > 0);
		fclose(l4_portfile);	
	}

	// open protocol statistics file
	if ((protocolStatsFile = fopen(protocolStatsFileName, "w")) == NULL) {
		printf("Protstat: Can't open file %s in write mode\n", protocolStatsFileName);
		return;
	}

#if IPV6_ACTIVATE == 1
 	i = 6;
	num = numOfV6Packets;
#else // IPV6_ACTIVATE == 0
	i = 4;
	num = numOfV4Packets;
#endif // IPV6_ACTIVATE
	fprintf(protocolStatsFile, "Total IPv%d packets captured: %lu\n", i, num);
	fprintf(protocolStatsFile, "# L4 Protocol\t# Packets\tRelative Frequency[%%]\tProtocol description\n\n");
	

	// print protocol usage
	for ( i = 0; i < IPPROTMAX; i++ ) {
		if ( numPackets[i] > 0 ) fprintf( protocolStatsFile, "%u\t%lu\t%11f\t%s\n", i, numPackets[i], 100.0f / (double)num * numPackets[i], ip_protname[i] );
	}

	// print tcp port usage
	fprintf(protocolStatsFile, "\n\n\nTotal TCP packets: %lu\n", numPackets[6]);
	fprintf(protocolStatsFile, "# Port\t# Packets\tRelative Frequency[%%]\tProtocol description\n\n");
	for ( i = 0; i < MAX_PORTS; i++ ) {
		if ( tcpPortUsage[i] > 0 ) fprintf(protocolStatsFile, "%u\t%lu\t%11f\t%s\n", i, tcpPortUsage[i], 100.0f/(double)numPackets[6] * tcpPortUsage[i], l4_portname[1][i]);
	}

	// print udp port usage
	fprintf(protocolStatsFile, "\n\n\nTotal UDP packets: %lu\n", numPackets[17]);
	fprintf(protocolStatsFile, "# Port\t# Packets\tRelative Frequency[%%]\tProtocol description\n\n");
	for ( i = 0; i < MAX_PORTS; i++ ) {
		if ( udpPortUsage[i] > 0 ) fprintf(protocolStatsFile, "%u\t%lu\t%11f\t%s\n", i, udpPortUsage[i], 100.0f/(double)numPackets[17] * udpPortUsage[i], l4_portname[0][i]); 
	}
	
	// print sctp port usage
	fprintf(protocolStatsFile, "\n\n\nTotal SCTP packets: %lu\n", numPackets[132]);
	fprintf(protocolStatsFile, "# Port\t# Packets\tRelative Frequency[%%]\tProtocol description\n\n");
	for ( i = 0; i < MAX_PORTS; i++ ) {
		if (scpPortUsage[i] > 0) fprintf(protocolStatsFile, "%u\t%lu\t%11f\t%s\n", i, udpPortUsage[i], 100.0f/(double)numPackets[132] * scpPortUsage[i], l4_portname[1][i]); 
	}

	fclose(protocolStatsFile);
}
