/*	regex_pattern_matcher.h
*	Mischa Lehmann
*	ducksource@duckpond.ch
*	Version:1.0
*
*	regex_pattern_matcher
*	Require:
*		- stdio.h
*		- stdlib.h
*		- string.h
*		- pcre.h
*		- tcp.h
*		- netinet/in.h
*		- packetCapture.c
*		- sys/stat.h
*		- sys/types.h
*		- stdbool.h 
*               
*
*
*	Licence:
*	You're allowed to edit and publish my source in all of your free and open-source projects.
*	Please send me an e-mail if you'll implement this source in one of your commercial or proprietary projects.
*	Leave this Header untouched!
*
*	Warranty:
*	Warranty void if signet is broken
*	=================== / /===================
*	[       Warranty   / /   Signet          ]
*	==================/ /=====================
*	!!NOWAY!!
*
*/

/*==============================*/
/*             TODO's           */
/*==============================*/
/*
*
*	- no globals -> Tranalyzer 3 ready
*	- masks -> become unsigned int's more effective
*
*
*
*/

/*==============================*/
/*           includes           */
/*==============================*/
#ifndef _REGEX_PATTERN_MATCHER_H

#define _REGEX_PATTERN_MATCHER_H 1

/*all Libraries*/
// global includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pcre.h>

#include <netinet/tcp.h>
#include <netinet/in.h>
#include <packetCapture.c>       

#include <sys/stat.h>
#include <sys/types.h>

#include <stdbool.h> // bool

// local includes
#include "utils/utils.h"
#include "networkHeaders.h"
#include "main.h"
#include "outputBuffer.h"
#include "binaryValue.h"

/*==============================*/
/*enums declaration + definition*/
/*==============================*/

#define REGEX_PLUGINNAME 		"REGEX_PATTERN_MATCHER"			// the plugin name
#define REGEX_LONG_NAME			"Regex pattern matcher"			// long name of the plugin
#define REGEX_SHORT_NAME		"regexp"				// short name of the plugin
#define REGEX_PLUGIN_VERSION		"0.2.3"					// the current plugin version
#define REGEX_CONFIG_FILE 		"/.tranalyzer/plugins/regex_config.txt"	// the config file
#define REGEX_FLOWDATA_FILES 		"/.tranalyzer/flowdata/"		// directory for flowdumps (see REGEX_FILE_FORMAT_DUMP)
#define REGEX_FLOWDATA_FILEPREFIX 	"flow_"					// flowdump prefix
#define REGEX_FLOWDATA_FILESUFFIX 	".bin.XXXXXX"				// flowdump suffix (replace XXXXXX with random characters)
#define REGEX_COMPILE_FLAGS 		0x0 					// default flags which are used during regex compilation
#define REGEX_FILE_FORMAT_RULEDEF 	"#%u;%[^\n]" 				// Note: first char has to be a unique identifyer
#define REGEX_FILE_FORMAT_DUMP 		"dump:%u"				// dump the flow on rulematch
#define REGEX_FILE_FORMAT_COMPILE	"compile:%x"				// compile the rule with the given options
#define REGEX_FILE_FORMAT_STUDY		"study:%x"				// study the rule with the gien opetions
#define REGEX_FILE_FORMAT_CMD 		"cmd:%[^\n]"				// execute command on flow terminate after match
#define REGEX_FILE_FORMAT_PREV 		"prev:%u"				// dependency -> require previous matched rule

#ifndef REGEX_DEBUG
#define REGEX_DEBUG 0								// DEBUG_FLAG 1:= MAKE DEBUG PRINTS, 0:= no debug
#endif

/*==============================*/
/*     global declarations      */
/*==============================*/

// A regex flow
typedef struct rexFlow_s {
	char *idstr;				// id of flow as a string (for string replace)
	char *mask;				// mask of the flow -> for previous pattern match check
	bool dumpToFile;			// dump all the packets of flow to file
	unsigned int patternMatchCount;		// Count of matched patterns for this flow
	bool flowOutFile_init;			// is the flowoutfile initialized
	char *flowOutFile_path;			// path to the flow out file
	FILE *flowOutFile;			// the file descriptor for the flow out file
}rexFlow_t;

// a single pattern
typedef struct pattern_s{
	unsigned int id;			// the unique rule number
	size_t regex_len;			// length of the regex pattern + trailing '\0'
	char *regex;				// the pattern as a string
	bool regexCompiled_init;		// is the compiled regex ready to use?
	pcre *regexCompiled;			// the compiled pattern 
	char *mask;				// the mask to apply to flow after this rule has matched
	char *prev_mask;			// previous mask which has to match in order to make this rule match
	bool studyRegex_init;			// has a studied regex pattern?
	pcre_extra *studyRegex;			// the studied pattern
	bool command_init;			// is there a valid command?
	char *command;				// linked command
	size_t command_len;			// linked command len + trailing '\0'
	bool dodump;				// a match on this regex causes a dump
}pattern_t;

// regex engine data
typedef struct regexdata_s{
	bool initialzed;			// is the plugin initialized?
	rexFlow_t *rexFlowTable;		// all the regex flows
	unsigned int patterns_c;		// amount of loaded patterns
	unsigned int max_patterns_c;		// maximum number of patterns which the patterns array can hold
	pattern_t *patterns;			// all the patterns
	unsigned int mask_size;			// size of mask in bytes
	char *flowfiles_path;			// destination for all the output flowfiles
	unsigned int flowfiles_path_len;	// path len of destination folder
	char *regexfile_path;			// file of the regex rules
	unsigned int regexfile_path_len;	// path len of destination folder
}regexdata_t;

// Globals
unsigned long plugin_REGEX_PLUGINNAME_Index;	// the plugin Index (TODO: do we really need a long type here?)
regexdata_t regexData;				// All the relevant plugin data

/*==============================*/
/*    function declarations     */
/*==============================*/
/*
*       funcname(args): <desc>
*       Arguments:
*               - <arg1> : <desc1> 
*/

/*
*       Tranalyzer plugin functions
*
*/
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
char* get_dependencies();
void initialize();
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer2Information(packet_t* packet);
void claimLayer3Information(packet_t* packet);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();
void bufferToSink(outputBuffer_t *buffer);

#endif
