/*	utils.h
*	Mischa Lehmann
*	ducksource@duckpond.ch
*	Version:1.0
*
*	Utilities for daily work
*	Require:
*		- stdlib.h
*		- stdio.h
*		- string.h
*		- math.h
*		- stdbool.h
*		- sys/stat.h
*		- sys/types.h
*
*
*	Licence:
*	You're allowed to edit and publish my source in all of your free and open-source projects.
*	Please send me an e-mail if you'll implement this source in one of your commercial or proprietary projects.
*	Leave this Header untouched!
*
*	Warranty:
*	Warranty void if signet is broken
*	=================== / /===================
*	[       Warranty   / /   Signet          ]
*	==================/ /=====================
*	!!NOWAY!!
*
*/

/*==============================*/
/*             TODO's           */
/*==============================*/
/*
*
*
*
*
*
*
*/

/*==============================*/
/*           includes           */
/*==============================*/
#ifndef _UTILS_H
	#define _UTILS_H 1

	/*all Libraries*/
	#include <stdlib.h>
	#include <stdio.h>
	#include <string.h>
	#include <math.h>
	#include <stdbool.h>
	#include <sys/stat.h>
	#include <sys/types.h>


/*==============================*/
/*         Cool stuff           */
/*==============================*/
#ifndef CURSystem
	#ifdef __unix__ 
		/*Unix-Systes*/
		#define CURSystem printf("Hello Unix-user\n You're elite!\n");
	#elif __BORLANDC__ && __MSDOS__
		/*Borland c-compiler and MSDOS*/
		#define CURSystem printf("Hello BorlandC-complier or MSDOS user!\n");
	#elif __WIN32__ || _MSC_VER
		/*Windows 32-Syste or visual C*/
		#define CURSystem printf("Hello windows user!\n");
	#else
		#define CURSystem printf("This programm is not compatible with your System, sorry -> EXIT\n");return EXIT_FAILURE;
	#endif
#endif

// stringification
#define xstr(a) str(a)
#define str(a) #a

// printf formatting
#define printf(format, args...) printf("%s> " format, __FILE__, ##args)
/*==============================*/
/*enums declaration + definition*/
/*==============================*/

/*==============================*/
/*     global declarations      */
/*==============================*/

/*==============================*/
/*    function declarations     */
/*==============================*/

/*
*	funcname(args): <desc>
*	Arguments:
*		- <arg1> : <desc1> 
*/

/*
*	printerr(char *errmsg,int errcode): Prints an error
*	Arguments:
*		errmsg = the message
*		errcode = the error code
*/
void printerr(char *errmsg,int errcode); // Error-Printing
#define printerr(errmsg, errcode) printerr(__FILE__ "("xstr(__LINE__)")> "errmsg,errcode)

/*
*	util_isbigendian(): Deremits of the systems endianness
*	Arguments:
*	Return: true = is big endian system
*		false = is little endian system
*/
bool util_isbigendian();

/*
*	void util_convertendian(void *data, size_t size): switch between endian
*	Arguments:
*		- data : the switching data
*		- size : size of the switching data 
*/
void util_convertendian(void *data, size_t size);

/*
*	int util_write_to_file(FILE *fd,void *data,size_t size,bool save_bigendian): saves data to a file (swtich endian if needed)
*	Arguments:
*		- fd: the filedescriptor
*		- data: the data
*		- size: size of the data (in bytes)
*		- save_bigendian: if true saves in bigendian if not saves in little-endian
*/
int util_write_to_file(FILE *fd,void *data,size_t size,bool save_bigendian);


/*
*	int util_read_from_file(FILE *fd,void* data,size_t size,bool read_bigendian): read data from a file (swtich endian if needed)
*	Arguments:
*		- fd: the filedescriptor
*		- data: the data
*		- size: size of the data (in bytes)
*		- read_bigendian: if true data read is in bigendian if not data is in little-endian
*/
int util_read_from_file(FILE *fd,void* data,size_t size,bool read_bigendian);

/*
*	util_mk_2dmatrix: malloc a 2d matrix
*	Arguments:
*		- l1_size : size of first dimension
*		- l2_size : size of second dimension
*		- element_size : size of an element of the matrix
*		- matrix : pointer to the returning matrix
*/
inline int util_mk_2dmatrix(unsigned int l1_size,unsigned int l2_size,size_t element_size,void ***matrix);

/*
*	util_set_bit(field,state,bitnum): set a bit in a field of one character to the given state
*	Arguments:
*		- field : the affected field
*		- state : the new bit state true := 1, false := 0
*		- bitnum : the number of the bit
*/
inline int util_set_bit(char *field,bool state,unsigned char bitnum);

/*
*	util_set_bit_mask(mask,state,bitnum): seta bit in a maks of x characters to the given state (make sure that mask is big enough)
*	Arguments:
*		- mask : the affected mask
*		- state : the new bit state true := 1, false := 0
*		- bitnum : the number of the bit
*/
int util_set_bit_mask(char *mask,bool state,unsigned int bitnum);

/*
*	util_or_bit_mask(mask,mask2,ret_mask,mask_size): Do or operation for each byte of mask and mask2 return result in ret_mask
*	Arguments:
*		- mask : mask 1 for operation
*		- mask2 : mask 2 for operation
*		- ret_mask : mask1 | mask2
*		- mask_size : size of all the three masks or at least size of the smallest mask
*/
inline int util_or_bit_mask(char *mask,char *mask2,char *ret_mask,unsigned int mask_size);

/*
*	util_cmp_bit_mask(mask,cmp_mask,mask_size): Check if all the bits which are set in cmp_mask are also set in mask
*	Arguments:
*		- mask : the origianal mask to compare
*		- cmp_mask : the masking, mask. Check if all the bits which are set in this mask are also set in the other mask
*		- mask_size : size of the both masks or at least size of the smaller mask
*/
bool util_cmp_bit_mask(char *mask,char *cmp_mask,unsigned int mask_size);

/*
*	util_dump_bit_mask(mask,mask_size): dump (print) a bit mask of the given size
*	Arguments:
*		- mask : the mask to dump
*		- mask_size : the size of the mask
*/
inline void util_dump_bit_mask(char *mask,unsigned int mask_size);

/*
*	util_mk_progress(count,max,print_percent): print progress
*	Arguments:
*		- count : number of element
*		- max : maximum number of elements
*		- print_percent : print every 'print_percent'
*/
void util_mk_progress(unsigned int count,unsigned int max,unsigned short print_percent);

/*
*	util_recursive_mkdir(dir): create a directory and all it's parents
*	Arguments:
*		- dir : the directory tree to create
*/
void util_recursive_mkdir(const char *dir);
#endif
