/*
 * Snort Regex and Scan detector 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
 */

#include "regex_pcre.h"
//#include "tcpFlags.h"

char* get_plugin_name() {
	return "regex_pcre";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

/*char* get_dependencies() {
        return "tcpFlags";
}*/

void initialize(unsigned long pPluginIndex){
	char filename[256];
	
	plugin_PLUGINNAME_Index = pPluginIndex;
	rexFlowTable = (rexFlow_t*)malloc(mainHashMap->hashChainTableSize * sizeof(rexFlow_t));
        
	memset(filename,'\0', 256);
        strncpy(filename, getenv("HOME"),255);
        strncat(filename, "/", 1);
        strncat(filename, PLUGIN_FOLDER, 255 - strlen(filename));
 //       strncat(filename, "/", 1);
        strncat(filename, REXPOSIX_FILE, 255 - strlen(filename));

	rex_tableP = (rex_table_t*)malloc(sizeof(rex_table_t));
	if ( rex_load( filename, rex_tableP) ) exit(-1);
	

}

binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;

	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp match count", "RgxCnt", 0, 1, bt_uint_16));
#if EXPERTMODE == 1
#if PKTTIME == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp: time_Pkt_Byte Position_Regfile ID_Snort ID_Flags_Classtyp_Severity Match", "Rgx:t_N_B_RID_Amsk_F_CT_Sv", 1, 8, bt_unix_time, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_16, bt_hex_8, bt_uint_8, bt_uint_8));
#else // PKTTIME == 0
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp: Pkt_Byte Position_Regfile ID_Snort ID_Flags_Classtyp_Severity Match", "Rgx:N_B_RID_Amsk_F_CT_Sv", 1, 7, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_16, bt_hex_8, bt_uint_8, bt_uint_8));
#endif // PKTTIME
#else // EXPERTMODE == 0
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp classtype ", "Rgx_ClTyp", 0, 1, bt_uint_8));
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp severity ", "Rgx_sev", 0, 1, bt_uint_8));
#endif // EXPERTMODE
//	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp classtype_severity ", "Rgx_ClTyp_sev", 0, 2, bt_uint_8, bt_uint_8));

	return main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
	rexFlow_t *rexFlowP = &(rexFlowTable[flowIndex]);
	memset(rexFlowP, '\0', sizeof(rexFlow_t));	
}




void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	rexFlow_t *rexFlowP = &(rexFlowTable[flowIndex]);
	if ( rexFlowP->count >= MAXREGPOS ) return;

	flow_t *currFlowP = &flows[flowIndex];
	int ovector[OVECCOUNT]; 
	register unsigned int i, k, l, j, icount, istate;
	uint16_t length, hdrSel[4];
	char *regstart;
       
	regstart = (char*)packet->layer7Header;
	length = packet->snapL7Length; 

#if IPV6_ACTIVATE == 1
	if ( length < 40 ) return;  
#else // IPV6_ACTIVATE == 0
	if ( length < 20 ) return;  
#endif // IPV6_ACTIVATE 
	rexFlowP->pktN++;
	icount = rexFlowP->count;
	hdrSel[0] = currFlowP->findex;
	hdrSel[1] = currFlowP->layer4Protocol;
	hdrSel[2] = currFlowP->srcPort;
	hdrSel[3] = currFlowP->dstPort;
		
	// match all loaded patterns
	for ( i = 0; i < rex_tableP->count; i++ ) {
		k = 0x08;
		l = rex_tableP->hdrSel[4][i];
		for (j = 0; j <= 3; j++) {
			if ( ( l & k) && (hdrSel[j] != rex_tableP->hdrSel[j][i]) ) 
				goto nxtrex; 
			k >>= 1;
		} 

		switch ( l & 0xf0 ) {
			case 0x80:
				regstart = (char*)packet->layer7Header + rex_tableP->offset[i];
				length = packet->snapL7Length - rex_tableP->offset[i]; 
				break;
			case 0x40:
				regstart = (char*)packet->layer4Header + rex_tableP->offset[i];
				length = packet->snapL4Length - rex_tableP->offset[i]; 
				break;
			case 0x20:
				regstart = (char*)packet->layer3Header + rex_tableP->offset[i];
				length = packet->snapL3Length - rex_tableP->offset[i]; 
				break;
			case 0x10:
				regstart = (char*)packet->layer2Header + rex_tableP->offset[i];
				length = packet->snapL2Length - rex_tableP->offset[i]; 
				break;
		}
 

#if RULE_OPTIMIZE == 1
		if ( pcre_exec(rex_tableP->compRex[i], rex_tableP->studyRex[i], regstart, length, 0, 0, ovector, OVECCOUNT ) < 0 ) continue;
#else // RULE_OPTIMEZE == 0
		if ( pcre_exec(rex_tableP->compRex[i], 0, regstart, length, 0, 0, ovector, OVECCOUNT ) < 0 ) continue;
#endif // RULE_OPTIMIZE  
		if ( ovector[1] <= ovector[0] ) continue;

		istate = rex_tableP->isstate[i];
		if ( istate ) {
			j = 0;
			for ( l = 0; l < icount; l++ ) {
				if ( rexFlowP->pregID[l] == istate ) { 
					j |= rexFlowP->andPin[l];
					if ( j != rex_tableP->andPin[i] ) continue;
					else {	
						rexFlowP->flags[l] |= REG_F_PRE;
//						if ( rexFlowP->flags[l] & REG_F_PRES ) l = icount; 
						goto predfound;
					} 
				} 
			}
			continue;
		} 

predfound:	l = icount;

#if EXPERTMODE == 1
#if PKTTIME == 1
		rexFlowP->time[l] = currFlowP->lastSeen;
#endif //  PKTTIME == 1	
		rexFlowP->andMsk[l] = rex_tableP->andMsk[i];
		rexFlowP->andPin[l] = rex_tableP->andPin[i];
		rexFlowP->pregPos[l] = ovector[0];
		rexFlowP->pkt[l] = rexFlowP->pktN;
#endif // EXPERTMODE == 1
		rexFlowP->flags[l] |= rex_tableP->flags[i];
		rexFlowP->pregID[l] = rex_tableP->class[i];
		rexFlowP->alarmcl[l] = rex_tableP->alarmcl[i];
		rexFlowP->severity[l] = rex_tableP->severity[i];
				
		if ( ++icount >= MAXREGPOS ) break;

nxtrex:		continue;
	}

	rexFlowP->count = icount;

}


void onFlowTerminate(unsigned long flowIndex) {
	rexFlow_t *rexFlowP = &(rexFlowTable[flowIndex]);
	uint32_t i = 0, k = 0, j = rexFlowP->count;

#if LABELSCANS == 1
	rexFlow_t *rexFlowPO = NULL;
	tcpFlagsPerFlow_t* tcpFlagsP = &tcpFlagsPerFlow[flowIndex];

	if ( flows[flowIndex].oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) { 
		rexFlowPO = &rexFlowTable[flows[flowIndex].oppositeFlowIndex];
		
		if ( ( rexFlowP->pktN < 3  && rexFlowPO->pktN < 2) && ( tcpFlagsP->tcpFlagsT == 0x02 || tcpFlagsP->tcpFlagsT == 0x06 || (tcpFlagsP->tcpSpecialFlagsT & SCANMASK) ) && !(tcpFlagsP->tcpSpecialFlagsT & TCPRETRIES) ){
			rexFlowP->alarmcl[j] = 102; // response to scan
			rexFlowP->severity[j] = 1;
			j++;
		}
	//	printf ("Flags: 0x%02x\t0x%04x\t%d\t%d\t%d_%d\n",tcpFlagsP->tcpFlagsT, tcpFlagsP->tcpSpecialFlagsT, rexFlowP->pktN, rexFlowPO->pktN, rexFlowP->alarmcl[j], rexFlowP->severity[j] );	
	} else {
		if ( ( rexFlowP->pktN < 2 ) && ( tcpFlagsP->tcpFlagsT == 0x02 || tcpFlagsP->tcpFlagsT == 0x06 || (tcpFlagsP->tcpSpecialFlagsT & SCANMASK) ) && !(tcpFlagsP->tcpSpecialFlagsT & TCPRETRIES) ) {
			rexFlowP->alarmcl[j] = 101; // no response to scan
                        rexFlowP->severity[j] = 1;
                        j++;
		}
	}
#endif // LABELSCANS
 
#if EXPERTMODE == 0
	uint32_t sevim = 0, sevmax = rexFlowP->severity[0];
	for ( i = 1; i < j; i++ ) {
		if ( rexFlowP->severity[i] > sevmax ) sevim = i;   
	}
	i = sevim;
#else //EXPERTMODE == 1

#if ALARMS == 1
	for ( i = 0; i < j; i++ ) {
		if ( rexFlowP->flags[i] & REG_F_ALRM ) k++;
	}

	outputBuffer_append(main_output_buffer, (char*) &k, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &k, sizeof(uint32_t));
#else // ALARMS == 0
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
#endif // ALARMS

	for (i = 0; i < j; i++) {
#if ALARMS == 1
		if ( rexFlowP->flags[i] & REG_F_ALRM ) {
#endif // ALARMS

#if PKTTIME == 1
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->time[i].tv_sec), sizeof(uint64_t));
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->time[i].tv_usec), sizeof(uint32_t));
#endif // PKTTIME == 1
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->pkt[i]), sizeof(uint16_t));
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->pregPos[i]), sizeof(uint16_t));
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->pregID[i]), sizeof(uint16_t));
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->andMsk[i]), sizeof(uint16_t));
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->flags[i]), sizeof(uint8_t));

#endif //EXPERTMODE 
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->alarmcl[i]), sizeof(uint8_t));
			outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->severity[i]), sizeof(uint8_t));
#if EXPERTMODE == 1
#if ALARMS == 1
		}
#endif // ALARMS
	}
#endif //EXPERTMODE == 1

}

void onApplicationTerminate() {
	
	if (rexFlowTable) free(rexFlowTable);
	rexFlowTable = NULL;

	if (rex_tableP) {
#if EXPERTMODE == 1
		free(rex_tableP->class);
		free(rex_tableP->isstate);
		free(rex_tableP->flags);
		free(rex_tableP->compRex[0]);
		free(rex_tableP->compRex[1]);
#endif //EXPERTMOD == 1
		free(rex_tableP->alarmcl);
		free(rex_tableP->severity);
		free(rex_tableP);
	}
	rex_tableP = NULL;
	
}

