/*
 * Snort Regex and Scan detector 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
 */

#ifndef REGEX_PCRE_H
#define REGEX_PCRE_H

// global includes

// local includes
#include "main.h"
#include "regfile_pcre.h"

// User configuration 
#define	EXPERTMODE	1	//
#define	PKTTIME		0	//
#define LABELSCANS	0	//
#define ALARMS		1	//

// defines Regex
#define OVECCOUNT	3	// value % 3
#define REGSTART	0	// default 0, not used
 
#define MAXREGPOS	30	// Maximal # of matches stored / flow
#define REGOPT		0	// 0: No optimization, 1: execute regex optimization, requires RULE\_OPTIMIZE = 1 in regfile_snort_pcre.h
#define REXPOSIX_FILE "regexfile.txt"	// regexfile name under .tranalyzer/plugins

// defines scans
#define SCANMASK 0x00ff 
#define TCPRETRIES 0x0100 

// future flag definition
#define REG_F_PRE	0x10	//	
#define REG_F_SUCC	0x20	//	
#define REG_F_PRES	0x40	//
#define REG_F_ALRM	0x80	//


// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
char* get_dependencies();
void initialize();
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

// variables
unsigned long plugin_PLUGINNAME_Index;

rex_table_t *rex_tableP;

typedef struct rexFlow_s {
	uint16_t pregID[MAXREGPOS]; 	// pattern match Regex ID 
	uint16_t andMsk[MAXREGPOS]; 	// and mask
	uint16_t andPin[MAXREGPOS]; 	// and mask
	uint16_t pktN;
#if EXPERTMODE == 1
#if PKTTIME == 1
	struct timeval time[MAXREGPOS];
#endif // PKTTIME == 1
	uint16_t pregPos[MAXREGPOS];
	uint16_t pkt[MAXREGPOS];
#endif // EXPERTMODE == 1
	uint16_t count;
	uint8_t alarmcl[MAXREGPOS];
	uint8_t flags[MAXREGPOS];
	uint8_t severity[MAXREGPOS];
}rexFlow_t;

rexFlow_t *rexFlowTable;


#endif /* regexp_pattern_matcher_H_ */

