/*
 * Snort Regex and Scan detector 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
 */

#include "regfile_pcre.h"
#include <string.h>

int16_t rex_load(char *rex_file, rex_table_t *preg_table) {
	FILE *regexp_fp;
	int cl, z, isstate, flags, erroffset, andMsk, andPin, alarmcl, severity, hdrSel, findex, proto, srcPort, dstPort, offset;	
	uint32_t i, j = 0, k, pcount = 0, mode;
	char regline[MAXREGEXPR];
	const char *errPtr;

	preg_table->count = 0;
	
	// open regex-file
	if( (regexp_fp = fopen(rex_file,"r") ) == NULL){
		printf("Regex - file error: %s\n", rex_file);
		return -1;
	}
	
	while ( fgets(regline, MAXREGEXPR, regexp_fp) ) {
		if ( regline[0] == '#' || regline[0] == '\n' || regline[0] == ' ' ) continue;
		else pcount ++; 
	}
	rewind(regexp_fp);
 
	preg_table->count = pcount;
	preg_table->compRex = malloc(sizeof(size_t*) * pcount);
	preg_table->class = malloc(sizeof(uint16_t) * pcount);
	preg_table->andMsk = malloc(sizeof(uint16_t) * pcount);
	preg_table->andPin = malloc(sizeof(uint16_t) * pcount);
	preg_table->isstate = malloc(sizeof(uint16_t) * pcount);
	preg_table->flags = malloc(sizeof(uint8_t) * pcount);
	preg_table->alarmcl = malloc(sizeof(uint8_t) * pcount);
	preg_table->severity = malloc(sizeof(uint8_t) * pcount);
#if RULE_OPTIMIZE == 1
	preg_table->studyRex = malloc(sizeof(size_t*) * pcount);
#endif // RULE_OPTIMIZE == 1
	for ( i = 0; i < HDRSELMX; i++ ) {
		preg_table->hdrSel[i] = malloc(sizeof(uint16_t) * pcount);
	}
	preg_table->offset = malloc(sizeof(uint16_t) * pcount);

	while ( fgets(regline, MAXREGEXPR, regexp_fp) ) {
		
		if ( regline[0] == '#' || regline[0] == '\n' || regline[0] == ' ' ) continue;

		z = sscanf(regline, "%d\t%d\t%*2c%02x\t%*2c%04x\t%*2c%04x\t%d\t%d\t%*2c%02x\t%*2c%04x\t%d\t%d\t%d\t%d\t%[^\n\t]", &cl, &isstate, &flags, &andMsk, &andPin, &alarmcl, &severity, &hdrSel, &findex, &proto, &srcPort, &dstPort, &offset, &regline[0]);
		if (z == 0) { 
			printf("Regex_pcre error: %d\t%d\t0x%02x\t0x%04x\t0x%04x\t%d\t%d\t0x%02x\t0x%04x\t%d\t%d\t%d\t%d\t%s\n", cl, isstate, flags, andMsk, andPin, alarmcl, severity, hdrSel, findex, proto, srcPort, dstPort, offset, regline );
			exit(-1);
		}

		mode = ((flags & 0x0f) | REGEX_MODE);

		if ( ( preg_table->compRex[j] = pcre_compile( regline, mode, &errPtr, &erroffset, NULL ) ) ) {
#if RULE_OPTIMIZE == 1
			preg_table->studyRex[j] = pcre_study( preg_table->compRex[j], 0, &errPtr );
			if(errPtr != NULL) {
                               	printf("Regex_pcre study rule ignored: # %d, @ %d, Error: %s\n", i+1, cl, errPtr );
                               	continue;
                       	}
#endif // RULE_OPTIMIZE == 1
			preg_table->class[j] = (uint16_t)cl;
			preg_table->isstate[j] = (uint16_t)isstate;
			preg_table->flags[j] = (uint8_t)flags;
			preg_table->andMsk[j] = (uint16_t)andMsk;
			preg_table->andPin[j] = (uint16_t)andPin;
			preg_table->alarmcl[j] = (uint8_t)alarmcl;
			preg_table->severity[j] = (uint8_t)severity;
			preg_table->hdrSel[0][j] = (uint16_t)findex;
			preg_table->hdrSel[1][j] = (uint16_t)proto;
			preg_table->hdrSel[2][j] = (uint16_t)srcPort;
			preg_table->hdrSel[3][j] = (uint16_t)dstPort;
			preg_table->hdrSel[4][j] = (uint16_t)hdrSel;
			preg_table->offset[j] = (uint16_t)offset;
			j++;
			
		} else {
                       	printf("PCRE ignored - # %d, class %d, @ %d, Error: %s\n", i+1, cl, erroffset, errPtr );
			continue;
		}
	}

	fclose(regexp_fp);

	preg_table->count = j;
	printf(" # of valid rules: %d\n",j);
	if (j < pcount) {
		printf("Regex: %d rules have no predecessor\n", pcount-j);
		return -2;
	}
	else return 0; 
}
