/*
 * Snort Regex and Scan detector 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
 */

#include "regexp_snort_pcre.h"
//#include "tcpFlags.h"

char* get_plugin_name() {
	return "regexp_snort_pcre";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

/*char* get_dependencies() {
        return "tcpFlags";
}*/

void initialize(unsigned long pPluginIndex){
	char filename[256];
	
	plugin_PLUGINNAME_Index = pPluginIndex;
	rexFlowTable = (rexFlow_t*)malloc(mainHashMap->hashChainTableSize * sizeof(rexFlow_t));
        
	memset(filename,'\0', 256);
        strncpy(filename, getenv("HOME"),255);
        strncat(filename, "/", 1);
        strncat(filename, PLUGIN_FOLDER, 255 - strlen(filename));
 //       strncat(filename, "/", 1);
        strncat(filename, REXPOSIX_FILE, 255 - strlen(filename));

	rex_tableP = (rex_table_t*)malloc(sizeof(rex_table_t));
	if( rex_load( filename, rex_tableP) < 0 ) {
		printf("Error: No Rules found\n");
	}

}

binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;

	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp match count", "Rgx_Cnt", 0, 1, bt_uint_16));
#if EXPERTMODE == 1
#if PKTTIME == 1
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp: time_Pkt_Byte Position_Regfile ID_Snort ID_Flags_Classtyp_Severity Match", "Rgx:t_N_B_RID_SID_F_CT_Sv", 1, 8, bt_unix_time, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_32, bt_hex_8, bt_uint_8, bt_uint_8));
#else // PKTTIME == 0
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp: Pkt_Byte Position_Regfile ID_Snort ID_Flags_Classtyp_Severity Match", "Rgx:N_B_RID_SID_F_CT_Sv", 1, 7, bt_uint_16, bt_uint_16, bt_uint_16, bt_uint_32, bt_hex_8, bt_uint_8, bt_uint_8));
#endif // PKTTIME
#else // EXPERTMODE == 0
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp/Scan classtype ", "Rgx_ClTyp", 0, 1, bt_uint_8));
	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp/Scan severity ", "Rgx_sev", 0, 1, bt_uint_8));
#endif // EXPERTMODE
//	main_bv = bv_append_bv(main_bv, bv_new_bv("Regexp/Scan classtype_severity ", "Rgx_ClTyp_sev", 0, 2, bt_uint_8, bt_uint_8));

	return main_bv;
}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
	rexFlow_t *rexFlowP = &(rexFlowTable[flowIndex]);
	memset(rexFlowP, '\0', sizeof(rexFlow_t));	
}




void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	rexFlow_t *rexFlowP = &(rexFlowTable[flowIndex]);
	if ( rex_tableP->count < 1 ) return; // error in regex, stop processing

	flow_t *currFlowP = &flows[flowIndex];
	int ovector[OVECCOUNT]; 
	unsigned int i, j = rexFlowP->count, k, l;
	uint16_t length[2], istate;
	char *regstart[2];
	char proto = packet->layer3Header->ipHeader.ip_p;
        length[1] = packet->snapL7Length; 
	length[0] = packet->snapL3Length - length[1]; 
	regstart[0] = (char*)packet->layer3Header;
        regstart[1] = (char*)packet->layer7Header;
	rexFlowP->pktN++;

	if ( j >= MAXREGPOS || length[0] < 20 || length[1] < 1 ) return; 

		// match all loaded patterns
	for ( i = 0; i < rex_tableP->count && rexFlowP->count < MAXREGPOS; i++ ) {
		k = rex_tableP->class[i] / 10000 - 1;
		if ( k >= PROTOMAX || proto != protocl[k] ) continue; 

		for ( k = 0; k <= 1; k++ ) {
#if RULE_OPTIMIZE == 1
			if ( pcre_exec(rex_tableP->compRex[k][i], rex_tableP->studyRex[k][i], regstart[k], length[k], 0, 0, ovector, OVECCOUNT ) < 0 ) break;
#else // RULE_OPTIMEZE == 0
			if ( pcre_exec(rex_tableP->compRex[k][i], 0, regstart[k], length[k], 0, 0, ovector, OVECCOUNT ) < 0 ) break;
#endif // RULE_OPTIMIZE  
			if ( ovector[1] <= ovector[0] ) break;

			if (k) {
				istate = rex_tableP->isstate[i];
				if ( istate ) {
					for ( l = 0; l < rexFlowP->count; l++ ) {
						if ( (rexFlowP->flags[l] & 0x40) && rexFlowP->pregID[l] == istate ) 
								break;
					}
				} else l = rexFlowP->count;;

#if EXPERTMODE == 1
#if PKTTIME == 1
				rexFlowP->time[l] = currFlowP->lastSeen;
#endif //  PKTTIME == 1	
				rexFlowP->sID[l] = rex_tableP->sID[i];
				rexFlowP->pregPos[l] = ovector[0];
				rexFlowP->pkt[l] = rexFlowP->pktN;
#endif // EXPERTMODE == 1
				rexFlowP->flags[l] = rex_tableP->flags[i];
				rexFlowP->pregID[l] = rex_tableP->class[i];
				rexFlowP->alarmcl[l] = rex_tableP->alarmcl[i];
				rexFlowP->severity[l] = rex_tableP->severity[i];
				
				if ( l == rexFlowP->count ) rexFlowP->count++;
					
			}
		}
	}
//	rexFlowP->count = j;

}


void onFlowTerminate(unsigned long flowIndex) {
	rexFlow_t *rexFlowP = &(rexFlowTable[flowIndex]);
	uint32_t i=0, j = rexFlowP->count;

#if LABELSCANS == 1
	rexFlow_t *rexFlowPO = NULL;
	tcpFlagsPerFlow_t* tcpFlagsP = &tcpFlagsPerFlow[flowIndex];

	if ( flows[flowIndex].oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) { 
		rexFlowPO = &rexFlowTable[flows[flowIndex].oppositeFlowIndex];
		
		if ( ( rexFlowP->pktN < 3  && rexFlowPO->pktN < 2) && ( tcpFlagsP->tcpFlagsT == 0x02 || tcpFlagsP->tcpFlagsT == 0x06 || (tcpFlagsP->tcpSpecialFlagsT & SCANMASK) ) && !(tcpFlagsP->tcpSpecialFlagsT & TCPRETRIES) ){
			rexFlowP->alarmcl[j] = 102; // response to scan
			rexFlowP->severity[j] = 1;
			j++;
		}
	//	printf ("Flags: 0x%02x\t0x%04x\t%d\t%d\t%d_%d\n",tcpFlagsP->tcpFlagsT, tcpFlagsP->tcpSpecialFlagsT, rexFlowP->pktN, rexFlowPO->pktN, rexFlowP->alarmcl[j], rexFlowP->severity[j] );	
	} else {
		if ( ( rexFlowP->pktN < 2 ) && ( tcpFlagsP->tcpFlagsT == 0x02 || tcpFlagsP->tcpFlagsT == 0x06 || (tcpFlagsP->tcpSpecialFlagsT & SCANMASK) ) && !(tcpFlagsP->tcpSpecialFlagsT & TCPRETRIES) ) {
			rexFlowP->alarmcl[j] = 101; // no response to scan
                        rexFlowP->severity[j] = 1;
                        j++;
		}
	}
#endif // LABELSCANS
 
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint16_t));
#if EXPERTMODE == 0
	uint32_t sevim = 0, sevmax = rexFlowP->severity[0];
	for ( i = 1; i < j; i++ ) {
		if ( rexFlowP->severity[i] > sevmax ) sevim = i;   
	}
	i = sevim;
#else //EXPERTMODE == 1
	outputBuffer_append(main_output_buffer, (char*) &j, sizeof(uint32_t));
	for (i = 0; i < j; i++) {
#if PKTTIME == 1
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->time[i].tv_sec), sizeof(uint64_t));
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->time[i].tv_usec), sizeof(uint32_t));
#endif // PKTTIME == 1
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->pkt[i]), sizeof(uint16_t));
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->pregPos[i]), sizeof(uint16_t));
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->pregID[i]), sizeof(uint16_t));
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->sID[i]), sizeof(uint32_t));
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->flags[i]), sizeof(uint8_t));

#endif //EXPERTMODE 
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->alarmcl[i]), sizeof(uint8_t));
		outputBuffer_append(main_output_buffer, (char*) &(rexFlowP->severity[i]), sizeof(uint8_t));
#if EXPERTMODE == 1
	}
#endif //EXPERTMODE 

}

void onApplicationTerminate() {
	
	if (rexFlowTable) free(rexFlowTable);
	rexFlowTable = NULL;

	if (rex_tableP) {
#if EXPERTMODE == 1
		free(rex_tableP->class);
		free(rex_tableP->isstate);
		free(rex_tableP->flags);
		free(rex_tableP->compRex[0]);
		free(rex_tableP->compRex[1]);
#endif //EXPERTMOD == 1
		free(rex_tableP->alarmcl);
		free(rex_tableP->severity);
		free(rex_tableP);
	}
	rex_tableP = NULL;
	
}

