/*
 * Snort Regex and Scan detector 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
 */

#ifndef regexp_pattern_matcher_H_
#define regexp_pattern_matcher_H_

// global includes

// local includes
#include "main.h"
#include "regfile_snort_pcre.h"

// User configuration 
#define	EXPERTMODE	0	//
#define	PKTTIME		0	//
#define LABELSCANS	0	//

// defines Regex
#define OVECCOUNT	3	//
#define REGSTART	0	//
#define REGOFFSET	0	//
 
#define MAXREGPOS	30	// Maximal # of matches stored / flow
#define REGOPT		0	// 0: No optimization, 1: execute regex optimization, requires RULE\_OPTIMIZE = 1 in regfile_snort_pcre.h
#define REXPOSIX_FILE "regexfile.txt"	// regexfile name under .tranalyzer/plugins

// defines scans
#define SCANMASK 0x00ff 
#define TCPRETRIES 0x0100 

#define PROTOMAX 4
const uint16_t protocl[PROTOMAX] = { 0, 6, 17, 1 };

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
char* get_dependencies();
void initialize();
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();

// variables
unsigned long plugin_PLUGINNAME_Index;

rex_table_t *rex_tableP;

typedef struct rexFlow_s {
	uint32_t sID[MAXREGPOS]; 	// pattern match Snort ID 
	uint16_t pregID[MAXREGPOS]; 	// pattern match Regex ID 
	uint16_t pktN;
#if EXPERTMODE == 1
#if PKTTIME == 1
	struct timeval time[MAXREGPOS];
#endif // PKTTIME == 1
	uint16_t pregPos[MAXREGPOS];
	uint16_t pkt[MAXREGPOS];
#endif // EXPERTMODE == 1
	uint16_t count;
	uint8_t alarmcl[MAXREGPOS];
	uint8_t flags[MAXREGPOS];
	uint8_t severity[MAXREGPOS];
}rexFlow_t;

rexFlow_t *rexFlowTable;


#endif /* regexp_pattern_matcher_H_ */

