/*
 * Snort Regex and Scan detector 
 * Copyright (C), RUAG AG, RUAG Defence AG, Switzerland
 * Author: Stefan Burschka, RDNIZ
 *
 *
 */

#include "regfile_snort_pcre.h"
#include <string.h>

int16_t rex_load(char *rex_file, rex_table_t *preg_table) {
	FILE *regexp_fp;
	int cl, z, isstate, flags, erroffset, sid, alarmcl, severity;	
	uint32_t i, j = 0, k, pcount = 0, mode;
	char regline[2][MAXREGEXPR], *p;
	const char *errPtr;

	preg_table->count = 0;
	
	// open regex-file
	if( (regexp_fp = fopen(rex_file,"r") ) == NULL){
		printf("Regex - file error: %s\n", rex_file);
		return -1;
	}
	
	while ( fgets(regline[0], MAXREGEXPR, regexp_fp) ) {
		if ( regline[0][0] == '\n' || regline[0][0] == ' ' ) continue;
		else pcount ++; 
	}

	rewind(regexp_fp);
 
	preg_table->count = pcount;
	preg_table->compRex[0] = malloc(sizeof(size_t*) * pcount);
	preg_table->compRex[1] = malloc(sizeof(size_t*) * pcount);
	preg_table->class = malloc(sizeof(uint16_t) * pcount);
	preg_table->sID = malloc(sizeof(uint32_t) * pcount);
	preg_table->isstate = malloc(sizeof(uint16_t) * pcount);
	preg_table->flags = malloc(sizeof(uint8_t) * pcount);
	preg_table->alarmcl = malloc(sizeof(uint8_t) * pcount);
	preg_table->severity = malloc(sizeof(uint8_t) * pcount);
#if RULE_OPTIMIZE == 1
	preg_table->studyRex[0] = malloc(sizeof(size_t*) * pcount);
	preg_table->studyRex[1] = malloc(sizeof(size_t*) * pcount);
#endif // RULE_OPTIMIZE == 1
	
	for ( i = 0; i < pcount; i++ ) {

		z = fscanf(regexp_fp, "%d\t%d\t%*2c%02x\t%d\t%d\t%d\t%[^\n\t]\t%[^\n\t]", &cl, &isstate, &flags, &sid, &alarmcl, &severity, &regline[0][0], &regline[1][0]);
//		if ( regline[0][0] == '\n' || regline[0][0] == ' ' ) continue;
		if (z == 0) { 
			printf("%d: %d %d 0x%02x\t%d\t%d\t%d\t%s\t%s\n",z, cl, isstate, flags, sid, alarmcl, severity, regline[0], regline[1] );
			exit(1);
		}

		mode = ((flags & 0x0f) | REGEX_MODE);

		for (k = 0; k <= 1; k++) {
			if ( ( preg_table->compRex[k][j] = pcre_compile( regline[k], mode, &errPtr, &erroffset, NULL ) ) ) {
#if RULE_OPTIMIZE == 1
				preg_table->studyRex[k][j] = pcre_study(preg_table->compRex[k][j], 0, &errPtr);
				if(errPtr != NULL) {
                                	printf("PCRE study ignored - # %d, @ %d, Error: %s\n", i+1, cl, errPtr );
                                	break;
                        	}
#endif // RULE_OPTIMIZE == 1
				if ( k ) {
					preg_table->class[j] = (uint16_t)cl;
					preg_table->isstate[j] = (uint16_t)isstate;
					preg_table->flags[j] = (uint8_t)flags;
					preg_table->sID[j] = (uint32_t)sid;
					preg_table->alarmcl[j] = (uint8_t)alarmcl;
					preg_table->severity[j] = (uint8_t)severity;
					j++;
				}
			
			} else {
                        	printf("PCRE ignored - # %d, class %d, @ %d, Error: %s\n", i+1, cl, erroffset, errPtr );
				break;
			}
		}
	}

	fclose(regexp_fp);

	preg_table->count = j;
printf("%d\n",j);
	if (j < pcount) return -2;
	else return 0; 
}
