/*
 * standardFileSink.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "standardFileSink.h"

char* get_plugin_name() {
	return "standardFileSink";
}


char* get_plugin_version() {
	return "0.5.0";
}


unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}


unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}


void initialize(unsigned long pPluginIndex) {
	uint32_t *bP;
	register uint32_t i = 0;

#ifdef __APPLE__
	baseFileName = getBaseFileName();
	main_header_bv = getMainHeaderBv();
#endif // __APPLE__
	pluginStandardFileSinkIndex = pPluginIndex;	
	
	// setup output file names
	if (baseFileName) {
		strncpy(outputFileName, baseFileName, MAX_FILENAME_LEN);
		strncat(outputFileName, FLOWS_SUFFIX, MAX_FILENAME_LEN-strlen(baseFileName));

		// open flow output file
		if ((outputFile = fopen(outputFileName, "w")) == NULL) {
			fprintf(stderr, "Error: Can't open file %s in write mode! Maybe incorrect permissions?\n", outputFileName);
			terminate();
		}
	} else outputFile = stdout;

	// generate and write header in flow file
	// build binary header from binary values
	binary_header_t *header = build_header(main_header_bv);

	bP = header->header;
	i = header->header_length * 4;

#if BUF_DATA_SHFT > 0
	bP[0] = i;
#if BUF_DATA_SHFT > 1
	bP[1] = 0;
	bP[1] = Checksum(bP, i);
#endif // BUF_DATA_SHFT > 1
#endif // BUF_DATA_SHFT > 0

	fwrite(header->header, sizeof(uint32_t), header->header_length, outputFile);
	free(header->header);
	free(header);
}


void bufferToSink(outputBuffer_t *buffer) {
#if ALARM_MODE == 1
	if (supOut) {
		return;
	}
#endif // ALARM_MODE ==1

#if BUF_DATA_SHFT > 0
	uint32_t i, *ubP;
	char *bP;

	bP = buffer->buffer - STD_BUFSHFT;
	i = buffer->act_pos + STD_BUFSHFT;

	ubP = (uint32_t*)bP;
	ubP[0] = buffer->act_pos;
#if BUF_DATA_SHFT > 1
	ubP[1] = 0;
	ubP[1] = Checksum(ubP, i);
#endif // BUFi_DATA_SHFT > 1
	fwrite(bP, sizeof(char), i, outputFile);
#else // BUF_DATA_SHFT == 0
	fwrite(buffer->buffer, sizeof(char), buffer->act_pos, outputFile);
#endif // BUF_DATA_SHFT
}


void onApplicationTerminate() {
 fclose(outputFile);
}


#if BUF_DATA_SHFT > 1
// Checksum calculation

static inline uint32_t Checksum(uint32_t *data, uint32_t byteLen) {
	uint64_t chkSum, i;
	uint32_t wordLen = byteLen >> 2;

	while (wordLen--) chkSum += *data++;

	wordLen = (3 - (byteLen & 0x3)) * 8;
	chkSum += *data & (0x00ffffff >> wordLen);

	while (i = (chkSum >> 32)) chkSum = (chkSum & 0xffffffff) + i;

	return chkSum;
}
#endif // BUF_DATA_SHFT > 1
