/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 * 
 * Stefan Burschka: stbf104g
 */

#include "tranalyzer.h"
#include "tcpFlags.h"
#include <limits.h>

char* get_plugin_name() {
	return "tcpFlags";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
#ifdef __APPLE__
	mainHashMap = getMainHashMap();
	baseFileName = getBaseFileName();
	specialOutput = getSpecialOutput();
#endif // __APPLE__
	pluginTCPFlagsIndex = pPluginIndex;

	char sPktFileName[MAX_FILENAME_LEN], *temp;
	tcpFlagsPerFlow = malloc(sizeof(tcpFlagsPerFlow_t) * mainHashMap->hashChainTableSize);
	memset(tcpFlagsPerFlow, '\0', sizeof(tcpFlagsPerFlow_t) * mainHashMap->hashChainTableSize); 

	if (specialOutput == 1) {
		if (baseFileName) {
			temp = baseFileName;
		} else if (captureFileName) {
			temp = captureFileName;
		} else if (captureListFileName) {
			temp = captureListFileName;
		} else if (captureInterface) {
			temp = captureInterface;
		} else terminate();
		
		strncpy(sPktFileName, temp, strlen(temp)+1);
		strncat(sPktFileName, PACKETSFILE_SUFFIX, sizeof(PACKETSFILE_SUFFIX));
		if ((sPktFile = fopen(sPktFileName, "w")) == NULL) {
			printf("%s Error: Can't open file %s in write mode..\n", ERROR_PREFIX, sPktFileName);
			terminate();
		}
		fprintf(sPktFile,"Time\tpktInterdis\tFlow_Index\tEtherSrc_EtherDst\tsrcIP\tsrcPort\tdstIP\tdstport\tproto\tipTOS\tipID\tipIDDiff\tipFrag\tipTTL\tipHdrChkSum\tipCalChkSum\tL4HdrChkSum\tL4CalChkSum\tipFlags\tpktLen\tipOptLen\tipOpts\tseq\tack\tseqDiff\tackDiff\tSeqPktLen\tackPktLen\ttcpFlags\tSpecialFlags\ttcpWin\ttcpOptLen\ttcpOpts\n");
	}
}

binary_value_t* printHeader() {

	binary_value_t *main_bv = NULL;

	binary_value_t *act_bv = bv_new_bv("IP Minimum delta IP ID", "ipMindIPID", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("IP Maximum delta IP ID", "ipMaxdIPID", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("IP Minimum TTL", "ipMinTTL", 0, 1, bt_uint_8);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("IP Maximum TTL", "ipMaxTTL", 0, 1, bt_uint_8);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("IP TTL Change count", "ipTTLChg", 0, 1, bt_uint_8);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("IP Type of Service", "ipTOS", 0, 1, bt_hex_8);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("IP aggregated flags", "ipFlags", 0, 1, bt_hex_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("IP options count", "ipOptCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

//	act_bv = bv_new_bv("IP options Packet count", "ip_OptPktCnt", 0, 1, bt_uint_16);
//	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("IP aggregated options, copy-class & number", "ipOptCpCl_Num", 0, 2, bt_hex_8, bt_hex_32);
	main_bv = bv_append_bv(main_bv, act_bv);

#if SEQ_ACK_NUM == 1
	act_bv = bv_new_bv("TCP packet seq count", "tcpPSeqCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP sent seq diff bytes", "tcpSeqSntBytes", 0, 1, bt_uint_64);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP sequence number fault count", "tcpSeqFaultCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP packet ack count", "tcpPAckCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP flawless ack received bytes", "tcpFlwLssAckRcvdBytes", 0, 1, bt_uint_64);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP ack number fault count", "tcpAckFaultCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // SEQ_ACK_NUM 

#if WINDOWSIZE == 1
	act_bv = bv_new_bv("TCP initial effective window size", "tcpInitWinSz", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP average effective window size", "tcpAveWinSz", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP minimum effective window size", "tcpMinWinSz", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP maximum effective window size", "tcpMaxWinSz", 0, 1, bt_uint_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP effective window size change down count", "tcpWinSzDwnCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP effective window size change up count", "tcpWinSzUpCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP effective window size direction change count", "tcpWinSzChgDirCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // WINDOWSIZE

	act_bv = bv_new_bv("TCP aggregated protocol flags (cwr, ecn, urgent, ack, push, reset, syn, fin)", "tcpAggrFlags", 0, 1, bt_hex_8);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP aggregated header anomaly flags", "tcpAggrAnomaly", 0, 1, bt_hex_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP options Packet count", "tcpOptPktCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP options count", "tcpOptCnt", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP aggregated options", "tcpAggrOptions", 0, 1, bt_hex_32);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP Maximum Segment Length", "tcpMSS", 0, 1, bt_uint_16);
	main_bv = bv_append_bv(main_bv, act_bv);

	act_bv = bv_new_bv("TCP Window Scale", "tcpWS", 0, 1, bt_uint_8);
	main_bv = bv_append_bv(main_bv, act_bv);
	
#if RTT_ESTIMATE == 1
	act_bv = bv_new_bv("TCP Trip Time Syn, Syn-Ack Destination | Syn-Ack, Ack Source","tcpS-SA/SA-ATrip", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP Ack Trip Min","tcpRTTAckTripMin", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP Ack Trip Max","tcpRTTAckTripMax", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP Ack Trip Average","tcpRTTAckTripAve", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP Ack Trip Jitter Average","tcpRTTAckTripJitAve", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP Round Trip Time Syn, Syn-Ack, Ack | TCP Ack-Ack RTT","tcpRTTSseqAA", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
	
	act_bv = bv_new_bv("TCP Ack Round trip average Jitter","tcpRTTAckJitAve", 0, 1, bt_float);
	main_bv = bv_append_bv(main_bv, act_bv);
#endif // RTT_ESTIMATE
	
	return main_bv;

}


void onFlowGenerated(packet_t *packet, unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *actFlowPtr = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
	ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE
	
	tcpFlagsPerFlow_t *tcpFlagsPtr;
	tcpFlagsPtr = &tcpFlagsPerFlow[flowIndex];

	memset(tcpFlagsPtr, '\0', sizeof(tcpFlagsPerFlow_t)); 

	tcpFlagsPtr->lastPktTime.tv_sec = actFlowPtr->lastSeen.tv_sec;
	tcpFlagsPtr->lastPktTime.tv_usec = actFlowPtr->lastSeen.tv_usec;
	tcpFlagsPtr->ipFlagsT = IP_INT_DISSTATE; 

#if IPV6_ACTIVATE == 1
	tcpFlagsPtr->ipIDT = 0;
#else // IPV6_ACTIVATE == 0
	tcpFlagsPtr->ipIDT = ntohs(ipHeader->ip_id);
#endif // IPV6_ACTIVATE

	tcpFlagsPtr->ipMinIDT = 65535;
	tcpFlagsPtr->ipTTLT = ipHeader->ip_ttl;
	tcpFlagsPtr->ipMinTTLT = ipHeader->ip_ttl;
	tcpFlagsPtr->ipMaxTTLT = tcpFlagsPtr->ipMinTTLT;

#if RTT_ESTIMATE == 1
	tcpFlagsPtr->tcpRTTAckTripMin = 65535;
#endif // RTT_ESTIMATE

	if (packet->layer4Type == 6) {
		tcpHeader_t *tcpHeader = (tcpHeader_t*) packet->layer4Header;
		uint16_t tcpWin = ntohs(tcpHeader->window);
#if WINDOWSIZE == 1
		tcpFlagsPtr->tcpWinMinT = 4294967295;
		tcpFlagsPtr->tcpWinInitT = tcpWin; // save initial window size 
		tcpFlagsPtr->tcpWinLastT = tcpWin; // first is the last Window 
		tcpFlagsPtr->tcpWinAveT = tcpWin; // start with average
#endif // WINDOWSIZE

#if SEQ_ACK_NUM == 1
		tcpFlagsPtr->tcpSeqT = ntohl(tcpHeader->seq);
		tcpFlagsPtr->tcpAckT = ntohl(tcpHeader->ack_seq);
#endif //SEQ_ACK_NUM
	} 
}


void claimLayer4Information(packet_t *packet, unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	globalWarn = getGlobalWarn();
#endif // __APPLE__
	flow_t *actFlowPtr = &flows[flowIndex];
	tcpFlagsPerFlow_t *tcpFlagsPtr, *tcpFlagsPtrO;

#if IPV6_ACTIVATE == 1
	ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

	tcpHeader_t *tcpHeader = (tcpHeader_t*)packet->layer4Header;
	udpHeader_t *udpHeader = (udpHeader_t*)packet->layer4Header;

	unsigned int i;
	int16_t ipOptLen = 0;
	uint16_t *L4Header;
	uint16_t ipHdrChkSum = 0, ipCalChkSum = 0;	
	uint16_t L4HdrChkSum = 0, L4CalChkSum = 0;	
	uint16_t ipFlags = 0, ipID = 0, ipIDDiff = 0;
	uint16_t l3Len, l4Len;
	uint16_t l3HDLen, l4HDLen = 8; // L4HDLen standard = 8
	uint16_t ipFrag = 0;
	uint16_t srcPort = 0, dstPort = 0;
	
	char srcIP[MAX_IP_PRINT_LEN];
	char dstIP[MAX_IP_PRINT_LEN];

	unsigned char *l2Header;
	unsigned char tcpFlags = 0; 
	uint8_t *ipOpt;
	
	tcpFlagsPtr = &tcpFlagsPerFlow[flowIndex];

	if ( actFlowPtr->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) tcpFlagsPtrO = &tcpFlagsPerFlow[actFlowPtr->oppositeFlowIndex];
	else tcpFlagsPtrO = NULL;

	l2Header = (uint8_t*) packet->layer2Header;
	L4Header = (uint16_t*)packet->layer4Header;

#if IPV6_ACTIVATE == 1
	if (packet->ip6FragHdr) {
		ipFlags = (packet->ip6FragHdr->frag_off << 1) & IP_FRAG_BITS; // fragmentation bits (Res ,MF), | tcpFlags->ipFlagsT;
		ipFrag = ntohs(packet->ip6FragHdr->frag_off) >> 3;
	}
	l3Len = ntohs(ipHeader->payload_len) + 40;
	l3HDLen = packet->l3HdrLen;
	l4Len = l3Len - l3HDLen;
	if (packet->ip6OptHdr) {
		ipOptLen = l3HDLen - 42;	
		ipOpt =  (uint8_t*)&packet->ip6OptHdr->options;
	}
#else // IPV6_ACTIVATE == 0	
	ipFlags = (uint16_t) *((char*)ipHeader + 6) & IP_FRAG_BITS; // fragmentation bits (Res, DF,MF), | tcpFlags->ipFlagsT;
	l3Len = ntohs(ipHeader->ip_len);
	l3HDLen = IP_HL(ipHeader) << 2;
	l4Len = l3Len - l3HDLen;
	ipOptLen = l3HDLen - 20;	
	ipOpt =  ((char*)ipHeader + 20);
	ipFrag = ipHeader->ip_off & FRAGIDM_N;
#endif // IPV6_ACTIVATE
	
	if (ipOptLen > 0) {
		if ( packet->snapL3Length >= l3HDLen && l3Len >= l3HDLen ) { // option field truncated or crafted packet?
#if IPV6_ACTIVATE == 1
			for ( i = 0; i < ipOptLen && ipOpt[i] > 0; i += (ipOpt[i] > 0) ? ipOpt[i+1]+2: 1 ) {
#else // IPV6_ACTIVATE == 0
			for ( i = 0; i < ipOptLen && ipOpt[i] > 0; i += (ipOpt[i] > 1) ? ipOpt[i+1]: 1 ) {
#endif // IPV6_ACTIVATE
				tcpFlagsPtr->ipCpClT |= 1 << (ipOpt[i] >> 5); // copy & class
				tcpFlagsPtr->ipOptionsT |= 1 << (ipOpt[i] & 0x1F); // ipOptions < 32
				tcpFlagsPtr->ipOptCntT++;
#if IPV6_ACTIVATE == 0
				if (ipOpt[i+1] == 0) break;
#endif // IPV6_ACTIVATE
			}
//			tcpFlagsPtr->ipOptPktCntT++;
		} else ipFlags |= IP_OPT_CORRPT; // warning: crafted packet or option field not acquired
	}


	// minTTL and maxTTL, os guessing and network reversing 

	if (ipHeader->ip_ttl != tcpFlagsPtr->ipTTLT ) tcpFlagsPtr->ipTTLChgT++;
	if (ipHeader->ip_ttl < tcpFlagsPtr->ipMinTTLT) tcpFlagsPtr->ipMinTTLT = ipHeader->ip_ttl;
	if (ipHeader->ip_ttl > tcpFlagsPtr->ipMaxTTLT) tcpFlagsPtr->ipMaxTTLT = ipHeader->ip_ttl;

	// minIPID and maxIPID, good estimate for windows about load state of the source machine

#if IPV6_ACTIVATE == 1
	tcpFlagsPtr->ipTosT |= (unsigned char) (ntohs(*(uint16_t*)ipHeader) >> 4); // get TOS byte
#else // IPV6_ACTIVATE == 0
	tcpFlagsPtr->ipTosT |= (unsigned char) *((char*)ipHeader + 1); // get TOS byte
	ipID = ntohs(ipHeader->ip_id);
	
	ipIDDiff = ipID - tcpFlagsPtr->ipIDT;
	if ( ipID < tcpFlagsPtr->ipIDT ) {
		if ( ipIDDiff < IP_ID_RLLOVR ) ipFlags |= IP_ID_ROLL_OVER; // rollover
		else ipFlags |= IP_ID_OUT_ORDER; // messy packet order
	}

	if (ipIDDiff < tcpFlagsPtr->ipMinIDT) tcpFlagsPtr->ipMinIDT = ipIDDiff;
	if (ipIDDiff > tcpFlagsPtr->ipMaxIDT) tcpFlagsPtr->ipMaxIDT = ipIDDiff;

// ip checksum processing
	ipHdrChkSum = ntohs(ipHeader->ip_sum);

 	if ( packet->snapL3Length < l3HDLen )  ipFlags |= (IP_SNP_HLEN_WRN | IP_L3CHK_SUMERR); 
#if IPCHECKSUM > 0	
 	else {
		ipCalChkSum = ntohs(~(Checksum((uint16_t*)ipHeader, 0, l3HDLen, 5)));
		if (ipHdrChkSum != ipCalChkSum) ipFlags |= IP_L3CHK_SUMERR;
 	}
#endif // IPCHECKSUM > 0

#endif // IPV6_ACTIVATE

	// L4 Checksum processing

	if ( (ipFrag & FRAGID_N) == 0x0000 ) { // only the first header fragment has a L4 header
		
		uint16_t chkSumWrdPos = 0;
	
		switch (packet->layer4Type) {
			case 1:
				chkSumWrdPos = 1;
                                break;
			case 2:
				chkSumWrdPos = 3;
                                break;
#if IPV6_ACTIVATE == 1
			case 58:
				chkSumWrdPos = 1;
				l4HDLen = 4;
				break;
#endif // IPV6_ACTIVATE == 0
			case 6:
				chkSumWrdPos = 8;
				l4HDLen = tcpHeader->doff<<2;
				srcPort = ntohs(tcpHeader->source);
				dstPort = ntohs(tcpHeader->dest);
				break;
			case 17:
				chkSumWrdPos = 3;
				srcPort = ntohs(udpHeader->source);
				dstPort = ntohs(udpHeader->dest);
				break;
			case 47: // check L4Header & 0x0080
				if ( *L4Header & 0x0080 ) chkSumWrdPos = 2;
				else l4HDLen = 4;
				break;
			case 89:
				chkSumWrdPos = 6;
				l4HDLen = 16;
				break;
			case 132:
				chkSumWrdPos = 4;
				l4HDLen = 12;
				srcPort = ntohs(tcpHeader->source);
				dstPort = ntohs(tcpHeader->dest);
				break;
			default:
				break;
		}

		L4HdrChkSum = ntohs(L4Header[chkSumWrdPos]);

		if ( packet->snapL3Length < l3Len ) ipFlags |= (IP_SNP_HLEN_WRN | IP_L4CHK_SUMERR); 
#if IPCHECKSUM > 1 
		else if ( chkSumWrdPos ) {
			if ( packet->layer4Type > 2) {      // ICMP and IGMP use no psydo header
				psyL3Header_t psyL3Header;
				psyL3Header.ip_src = ipHeader->ip_src;
				psyL3Header.ip_dst = ipHeader->ip_dst;
#if IPV6_ACTIVATE == 0
				psyL3Header.ip_p = packet->layer4Type << 8;
				psyL3Header.l4_len = htons(l4Len);
#else // IPV6_ACTIVATE == 1
				psyL3Header.ip_p = packet->layer4Type << 24;
				psyL3Header.l4_len = htonl(l4Len);
#endif // IPV6_ACTIVATE
				L4CalChkSum = Checksum((uint16_t*)&psyL3Header, 0, sizeof(psyL3Header), 0);
			}
		
			L4CalChkSum = ntohs(~(Checksum(L4Header, (uint32_t)L4CalChkSum, l4Len, chkSumWrdPos))); // use the largest structure pointer: tcp
			if (L4HdrChkSum != L4CalChkSum) ipFlags |= IP_L4CHK_SUMERR;
			
		}

#endif // IPCHECKSUM > 1
	}

	float pktInterDis;

	pktInterDis = actFlowPtr->lastSeen.tv_sec - tcpFlagsPtr->lastPktTime.tv_sec + (float)(actFlowPtr->lastSeen.tv_usec - tcpFlagsPtr->lastPktTime.tv_usec) / 1000000.0f;
	if (tcpFlagsPtr->ipFlagsT & IP_INT_DISSTATE)  tcpFlagsPtr->ipFlagsT &= ~IP_INT_DISSTATE; // 1. Packet not suitable for interdistance assessment, later ipFlags
	else { 
		if (pktInterDis < 0) {
			ipFlags |= IP_PKT_INTDISN;
			globalWarn |= TIMEJUMP;
		}
/*		tcpFlagsPtr->avePktInterDis = tcpFLagsPtr->avePktInterDis*0.7 + pktInterDis*0.3;
		if (pktInterDis < 0 || fabs(pktInterDis - tcpFLagsPtr->avePktInterDis) > PKTINTDISTOLL) {
			ipFlags |= 0x4000;
			globalWarn |= TIMEJUMP;
		}  */
		if (pktInterDis == 0) ipFlags |= IP_PKT_INTDIS; 
	}
	
	tcpFlagsPtr->lastPktTime.tv_sec = actFlowPtr->lastSeen.tv_sec;
	tcpFlagsPtr->lastPktTime.tv_usec = actFlowPtr->lastSeen.tv_usec;

#if FRAG_ANALYZE == 1	
	if (ipFrag) { // fragments with and without L4 header
		ipFrag = ntohs(ipFrag) & IPFRAGPKTSZMAX;
		if ( l3Len < IPFRAGPKTSZMIN ) ipFlags |= IP_FRAG_BLW_MIN; // below minimum RFC fragments 
		if ( ipFrag - tcpFlagsPtr->ipNxtFragBgnExp > 1) ipFlags |= IP_FRAG_NXTPPOS; // fragments not at the expected position, pkt loss or attack
		tcpFlagsPtr->ipNxtFragBgnExp = ipFrag + (l4Len >> 3);
		if ( tcpFlagsPtr->ipNxtFragBgnExp > IPFRAGPKTSZMAX ) ipFlags |= IP_FRAG_OUT_RNG; // fragments out of buffer range, possible teardrop
	}
#endif // FRAG_ANALYZE

#if IPV6_ACTIVATE == 0
	if (sPktFile) {
		
		inet_ntop(AF_INET, &(ipHeader->ip_src), srcIP, MAX_IP_PRINT_LEN);
		inet_ntop(AF_INET, &(ipHeader->ip_dst), dstIP, MAX_IP_PRINT_LEN);

#if SPKTFILE_DATE_TIME > 0
		char timeBuf[36]="\0";
		strftime(timeBuf, 35, "%X", localtime(&flows[flowIndex].lastSeen.tv_sec));
		fprintf(sPktFile,"%s"
#else // SPKTFILE_DATE_TIME == 0		
		fprintf(sPktFile,"%ld"
#endif // SPKTFILE_DATE_TIME
		".%06ld\t%f\t%"PRIu32"\t%02X:%02X:%02X:%02X:%02X:%02X_%02X:%02X:%02X:%02X:%02X:%02X\t%s\t%"PRIu16"\t%s\t%"PRIu16"\t%"PRIu8"\t0x%02X\t0x%04X\t%"PRIu16"\t0x%04X\t%"PRIu8"\t0x%04X\t0x%04X\t0x%04X\t0x%04X\t0x%04X\t%"PRIu16"\t%d", 
#if SPKTFILE_DATE_TIME > 0
		timeBuf,
#else // SPKTFILE_DATE_TIME == 0		
		flows[flowIndex].lastSeen.tv_sec,
#endif // SPKTFILE_DATE_TIME
		actFlowPtr->lastSeen.tv_usec, pktInterDis, actFlowPtr->findex, l2Header[6], l2Header[7], l2Header[8], l2Header[9], l2Header[10], l2Header[11], l2Header[0], l2Header[1], l2Header[2], l2Header[3], l2Header[4], l2Header[5], srcIP, srcPort, dstIP, dstPort, packet->layer4Type, tcpFlagsPtr->ipTosT, ipID, ipIDDiff, ntohs(ipHeader->ip_off), ipHeader->ip_ttl, ipHdrChkSum, ipCalChkSum, L4HdrChkSum, L4CalChkSum, ipFlags, packet->packetLength, ipOptLen);

		if ( ipOptLen < 1 ) fprintf(sPktFile,"\t");
		else for ( i = 0; i < ipOptLen; i++ ) fprintf(sPktFile, " 0x%02X", ipOpt[i]); // print all tcpOptions

		if ( packet->layer4Type != 6 || ipFrag & FRAGID_N ) fprintf(sPktFile,"\n"); // nothing follows if not tcp and not 1. Fragment 
	}		

#endif // IPV6_ACTIVATE == 0

	tcpFlagsPtr->ipFlagsT |= ipFlags;
	tcpFlagsPtr->ipIDT = ipID;
	tcpFlagsPtr->ipTTLT = ipHeader->ip_ttl;

	if (packet->layer4Type != 6 || ipFrag & FRAGID_N ) return; // return if not tcp and not 1. Fragment 

// only TCP is processed here

	uint32_t tcpWin = 0;
	uint16_t tcpSpecialFlags = 0;	

	tcpFlags = (unsigned char) *((char*)tcpHeader + 13); 


// sequence number processing

#if SEQ_ACK_NUM == 1
	
	uint32_t seqDiff = 0;
	uint32_t ackDiff = 0;
	uint32_t seq = ntohl(tcpHeader->seq);
        uint32_t ack = ntohl(tcpHeader->ack_seq);	
	
	if ( (tcpFlags & TH_ACK) == TH_ACK && tcpFlagsPtr->tcpSeqT != 0 ) { // problem at rollover
		if ( seq <= tcpFlagsPtr->tcpSeqT ) {
			if ( tcpFlagsPtrO && (tcpFlagsPtrO->tcpSpecialFlagsT & TCP_ACK_PKTLOSS) ) { // same length than before then retransmit
				tcpFlagsPtr->tcpSeqFaultCntT++;
				tcpSpecialFlags |= TCP_SEQ_RETRY; // retransmit
				seqDiff = 0;
			} else if ( (seqDiff = seq + 1 + ~tcpFlagsPtr->tcpSeqT ) <= tcpWin ) { // rollover
				tcpFlagsPtr->tcpPSeqCntT++; // only good if all packets are acked
			} else {
				tcpSpecialFlags |= TCP_SEQ_OUTORDR; // mess in flow packet order, or packetloss at receiver
				seqDiff = 0;
			}
		} else if ( seq > tcpFlagsPtr->tcpSeqT ) {
			seqDiff = seq - tcpFlagsPtr->tcpSeqT;
			tcpFlagsPtr->tcpPSeqCntT++; // only good if all packets are acked
		} 
		
		tcpFlagsPtr->tcpOpSeqPktLength += seqDiff;
	
		if ( tcpFlagsPtrO && tcpFlagsPtrO->tcpAckT > seq  ) {
                        tcpSpecialFlags |= TCP_SEQ_PLSSMS; // mess in flow order, rather pcap packet loss
                }
	
	}

	tcpFlagsPtr->tcpSpecialFlagsT &= ~TCP_ACK_PKTLOSS; // reset packet loss on the wire

// Acknowledge number processing

	if ( (tcpFlags & TH_ACK) == TH_ACK && tcpFlagsPtr->tcpAckT != 0 ) { // problem at rollover
		if ( tcpFlagsPtr->tcpAckT == ack ) {
			if (tcpFlagsPtrO && tcpFlagsPtrO->tcpSeqT >= ack && seqDiff == 0 && packet->packetLength == 0) {
				tcpSpecialFlags |= TCP_ACK_PKTLOSS; // packet loss
				tcpFlagsPtr->tcpAckFaultCntT++;
			}
			ackDiff = 0;
		} else if ( ack > tcpFlagsPtr->tcpAckT ) {
			ackDiff = ack - tcpFlagsPtr->tcpAckT;
			tcpFlagsPtr->tcpPAckCntT++; // only good if all packets are acked 
		} else if ( (ackDiff = ack + 1 + ~tcpFlagsPtr->tcpAckT ) <= (uint32_t)tcpWin ) { // rollover
			tcpFlagsPtr->tcpPAckCntT++; // only good if all packets are acked 
		} else {
			tcpSpecialFlags |= TCP_ACK_OUTORDR; // mess in flow packet order
			ackDiff = 0;
		}
		
		tcpFlagsPtr->tcpOpAckPktLength += ackDiff;
			
	}

#endif // SEQ_ACK_NUM 
	
	// aggregated anomaly Flags and RTT statemachine simplest version @ init tcp handshake and ack rtt
	
	flow_t* actrevFlowPtr = NULL;

#if RTT_ESTIMATE == 1
	float tcpRTTemp, fac; 
	if ( actFlowPtr->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) actrevFlowPtr = &flows[actFlowPtr->oppositeFlowIndex];
	tcpFlagsPtr->tcpPktCnt += 1.0;
#endif // RTT_ESTIMATE

	switch (tcpFlags & 0x17) { // grab only relevant tcp flag bits
		case TH_SYN: // Syn
			if (tcpFlagsPtr->tcpRTTFlag == TCP_RTT_SYN_ST) tcpSpecialFlags |= TCP_SYN_RETRY; // syn retransmit detected
			tcpFlagsPtr->tcpRTTFlag = TCP_RTT_SYN_ST;
			break; 
#if RTT_ESTIMATE == 1
		case TH_ACK: // Ack
			if (actrevFlowPtr) {
				switch (tcpFlagsPtrO->tcpRTTFlag) {
				case TCP_RTT_SYN_ACK: 
					tcpRTTemp = tcpFlagsPtr->tcpRTTtrip = actFlowPtr->lastSeen.tv_sec - actrevFlowPtr->lastSeen.tv_sec + ((float)actFlowPtr->lastSeen.tv_usec - (float)actrevFlowPtr->lastSeen.tv_usec) / 1000000.0f; 
					if (tcpRTTemp < tcpFlagsPtr->tcpRTTAckTripMin) tcpFlagsPtr->tcpRTTAckTripMin = tcpRTTemp;
					if (tcpRTTemp > tcpFlagsPtr->tcpRTTAckTripMax) tcpFlagsPtr->tcpRTTAckTripMax = tcpRTTemp;
					tcpFlagsPtr->tcpRTTAckTripAve = tcpRTTemp;
					tcpFlagsPtr->tcpRTTFlag = TCP_RTT_ACK_A;
					tcpFlagsPtrO->tcpRTTFlag = TCP_RTT_ACK_B;
					break;
				case TCP_RTT_ACK_A:
					tcpRTTemp = actFlowPtr->lastSeen.tv_sec - actrevFlowPtr->lastSeen.tv_sec + ((float)actFlowPtr->lastSeen.tv_usec - (float)actrevFlowPtr->lastSeen.tv_usec) / 1000000.0f;
                                	if (tcpRTTemp < tcpFlagsPtr->tcpRTTAckTripMin) tcpFlagsPtr->tcpRTTAckTripMin = tcpRTTemp;
                                	if (tcpRTTemp > tcpFlagsPtr->tcpRTTAckTripMax) tcpFlagsPtr->tcpRTTAckTripMax = tcpRTTemp;
                                	fac = 1.0 / tcpFlagsPtr->tcpPktCnt;
					tcpFlagsPtr->tcpRTTAckTripAve = tcpFlagsPtr->tcpRTTAckTripAve * (1.0 - fac) + tcpRTTemp * fac;
					tcpRTTemp -= tcpFlagsPtr->tcpRTTAckTripAve;
					tcpFlagsPtr->tcpRTTAckTripJitAve =  tcpFlagsPtr->tcpRTTAckTripJitAve * (1.0 - fac) + tcpRTTemp * tcpRTTemp * fac;
					tcpFlagsPtr->tcpRTTFlag = TCP_RTT_ACK_A;
					tcpFlagsPtrO->tcpRTTFlag = TCP_RTT_ACK_B | TCP_RTT_FLTCST;
					break;
				case TCP_RTT_NO_SYN:
					tcpRTTemp = tcpFlagsPtr->tcpRTTtrip = actFlowPtr->lastSeen.tv_sec - actrevFlowPtr->lastSeen.tv_sec + ((float)actFlowPtr->lastSeen.tv_usec - (float)actrevFlowPtr->lastSeen.tv_usec) / 1000000.0f; 
                                	if (tcpRTTemp < tcpFlagsPtr->tcpRTTAckTripMin) tcpFlagsPtr->tcpRTTAckTripMin = tcpRTTemp;
                                	if (tcpRTTemp > tcpFlagsPtr->tcpRTTAckTripMax) tcpFlagsPtr->tcpRTTAckTripMax = tcpRTTemp;
					tcpFlagsPtr->tcpRTTAckTripAve = tcpRTTemp;
					tcpFlagsPtr->tcpRTTFlag = TCP_RTT_ACK_A;
					tcpFlagsPtrO->tcpRTTFlag = TCP_RTT_ACK_B;
                                        break;
				}
			}
			break; 
#endif // RTT_ESTIMATE
		case TH_FIN_ACK: // Fin-Ack
			tcpSpecialFlags |= TCP_FIN_ACK;
			break;
		case TH_SYN_ACK: // Syn-Ack
			tcpSpecialFlags |= TCP_SYN_ACK;
#if RTT_ESTIMATE == 1
			if ( actrevFlowPtr ) {
				if ( tcpFlagsPtrO->tcpRTTFlag == TCP_RTT_SYN_ST) tcpFlagsPtr->tcpRTTtrip = actFlowPtr->lastSeen.tv_sec - actrevFlowPtr->lastSeen.tv_sec + ((float)actFlowPtr->lastSeen.tv_usec - (float)actrevFlowPtr->lastSeen.tv_usec) / 1000000.0f;
				tcpFlagsPtr->tcpRTTAckTripAve = tcpFlagsPtr->tcpRTTtrip;
				tcpFlagsPtr->tcpRTTFlag = TCP_RTT_SYN_ACK;
			}			
#endif // RTT_ESTIMATE
			break;
		case TH_RST_ACK: // Rst-Ack
			tcpSpecialFlags |= TCP_RST_ACK;
			break;
		case TH_SYN_FIN: // Syn-Fin
			tcpSpecialFlags |= TCP_SYN_FIN;
			break;
		case TH_SYN_FIN_RST: // Syn-Fin-Rst
			tcpSpecialFlags |= TCP_SYN_FIN_RST;
			break;
		case TH_RST_FIN: // Rst-Fin
			tcpSpecialFlags |= TCP_RST_FIN;
			break;
		case TH_NULL: // possible Null Scan
			tcpSpecialFlags |= TCP_NULL;
			break;
		case TH_XMAS: // possible Xmas Scan
			tcpSpecialFlags |= TCP_XMAS;
			break;
	}

#if RTT_ESTIMATE == 1
//	if ( (tcpFlags & TH_RST_FIN) ) tcpFlagsPtr->tcpRTTFlag |= TCP_RTT_STOP; 
#endif

// Tcp Options

	int16_t tcpOptLen = l4HDLen - 20;
	unsigned char *tcpOpt =  ((char*)tcpHeader + 20);

	if (tcpOptLen > 0) { // consider all tcpOptions and set flag bits 
		if ( packet->snapL3Length >= (l3HDLen + l4HDLen) && l4Len >= l4HDLen ) { // option field exists or crafted packet?
			for ( i = 0; i < tcpOptLen && tcpOpt[i] > 0; i += (tcpOpt[i] > 1) ? tcpOpt[i+1]: 1 ) {
				tcpFlagsPtr->tcpOptionsT |= 1 << (tcpOpt[i] & 0x1F); // tcpOptions < 32
				if (tcpOpt[i] == 2) tcpFlagsPtr->tcpMssT = (tcpOpt[i+2] << 8) + tcpOpt[i+3]; // save the last MSS 
				if (tcpOpt[i] == 3) tcpFlagsPtr->tcpWST = tcpOpt[i+2]; // save the Window Scale TODO: max MSS, ave etc bandwidth and pipe length estimation
				tcpFlagsPtr->tcpOptCntT++;
				if (tcpOpt[i+1] == 0) break;
                        }
			tcpFlagsPtr->tcpOptPktCntT++;
                } else tcpSpecialFlags |= TCP_L4OPTCORRPT; // warning: crafted packet or option field not acquired
        }



#if WINDOWSIZE == 1
//	tcp Window Size processing engine
	
	tcpWin = ntohs(tcpHeader->window) * (1 << tcpFlagsPtr->tcpWST);
	
	if ( tcpFlags & TH_SYN ) {
//		tcpFlagsPtr->tcpWinMinT = 4294967295;
		tcpFlagsPtr->tcpWinInitT = tcpWin; // save initial window size 
		tcpFlagsPtr->tcpWinLastT = tcpWin; // first is the last Window 
		tcpFlagsPtr->tcpWinAveT = tcpWin; // start with average
	}

	tcpFlagsPtr->tcpWinAveT = tcpFlagsPtr->tcpWinAveT * 0.7 + (float)tcpWin * 0.3; // IIR filter for winsize

	switch ( tcpSpecialFlags & TCP_WIN_UP ) {
		case TCP_WIN_DWN: //tcpWin decreases
			if (tcpWin <= tcpFlagsPtr->tcpWinLastT) {
				if (tcpWin < tcpFlagsPtr->tcpWinLastT){ 
					tcpFlagsPtr->tcpWdwnCntT++;
					if (tcpWin < tcpFlagsPtr->tcpWinMinT) tcpFlagsPtr->tcpWinMinT = tcpWin;
				}
			} else {
				tcpFlagsPtr->tcpWchgCntT++;
				tcpFlagsPtr->tcpWupCntT++;
				if (tcpWin > tcpFlagsPtr->tcpWinMaxT) tcpFlagsPtr->tcpWinMaxT = tcpWin;
				tcpSpecialFlags |= TCP_WIN_UP;
			}
			break;	
		case TCP_WIN_UP: // tcpWin increases
			if (tcpWin >= tcpFlagsPtr->tcpWinLastT) {
				if (tcpWin > tcpFlagsPtr->tcpWinLastT) {
					tcpFlagsPtr->tcpWupCntT++;
					if (tcpWin > tcpFlagsPtr->tcpWinMaxT) tcpFlagsPtr->tcpWinMaxT = tcpWin;
				}
			} else {
				tcpFlagsPtr->tcpWchgCntT++;
				tcpFlagsPtr->tcpWdwnCntT++;
				if (tcpWin < tcpFlagsPtr->tcpWinMinT) tcpFlagsPtr->tcpWinMinT = tcpWin;
				tcpSpecialFlags &= ~TCP_WIN_UP;
			}
	}
	
#endif  // WINDOWSIZE	

 	if (sPktFile) {

#if SEQ_ACK_NUM == 1
		fprintf(sPktFile,"\t0x%08X\t0x%08X\t%"PRIu32"\t%"PRIu32"\t%"PRIu64"\t%"PRIu64"\t0x%02X\t0x%04X\t%"PRIu32"\t%d\t",seq, ack, seqDiff, ackDiff, tcpFlagsPtr->tcpOpSeqPktLength, tcpFlagsPtr->tcpOpAckPktLength, tcpFlags, tcpSpecialFlags, tcpWin, tcpOptLen);
#else // SEQ_ACK_NUM
		fprintf(sPktFile,"\t0x%08X\t0x%08X\t0x%02X\t0x%04X\t%"PRIu16"\t%d\t", ntohl(tcpHeader->seq), ntohl(tcpHeader->ack_seq), tcpFlags, tcpSpecialFlags, tcpWin, tcpOptLen);
#endif // SEQ_ACK_NUM
               
		for ( i = 0; i < tcpOptLen; i++) fprintf(sPktFile, " 0x%02X", tcpOpt[i]); // print all tcpOptions
#if FLGS_PCNT == 1
		fprintf(sPktFile, " %s", packet->layer7Header); // print content 
#endif // FLGS_PCNT
		
                fprintf(sPktFile,"\n");
        }

	if ((tcpFlags & TH_RST) == 0) {
#if SEQ_ACK_NUM == 1
		tcpFlagsPtr->tcpSeqT = seq;
		tcpFlagsPtr->tcpAckT = ack;
#endif // SEQ_ACK_NUM

#if WINDOWSIZE == 1
		tcpFlagsPtr->tcpWinLastT = tcpWin;
#endif // WINDOWSIZE	
	}

	tcpFlagsPtr->tcpFlagsT |= tcpFlags;
	tcpFlagsPtr->tcpSpecialFlagsT |= tcpSpecialFlags;

#if FRAG_ANALYZE == 1	
	if ( actrevFlowPtr && (actrevFlowPtr->status & IPV4_FRAG_PENDING) && (tcpFlags & TH_RST_FIN) ) tcpFlagsPtrO->ipFlagsT |= IP_FRAG_SEQERR;
#endif // FRAG_ANALYZE
}

void onFlowTerminate(unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__

	flow_t *flowP = &flows[flowIndex];
	tcpFlagsPerFlow_t *tcpFlagsPtr = &tcpFlagsPerFlow[flowIndex], *tcpFlagsPtrO = NULL;

	if ( flowP->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND ) tcpFlagsPtrO = &tcpFlagsPerFlow[flowP->oppositeFlowIndex];

#if FRAG_ANALYZE == 1
	if (flowP->status & IPV4_FRAG_ERR)  tcpFlagsPtr->ipFlagsT |= IP_FRAG_SEQERR;
#endif // FRAG_ANALYZE

#if SEQ_ACK_NUM == 1	
//	if ((tcpFlagsPtr->tcpFlagsT & TH_FIN) && tcpFlagsPtr->tcpOpSeqPktLength > 0) tcpFlagsPtr->tcpOpSeqPktLength--;
	if ((tcpFlagsPtr->tcpFlagsT & TH_SYN) && tcpFlagsPtr->tcpOpSeqPktLength > 0) tcpFlagsPtr->tcpOpSeqPktLength--;
	if ((tcpFlagsPtr->tcpSpecialFlagsT & TH_FIN) && tcpFlagsPtr->tcpOpAckPktLength > 0) tcpFlagsPtr->tcpOpAckPktLength--;
//	if ((tcpFlagsPtr->tcpSpecialFlagsT & TH_SYN) && tcpFlagsPtr->tcpOpAckPktLength > 0) tcpFlagsPtr->tcpOpAckPktLength--;
#endif //SEQ_ACK_NUM

	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipMinIDT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipMaxIDT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipMinTTLT, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipMaxTTLT, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipTTLChgT, sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipTosT, sizeof(uint8_t)); // bt_hex_8
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipFlagsT, sizeof(uint16_t)); // bt_hex_16
//	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipOptPktCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipOptCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipCpClT, sizeof(uint8_t)); // bt_hex_8: copy,class
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->ipOptionsT, sizeof(uint32_t)); // bt_hex_32: ip option

#if SEQ_ACK_NUM == 1
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpPSeqCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpOpSeqPktLength, sizeof(uint64_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpSeqFaultCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpPAckCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpOpAckPktLength, sizeof(uint64_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpAckFaultCntT, sizeof(uint16_t));
#endif //SEQ_ACK_NUM

#if WINDOWSIZE == 1
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWinInitT, sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWinAveT, sizeof(float));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWinMinT, sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWinMaxT, sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWdwnCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWupCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWchgCntT, sizeof(uint16_t));
#endif // WINDOWSIZE

	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpFlagsT, sizeof(uint8_t)); // bt_hex_8
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpSpecialFlagsT, sizeof(uint16_t)); // bt_hex_16
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpOptPktCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpOptCntT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpOptionsT, sizeof(uint32_t)); // bt_hex_32
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpMssT, sizeof(uint16_t));
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpWST, sizeof(uint8_t));

#if RTT_ESTIMATE == 1
	float dum;
	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpRTTtrip, 4*sizeof(float));
//	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpRTTAckTripMin, sizeof(float));
//	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpRTTAckTripMax, sizeof(float));
//	outputBuffer_append(main_output_buffer, (char*) &tcpFlagsPtr->tcpRTTAckTripAve, sizeof(float));
	dum = sqrt(tcpFlagsPtr->tcpRTTAckTripJitAve);
	outputBuffer_append(main_output_buffer, (char*) &dum, sizeof(float));
	if ( tcpFlagsPtrO ) {
		if (flowP->status & L3FLOWINVERT) dum = tcpFlagsPtr->tcpRTTAckTripAve + tcpFlagsPtrO->tcpRTTAckTripAve;
		else dum = tcpFlagsPtr->tcpRTTtrip + tcpFlagsPtrO->tcpRTTtrip;
	} else dum = tcpFlagsPtr->tcpRTTtrip; 
	outputBuffer_append(main_output_buffer, (char*) &dum, sizeof(float));
	if ( tcpFlagsPtrO && (flowP->status & L3FLOWINVERT) ) dum = sqrt(tcpFlagsPtr->tcpRTTAckTripJitAve + tcpFlagsPtrO->tcpRTTAckTripJitAve);
	else dum = -1.0;
	outputBuffer_append(main_output_buffer, (char*) &dum, sizeof(float));
#endif //RTT_ESTIMATE

}


void onApplicationTerminate() {
 	if (sPktFile) fclose(sPktFile);
	if (tcpFlagsPerFlow) {
		free(tcpFlagsPerFlow);
		tcpFlagsPerFlow = NULL;
	}
}	


#if IPCHECKSUM > 0
// Checksum calculation, according to RFC 

static inline uint16_t Checksum(uint16_t *data, uint32_t chkSum, uint16_t byteLen, uint16_t chkSumWrdPos) {
	uint32_t i;
	uint32_t wordLen = byteLen >> 1;

	if (chkSumWrdPos > 0) chkSum -= data[chkSumWrdPos];

	while (wordLen--) chkSum += *data++;

	if (byteLen & 1) chkSum += *data & 0x00FF;

	while (i = (chkSum >> 16)) chkSum = (chkSum & 0xFFFF) + i;

	return chkSum;
}
#endif // IPCHECKSUM
