/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __TCPFLAGS_H__
#define __TCPFLAGS_H__


// includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <pcap.h>

// local includes
#include "main.h"
#include "getter.h"

// local defines
#define	ERROR_PREFIX	"tcpFlags"
#define	PACKETSFILE_SUFFIX	"_"ERROR_PREFIX""PACKETS_SUFFIX".txt"

// User configuration: switches for plugin features
#define SPKTFILE_DATE_TIME	0	// -s option timestamp format: 0: Unix timestamp , 1: Date Time
#define FLGS_PCNT	0		// -s option tcp print L7 content

#define RTT_ESTIMATE	1	// 1: Round trip time estimation
#define IPCHECKSUM	2	// 1: Calculation of L3 (IP) Header Checksum, 2: L3 + L4 (TCP,UDP) Checksum 
#define WINDOWSIZE	1	// 1: Calculation of tcp window size parameters
#define SEQ_ACK_NUM	1	// 1: Sequence / Acknowledge Number features 
#define FRAG_ANALYZE	1	// 1: Fragmentation analysis

// Local constants
#define	IPFRAGPKTSZMIN	48
#define	IPFRAGPKTSZMAX	0x1fff
#define PKTINTDISTOLL	0.01
#define	RTTFILTERCONST	0.6
#define IP_ID_RLLOVR	0x2000

// ipFlagsT
#define IP_OPT_CORRPT	0x0001 //
#define IP_ID_OUT_ORDER	0x0002 //
#define IP_ID_ROLL_OVER	0x0004 //
#define IP_FRAG_BLW_MIN	0x0008 //
#define IP_FRAG_OUT_RNG	0x0010 //
#define IP_FRAG_MF	0x0020 //
#define IP_FRAG_DF	0x0040 // IPv6: Res
#define IP_FRAG_RES	0x0080 //
#define IP_FRAG_NXTPPOS	0x0100 //
#define IP_FRAG_SEQERR	0x0200 //
#define IP_L3CHK_SUMERR	0x0400 //
#define IP_L4CHK_SUMERR	0x0800 //
#define IP_SNP_HLEN_WRN	0x1000 //
#define IP_PKT_INTDIS	0x2000 //
#define IP_PKT_INTDISN	0x4000 //
#define IP_INT_DISSTATE	0x8000 //

#define IP_FRAG_BITS	0x00E0
#define IP_FRAG_NO_HDR	(IP_FRAG_SEQERR | IP_FRAG_OUT_RNG | IP_FRAG_NXTPPOS)

// tcpSpecialFlagsT
#define TCP_FIN_ACK 	0x0001 //
#define TCP_SYN_ACK	0x0002 //
#define TCP_RST_ACK	0x0004 //
#define TCP_SYN_FIN	0x0008 //
#define TCP_SYN_FIN_RST	0x0010 //
#define TCP_RST_FIN	0x0020 //
#define TCP_NULL	0x0040 //
#define TCP_XMAS	0x0080 //
#define TCP_SEQ_RETRY	0x0100 //
#define TCP_SEQ_OUTORDR	0x0200 //
#define TCP_SEQ_PLSSMS	0x0400 //
#define TCP_L4OPTCORRPT	0x0800 //
#define TCP_SYN_RETRY	0x1000 //
#define TCP_ACK_OUTORDR	0x2000 //
#define TCP_ACK_PKTLOSS	0x4000 //
#define TCP_WIN_UP	0x8000 //

#define TCP_WIN_DWN	0x0000 //
#define TCP_SPEC_FLAGS	0x00FF //

// round trip flags
#define	TCP_RTT_NO_SYN	0x00 //
#define	TCP_RTT_ACK_A	0x04 //
#define	TCP_RTT_ACK_B	0x08 //
#define	TCP_RTT_SYN_ST	0x01 //
#define	TCP_RTT_SYN_ACK	0x02 //
#define	TCP_RTT_FLTCST	0X10 //
#define	TCP_RTT_STOP	0X80 //

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
void onApplicationTerminate();
static uint16_t Checksum(uint16_t *data, uint32_t chkSum, uint16_t byteLen, uint16_t chkSumWrdPos);

// structs
typedef struct tcpFlagsPerFlow_s { 
	struct timeval lastPktTime;
#if SEQ_ACK_NUM == 1
	uint64_t tcpOpSeqPktLength;
	uint64_t tcpOpAckPktLength;
	uint32_t tcpSeqT;
	uint32_t tcpAckT;
#endif // SEQ_ACK_NUM == 1

	uint32_t ipOptionsT;
	uint32_t tcpOptionsT;

#if RTT_ESTIMATE == 1
	float tcpRTTtrip;
	float tcpRTTAckTripMin;
	float tcpRTTAckTripMax;
	float tcpRTTAckTripAve;
	float tcpRTTAckTripJitAve;
	float tcpPktCnt;
#endif // RTT_ESTIMATE == 1

#if WINDOWSIZE == 1
	float tcpWinAveT;
	uint32_t tcpWinInitT;
	uint32_t tcpWinLastT;
	uint32_t tcpWinMinT;
	uint32_t tcpWinMaxT;
	uint16_t tcpWupCntT;
	uint16_t tcpWdwnCntT;
	uint16_t tcpWchgCntT;
#endif // WINDOWSIZE == 1

#if SEQ_ACK_NUM == 1
	uint16_t tcpPSeqCntT;
	uint16_t tcpSeqFaultCntT; 
	uint16_t tcpPAckCntT;
	uint16_t tcpAckFaultCntT; 
#endif // SEQ_ACK_NUM == 1
	uint16_t tcpSpecialFlagsT;
	uint16_t tcpMssT;
	uint16_t tcpOptCntT;
	uint16_t tcpOptPktCntT;
	uint16_t ipIDT;
	uint16_t ipMinIDT;
	uint16_t ipMaxIDT;
	uint16_t ipFlagsT;
	uint16_t ipNxtFragBgnExp;
	uint16_t ipOptCntT;
//	uint16_t ipOptPktCntT;
	uint8_t ipCpClT;
	uint8_t tcpWST;
	uint8_t tcpFlagsT;
	uint8_t ipTosT;
	uint8_t ipMinTTLT;
	uint8_t ipMaxTTLT;
	uint8_t ipTTLChgT;
	uint8_t ipTTLT;
	uint8_t tcpRTTFlag;
	//uint8_t tcpRTTAggFlag;
} tcpFlagsPerFlow_t;

// pseudo header for checksum calculation, network order

typedef struct psyL3Header_s { // netowrk order
#if IPV6_ACTIVATE == 1
	ipAddr_t ip_src;	// source address
	ipAddr_t ip_dst;	// dest address
	uint32_t l4_len;	// total l4 length 
	uint32_t ip_p;		// high byte = protocol 
#else // IPV6_ACTIVATE == 0
	struct in_addr ip_src;	// source address
	struct in_addr ip_dst;	// dest address
	uint16_t ip_p;		// high byte = protocol
	uint16_t l4_len;	// total l4 length 
#endif // IPV6_ACTIVATE
} psyL3Header_t;

// variables
FILE *sPktFile = NULL; 
extern FILE *flowStatsFile;
extern pcap_t* captureDescriptor;
unsigned long pluginTCPFlagsIndex;

tcpFlagsPerFlow_t *tcpFlagsPerFlow;
#ifdef __APPLE__
// global variables
hashMap_t* mainHashMap;
char *baseFileName;
uint16_t globalWarn;
char  specialOutput;
outputBuffer_t *main_output_buffer;
flow_t *flows;
#endif // __APPLE__
#endif // __TCPFLAGS_H__
