/*
 * tranalyzer_plugin_skeleton.h
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef TCPSTATES_H_
#define TCPSTATES_H_

#include <stdlib.h>
#include <stdio.h>
#include "tranalyzer.h"
#include "main.h"
#include "binaryValue.h"
#include "networkHeaders.h"
#include "getter.h"

// User defines
#define TCPSTATES_PRINT_OUT 1

// Local definition of the states
#define STATE_NEW 0
#define STATE_ESTABLISHED 1
#define STATE_CLOSING 2
#define STATE_CLOSED 3
#define STATE_RESET 4
#define STATE_BOGUS 255

// Local definition of the timeouts
#define TIMEOUT_RESET 0.1f // 100 ms
#define TIMEOUT_NEW 120.0f // 2 minutes
#define TIMEOUT_ESTABLISHED 610.0f // 10 minutes 10 seconds // 3600.0f // one hour
#define TIMEOUT_CLOSING 120.0f // 2 minutes
#define TIMEOUT_CLOSED 10.0f // 10 seconds

// bogus_bitField 
#define MAL_CON_EST	0x01	// Malformed connection establishement
#define MAL_TEARDWN	0x02	// Malformed teardown
#define MAL_FLG2_EST	0x04	// Malformed flags during established connection
#define PKTS_TERM	0x08	// More packets after teardown
#define PKTS_RST	0x10	// More packets after reset seen
#define RST_TRANS	0x40	// Reset from sender seen
#define EVIL		0x80	// Potential evil behavior

// basic struct to recognize the states
typedef struct tcp_connection_states_s {
	uint64_t sender_syn_sequ_nr;
	uint64_t sender_fin_sequ_nr;
	uint8_t bogus_bitField; // indicates anomalies in tcp states:
	unsigned char connection_state;
	unsigned char sender_syn_seen:1;
	unsigned char sender_syn_ackd:1;
	unsigned char sender_fin_seen:1;
	unsigned char sender_fin_ackd:1;
	unsigned char sender_fin_scan:1;
} tcp_connection_states_t;

/* variables */
tcp_connection_states_t *tcpStates_connection_table;
unsigned long pluginTCPStatesIndex;
#ifdef __APPLE__
//global variables 
hashMap_t *mainHashMap;
outputBuffer_t *main_output_buffer;
flow_t *flows;
#endif // __APPLE__
// char tcpFlags_srcIP[MAX_IP_PRINT_LEN]; // used for debug
// char tcpFlags_dstIP[MAX_IP_PRINT_LEN]; // used for debug

/* functions */
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
void initialize(unsigned long pPluginIndex);
void onFlowGenerated(packet_t *packet, unsigned long flowIndex);
void claimLayer4Information(packet_t *packet, unsigned long flowIndex);

# if TCPSTATES_PRINT_OUT != 0
binary_value_t* printHeader();
void onFlowTerminate(unsigned long flowIndex);
#endif /* TCPSTATES_PRINT_OUT != 0 */

void onApplicationTerminate();

#endif /* TCPSTATES_H_ */
