/*
 * binaryToText.c
 *
 *  Created on: Apr 23, 2010
 *  Author: Torben Rühl, RDNIZ
 *  Co-Author: Stefan Burschka, RDNIZ
 */


#include "binaryToText.h"
#include <string.h>

int type_of_input;
char* input_buffer = NULL;
uint32_t input_buffer_pos = 0;

//uint32_t num_flows_proc = 0;

FILE* input_file = NULL;
uint32_t (*get_val_fkt)(void*, size_t, size_t);

void parse_subheader_type(binary_subvalue_t *sv, FILE* output_file);
void parse_subval(binary_subvalue_t *sv, int depth, FILE* output_file);
void parse_binary_value(uint32_t type, FILE* output_file);
uint32_t get_val_from_input_buffer(void* dest, size_t size, size_t n);
uint32_t get_val_from_input_file(void* dest, size_t size, size_t n);


/*
 * selects the input source
 * 0: a character buffer in memory
 * 1: a file on disk
 */

void set_input_source(int new_type_of_input, ...) {
    type_of_input = new_type_of_input;

    if (type_of_input == INPUT_TYPE_BUFFER) {
        input_file = NULL;
        va_list argPtr;
        va_start(argPtr, new_type_of_input);
        input_buffer = va_arg(argPtr, char*);
        input_buffer_pos = 0;
        get_val_fkt = get_val_from_input_buffer;
        va_end(argPtr);
    }

    if (type_of_input == INPUT_TYPE_FILE) {
        input_buffer = NULL;
        input_buffer_pos = 0;
        va_list argPtr;
        va_start(argPtr, new_type_of_input);
        input_file = va_arg(argPtr, FILE*);
        get_val_fkt = get_val_from_input_file;
        va_end(argPtr);
    }
}


uint32_t get_val_from_input_buffer(void* dest, size_t size, size_t n) {
    memcpy(dest, input_buffer + input_buffer_pos, size * n);
    input_buffer_pos += size * n;
    return size * n;
}


uint32_t get_val_from_input_file(void* dest, size_t size, size_t n) {
    return fread(dest, size, n, input_file);
}


/*
 * parse the header strings
 * header_type:
 * 0: long strings with column number, types and repeat, as rows
 * 1: short strings aggregated to one row with '#' as prefix
 */

uint32_t parse_binary_header2text(binary_value_t *bv, int header_type, FILE* output_file) {
    uint32_t i, j = 0;
    binary_value_t *act_bv = bv;

    if (header_type == 0) {
        uint32_t colNr = 1;

        fprintf(output_file, "# Col No.\tType\tName");
        while (act_bv) {
            fprintf(output_file, "\n%"PRIu32"\t", colNr++);
            for (i = 0; i < act_bv->num_values; i++) {
                if (i != 0) fprintf(output_file, "_");
                if (act_bv->subval[i].value_type == 0) parse_subheader_type(&(act_bv->subval[i]), output_file);
                else fprintf(output_file, "%"PRIu32, act_bv->subval[i].value_type);
            }
            if (act_bv->is_repeating) fprintf(output_file, ":R");
            else fprintf(output_file, ":N");
            fprintf(output_file, "\t%s", act_bv->name_value_long);
            act_bv = act_bv->next;
            j++;
        }
        return j;

    } else if (header_type == 1) {
        while (act_bv) {
            j++;
            act_bv = act_bv->next;
        }



        act_bv = bv;

#if TEXT_FILE_SINK_EXTENDED_HEADER == 1
        fprintf(output_file, "%%                     \n"); // We need 21 placeholders to later insert the number of flows, which can be UINT_64_MAX, which are 20 digits
        fprintf(output_file, "%% %d\n", j); // print number of bv's
        // print extended header
        fprintf(output_file, "%% ");
        while (act_bv) {
            if (act_bv != bv) fputc('\t', output_file);
            for (i = 0; i < act_bv->num_values; i++) {
                if (i != 0) fprintf(output_file, "_");
                if (act_bv->subval[i].value_type == 0) parse_subheader_type(&(act_bv->subval[i]), output_file);
                else fprintf(output_file, "%"PRIu32, act_bv->subval[i].value_type);
            }
            if (act_bv->is_repeating) fprintf(output_file, ":R");
            else fprintf(output_file, ":N");
            act_bv = act_bv->next;
        }
        fputc('\n', output_file);
        act_bv = bv;
#endif // TEXT_FILE_SINK_EXTENDED_HEADER == 1

        fprintf(output_file, "%%");
        while (act_bv) {
            if (act_bv != bv) fprintf(output_file, "\t");
            fprintf(output_file, "%s", act_bv->name_value_short);
            act_bv = act_bv->next;
        }
        return j;
    } else {
        return 0;
    }
}


void parse_subheader_type(binary_subvalue_t *sv, FILE* output_file) {
    fprintf(output_file, "(");
    uint32_t i;
    for (i = 0; i < sv->num_values; i++) {
        if (i != 0) fprintf(output_file, "_");
        if (sv->subval[i].value_type == 0) parse_subheader_type(&(sv->subval[i]), output_file);
        else fprintf(output_file, "%"PRIu32, sv->subval[i].value_type);
    }
    if (sv->is_repeating) fprintf(output_file, ":R");
    else fprintf(output_file, ":N");
    fprintf(output_file, ")");
}


void parse_binary2text(binary_value_t *bv, FILE* output_file) {
    binary_value_t *act_bv = bv;
    input_buffer_pos = 0;
    while(act_bv) {
        // before each new column, there has to be a tab
        if (act_bv != bv) {
            fprintf(output_file, "\t");
        }

        // check if output can be repeated. If yes, read amount of repeats, if no set repeat_val to 1
        uint32_t repeat_val;
        if (act_bv->is_repeating) {
            get_val_fkt(&repeat_val, sizeof(uint32_t), 1);
        } else
            repeat_val = 1;

        int repeat_i;

        for (repeat_i = 0; repeat_i < repeat_val; repeat_i++) {
            int num_val_i;

            // for each output val:
            // check type and write it out, if zero then it contains subvals
            for (num_val_i = 0; num_val_i < act_bv->num_values; num_val_i++) {
                if (act_bv->subval[num_val_i].value_type != 0) {
                    parse_binary_value(act_bv->subval[num_val_i].value_type, output_file);
                } else {
                    parse_subval(&act_bv->subval[num_val_i], 1, output_file);
                }

                // write value delimeter
                if (num_val_i < act_bv->num_values - 1) fprintf(output_file, "_");
            }
            // write repeat delimeter
            if (repeat_i < repeat_val - 1) fprintf(output_file, ";");

        }

        act_bv = act_bv->next;
    }
    //num_flows_proc++;
}


void parse_subval(binary_subvalue_t *sv, int depth, FILE* output_file) {
    fprintf(output_file, "(");
    // check if output can be repeated. If yes, read amount of repeats, if no set repeat_val to 1
    uint32_t repeat_val;
    if (sv->is_repeating)
        get_val_fkt(&repeat_val, sizeof(uint32_t), 1);
    else
        repeat_val = 1;

    int repeat_i;
    for (repeat_i = 0; repeat_i < repeat_val; repeat_i++) {
        int num_val_i;

        // for each output val:
        // check type and write it out, if zero then it contains subvals
        for (num_val_i = 0; num_val_i < sv->num_values; num_val_i++) {
            if (sv->subval[num_val_i].value_type != 0) {
                parse_binary_value(sv->subval[num_val_i].value_type, output_file);
            } else {
                parse_subval(&sv->subval[num_val_i], depth + 1, output_file);
            }

            // write value delimeter
            if (num_val_i < sv->num_values - 1) fprintf(output_file, "_");
        }
        // write repeat delimeter
        if (repeat_i < repeat_val - 1) fprintf(output_file, ";");
    }
    fprintf(output_file, ")");
}


void parse_binary_value(uint32_t type, FILE* output_file) {
    // specify buffer
    uint8_t val[16];
    uint32_t i,j, sw=0;

    switch (type) {
    case bt_int_8:
        // read value
        if (get_val_fkt(val, sizeof(int8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRId8, *((int8_t*)val));
        break;

    case bt_int_16:
        // read value
        if (get_val_fkt(val, sizeof(int16_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRId16, *((int16_t*)val));
        break;

    case bt_int_32:
        // read value
        if (get_val_fkt(val, sizeof(int32_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRId32, *((int32_t*)val));
        break;

    case bt_int_64:
        // read value
        if (get_val_fkt(val, sizeof(int64_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRId64, *((int64_t*)val));
        break;

    case bt_uint_8:
        // read value
        if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRIu8, *((uint8_t*)val));
        break;

    case bt_uint_16:
        // read value
        if (get_val_fkt(val, sizeof(uint16_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRIu16, *((uint16_t*)val));
        break;

    case bt_uint_32:
        // read value
        if (get_val_fkt(val, sizeof(uint32_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRIu32, *((uint32_t*)val));
        break;

    case bt_uint_64:
        // read value
        if (get_val_fkt(val, sizeof(uint64_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%"PRIu64, *((uint64_t*)val));
        break;

    case bt_hex_8:
        // read value
        if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
#if HEX_CAPITAL == 1
        fprintf(output_file, "0x%02"PRIX8, *((uint8_t*)val));
#else
        fprintf(output_file, "0x%02"PRIx8, *((uint8_t*)val));
#endif
        break;

    case bt_hex_16:
        // read value
        if (get_val_fkt(val, sizeof(uint16_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
#if HEX_CAPITAL == 1
        fprintf(output_file, "0x%04"PRIX16, *((uint16_t*)val));
#else
        fprintf(output_file, "0x%04"PRIx16, *((uint16_t*)val));
#endif
        break;

    case bt_hex_32:
        // read value
        if (get_val_fkt(val, sizeof(uint32_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
#if HEX_CAPITAL == 1
        fprintf(output_file, "0x%08"PRIX32, *((uint32_t*)val));
#else
        fprintf(output_file, "0x%08"PRIx32, *((uint32_t*)val));
#endif
        break;

    case bt_hex_64:
        // read value
        if (get_val_fkt(val, sizeof(uint64_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
#if HEX_CAPITAL == 1
        fprintf(output_file, "0x%016"PRIX64, *((uint64_t*)val));
#else
        fprintf(output_file, "0x%016"PRIx64, *((uint64_t*)val));
#endif
        break;

    case bt_float:
        // read value
        if (get_val_fkt(val, sizeof(float), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%f", *((float*)val));
        break;

    case bt_double:
        // read value
        if (get_val_fkt(val, sizeof(double), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%f", *((double*)val));
        break;

    case bt_long_double:
        // read value
        if (get_val_fkt(val, sizeof(long double), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fprintf(output_file, "%Lf", *((long double*)val));
        break;

    case bt_char:
        // read value
        if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        // print value
        fputc(val[0], output_file);
        // fprintf(output_file, "%c", *((uint8_t*)val));
        break;

    case bt_string_class:
        // read value
        if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        while (*val != '\0') {
            fputc(val[0], output_file);
            if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
                fprintf(stderr, "binaryToText: Error: Input is corrupted!");
                exit(1);
            }
        }
        break;

    case bt_string:
        // print startsign '"'
        fputc(34, output_file);
        // read value
        if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        while (*val != '\0') {
            // print value
            switch (*val) {
            case '\t':
                fprintf(output_file, "\\t");
                break;
            case '\n':
                fprintf(output_file, "\\n");
                break;
            case '\\':
            case '"':
                fputc(92, output_file); // '\'
            default:
                fputc(*val, output_file);
                break;
            }
            // fputc(*val, output_file);
            // fprintf(output_file, "%c", *((uint8_t*)val));
            if (get_val_fkt(val, sizeof(uint8_t), 1) == 0) {
                fprintf(stderr, "binaryToText: Error: Input is corrupted!");
                exit(1);
            }
        }
        // print stopsign '"'
        fputc(34, output_file);
        break;
    case bt_mac_addr:
        // read value
        if (get_val_fkt(val, l_bt_mac_addr * sizeof(u_int8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }

#if HEX_CAPITAL == 1
        fprintf(output_file, "%02"PRIX8":%02"PRIX8":%02"PRIX8":%02"PRIX8":%02"PRIX8":%02"PRIX8, val[0], val[1], val[2], val[3], val[4], val[5]);
#else
        fprintf(output_file, "%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8, val[0], val[1], val[2], val[3], val[4], val[5]);
#endif
        break;
    case bt_ip4_addr:
        // read value
        if (get_val_fkt(val, l_bt_ip4_addr * sizeof(u_int8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }

#if IP_PRINT_NORMALIZE == 1			
        fprintf(output_file, "%03"PRIu8".%03"PRIu8".%03"PRIu8".%03"PRIu8, val[0], val[1], val[2], val[3]);
#else
        fprintf(output_file, "%"PRIu8".%"PRIu8".%"PRIu8".%"PRIu8, val[0], val[1], val[2], val[3]);
#endif
        break;
    case bt_ip6_addr:
        // read value
        if (get_val_fkt(val, l_bt_ip6_addr * sizeof(u_int8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }

#if HEX_CAPITAL == 1
#if IP6_COMPRESS == 0	
        fprintf(output_file, "%04"PRIX16":%04"PRIX16":%04"PRIX16":%04"PRIX16":%04"PRIX16":%04"PRIX16":%04"PRIX16":%04"PRIX16,
                ntohs(((uint16_t*)val)[0]), ntohs(((uint16_t*)val)[1]), ntohs(((uint16_t*)val)[2]), ntohs(((uint16_t*)val)[3]),
                ntohs(((uint16_t*)val)[4]), ntohs(((uint16_t*)val)[5]), ntohs(((uint16_t*)val)[6]), ntohs(((uint16_t*)val)[7]) );
#else // IP6_COMPRESS == 1	
        for ( i = 0; i < 8; i++ ) {
            j = ((uint16_t*)val)[i];
            if ( j ) {
                if ( sw ) fprintf(output_file, ":");
                fprintf(output_file, "%X", ntohs(j));
                sw = 1;
            } else if ( sw ) {
                if ( !((uint16_t*)val)[i+1] && sw == 1) {
                    fprintf(output_file, "::");
                    sw = 0;
                }
                else fprintf(output_file, ":0");
            }
        }
        if ( !sw ) fprintf(output_file, "::");
#endif // IP6_COMPRESS

#else // HEX_CAPITAL == 0
#if IP6_COMPRESS == 0
        fprintf(output_file, "%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16":%04"PRIx16,
                ntohs(((uint16_t*)val)[0]), ntohs(((uint16_t*)val)[1]), ntohs(((uint16_t*)val)[2]), ntohs(((uint16_t*)val)[3]),
                ntohs(((uint16_t*)val)[4]), ntohs(((uint16_t*)val)[5]), ntohs(((uint16_t*)val)[6]), ntohs(((uint16_t*)val)[7]) );

#else // IP6_COMPRESS == 1
        for ( i = 0; i < 8; i++ ) {
            j = ((uint16_t*)val)[i];
            if ( j ) {
                if ( sw ) fprintf(output_file, ":");
                fprintf(output_file, "%x", ntohs(j));
                sw = 1;
            } else if ( sw ) {
                if ( !((uint16_t*)val)[i+1] && sw == 1) {
                    fprintf(output_file, "::");
                    sw = 0;
                }
                else fprintf(output_file, ":0");
            }
        }
        if ( !sw ) fprintf(output_file, "::");
#endif // IP6_COMPRESS

#endif // HEX_CAPITAL
        break;

    case bt_unix_time:
    case bt_time:
        // read seconds
        if (get_val_fkt(val, sizeof(u_int64_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }

        if (type == bt_time) {
            char timeBuf[36];
            strftime(timeBuf, 35, "%d.%m.%Y_%X", localtime((time_t*)val));
            fprintf(output_file, "%s.", timeBuf);
        } else {
            fprintf(output_file, "%"PRIu64".", *((uint64_t*)val));
        }

        // read nanoseconds
        if (get_val_fkt(val, sizeof(u_int32_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
#if B2T_TIME_IN_MICRO_SECS != 0
        fprintf(output_file, "%06"PRIu32, *((uint32_t*)val) / 1000);
#else
        fprintf(output_file, "%09"PRIu32, *((uint32_t*)val));
#endif /* B2T_TIME_IN_MICRO_SECS != 0 */
        break;

    case bt_flow_direction:
        // read value
        if (get_val_fkt(val, sizeof(u_int8_t), 1) == 0) {
            fprintf(stderr, "binaryToText: Error: Input is corrupted!");
            exit(1);
        }
        if (*val == 0) fprintf(output_file, "A");
        else fprintf(output_file, "B");
        break;

    default:
        fprintf(stderr, "ERROR: couldn't determine type of output or output type is not implemented!\nType is: %"PRIu32"\n", type);
        exit(1);
    }
}
