/*
 * textFileSink.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "textFileSink.h"
#include "binaryToText.h"

int first_line = 1;

char* get_plugin_name() {
	return "textFileSink";
}

char* get_plugin_version() {
	return "0.5.8";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

void initialize(unsigned long pPluginIndex) {
	uint32_t j = 0;

	pluginTextFileSinkIndex = pPluginIndex;
	baseFileName = getBaseFileName();
	main_output_buffer = getMainOutputBuffer();
	main_header_bv = getMainHeaderBv();
	
	set_input_source(0, main_output_buffer->buffer);

	// setup output file names
	if (baseFileName) {
		strncpy(txt_output_file_name, baseFileName, MAX_FILENAME_LEN);
		strncat(txt_output_file_name, FLOWS_TEXT_SUFFIX, MAX_FILENAME_LEN-strlen(baseFileName));

		// open flow output file
		if ((txt_output_file = fopen(txt_output_file_name, "wt")) == NULL) {
			fprintf(stderr, "textFileSink: Error: Can't open file %s in write mode! Maybe incorrect permissions?\n", txt_output_file_name);
			terminate();
		} 
	} else txt_output_file = stdout;

#if TEXT_FILE_SINK_PRINT_HEADER_FILE == 1
	// setup output file names
	if (!baseFileName) {
		if (captureFileName) {
			baseFileName = captureFileName;
			strncpy(txt_output_file_name, captureFileName, MAX_FILENAME_LEN);
		} else if (captureListFileName) {
			baseFileName = captureListFileName;
			strncpy(txt_output_file_name, captureListFileName, MAX_FILENAME_LEN);
		} else if (captureInterface) { 
			baseFileName = captureInterface;
			strncpy(txt_output_file_name, captureInterface, MAX_FILENAME_LEN);
		} else terminate();
			
	}

	strncpy(header_output_file_name, baseFileName, MAX_FILENAME_LEN);
	strncat(header_output_file_name, HEADER_SUFFIX, MAX_FILENAME_LEN-strlen(baseFileName));
	
	// open header output file
	if ((header_output_file = fopen(header_output_file_name, "wt")) == NULL) {
		fprintf(stderr, "textFileSink: Error: Can't open header file %s in write mode! Maybe incorrect permissions?\n", header_output_file_name);
		terminate();
	} else {
		// write headers
		fprintf(header_output_file, "# Header file for flow file\n# %s\n", txt_output_file_name);
		j = parse_binary_header2text(main_header_bv, 0, header_output_file);
		fclose(header_output_file);
	}
#endif // TEXT_FILE_SINK_PRINT_HEADER_FILE == 1

	
	// write header in flow file
#if TEXT_FILE_SINK_PRINT_HEADER == 1
	parse_binary_header2text(main_header_bv, 1, txt_output_file);
#endif // TEXT_FILE_SINK_PRINT_HEADER == 1

}


void bufferToSink(outputBuffer_t* buffer) {
	if (supOut) {
		return;
	}

	if (first_line) {
		first_line = 0;
		if (TEXT_FILE_SINK_PRINT_HEADER) fputc('\n', txt_output_file);
	}
	else fputc('\n', txt_output_file);

	parse_binary2text(main_header_bv, txt_output_file);
}


void onApplicationTerminate() {
#if TEXT_FILE_SINK_PRINT_HEADER == 1 && TEXT_FILE_SINK_EXTENDED_HEADER == 1
	if (baseFileName) {
		fseek(txt_output_file, 0, SEEK_SET);
		fprintf(txt_output_file, "%% %lu", totalFlows);
	}
#endif
	fclose(txt_output_file);
}
