/*
 * tranalyzer-b2t.c
 *
 *  Created on: May 3, 2010
 *      Author: torben
 */


#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include "binaryToText.h"

int op; // parameter option

char* input_file_name = NULL; 	// the name for the input file
char* output_file_name = NULL;	// the name for the output file
FILE* b2t_input_file = NULL;		// the input file
FILE* output_file = NULL;		// the output file
int write_header = 0;
int new_line_at_end = 0;
void print_usage(char* binary);

int main(int argc, char *argv[]) {
	while ((op = getopt(argc, argv, "r:w:cn")) != EOF) {
		switch (op) {
			case 'r':
				input_file_name = optarg;
				break;
			case 'w':
				output_file_name = optarg;
				break;
			case 'c':
				write_header = 1;
				break;
			case 'n':
				new_line_at_end = 1;
				break;
			default:
				print_usage(argv[0]);
				exit(1);
			}
	}

	// if no input file is given, print usage and terminate
	// else try to open input file in read mode
	if (!input_file_name) {
		print_usage(argv[0]);
		exit(1);
	} else {
		if ((b2t_input_file = fopen(input_file_name, "rt")) == NULL) {
			fprintf(stderr, "ERROR: Cannot open input file %s!\n", input_file_name);
			exit(1);;
		}
	}

	// if no output file is given, set output file to standard output
	// else try to open output file in write mode
	if (!output_file) output_file = stdout;
	else {
		if ((output_file = fopen(output_file_name, "wt")) == NULL) {
			fprintf(stderr, "ERROR: Cannot open input file %s!\n", input_file_name);
			exit(1);
		}
	}

	// estimate the length of the header
	uint32_t headerLength = 0;
	uint32_t value;

	for (;;) {
		if (fread(&value, sizeof(uint32_t), 1, b2t_input_file) != 1) {
			fprintf(stderr, "ERROR: Incorrect header in file %s\n", argv[1]);
			return 1;
		}
		headerLength++;
		if (value == UINT32_MAX) break; // header end reached
		if (headerLength == UINT32_MAX) {
			fprintf(stderr, "ERROR: header is too long. Seems to be incorrect\n");
			return 1;
		}
	}

	// read the header
	rewind(b2t_input_file);
	uint32_t headerBuffer[headerLength];
	fread(headerBuffer, sizeof(uint32_t), headerLength, b2t_input_file);


	binary_header_t *header = malloc(sizeof(binary_header_t));
	header->header = headerBuffer;
	header->header_length = headerLength;

	binary_value_t *bv = analyseHeader(header);

	if (write_header) {
		parse_binary_header2text(bv, 1, output_file);
		fprintf(output_file, "\n");
	}

	// set input source
	set_input_source(INPUT_TYPE_FILE, b2t_input_file);
	int first_line = 1;
	char tempChar;
	while (fread(&tempChar, sizeof(char), 1, b2t_input_file) != 0) {
		if (first_line) first_line = 0;
		else fprintf(output_file, "\n");
		ungetc(tempChar, b2t_input_file);
		parse_binary2text(bv, output_file);
	}
	if (new_line_at_end) fprintf(output_file, "\n");
	fclose(b2t_input_file);
	fclose(output_file);
	return 0;
}

void print_usage(char* binary) {
	fprintf(stderr, "Usage: %s -r INPUT [-w OUTPUT]\n"
					"transforms a binary into a text based tranalyzer flow file\n"
					"OPTIONS:\n"
					"-r INPUT       binary tranalyzer flow file\n\n"
					"OPTIONAL:\n"
					"-w OUTPUT      text output\n"
					"-c             write header into first column\n"
					"-n             append newline char at end of row\n", binary);
}
