#!/bin/bash


# ------------------------------------------
#-------------- CONFIG PART ----------------
# ------------------------------------------

PKG_NAME="tranalyzer"



# ------------------------------------------
#-------------- SCRIPT PART ----------------
# ------------------------------------------

DEBUG_MODE=0
NOCONFIG=0

function USAGE ()
{
    echo ""
    echo "USAGE:"
    echo "    autogen.sh [-ndh]"
    echo ""
    echo "OPTIONS:"
    echo "    -n  only execute make and copy plugins"
    echo "    -d  compile in debug and profiling mode"
    echo "    -h  show this usage information"
    echo ""
    exit $E_OPTERROR    # Exit and explain usage, if no argument(s) given.
}


#PROCESS ARGS
while getopts "dhn" Option
do
    case $Option in
        n    ) NOCONFIG=1;;
        d    ) DEBUG_MODE=1;;
        h    ) USAGE
               exit 0;;
        *    ) echo ""
               echo "Unimplemented or incomplete option chosen!"
               USAGE   # DEFAULT
    esac
done

shift $(($OPTIND - 1))
#  Decrements the argument pointer so it points to next argument.
#  $1 now references the first non option item supplied on the command line
#+ if one exists.

if [ "$NOCONFIG" != 1 ]; then
	if [ ! -d "m4" ]; then
        	mkdir m4
	fi

	aclocal
	automake --add-missing
	autoconf
	autoreconf --warnings=all
	if [ "$DEBUG_MODE" == 1 ]; then
		echo "Compiling in debug and profiling mode..."
		CFLAGS="-O0 -p -pg -g" ./configure "$@" && echo "Now type 'make' to compile $PKG_NAME" || exit 1
	else
	./configure "$@"
	fi
fi

make clean
make
