/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */
 
 /*
 * Inspired by http://msmvps.com/blogs/vandooren/archive/2007/01/05/creating-a-thread-safe-producer-consumer-queue-in-c-without-using-locks.aspx
 */
#if ENABLE_IO_BUFFERING != 0
#ifndef __IOBUFFER_H__
#define __IOBUFFER_H__

// includes
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <pcap.h>
#include <pthread.h>
#include "main.h"

//structs
typedef struct mtuPacket_s {
	u_char data[MAX_MTU];
} mtuPacket_t;

// functions
extern void terminate();
extern void cycleLRUList();

void ioBufferInitialize();
void *ioBufferThreadFunction(void *args);
void ioBufferPush(u_char* inqueue_uchar, const struct pcap_pkthdr *pcapHeader, const u_char* packet);
int ioBufferPop(struct pcap_pkthdr **header, u_char **packet);
void ioBufferTerminate();

// variables
volatile unsigned long writePos;
volatile unsigned long readPos;
volatile mtuPacket_t packetBuffer[IO_BUFFER_SIZE];
volatile struct pcap_pkthdr headerBuffer[IO_BUFFER_SIZE];
volatile int threadTerminated;
pthread_t ioBufferThread;

#endif
#endif

