/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "main.h"
#include "../config.h"

int main(int argc, char *argv[]) {
	// TODO: check for root rights, and - if we have them - print a warning msg

	// load ascii logo
	// tranalyzer_logo = loadAscii(LOGO_FILE_NAME);

	// load ascii ants
	// ants_ascii = loadAscii(ANTS_FILENAME);

	fprintf(
			stdout,
			"================================================================================\n"); // 80 chars
	fprintf(stdout, "%s %s (%s), %s. PID: %d\n", APPNAME, APPVERSION, CODENAME,
			RELEASE_TYPE, getpid());
	if (tranalyzer_logo != NULL)
		fprintf(stdout, "%s", tranalyzer_logo);// print logo
	fprintf(
			stdout,
			"================================================================================\n"); // 80 chars

	int op; // parameter option
	running = 1; // running state of the application. Set it to 0 and it terminates
	captureInterface = NULL; // the interface from which the packets get sniffed
	captureFileName = NULL; // if not read from an interface, read from a stored dump file
	esomFileName = NULL; // 
	captureListFileName = NULL;
	pluginFolder = NULL; // optional plugin folder name
	baseFileName = NULL; // the base output file name
	bpfCommand = NULL; // bpf filter command
	captureDescriptor = NULL; // pcap interface
	mainHashMap = NULL;
#if FRAGMENTATION >= 1	
	fragPendMap = NULL;
#endif //FRAGMENTATION	
	flows = NULL;
	tranalyzer_plugins = NULL; // tranalyzer plugins
	specialOutput = 0;
	startTS = 0;
	supOut = 0;
	totalfIndex = 0;
	maxNumFlows = 0;
	maxNumFlowsPeak = 0;
	main_header = NULL;
	main_header_bv = NULL;

	// read arguments passed to the application
	if (argc <= 2) {
		printUsage();
		exit(1);
//		terminate();
	}
	
	while ((op = getopt(argc, argv, "i:r:R:w:p:e:s")) != EOF) {
		switch (op) {
		case 'r':
			captureFileName = optarg;
			break;
		case 'R':
			captureListFileName = optarg;
			break;
		case 'w':
			baseFileName = optarg;
			break;
		case 'i':
			captureInterface = optarg;
			break;
		case 'p':
			pluginFolder = optarg;
			break;
		case 'e':
			esomFileName = optarg;
			break;
		case 's':
			specialOutput = 1;
			break;
		default:
			printUsage();
		//	terminate();
			exit(1);
			break;
		}
	}

	// post argument processing, check if at least one input AND the output is set
	if ((captureFileName && captureInterface) || (captureFileName && captureListFileName) || (captureListFileName && captureInterface)) {
		fprintf(stdout, "\nERROR: the '-r' '-R' and '-i' options can only be used exclusively!\n");
		printUsage(); // we don't know from which source to get the packets
		exit(1);
	}

	if (!captureFileName && !captureListFileName && !captureInterface) {
	        fprintf(stdout, "\nERROR: Ethernet Interface unknown!\n");
		printUsage(); // we don't know from which source to get the packets
		exit(1);
	}

	if (!baseFileName) {
	//	baseFileName = captureFileName;
//		fprintf(stdout, "\nOutput file missing!\n");
//		printUsage(); // user forgot to supply the base file name for the output (flows, protocols, packets)
//		exit(1);
	}
	
	// hook interrupts
	signal(SIGINT, catchInterrupt);
	signal(SIGTERM, catchInterrupt);
	signal(SIGUSR1, userInterrupt);
	signal(SIGUSR2, userInterrupt);

	// all remaining parameters belong to the bpf string
	bpfCommand = copy_argv(&argv[optind]);

	// prepare for sniffing
	prepareSniffing();

	// TODO: find better place to invoke gettimeofday functions

	// begin counting ticks
	gettimeofday(&startTime, NULL);

	mainLoop(); // begin sniffing

	gettimeofday(&endTime, NULL);
	timersub(&endTime, &startTime, &duration);

	fprintf(stdout, "Finished processing, %s: %ld.%02ld seconds\n", (captureInterface) ? "capture duration" : "duration of file reading", (long int)duration.tv_sec, (long int)duration.tv_usec);
	terminate();
	return 0;
}

void mainLoop() {
	fprintf(stdout, "Start ");

	if ( captureFileName ) fprintf(stdout, "processing file: %s\n", captureFileName);
	if ( captureListFileName ) fprintf(stdout, "processing list file: %s\n", captureListFileName);
	if ( captureInterface ) fprintf(stdout, "live capture on interface: %s\n", captureInterface);
	
	fprintf(stdout, "BPF: %s\n", bpfCommand);

#if ENABLE_IO_BUFFERING == 0
	int pcapDispatchReturnCode;
	if (captureListFileName && VERBOSE != 0) {
		fprintf(stdout, "processing file no. %lu of %lu: %s\n", act_capture_file_index + 1, capture_file_list->number_of_files, act_capture_list_element->file_name);
	}
	while (running) {
		// dispatch

#if DISPATCH_SLOW_DOWN > 0
		usleep(DISPATCH_SLOW_DOWN);
#endif /* DISPATCH_SLOW_DOWN > 0 */

		pcapDispatchReturnCode = pcap_dispatch(captureDescriptor, PACKETS_PER_BURST, perPacketCallback, NULL);
		switch (pcapDispatchReturnCode) {
		case -1:
			// An error occured, terminate
			running = 0;
			break;
		case 0:
			// Terminate if we are reading from a file rather than live-sniffing
			if (captureFileName) {
				running = 0;
			} else if (captureListFileName) {
				// set descriptor to next file
				if (act_capture_list_element->next) {
					pcap_close(captureDescriptor);
					act_capture_list_element = act_capture_list_element->next;
					act_capture_file_index++;
					if ((captureDescriptor = pcap_open_offline(act_capture_list_element->file_name, errbuf)) == NULL) {
						fprintf(stdout, "ERROR: pcap open %s: %s\n", act_capture_list_element->file_name, errbuf);
						terminate();
					}
					if (VERBOSE != 0) {
						fprintf(stdout, "Processing file no. %lu of %lu: %s\n", act_capture_file_index + 1, capture_file_list->number_of_files, act_capture_list_element->file_name);
					}


				} else {
					// there is no next file -> terminate
					running = 0;
				}
			} else {
				// reading from live-interface. As we're using non-blocking mode, pcap_dispatch returns zero immediately
				// if no packets are to be read at the moment. This would cost much cpu consumption when no packets can be read,
				// therefore sleep some time
				usleep(NO_PACKETS_WAIT_MICROSEC);
			}
			break;
		}
	}
#else
	struct pcap_pkthdr *pcapHeader;
	u_char* packet;

	while (running) {
		if (!ioBufferPop(&pcapHeader, &packet)) break;
		actTime.tv_sec = pcapHeader->ts.tv_sec;
		actTime.tv_usec = pcapHeader->ts.tv_usec;
		cycleLRULists();
		processPacket(NULL, pcapHeader, packet);
	}
#endif

}

void removeFlow(flow_t *flow) {
	// prevFlow and nextFlow are necessary because the flow and its opposite flow could be two following flows.
	// if so, the pointer we are using in our main function would then point to null,
	// because we are going to overwrite the current flow position.
	if (flow && hashTable_remove(mainHashMap, (char*) &flow->srcIP) == 0) {
#if FRAGMENTATION >= 1
		if (flow->status & IPV4_FRAG_PENDING) {
			flow->srcPort = flow->lastFragIPID;
			hashTable_remove(fragPendMap, (char*) &flow->srcIP); // remove fragPend
			flow->status = 0;
		}
#endif //FRAGMENTATION
		flow->lruNextFlow->lruPrevFlow = flow->lruPrevFlow;
		flow->lruPrevFlow->lruNextFlow = flow->lruNextFlow;
		if (lruPrevFlow == flow) lruPrevFlow = flow->lruPrevFlow;
		if (lruNextFlow == flow) lruNextFlow = flow->lruNextFlow;

		// remove the opposite flow as well
		unsigned long reverseFlowIndex = flow->oppositeFlowIndex;
		if (reverseFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
			flow_t *reverseFlow = &flows[reverseFlowIndex];
		
			if (hashTable_remove(mainHashMap, (char*) &reverseFlow->srcIP) == 0) {
#if FRAGMENTATION >= 1		
				if (reverseFlow->status & IPV4_FRAG_PENDING) {
					reverseFlow->srcPort = reverseFlow->lastFragIPID;
					hashTable_remove(fragPendMap, (char*) &reverseFlow->srcIP); // remove fragPend
					reverseFlow->status = 0;
				}
#endif //FRAGMENTATION
				reverseFlow->lruNextFlow->lruPrevFlow = reverseFlow->lruPrevFlow;
				reverseFlow->lruPrevFlow->lruNextFlow = reverseFlow->lruNextFlow;
				if (lruPrevFlow == reverseFlow) lruPrevFlow = reverseFlow->lruPrevFlow;
				if (lruNextFlow == reverseFlow) lruNextFlow = reverseFlow->lruNextFlow;
			}
			maxNumFlows--;
		}
		maxNumFlows--;
	}
}

void cycleLRULists() {
	// printf("**************** Cycle Nr. %lu ****************\n", numOfPackets);
	main_timeout_handler = main_timeout_manager.timeout_handler_list;
	while (main_timeout_handler != NULL) {
		cycleLRUList(main_timeout_handler);
		main_timeout_handler = main_timeout_handler->next;
	}
}

// This is a Debug function
void printLru() {
	lruPointer = &lruHead;
	while (lruPointer) {
		printf("[%p", lruPointer);
		if (lruPointer == &lruHead)
			printf(", (H)]");
		else if (lruPointer == &lruTail)
			printf(", (T)]");
		else if (lruPointer->timeout == INFINITY) {
			timeout_handler_t *th = main_timeout_manager.timeout_handler_list;
			while (th) {
				if (&(th->flow) == lruPointer) {
					printf(", (S:%.2f)]", th->timeout);
					break;
				}
				th = th->next;
			}
		} else {
			timeDiff = actTime.tv_sec - lruPointer->lastSeen.tv_sec;
			timeDiff += (((float) actTime.tv_usec - (float) lruPointer->lastSeen.tv_usec) / 1000000.0f);
			printf("(dT=%.2f),(t_out=%f)]", timeDiff, lruPointer->timeout);
		}

		if (lruPointer->lruNextFlow)
			printf(" -> ");
		lruPointer = lruPointer->lruNextFlow;
	}

	printf("\n");

	lruPointer = &lruTail;
	while (lruPointer) {
		printf("[%p", lruPointer);
		if (lruPointer == &lruHead)
			printf(", (H)]");
		else if (lruPointer == &lruTail)
			printf(", (T)]");
		else if (lruPointer->timeout == INFINITY) {
			timeout_handler_t *th = main_timeout_manager.timeout_handler_list;
			while (th) {
				if (&(th->flow) == lruPointer) {
					printf(", (S:%.2f)]", th->timeout);
					break;
				}
				th = th->next;
			}
		} else {
			timeDiff = actTime.tv_sec - lruPointer->lastSeen.tv_sec;
			timeDiff += (((float) actTime.tv_usec - (float) lruPointer->lastSeen.tv_usec) / 1000000.0f);

			printf("(dT=%.2f),(t_out=%f)]", timeDiff, lruPointer->timeout);
		}

		if (lruPointer->lruPrevFlow)
			printf(" -> ");
		lruPointer = lruPointer->lruPrevFlow;
	}
}

void cycleLRUList(timeout_handler_t *th) {

	if (captureInterface) {
		// when reading from a live interface and only one flow is in the list, we cannot check if the newest and oldest
		// flows have a difference smaller than FLOW_TIMEOUT, because the difference would be zero! Take the current time for
		// that special case
		gettimeofday(&actTime, NULL);
	}

	lruPrevFlow = NULL;
	lruNextFlow = NULL;

	// from the timeout handler work backwards (commit and remove flows) until we hit a flow which
	// is younger than the timeout value of the timeout handler

#if FORCE_MODE == 1
	while (num_rm_flows) {
		lruPointer = rm_flows[--num_rm_flows];
		lruFlowPrintout();
		removeFlow(lruPointer);
	}
#endif // FORCE_MODE == 1

	lruPointer = th->flow.lruPrevFlow;
	while (lruPointer != &lruHead) {
		// if flow is a sentinel skip it
		if (lruPointer->lastSeen.tv_sec == 0) {
			lruPointer = lruPointer->lruPrevFlow;
			continue;
		}

		// check if flow would be too young and could not have timed-out
		timeDiff = actTime.tv_sec - lruPointer->lastSeen.tv_sec;
		timeDiff += (((float)actTime.tv_usec - (float)lruPointer->lastSeen.tv_usec) / 1000000.0f);

		if (timeDiff < th->timeout) break; // flows too young, break

		lruPrevFlow = lruPointer->lruPrevFlow;

		if (timeDiff >= lruPointer->timeout) {
			// only remove flow if the opposite flow has timed-out too

			if (lruPointer->oppositeFlowIndex == HASHTABLE_ENTRY_NOT_FOUND) {
				lruFlowPrintout();
				removeFlow(lruPointer);
			} else {
				timeDiff = actTime.tv_sec - flows[lruPointer->oppositeFlowIndex].lastSeen.tv_sec;
				timeDiff += ((float) (actTime.tv_usec - flows[lruPointer->oppositeFlowIndex].lastSeen.tv_usec)) / 1000000.0f;

				if (timeDiff >= flows[lruPointer->oppositeFlowIndex].timeout) {
					lruFlowPrintout();
					removeFlow(lruPointer);
				}
			}
		}
	
		lruPointer = lruPrevFlow;
	}

	if (lruPointer != th->flow.lruPrevFlow) {
		// move timeout handler sentinel flow behind last inspected flow (behind lruPointer)
		th->flow.lruNextFlow->lruPrevFlow = th->flow.lruPrevFlow; // take out, step 1
		th->flow.lruPrevFlow->lruNextFlow = th->flow.lruNextFlow; // take out, step 2

		th->flow.lruNextFlow = lruPointer->lruNextFlow; // place in, step 1
		th->flow.lruPrevFlow = lruPointer; // place in, step 2

		lruPointer->lruNextFlow->lruPrevFlow = &(th->flow); // connect, step 1
		lruPointer->lruNextFlow = &(th->flow); // connect, step 2
	}

	// if the lru list is empty and we want to stop the application from creating new flows, terminate
	if (mainHashMap->freeListSize == mainHashMap->hashChainTableSize && exitPriority > 0)
		terminate();
}

void lruFlowPrintout() { // first, find out if the connection was bi-directional or uni-directional

	if (lruPointer->oppositeFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) { // bi-directional (normal case)
		if ( lruPointer->status & L3FLOWINVERT ) {
			srvFlowIndex = lruPointer->oppositeFlowIndex; 
			clntFlowIndex = lruPointer->flowIndex;
		} else {
			srvFlowIndex = lruPointer->flowIndex;
			clntFlowIndex = lruPointer->oppositeFlowIndex;
		} 
	
		printFlow(srvFlowIndex, clntFlowIndex); // print flows and remove them

	} else printFlow(lruPointer->flowIndex, HASHTABLE_ENTRY_NOT_FOUND);
}

void printFlow(unsigned long serverFlowIndex, unsigned long clientFlowIndex) {

	uint8_t temp_dir;
	uint16_t i;
		

	if (serverFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		srvFlow = &flows[serverFlowIndex];
		timersub(&srvFlow->lastSeen, &srvFlow->firstSeen, &srvFlow->duration);
		
		supOut=0;
		
		// print direction
		temp_dir = 0;
		outputBuffer_append(main_output_buffer, (char*) &temp_dir, 1);

		// EVENT: onFlowTerminate
		for (i = 0; i < tranalyzer_plugins->num_of_func_pluginOnFlowTerminate; i++) {
			tranalyzer_plugins->pluginOnFlowTerminate[i](serverFlowIndex);
		}
		if (!supOut) {
			// EVENT: send buffer to sinks
			for (i = 0; i < tranalyzer_plugins->num_of_func_bufferToSink; i++) {
				tranalyzer_plugins->pluginBufferToSink[i](main_output_buffer);
			}
		}
		// reset buffer
		outputBuffer_reset(main_output_buffer);
	}

	if (clientFlowIndex != HASHTABLE_ENTRY_NOT_FOUND) {
		clntFlow = &flows[clientFlowIndex];
		timersub(&clntFlow->lastSeen, &clntFlow->firstSeen, &clntFlow->duration);
		
		supOut=0;
		
		// print direction
		temp_dir = 1;
		outputBuffer_append(main_output_buffer, (char*) &temp_dir,1);

		// EVENT: onFlowTerminate
		for (i = 0; i < tranalyzer_plugins->num_of_func_pluginOnFlowTerminate; i++) {
			tranalyzer_plugins->pluginOnFlowTerminate[i](clientFlowIndex);
		}
		if (!supOut) {

			// EVENT: send buffer to sinks
			for (i = 0; i < tranalyzer_plugins->num_of_func_bufferToSink; i++) {
				tranalyzer_plugins->pluginBufferToSink[i](main_output_buffer);
			}
		}
		// reset buffer
		outputBuffer_reset(main_output_buffer);
	}

}

void prepareSniffing() {
	uint16_t i;

	globalWarn = 0; //  initialize warning byte for snaplength and timing 
	bytesonWire = 0;
	numAlarms = 0;
	// prepare data source
	if (captureFileName) {
		// read the number of bytes the file consists of
		if (stat(captureFileName, &fileStats) == -1) {
			fprintf(stdout, "ERROR: Cannot get stats of file %s, errno =  %d\n", captureFileName, errno);
			// terminate();
			captureFileSize = 0;
		} else {
			captureFileSize = fileStats.st_size;
		}

		// capture from a dump file
		if ((captureDescriptor = pcap_open_offline(captureFileName, errbuf)) == NULL) {
			fprintf(stdout, "ERROR: Pcap open %s: %s\n", captureFileName, errbuf);
			terminate();
		}
	}

	if (captureInterface) {
		// open in promisc mode
		if ((captureDescriptor = pcap_open_live(captureInterface, SNAPLEN, 1, CAPTURE_TIMEOUT, errbuf)) == NULL) {
			fprintf(stdout, "ERROR: Pcap open %s: %s\n", captureInterface, errbuf);
			terminate();
		}

		// malloc memory for the pcap live statistics
		pcap_statistics = malloc(sizeof(pcap_stats));
	}

	if (captureListFileName) {
		FILE *listFile;
		// open file with list of pcap dump files in it
		if ((listFile = fopen(captureListFileName, "r")) == NULL){
			fprintf(stdout, "ERROR: Cannot open file %s, errno = %d\n", captureListFileName, errno);
			terminate();
		}
		if ((capture_file_list = get_capture_file_list(listFile)) == NULL) {
			fprintf(stdout, "ERROR: No files specified in %s\n", captureListFileName);
			terminate();
		}

		// start with the first file in list
		act_capture_list_element = capture_file_list->file_list;
		act_capture_file_index = 0;
		if ((captureDescriptor = pcap_open_offline(act_capture_list_element->file_name, errbuf)) == NULL) {
			fprintf(stdout, "ERROR: Pcap open %s: %s\n", act_capture_list_element->file_name, errbuf);
			terminate();
		}
	}
		
	bytesProcessed = 0; // no file, therefore no global header

	// setup the bpf filter
	if (bpfCommand && strlen(bpfCommand) > 0) {
		// try to compile the filter
		if (pcap_compile(captureDescriptor, &bpfProgram, bpfCommand, ENABLE_BPF_OPTIMIZATION, 0) == -1) {
			fprintf(stdout, "ERROR: Calling pcap_compile\n");
			terminate();
		}

		// set the compiled program as the filter
		if (pcap_setfilter(captureDescriptor, &bpfProgram) == -1) {
			fprintf(stdout, "ERROR: Setting filter\n");
			terminate();
		}
	}

	// TODO: put in separate function and check for file validity

	// reset lru list
	lruHead.lruNextFlow = &lruTail;
	lruHead.lruPrevFlow = NULL;
	lruTail.lruNextFlow = NULL;
	lruTail.lruPrevFlow = &lruHead;

	// initialize timeout manager with default timeout handler
	main_timeout_manager.timeout_handler_list = NULL;
	timeout_handler_add(FLOW_TIMEOUT);

	// initialize main buffer
	main_output_buffer = outputBuffer_initialize(MAIN_OUTPUT_BUFFER_SIZE);
	header_output_buffer = outputBuffer_initialize(MAIN_OUTPUT_BUFFER_SIZE);

	// misc
	totalFlows = 0; // no flows at the beginning
	exitPriority = 0; // while running, we don't want to terminate the application if not indicated by the user
	if (captureInterface) {
		if (pcap_setnonblock(captureDescriptor, NON_BLOCKING_MODE, errbuf) == -1) {
			fprintf(stdout, "ERROR: Could not set blocking mode %d: %s\n",
					NON_BLOCKING_MODE, errbuf);
			exit(1);
		}
	}

	// sanity check
	if (strlen(getenv("HOME")) + strlen(PLUGIN_FOLDER) > 254) {
		fprintf(stdout, "ERROR: Plugin folder name is too long!");
		exit(1);
	}

	char user_plugin_folder[MAX_FILENAME_LEN + 1];
	if (pluginFolder) {
		strncpy(user_plugin_folder, pluginFolder, MAX_FILENAME_LEN);
		if (user_plugin_folder[strlen(user_plugin_folder) - 1] != '/')
			strncat(user_plugin_folder, "/", MAX_FILENAME_LEN - strlen(user_plugin_folder));
	} else {
		strncpy(user_plugin_folder, getenv("HOME"), MAX_FILENAME_LEN);
		if (user_plugin_folder[strlen(user_plugin_folder) - 1] != '/')
			strncat(user_plugin_folder, "/", MAX_FILENAME_LEN - strlen(user_plugin_folder));
		strncat(user_plugin_folder, PLUGIN_FOLDER, MAX_FILENAME_LEN - strlen(user_plugin_folder));
		if (user_plugin_folder[strlen(user_plugin_folder) - 1] != '/')
			strncat(user_plugin_folder, "/", MAX_FILENAME_LEN - strlen(user_plugin_folder));
	}

	if ((tranalyzer_plugins = load_tranalyzer_plugins(user_plugin_folder)) == NULL) {
		fprintf(stdout, "ERROR: No plugins in folder '%s' or plugins couldn't be loaded!\n", user_plugin_folder);
		// terminate();
	} else {

		printf("Active plugins:\n");
		for (i = 0; i < tranalyzer_plugins->num_of_plugins; i++) {
			printf("  %02u: %s, version %s\n", i, tranalyzer_plugins->plugin[i].name, tranalyzer_plugins->plugin[i].version);
#if VERBOSE >= 2
			printf("     Functions:\n");
			if (tranalyzer_plugins->plugin[i].initialize != NULL) printf("     initialize\n");
			if (tranalyzer_plugins->plugin[i].printHeader != NULL) printf("     printHeader\n");
			if (tranalyzer_plugins->plugin[i].onFlowGenerated != NULL) printf("     onFlowGenerated\n");
			if (tranalyzer_plugins->plugin[i].claimLayer2Information != NULL) printf("     claimLayer2Information\n");
			if (tranalyzer_plugins->plugin[i].claimLayer3Information != NULL) printf("     claimLayer3Information\n");
			if (tranalyzer_plugins->plugin[i].claimLayer4Information != NULL) printf("     claimLayer4Information\n");
			if (tranalyzer_plugins->plugin[i].onFlowTerminate != NULL) printf("     onFlowTerminate\n");
			if (tranalyzer_plugins->plugin[i].onApplicationTerminate != NULL) printf("     onApplicationTerminate\n");
			if (tranalyzer_plugins->plugin[i].bufferToSink != NULL) printf("     bufferToSink\n");
			printf("\n");
#endif /* VERBOSE >= 2 */
		}
	}

	mainHashMap = hashTable_initialize(1.0f, ((char*) &lruHead.layer4Protocol - (char*) &lruHead.srcIP + sizeof(lruHead.layer4Protocol)));

	// initialize flow array
	flows = (flow_t*) malloc(sizeof(flow_t) * mainHashMap->hashChainTableSize);
	memset(flows, '\0', sizeof(flow_t) * mainHashMap->hashChainTableSize);

#if FRAGMENTATION >= 1

#if IPV6_ACTIVATE == 1
	fragPendMap = hashTable_initialize(1.0f, ((char*) &lruHead.srcPort - (char*) &lruHead.srcIP + sizeof(uint32_t)));
#else
	fragPendMap = hashTable_initialize(1.0f, ((char*) &lruHead.srcPort - (char*) &lruHead.srcIP + sizeof(uint16_t)));
#endif
	// initialize frag array
	fragPend = (unsigned long*) malloc(sizeof(unsigned long) * fragPendMap->hashChainTableSize);
	memset(fragPend, '\0', sizeof(unsigned long) * fragPendMap->hashChainTableSize);
#endif //FRAGMENTATION

	// build main header struct.
	buildHeaders();

	numOfPackets = 0; // initialize number of packets counter
	numOfV6Packets = 0; // initialize number of ipv4 packets counter
	numOfV6Packets = 0; // initialize number of ipv6 packets counter
	numOfFragV4Packets = 0; // init num of ipv4 frag packets counter
	numOfFragV6Packets = 0; // init num of ipv6 frag packets counter
	numOfARPPackets = 0;
	numOfRARPPackets = 0;
#if FORCE_MODE == 1
	num_rm_flows = 0;
#endif // FORCE_MODE == 1
	if (tranalyzer_plugins) {
		for (i = 0; i < tranalyzer_plugins->num_of_func_pluginInitialize; i++) {
			tranalyzer_plugins->pluginInitialize[i](i);
		}
	}

#if ENABLE_IO_BUFFERING != 0
	// init io buffer
	ioBufferInitialize();
#endif /* ENABLE_IO_BUFFERING */

}

/****************************************************************************
 *
 * Function: copy_argv(u_char **)
 *
 * Purpose: Copies a 2D array (like argv) into a flat string.
 *          Shamelessly stolen from TCPDump.
 *
 * Arguments: argv => 2D array to flatten
 *
 * Returns: Pointer to the flat string
 *
 ****************************************************************************/
char *copy_argv(char **argv) {
	char **p;
	unsigned int len = 0;
	char *buf;
	char *src, *dst;
	void ftlerr(char *, ...);

	p = argv;
	if (*p == 0) return 0;

	while (*p) len += strlen(*p++) + 1;

	buf = (char *) malloc(len);

	if (buf == NULL) {
		fprintf(stdout, "ERROR: malloc() failed: %s\n", strerror(errno));
	}
	p = argv;
	dst = buf;

	while ((src = *p++) != NULL) {
		while ((*dst++ = *src++) != '\0')
			;
		dst[-1] = ' ';
	}

	dst[-1] = '\0';

	return buf;
}

void printUsage() {
	fprintf(stdout, "Usage: %s [-r INPUTFILE | -R INPUTFILE | -i INTERFACE] [-w OUTPUTPREFIX]  [BPF]\n\n", APPNAME);
	fprintf(stdout, "Analyses a file INPUTFILE or analyses files listed in a file INPUTFILE or\n"
			"captures traffic from a given UNIX interface INTERFACE and writes the output\n"
			"to files with prefix OUTPUTPREFIX, if -w is omitted stdout is used.\n\n");
	fprintf(stdout, "Optional arguments:\n");
	fprintf(stdout, "-p PATH    load plugins from path PATH instead from ~/.tranalyzer/plugins\n"
			"-s         activates special packet oriented output of plugins, if implemented\n"
			"-e         Pcap File name, when pcap plugins are loaded\n"
			"[BPF]      Berkely Packet Filter command, as in tcpdump\n\n");
}

void terminate() {
	unsigned int i;
	float f;

	fflush(NULL); // commit all changes in all buffers
	fprintf(stdout, "Shutting down %s %s...\n", APPNAME, APPVERSION);

	strftime(timeBuf, MAX_TM_BUF, "%a %d %b %Y %X", localtime(&actTime.tv_sec));
	fprintf(stdout, "Dump stop: %ld.%06ld sec : %s.%06ld\n", (long int)actTime.tv_sec, (long int)actTime.tv_usec, timeBuf, (long int)actTime.tv_usec);
	
	timersub(&actTime, &startTStamp, &duration);
	fprintf(stdout, "Total dump duration: %ld.%06ld sec\n", (long int)duration.tv_sec, (long int)duration.tv_usec);

#if ENABLE_LOG_FILE !=0
	write_log_file();
#endif

	// terminate flows
	if (totalFlows > 0) {
		lruPointer = lruTail.lruPrevFlow;
		while (lruPointer != &lruHead) {
			// set correct nextpointer because we are going to delete the pointer with the removal of the flow(s) 	 
			lruNextFlow = lruPointer->lruPrevFlow;
			if (lruPointer->timeout != INFINITY) {
				lruFlowPrintout();
				removeFlow(lruPointer);
			}
			lruPointer = lruNextFlow;
		}
	}

	// terminate plugins
	if (tranalyzer_plugins) {
		for (i = 0; i < tranalyzer_plugins->num_of_func_pluginOnApplicationTerminate; i++) {
			tranalyzer_plugins->pluginOnApplicationTerminate[i]();
		}
		unload_tranalyzer_plugins(tranalyzer_plugins);
	}

	fprintf(stdout, "Number of processed packets: %lu\n", numOfPackets);
	// terminate timeout handlers
	timeout_handler_t *th_curr = main_timeout_manager.timeout_handler_list;
	timeout_handler_t *th_next;
	while (th_curr) {
		th_next = th_curr->next;
		free(th_curr);
		th_curr = th_next;
	}

#if ENABLE_IO_BUFFERING != 0
	ioBufferTerminate();
#endif /* ENABLE_IO_BUFFERING */

	fflush(NULL); // commit all changes in all buffers
	if (bpfCommand) free(bpfCommand);

	if (captureInterface && captureDescriptor) {
		pcap_stats(captureDescriptor, pcap_statistics);
		fprintf(stdout, "Total number of received packets: %u\n", pcap_statistics->ps_recv);
		fprintf(stdout, "Total number of dropped packets by kernel / interface: %u / %u\n", pcap_statistics->ps_drop, pcap_statistics->ps_ifdrop);
	}

	if (captureDescriptor) pcap_close(captureDescriptor);

	hashTable_destroy(mainHashMap);
	mainHashMap = NULL;
#if FRAGMENTATION >= 1
	hashTable_destroy(fragPendMap);
	fragPendMap = NULL;
#endif //FRAGMENTATION
	if (flows) free(flows);
	flows = NULL;

	if (tranalyzer_logo) free(tranalyzer_logo);
	tranalyzer_logo = NULL;

	if (ants_ascii) free(ants_ascii);
	ants_ascii = NULL;

	// destroy buffers
	outputBuffer_destroy(main_output_buffer);
	outputBuffer_destroy(header_output_buffer);
	
	killHeader(main_header_bv); // TODO: Does not fix the leak!!

	fprintf(stdout, "Number of processed traffic bytes: %llu\n", bytesProcessed);
	fprintf(stdout, "Number of ARP packets: %lu\n", numOfARPPackets);
	fprintf(stdout, "Number of RARP packets: %lu\n", numOfRARPPackets);
	fprintf(stdout, "Number of IPv4 fragmented packets: %lu\n", numOfFragV4Packets);
	fprintf(stdout, "Number of IPv6 fragmented packets: %lu\n", numOfFragV6Packets);
	fprintf(stdout, "Number of IPv4 packets: %lu\n", numOfV4Packets);
	fprintf(stdout, "Number of IPv6 packets: %lu\n", numOfV6Packets);
	i = (IPV6_ACTIVATE) ? 6 : 4;
	fprintf(stdout, "Number of processed IPv%d flows:\t %ld\n", i, totalFlows);
	fprintf(stdout, "max Number of IPv%d flows:\t %ld\n", i, maxNumFlowsPeak);
	fprintf(stdout, "Number of IPv%d flows plugin alarms: %ld\n", i, numAlarms);
	f = (float)(duration.tv_sec+duration.tv_usec/1000000.0f)*125.0f;
	if ( f == 0.0 ) f = FLT_MIN;  
	fprintf(stdout, "Average snapped Bandwidth: %10.3f KBit/s\n", (float)bytesProcessed/f);
	fprintf(stdout, "Average full IP Bandwidth: %10.3f KBit/s\n", (float)bytesonWire/f);
	fprintf(stdout, "Average full raw Bandwidth: %10.3f KBit/s\n", (float)rawbytesonWire/f);
	
	if (globalWarn & SNAPLENGTH) fprintf(stdout, "Warning: Reduced snap length\n");
	if (globalWarn & TIMEJUMP) fprintf(stdout, "Warning: Timestamp jump, probably due to multi path packet delay or NTP operation\n"); 
	if (globalWarn & IPV4_FRAG_HDSEQ_ERR) {
		fprintf(stdout, "Warning: IPv4 Fragmentation header packet missing");
		if (FRAG_HLST_CRFT) fprintf(stdout, "\n");
		else fprintf(stdout, ", trailing packets ignored\n"); 
	}
	if (globalWarn & L2_VLAN) fprintf(stdout, "Warning: VLAN encapsulation\n"); 
	if (globalWarn & L2_MPLS) fprintf(stdout, "Warning: MPLS encapsulation\n"); 
	if (globalWarn & L2_L2TP) fprintf(stdout, "Warning: L2TP encapsulation\n"); 
	if (globalWarn & L2_PPP) fprintf(stdout, "Warning: PPP encapsulation\n"); 
	if (globalWarn & L2_GRE) fprintf(stdout, "Warning: GRE encapsulation\n"); 
	
	exit(0);
}

void catchInterrupt(int signal) {
	// Shut down, needed when we're live-capturing
	exitPriority++;
	if (exitPriority >= PROCESS_SIGKILL_LEVEL_THRESHOLD) {
		fprintf(stdout, "\nUser requested immediate abortion, terminating now...\n");
		running = 0;
		if (captureDescriptor)
			pcap_breakloop(captureDescriptor);
	} else {
		fprintf(stdout, "\nUser requested abortion, stopping from creating new flows\n");
	}
}

void userInterrupt(int signal) {
	fprintf( stdout, "\n================================================================================\n"); // 80 chars
	fprintf(stdout, "Showing progress for %s %s (%s), %s. PID: %d\n", APPNAME,
			APPVERSION, CODENAME, RELEASE_TYPE, getpid());
	if (captureInterface) {
		pcap_stats(captureDescriptor, pcap_statistics);
		fprintf(stdout, "Received packets: %u\n", pcap_statistics->ps_recv);
		fprintf(stdout, "Dropped packets by kernel / interface: %u / %u\n",
				pcap_statistics->ps_drop, pcap_statistics->ps_ifdrop);
	}
	fprintf(stdout, "Current processed packets: %ld\n", numOfPackets);
	fprintf(stdout, "Current processed flows: %ld\n", totalFlows);

	if (captureFileName && captureFileSize)
		fprintf(stdout, "Total bytes to process: %llu\n", captureFileSize);

	if (captureListFileName) {
		fprintf(stdout, "Number of files to process: %lu\nCurrent file: %s\nFile index: %lu\n", capture_file_list->number_of_files, act_capture_list_element->file_name, act_capture_file_index + 1);
		if (capture_file_list->size_of_files) {
			fprintf(stdout, "Total bytes to process: %llu\n", capture_file_list->size_of_files);
			fprintf(stdout, "Current file size in bytes: %llu\n", act_capture_list_element->file_size);
		}
	}

	fprintf(stdout, "Total packet bytes processed so far: %llu\n", bytesProcessed);
	if (captureFileName && captureFileSize) {
		fprintf(stdout, "Total file bytes processed so far: %llu\n", 24 + bytesProcessed + (numOfPackets * 16)); // 24 = size of global pcap header, 16 = pcap header of every capture packet. (see http://wiki.wireshark.org/Development/LibpcapFileFormat)
		fprintf(stdout, "Percentage completed: %.2f\n", 100.0f / captureFileSize * (24 + bytesProcessed + (numOfPackets * 16)));
	}
	else if (captureListFileName && capture_file_list->size_of_files) {
		fprintf(stdout, "Percentage completed: %.2f\n", 100.0f / capture_file_list->size_of_files * ((24 * capture_file_list->number_of_files) + bytesProcessed + (numOfPackets * 16)));
	}

#if VERBOSE > 0
	fprintf(stdout, "Fill size of main hash map: %lu (%.2f%%)\n", mainHashMap->hashChainTableSize - mainHashMap->freeListSize, (float) (mainHashMap->hashChainTableSize - mainHashMap->freeListSize) / (float) (mainHashMap->hashChainTableSize) * 100.0);
#endif
	if (globalWarn & SNAPLENGTH) fprintf(stdout, "Warning: Reduced snap length\n");
	if (globalWarn & TIMEJUMP) fprintf(stdout, "Warning: Timestamp jump, probably due to multi path packet delay or NTP operation\n");
	if (globalWarn & IPV4_FRAG) fprintf(stdout, "Warning: IPv4 Fragmentation\n"); 
	if (globalWarn & IPV4_FRAG_HDSEQ_ERR) {
		fprintf(stdout, "Warning: IPv4 Fragmentation header packet missing");
		if (FRAG_HLST_CRFT) fprintf(stdout, "\n");
		else fprintf(stdout, ", trailing packets ignored\n"); 
	}
	if (globalWarn & L2_VLAN) fprintf(stdout, "Warning: VLAN encapsulation\n"); 
	if (globalWarn & L2_MPLS) fprintf(stdout, "Warning: MPLS encapsulation\n"); 
	if (globalWarn & L2_L2TP) fprintf(stdout, "Warning: L2TP encapsulation\n"); 
	if (globalWarn & L2_PPP) fprintf(stdout, "Warning: PPP encapsulation\n"); 
	if (globalWarn & L2_GRE) fprintf(stdout, "Warning: GRE encapsulation\n"); 

	fprintf(stdout, "================================================================================\n"); // 80 chars

	if (ants_ascii != NULL) fprintf(stdout, "%s", ants_ascii); // the ants
}

void buildHeaders() {
	uint32_t i;
	// get flow stats header
	// build binary value for the direction
	main_header_bv = bv_new_bv("Flow direction", "dir", 0, 1, bt_flow_direction);

	if (tranalyzer_plugins) {
		// get binary values from plugins
		for (i = 0; i < tranalyzer_plugins->num_of_func_pluginPrintHeader; i++) {
			bv_append_bv(main_header_bv, tranalyzer_plugins->pluginPrintHeader[i]());
		}
	}
}

/*
 * int timeout_handler_add(float *timeout_in_sec)
 * adds a new timeout handler struct to the main timeout manager struct
 *
 * Arguments:
 * float timeout_in_sec    the timeout in seconds
 *
 * Return values:
 * 0    success
 * 1    no success, not used until now
 * 2    timeout already in list (no real error, different situations can cause same timeout values)
 */
int timeout_handler_add(float timeout_in_sec) {
	// first check if there is already a timeout_handler in the list;
	timeout_handler_t *th = main_timeout_manager.timeout_handler_list;
	while (th != NULL) {
		if (th->timeout == timeout_in_sec)
			return 2; // timeout already in list
		th = th->next;
	}

	// timeout value not in list -> build new timeout handler
	timeout_handler_t *th_new = malloc(sizeof(timeout_handler_t));
	th_new->timeout = timeout_in_sec;
	th_new->flow.timeout = INFINITY; // a sentinel never times out
	th_new->flow.lastSeen.tv_sec = 0; // An impossible timestamp of zero seconds marks it as a sentinel (see cycleLRUList)
	th_new->next = NULL;

	// place it in front of lru_tail
	th_new->flow.lruPrevFlow = lruTail.lruPrevFlow;
	th_new->flow.lruNextFlow = &lruTail;

	lruTail.lruPrevFlow->lruNextFlow = &(th_new->flow);
	lruTail.lruPrevFlow = &(th_new->flow);

	if (main_timeout_manager.timeout_handler_list == NULL) {
		main_timeout_manager.timeout_handler_list = th_new; // the new timeout handler is the only one in the list, so place it at the list's head
		return 0;
	}

	// add it at the right position in the timeout handler list
	// timeout handler with biggest timeout first
	timeout_handler_t *th_prev = main_timeout_manager.timeout_handler_list;

	th = th_prev;
	while (th != NULL) {
		if (th_new->timeout > th->timeout) {
			// add new timeout handler in front of th
			th_new->next = th;
			if (th_prev == th) {
				// The new timeout is at the top of the list -> set entry point to new timeout handler
				main_timeout_manager.timeout_handler_list = th_new;
				return 0;
			} else {
				th_prev->next = th_new;
				return 0;
			}
		}
		if (th_prev != th)
			th_prev = th_prev->next;
		th = th->next;
	}

	// new timeout is at the end of the list
	th_prev->next = th_new;
	return 0;
}

void write_log_file() {
	int log_i;
	float f;

	//Stop the execution if no file or plugins exist.
	//Actually this kind of check should be performed before entering the function..
	if (baseFileName == NULL || tranalyzer_plugins == NULL) return;

	strncpy(logFileName, baseFileName, MAX_FILENAME_LEN);
	strncat(logFileName, LOG_SUFFIX, MAX_FILENAME_LEN - strlen(baseFileName));

	// open log file
	if ((logFile = fopen(logFileName, "w")) == NULL) {
		fprintf(stdout, "ERROR: Can't open file %s in write mode! Set appropriate permissions\n", logFileName);
	} else {

		fprintf(stdout, "Writing log information to file %s...\n", logFileName);

		/*****	SYSTEM INFO   *****/

		fprintf(logFile, "SYSTEM INFO:\n");

		//HOST NAME
		char info[MAX_FILENAME_LEN + 1];
		memset(info, '\0', MAX_FILENAME_LEN + 1);
		if (gethostname(info, MAX_FILENAME_LEN) == -1)
			fprintf(stdout, "ERROR: Cannot acquire host name \n");
		else
			fprintf(logFile, "\tHost:\t\t%s\n", info);

		//USER
		memset(info, '\0', MAX_FILENAME_LEN + 1);
		strncpy(info, getenv("USER"), MAX_FILENAME_LEN);
		fprintf(logFile, "\tUser:\t\t%s\n", info);

		//IP of the interfaces
		print_IP_interfaces(logFile);

		/**** 		TRANALYZER EXECUTABLE INFO  	*****/

		fprintf(logFile, "\nEXECUTABLE INFO:");

		//Name of the application and version
		fprintf(logFile, "\n\tName:\t\t%s\n\tVersion:\t%s", APPNAME, APPVERSION);

		//Path of the executable
		char path[MAX_FILENAME_LEN + 1];
		ssize_t len;
		if ((len = readlink("/proc/self/exe", path, MAX_FILENAME_LEN + 1)) == -1) {
			fprintf(stdout, "ERROR: Cannot acquire path of executable");
		} else {
			path[len] = '\0';
			dirname(path);
			strcat(path, "/");
			fprintf(logFile, "\n\tExe Path:\t%s", path);
		}

		//Mode, Input File:
		if (captureInterface && captureDescriptor)
			fprintf(logFile, "\n\tMode: \t\tLive\n\tInput: \t\t----");
		else if (captureFileName && captureDescriptor)
			fprintf(logFile, "\n\tMode: \t\tInput File\n\tInput: \t\t%s", captureFileName);
		else if (captureListFileName && captureDescriptor) {
			fprintf(logFile, "\n\tMode: \t\tInput File List\n\tInput: \t\t%s\n\tFiles in input file:", captureListFileName);
			file_list_element_t *temp_file_element = capture_file_list->file_list;

			while (temp_file_element) {
				fprintf(logFile, "\n\t\t\t");
				fprintf(logFile, "%s", temp_file_element->file_name);
				temp_file_element = temp_file_element->next;
			}
		}

		fprintf(logFile, "\n\tOutputprefix:\t%s", baseFileName);

		//TIME info
		char *s_time = ctime(&(startTime.tv_sec));
		fprintf(logFile, "\n\tStart Time:\t%s", s_time);

		s_time = ctime(&(endTime.tv_sec));
		fprintf(logFile, "\tEnd Time:\t%s\tDuration:\t%ld.%06ld secs\n", s_time, (long int)duration.tv_sec, (long int)duration.tv_usec);

		/*****	PLUGINS INFO 	*****/

		fprintf(logFile, "\n\nPLUGINS INFO:\n");
		char user_plugin_folder[MAX_FILENAME_LEN + 1];

		if (pluginFolder) {
			fprintf(logFile, "\tFolder:\t %s\n", pluginFolder);
		} else {
			memset(user_plugin_folder, '\0', MAX_FILENAME_LEN + 1);
			strncpy(user_plugin_folder, getenv("HOME"), MAX_FILENAME_LEN);
			strncat(user_plugin_folder, "/", 1);
			strncat(user_plugin_folder, PLUGIN_FOLDER, MAX_FILENAME_LEN - 1
					- strlen(getenv("HOME")));
			fprintf(logFile, "\tFolder:\t %s\n", user_plugin_folder);
		}

		for (log_i = 0; log_i < tranalyzer_plugins->num_of_plugins; log_i++) {
			fprintf(logFile, "\t%02u\tversion: %s\t  %s \n", log_i,
					tranalyzer_plugins->plugin[log_i].version,
					tranalyzer_plugins->plugin[log_i].name);
		}

		/*****		PROCESSING INFO		******/

		fprintf(logFile, "\nPROCESSING INFO:\n");

		if (captureInterface && captureDescriptor) {
			pcap_stats(captureDescriptor, pcap_statistics);
			fprintf(logFile, "\tReceived Packets:\t %u\n",
					pcap_statistics->ps_recv);
			fprintf(logFile,
					"\tDropped Packets by kernel / interface: %u / %u\n",
					pcap_statistics->ps_drop, pcap_statistics->ps_ifdrop);
		}

		fprintf(logFile, "Number of processed Packets:\t %lu\n", numOfPackets);
		fprintf(logFile, "Number of processed Bytes:\t %llu\n", bytesProcessed);
		fprintf(logFile, "Number of ARP packets: %lu\n", numOfARPPackets);
		fprintf(logFile, "Number of RARP packets: %lu\n", numOfRARPPackets);
		fprintf(logFile, "Number of IPv4 fragmented packets: %lu\n", numOfFragV4Packets);
		fprintf(logFile, "Number of IPv6 packets: %lu\n", numOfV6Packets);
		fprintf(logFile, "Number of IPV6 packets:\t %lu\n", numOfV6Packets);
		log_i = (IPV6_ACTIVATE) ? 6 : 4;
		fprintf(logFile, "Number of processed IPv%d flows:\t %ld\n", log_i, totalFlows);
 
		strftime(timeBuf, MAX_TM_BUF, "%a %d %b %Y %X", localtime(&actTime.tv_sec));
		fprintf(logFile, "Dump stop: %ld.%06ld sec : %s.%06ld\n", (long int)actTime.tv_sec, (long int)actTime.tv_usec, timeBuf, (long int)actTime.tv_usec);
	
		timersub(&actTime, &startTStamp, &duration);
		fprintf(logFile, "Total dump duration: %ld.%06ld sec\n", (long int)duration.tv_sec, (long int)duration.tv_usec);

		f = (float)(duration.tv_sec+duration.tv_usec/1000000.0f)*125.0f;
		if ( f == 0.0 ) f = FLT_MIN;  
		fprintf(logFile, "Average snapped Bandwidth: %10.3f KBit/s\n", (float)bytesProcessed/f);
		fprintf(logFile, "Average full IP Bandwidth: %10.3f KBit/s\n", (float)bytesonWire/f);
		
		if (globalWarn & SNAPLENGTH) fprintf(logFile, "Warning: Reduced snap length\n"); 
		if (globalWarn & TIMEJUMP) fprintf(logFile, "Warning: Timestamp jump, probably due to multi path packet delay or NTP operation\n");
		if (globalWarn & IPV4_FRAG_HDSEQ_ERR) {
			fprintf(logFile, "Warning: IPv4 Fragmentation header packet missing");
			if (FRAG_HLST_CRFT) fprintf(logFile, "\n");
			else fprintf(logFile, ", trailing packets ignored\n"); 
		}
		if (globalWarn & L2_VLAN) fprintf(logFile, "Warning: VLAN encapsulation\n"); 
		if (globalWarn & L2_MPLS) fprintf(logFile, "Warning: MPLS encapsulation\n"); 
		if (globalWarn & L2_L2TP) fprintf(logFile, "Warning: L2TP encapsulation\n"); 
		if (globalWarn & L2_PPP) fprintf(logFile, "Warning: PPP encapsulation\n"); 
		if (globalWarn & L2_GRE) fprintf(stdout, "Warning: GRE encapsulation\n"); 


		fclose(logFile);
	}
}
#ifndef __APPLE__
void print_IP_interfaces(FILE * file) {
	struct ifreq *ifr;
	struct ifconf ifc;
	char buf[1024];
	int nInterfaces;
	int i;

	//Open a socket
	int sd = socket(PF_INET, SOCK_STREAM, 0);
	if (sd == -1) {
		fprintf(stdout, "ERROR: Cannot retrieve the IPs of the machine (socket problem)\n");
		return;
	}

	//Retrieve the sniffing interface in live capture
	if (captureInterface && captureDescriptor) {

		struct ifreq sniff_ifr;
		strcpy(sniff_ifr.ifr_name, captureInterface);

		if (ioctl(sd, SIOCGIFADDR, &sniff_ifr) < 0) {
			fprintf(stdout, "ERROR: Cannot retrieve the IPs of the machine (Sniffing interfaces problem)\n");
			return;
		}

		fprintf(file, "\tSniffing If:\t%s\t%s", sniff_ifr.ifr_name, inet_ntoa(((struct sockaddr_in *) &sniff_ifr.ifr_addr)->sin_addr));

		if (ioctl(sd, SIOCGIFHWADDR, &sniff_ifr) < 0) {
			fprintf(stdout, "ERROR: Cannot retrieve the IPs of the machine (MAC Error)\n");
			return;
		}

		fprintf(file, "\t%s\n", ether_ntoa(
				(struct ether_addr *) sniff_ifr.ifr_hwaddr.sa_data));
	}

	/* Query available interfaces and find out the number*/
	ifc.ifc_len = sizeof(buf);
	ifc.ifc_buf = buf;
	if (ioctl(sd, SIOCGIFCONF, &ifc) < 0) {
		fprintf(stdout, "ERROR: Cannot retrieve the IPs of the machine (Number of interfaces problem)\n");
		return;
	}

	/* Iterate through the list of interfaces. */
	ifr = ifc.ifc_req;
	nInterfaces = ifc.ifc_len / sizeof(struct ifreq);
	for (i = 0; i < nInterfaces; i++) {
		struct ifreq *item = &ifr[i];

		//If we are in live mode don't print again the sniffing interface
		if (!(captureInterface && captureDescriptor && strcmp(captureInterface,
				item->ifr_name) == 0)) {

			/* Show the device name and IP address */
			fprintf(file, "\tIf (not used):\t%s\t%s", item->ifr_name,
					inet_ntoa(
							((struct sockaddr_in *) &item->ifr_addr)->sin_addr));

			/* Get the MAC address */
			if (ioctl(sd, SIOCGIFHWADDR, item) < 0) {
				fprintf(stdout, "ERROR: Cannot retrieve the IPs of the machine (MAC Error)\n");
				return;
			}

			fprintf(file, "\t%s\n", ether_ntoa(
					(struct ether_addr *) item->ifr_hwaddr.sa_data));
		}
	}
	close(sd);
}
#else
void print_IP_interfaces(FILE * file) {
	struct ifaddrs *if_addrs = NULL;
	struct ifaddrs *if_addr = NULL;
	void *tmp = NULL;
	char buf[INET6_ADDRSTRLEN];
	if (0 == getifaddrs(&if_addrs)) {
		//Retrieve the sniffing interface in live capture
		if (captureInterface && captureDescriptor) {
			for (if_addr = if_addrs; if_addr != NULL; if_addr = if_addr->ifa_next) 
			{
				if (if_addr->ifa_name == captureInterface){
					if(if_addr->ifa_addr->sa_family == AF_INET){
						tmp = &((struct sockaddr_in *)if_addr->ifa_addr)->sin_addr;
					} else {
						tmp = &((struct sockaddr_in6 *)if_addr->ifa_addr)->sin6_addr;
					}
					fprintf(file, "\tSniffing if:\t%s\t%s",if_addr->ifa_name, inet_ntop(if_addr->ifa_addr->sa_family,tmp,buf, sizeof(buf)));
					if (if_addr->ifa_addr != NULL && if_addr->ifa_addr->sa_family == AF_LINK) 
					{
						struct sockaddr_dl* sdl = (struct sockaddr_dl *)if_addr->ifa_addr;
						uint8_t mac[6];
						if(6 == sdl->sdl_alen)
						{
							memcpy(mac, LLADDR(sdl), sdl->sdl_alen);
							fprintf(file,"\t%02x:%02x:%02x:%02x:%02x:%02x\n",mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
						}
					}
				}
				continue;
			}
		}
		freeifaddrs(if_addrs);
		if_addrs = NULL;
	}
	/* Query available interfaces. */
	if (0 == getifaddrs(&if_addrs)) {
		for (if_addr = if_addrs; if_addr != NULL; if_addr = if_addr->ifa_next){
			if (!(captureInterface && captureDescriptor && if_addr->ifa_name == captureInterface))
			{
				//get IP 
				if(if_addr->ifa_addr->sa_family == AF_INET){
					tmp = &((struct sockaddr_in *)if_addr->ifa_addr)->sin_addr;
				} else {
					tmp = &((struct sockaddr_in6 *)if_addr->ifa_addr)->sin6_addr;
				}	
				fprintf(file, "\tIf (not used):\t%s\t%s",if_addr->ifa_name, inet_ntop(if_addr->ifa_addr->sa_family,tmp,buf, sizeof(buf)));
				//get MAC
				if (if_addr->ifa_addr != NULL && if_addr->ifa_addr->sa_family == AF_LINK) 
				{
					struct sockaddr_dl* sdl = (struct sockaddr_dl *)if_addr->ifa_addr;
					uint8_t mac[6];
					if(6 == sdl->sdl_alen)
					{
						memcpy(mac, LLADDR(sdl), sdl->sdl_alen);
						fprintf(file,"\t%02x:%02x:%02x:%02x:%02x:%02x\n",mac[0], mac[1], mac[2], mac[3], mac[4], mac[5]);
					}

				}	
			}
		}
	}
  	return;
}
#endif
file_list_t* get_capture_file_list(FILE *list_file) {
#if VERBOSE != 0
		fprintf(stdout, "checking list file...\n");
#endif
	if (feof(list_file)) return NULL; // file is empty

	char stats_possible = 1; // needed to remember if stats could be red out or not
	char *buffer = calloc(1024, sizeof(char));
	pcap_t* tempDescriptor;
	file_list_t *list = calloc(1, sizeof(file_list_t));
	file_list_element_t *act_element = list->file_list;

	while (fgets(buffer, 1024, list_file) != NULL) {
		// Test if comment line
		if (buffer[0] == '#') continue;

		// cut off newline char
		if (buffer[strlen(buffer) - 1] == '\n') buffer[strlen(buffer) - 1] = '\0';

#if VERBOSE != 0
		fprintf(stdout, "    checking file %s...\n", buffer);
#endif

		// Test if valid pcap file
		if ((tempDescriptor = pcap_open_offline(buffer, errbuf)) == NULL) {
			fprintf(stdout, "ERROR during loading file %lu: %s: %s\n", list->number_of_files, buffer, errbuf);
			terminate();
		}

		// file is valid, add it to list
		file_list_element_t *new_element = calloc(1, sizeof(file_list_element_t));
		new_element->file_name = calloc((strlen(buffer) + 1), sizeof(char));
		strncpy(new_element->file_name, buffer, strlen(buffer));
		if (!act_element) {
			list->file_list = new_element;
			act_element = new_element;
		} else {
			act_element->next = new_element;
			act_element = act_element->next;
		}

		if (!stats_possible) {
			// do nothing
		} else if (stat(new_element->file_name, &fileStats) == -1) {
			fprintf(stdout, "ERROR: Cannot get complete file stats!\nfile containing error: %s, errno =  %d\n", new_element->file_name, errno);
			new_element->file_size = 0;
			list->size_of_files = 0;
			stats_possible = 0;
		} else {
			new_element->file_size = fileStats.st_size;
			list->size_of_files += new_element->file_size;
		}
		list->number_of_files++;
	}

	return list;
}

