/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __NETWORK_HEADERS_H__
#define __NETWORK_HEADERS_H__

// Constant Definition

// user defines
#define IPV6_ACTIVATE	0

// Ether Protocols
#ifndef ETHERTYPE_IPV6
#define ETHERTYPE_IPV6	0x86DD
#endif

// Ether Type Defs network order
#define ETHERTYPE_VLANn		  0x0081		  	
#define ETHERTYPE_ARPn		  0x0608
#define ETHERTYPE_RARPn		  0x3580
#define ETHERTYPE_IPn		  0x0008
#define ETHERTYPE_IPV6n		  0xDD86
#define ETHERTYPE_MPLS_UNICASTn   0x4788
#define ETHERTYPE_MPLS_MULTICASTn 0x4888
#define ETHERTYPE_PPPoE_Dn	  0x6388
#define ETHERTYPE_PPPoE_Sn	  0x6488
#define ETHERTYPE_LLDP_n	  0xCC88

#define ETHERTYPE32MASKn 	  0xffff0000

// VLAN
#define VLAN_8021Qn32		  0x81000000	 
#define VLANID_MASK16		  0x0fff	 

// MPLS
#define	BTM_MPLS_STKn16		  0x0001
#define	BTM_MPLS_STKn32		  0x00010000

// L2TP
#define L2TP_V2	 0x0200
#define L2TP_V3  0x0300
#define L2TP_VER 0Xff00
#define L2TP_FLG 0X00ff

#define L2TP_TYP 0x0080
#define L2TP_LEN 0x0040
#define L2TP_SQN 0x0008
#define L2TP_OFF 0x0002
#define L2TP_PRI 0x0001

#define L2TP_PORT_N	0xa506

// L2TP_PPP
#define PPP_ADD_CTL 0x03ff // not numbered/listen to all
#define PPP_PROT_IP 0x2100 // IPv4 
#define PPP_PROT_IP 0x2100 // IPv6 

// GRE
#define GRE_CKSMn	0x00000080	// Checksum
#define GRE_RTn		0x00000040	// Routing offset
#define GRE_KEYn	0x00000020	// Key
#define GRE_SQn 	0x00000010	// Sequence Number
#define GRE_SSRn	0x00000080	// Strickt Source Routing

#define GRE_RECURn	0x00000007	// Recursion control
#define GRE_ACKn	0x00008000	// Acknowledge Number
#define GRE_FLAGSn	0x0000f800	// Flags
#define GRE_Vn		0x00000700	// Version 

#define GRE_PROTOn	0xffff0000	// encapsulated protocols

// GRE protocols
#define GRE_PPPn 	0x0b880000	// PPP

// GRE_PPP
#define GRE_PPP_CMPRSS 	0xfd	


// L3 header definitions:
#define L3_ICMP	0x01
#define L3_IGMP	0x02
#define L3_TCP	0x06
#define L3_UDP	0x11
#define L3_GRE	0x2f
#define L3_OSPF	0x59

// L4 Header definitions: TCP Flags

#define TH_FIN	0x01
#define TH_SYN	0x02
#define TH_RST	0x04
#define TH_PUSH	0x08
#define TH_ACK	0x10
#define TH_URG	0x20
#define TH_ECE	0x40
#define TH_CWR	0x80

#define TH_SYN_ACK	0x12
#define TH_FIN_ACK	0x11
#define TH_RST_ACK	0x14
#define TH_SYN_FIN	0x03
#define TH_RST_FIN	0x05
#define TH_SYN_FIN_RST	0x07
#define TH_ARSF		0x17
#define TH_NULL		0x00
#define TH_XMAS		0x29
#define TH_ALL_FLAGS	0x3f

#define FRAGID_N	0xff1f
#define MORE_FRAG_N	0x0020
#define FRAGIDM_N	0xff3f	
#define FRAGID_1P_N	0x0000	


// IPv6 definitions 
#define FRAG6ID_N	0xf8ff
#define MORE_FRAG6_N	0x0100
#define FRAG6IDM_N	0xf9ff	
#define FRAG6ID_1P_N	0x0000	


// includes
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include <time.h>

// local includes

// structs


// MPLS
typedef struct mplsHdrh_s {
	uint32_t TTL:8;
	uint32_t S:1;
	uint32_t Exp_ToS:3;
	uint32_t label:20;
} mplsHdrh_t;
	
typedef union mplsh_s {
	mplsHdrh_t mplshs;
	uint32_t mplshu;
} mplsh_t;

// ethernet header
typedef struct ethernetHeader_s {
	uint8_t  ether_dhost[6];	// destination eth addr
	uint8_t  ether_shost[6];	// source ether addr
	uint16_t ether_type;		// packet type ID field
} ethernetHeader_t;


// IP header
typedef struct ipHeader_s {
	uint8_t	ip_vhl;				// header length, version
	#define IP_V(ip)	(((ip)->ip_vhl & 0xf0) >> 4)
	#define IP_HL(ip)	((ip)->ip_vhl & 0x0f)
	uint8_t	ip_tos;				// type of service
	uint16_t	ip_len;			// total length
	uint16_t	ip_id;			// identification
	uint16_t	ip_off;			// fragment offset field
	#define	IP_DF 0x4000			// dont fragment flag
	#define	IP_MF 0x2000			// more fragments flag
	#define	IP_OFFMASK 0x1fff		// mask for fragmenting bits
	uint8_t	ip_ttl;				// time to live
	uint8_t	ip_p;				// protocol
	uint16_t	ip_sum;			// checksum
	struct in_addr ip_src, ip_dst;		// source and dest address
} ipHeader_t;


// tcp header
typedef struct tcpHeader_s {
	uint16_t source;
	uint16_t dest;
	uint32_t seq;
	uint32_t ack_seq;
	uint16_t res1:4;
	uint16_t doff:4;
	uint16_t fin:1;
	uint16_t syn:1;
	uint16_t rst:1;
	uint16_t psh:1;
	uint16_t ack:1;
	uint16_t urg:1;
	uint16_t res2:2;
	uint16_t window;
	uint16_t check;
	uint16_t urg_ptr;
} tcpHeader_t;

// udp header
typedef struct udpHeader_s {
  uint16_t source;
  uint16_t dest;
  uint16_t len;
  uint16_t check;
} udpHeader_t;


// icmp header
typedef struct icmpHeader_s {
	uint8_t type;				// message type
	uint8_t code;				// type sub-code
	uint16_t checksum;
	union {
		struct {
			uint16_t id;
			uint16_t sequence;
		} echo;				// echo datagram
		uint32_t gateway;		// gateway address
		struct {
			uint16_t __unused;
			uint16_t mtu;
		} frag;				// path mtu discovery
	} un;
} icmpHeader_t;


// vlan tag, also used for MPLS header
typedef struct _8021Q_s {
	uint16_t identifier;
	uint16_t vlanID;
} _8021Q_t;

// PPP header
typedef struct pppH_s {
	uint16_t addctl;
	uint16_t prot;
} pppHdr_t;

typedef union pppHu_s {
        pppHdr_t pppHdru;
        uint32_t pppHdrc;
} pppHu_t;


// Content definitions
// dhcp
typedef struct dhcpHeader_s {
	uint8_t opcode;
	uint8_t hwType;
	uint8_t hwAddrLen;
	uint8_t hopCnt;
	uint32_t transID;
	uint16_t num_sec;
	uint16_t flags;
	uint32_t clientIP;
	uint32_t yourIP;
	uint32_t servIP;
	uint32_t gwIP;
	uint32_t clientHWaddr[4];
	uint32_t servHostName[16];
	uint32_t bootFname[32];
	uint32_t optMagNum;
} dhcpHeader_t;

// ospf
typedef struct ospfHeader_s {
	uint8_t version;
	uint8_t type;
	uint16_t len;
	uint32_t routerID;
	uint32_t areaID;
	uint16_t chksum;
	uint16_t autType_res;
	uint64_t authfield;
	uint8_t data;
} ospfHeader_t;

typedef struct ospfType1_s {
	uint32_t netMask;
	uint16_t helloInt;
	uint8_t options;
	uint8_t routPrio;
	uint32_t routDeadInt;
	uint32_t desRtr;
	uint32_t backupRtr;
	uint32_t neigbors;
} ospfType1_t;

typedef struct ospfType2_s {
	uint16_t intMTU;
	uint8_t options;
	uint8_t flags;
	uint32_t DDSeqNum;
	uint8_t lsaHdr;
} ospfType2_t;

typedef struct ospfType3_s {
	uint32_t linkStType;
	uint32_t linkStID;
	uint32_t advRtr;
	uint8_t lsa;
} ospfType3_t;

typedef struct ospfLSA_s { // type 5 has multiple of these
	uint16_t lsAge;
	uint8_t options;
	uint8_t lsType;
	uint32_t lsaID;
	uint32_t lsSeqNum;
	uint16_t lsChksum;
	uint16_t chksum;
} ospfLSA_t;

typedef struct ospfType4_s {
	uint32_t numLSA;
	ospfLSA_t ospfLSAP;	
} ospfType4_t;


// IPv6 structs definitions


//#if IPV6_ACTIVATE == 1

typedef union ipAddr_s {
	uint32_t IPv4x[4];
	struct in_addr IPv4; 	// ipv4 address 
	struct in6_addr IPv6; 	// ipv6 address
	uint64_t IPv6L[2]; 	// ipv6 address 2*64 bit max chunk for masking ops
} ipAddr_t;


typedef struct ip6Header_s {
	uint32_t vtc_flw_lbl; 	// first word: ver, tcl, flow
	uint16_t payload_len; 	// payload length
	uint8_t next_header; 	// next protocol
	union {
		uint8_t hop_limit; 	// hop limit
		uint8_t ip_ttl; 	// hop limit
	};
	ipAddr_t ip_src; 	// source address
	ipAddr_t ip_dst; 	// destination address
} ip6Header_t;

typedef struct ip6OptHdr_s {
	uint8_t next_header; 
	uint8_t len;
	uint8_t options;
} ip6OptHdr_t;

typedef struct ip6FragHdr_s {
	uint8_t next_header; 
	uint8_t res;
	uint16_t frag_off;
	uint32_t id;
} ip6FragHdr_t;

typedef struct ip6RouteHdr_s {
	uint8_t next_header; 
	uint8_t len;
	uint8_t route_type;
	uint8_t seg_left;
} ip6RouteHdr_t;

//#endif 

// general Layer 2 - 4 headers
typedef union l2Header_s {
	ethernetHeader_t ethernetHeader;
} l2Header_t;

typedef union l3Header_s {
	ipHeader_t ipHeader;
#if IPV6_ACTIVATE == 1
	ip6Header_t ip6Header;
#endif
} l3Header_t;

typedef union l4Header_s {
	tcpHeader_t tcpHeader;
	udpHeader_t udpHeader;
	icmpHeader_t icmpHeader;
} l4Header_t;

// gre encapsulation v1,2
typedef struct greHeader_s {
	uint32_t hdrFlags:5; // 
	uint32_t recur:3; // 
	uint32_t ack:1; // 
	uint32_t flags:4; // 
	uint32_t ver:3; // 
	uint32_t proto:16; //
	uint16_t plength;
	uint16_t CallID;
} greHeader_t;

// layer 2 tunneling protocol v2
typedef struct l2tpv2Header_s {
	uint16_t res3:4;	// reserved
	uint16_t ver:4;		// l2tp version
	uint16_t type:1;	// message type (0: Data, 1: Control)
	uint16_t len:1;		// message length present
	uint16_t res:2;		// reserved
	uint16_t seq:1;		// sequence numbers present
	uint16_t res2:1;	// reserved
	uint16_t off:1; 	// offset number present
	uint16_t prio:1;	// priority (zero on Control messages)
	uint16_t length;	// length
	uint16_t tID;		// sequence number
	uint16_t sID;		// sequence number expected
	uint16_t sN; 		// sequence number
	uint16_t sNExp; 	// sequence number expected
	uint16_t offSize;	// offset size
	uint16_t offPad;	// offset pad
} l2tpv2Header_t;

/*
typedef union l7Header_s {
	uint8_t pld;
	l2tpv2Header_t *l2TPP; // L2TP pointer 
	dnsHeader_t dnsHeader;
	httpHeader_t httpHeader;
} l7Header_t;
*/




// a common packet pointer, holds pointers to each layer's header

typedef struct packet_s {
	const u_char* raw_packet; // Network order
	const struct pcap_pkthdr *pcapHeader; // Network order
	const l2Header_t *layer2Header; // network order
	const uint32_t *vlans; // Network order, 
	const uint32_t *mpls; // Network order, MPLS pointer
	const l3Header_t *layer3Header;
	const l4Header_t *layer4Header;
	const greHeader_t *greHdr; // Network order pointer to GRE v1,2 Header
	const l2tpv2Header_t *l2TPHdr; // Network order pointer to L2TP V2 Header
	const l3Header_t *greLayer3Hdr; // Network order
	const l3Header_t *l2tpLayer3Hdr; // Network order
	const pppHu_t *pppHdr; // Network order, pointer to PPP header
#if IPV6_ACTIVATE == 1
	const ip6OptHdr_t *ip6OptHdr;
	const ip6FragHdr_t *ip6FragHdr;
	const ip6RouteHdr_t *ip6RouteHdr;
#endif // IPV6_ACTIVATE
	const uint8_t *layer7Header; // pointer to payload
	uint16_t l2HdrLen; // Host order
	uint16_t l3HdrLen; // Host order
	uint16_t l4HdrLen; // Host order
	uint16_t snapL2Length; // Host order, includes layer2 header, can be truncated due to limited snaplength, needs uint32_t
	uint16_t snapL3Length; // Host order, includes layer3 header, can be truncated due to limited snaplength
	uint16_t snapL4Length; // Host order, includes layer4 header, can be truncated due to limited snaplength
	uint16_t snapL7Length; // Host order, only higher packet payload (layer 7), can be truncated due to limited snaplength
	uint16_t packetL2Length; // Host order, derived from IP Header length field + length of L2 header, needs uint32_t 
	uint16_t packetRawLength; // Host order, derived from pcap
	uint16_t packetLength; // Host order, derived from IP Header length field, defined by PACKETLENGTH var in packetCapture.h (0: including L2-4 header, including L3-4 header, 2: including L4 header, 3: Only payload L7)
	uint32_t status; // 
	uint16_t layer2Type; // ethernet, ...
	uint16_t layer3Type; // IPV4, IPV6, ...
	uint8_t layer4Type; // tcp, udp, icmp, igmp, ...
//	_8021Q_t vlanTag;
} packet_t;



typedef struct flow_s {
	struct flow_s *lruNextFlow, *lruPrevFlow; // pointers to the next and previous flow in lru list
	struct timeval lastSeen; 	// the earliest time we've seen this flow
	struct timeval firstSeen; 	// the first time we've seen this flow
	struct timeval duration; 	// lastSeen-firstSeen. NOTE: Not available before flow completely terminated
	// flow identification
#if IPV6_ACTIVATE == 1 
	ipAddr_t srcIP, dstIP;
#else // IPV6_ACTIVATE == 0
	struct in_addr srcIP, dstIP;
#endif
	uint16_t vlanID;
#if IPV6_ACTIVATE == 1 
	union {
#endif 
		struct { 
			uint16_t srcPort;
			uint16_t dstPort;
		};
#if IPV6_ACTIVATE == 1 
		uint32_t fragID6;
	};
	uint8_t IPVer; 		// Ip versions or other l3 transport in bit notation
#endif
	uint8_t layer4Protocol; // see http://en.wikipedia.org/wiki/List_of_IP_protocol_numbers
	// flow identification end
	uint32_t findex; 	// flow index
#if IPV6_ACTIVATE == 1 
	uint32_t lastFragIPID; // for fragPend hash cleanup when flow terminates unexpectedly
#else
	uint16_t lastFragIPID; // for fragPend hash cleanup when flow terminates unexpectedly
#endif 
	uint16_t lastIPID; // for duplicate IPID dection errors on switches and scanning tools 
	uint32_t status; // status of flow e.g. fragmentation processing
	unsigned long flowIndex;
	unsigned long oppositeFlowIndex;
	float timeout; //the timeout of this flow in seconds
} flow_t;

#endif // __NETWORK_HEADERS_H__

