/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __PACKET_CAPTURE_H__
#define __PACKET_CAPTURE_H__

#define PRINTABLE_ASCII(c) ( (c > 31 && c < 127) ? (1) : (0))

// includes
#include <pcap.h>
#include <string.h>
#include <net/ethernet.h>

// local includes
#include "networkHeaders.h"
#include "hashTable.h"

// flow l3Status variable see main.h golobal warn definition

// User Defines
// interpretation of packet length
// following plugins are influenced by PACKETLENGTH: nFirstPacketsStats, packetSizeInterArrivalTimeHisto, basicLayer4CalcStatistics, descriptiveStatistics
#define PACKETLENGTH 		3 // 0: including L2,L3 and L4 haeder, 1: including L3 and L4 header, 2: including L4 header, 3: Only higher layer payload (Layer 7)
#define FRGIPPKTLENVIEW 	1 // if (PACKETLENGTH == 1) { 0: IP header stays with 2nd++ fragmented packets, 1: IP header stripped from 2nd++ fragmented packets }
#define NOLAYER2		0 // artificial traffic: { 0: Automatic L3 Header discovery in Layer, 1: Manual L3 Header Positioning}
#define NOL2_L3HDROFFSET	0 // artificial traffic: if (NOLAYER2) Offset of L3 header
#define MULTIPKTSUP		0 // multi-packet supression

// special IPv6 user defines
#define MAXHDRCNT		5 // maximal header count IPv6, minimum 3


// functions
void perPacketCallback(u_char* inqueue_uchar, const struct pcap_pkthdr *pcapHeader, const u_char* packet);
extern void processPacket(u_char* inqueue_uchar, const struct pcap_pkthdr *pcapHeader, const u_char* packet);
extern void dissembleIPv4Packet(packet_t *packet);
extern void dissembleIPv6Packet(packet_t *packet);
extern unsigned long flowCreate(packet_t *packet);
extern unsigned long flowCreate6(packet_t *packet);
extern void updateLRUList(flow_t *flowToUpdate);
// uint32_t subnet_testP(subnettable_t* table, in_addr_t ip);

// variables
extern hashMap_t *mainHashMap;
packet_t newPacket; // dissemble structure for analyzing the packet
uint16_t srcPort, dstPort; // globally used in packerCapture.c
uint16_t innerVLANID; // globally used in packetCapture.c
ipHeader_t *ipHeader; // ip header 'shape'. Is in header file due to performance reasons
unsigned long totalFlows; // how many flows we processed in total

flow_t *flows; // where our real flows will be stored. the hashtable simply exists for easy flow lookup (? stbf104g, ist doch in der fktn def)
flow_t lruHead, lruTail; // front and tail lru flows. Are unused and don't contain values
flow_t hashHelper; 

#if FRAGMENTATION >= 1
extern  hashMap_t *fragPendMap;
unsigned long *fragPend; 
#endif //FRAGMENTATION

#endif
