/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef __TRANALYZER_H__
#define __TRANALYZER_H__

#include "../config.h"

// application
#define APPNAME PACKAGE_NAME
#define APPVERSION PACKAGE_VERSION
#define PLUGIN_FOLDER ".tranalyzer/plugins/"

#define LOGO_FILE_NAME RESOURCEDIR "anteater_ascii.txt"
#define ANTS_FILENAME RESOURCEDIR "ants_ascii.txt"
#define MAX_FILENAME_LEN 2048 // should be enough for the wildest path- and filenames

// Suffixes for the generated files
#define SMACC_SUFFIX "_smacc"
#define PACKETS_SUFFIX "_packets"
#define LOG_SUFFIX "_log"

/*
 *  The debug level:
 *  0: no debug output
 *  1: debug output which occurs only once or very seldom (i.e. initialize stuff, errors, ...)
 *  2: debug output which occurs in special situations, but not regulary (i.e. border conditions, restructure works, ...)
 *  3: debug output which occurs regulary (i.e. every packet)
 */
#define DEBUG 0

/*
 * The verbose level:
 * 0: no verbose output
 * 1: Basic additional output
 * 2: More additional output, not really interesting for 'normal' users
 */
#define VERBOSE 0

#define ENABLE_LOG_FILE 0

// std out reporting
//#define STDOUTREPORTS 3 // 0: only time no statistics ; 1: only statistics; No warnings 2: No embedded protocol warnings 3: full report including embedded protocol warnings

// timing ops
#define NO_PACKETS_WAIT_MICROSEC 1000 // if no packets are to be processed, sleep some time
#define NON_BLOCKING_MODE 1 // non blocking mode of pcap_dispatch is necessary when using the CTRL+C keystroke feature and complete flows are to be expected
#define DISPATCH_SLOW_DOWN 0 // a value in microsec to slow down the dispatching process, usually used for debugging

// Flow output buffer
#define MAIN_OUTPUT_BUFFER_SIZE	1000000 // the size of the main output buffer

// IO Buffering
#define ENABLE_IO_BUFFERING 0 // enables buffering of the packets in a queue. useful in life sniffing if there is a (short) packet burst
#define IO_BUFFER_QUEUE_FULL_WAIT_MICROSEC 2000
#define IO_BUFFER_SIZE 10000 // max packets that can be stored in the buffer

/* pcap */
#define SNAPLEN BUFSIZ
#define CAPTURE_TIMEOUT		1000
#define ENABLE_BPF_OPTIMIZATION 0
#define PACKETS_PER_BURST	1 // THIS VALUE SHOULD NOT BE CHANGED!!!

#define PROCESS_SIGKILL_LEVEL_THRESHOLD 3 // after n-times CTRL+C keystroke hit, kill the process


// Tranalyzer User Operational modes
#define SENSOR_ID 0x29A // sensor id for multisensor mode

// Operation modes
#define ALARM_MODE 0
#define FORCE_MODE 0

// Protocol stack 
#define L2TP 0 // L2TP processing on: 1, off: 0
#define GRE 0 // GRE processing on: 1, off: 0
#define FRAGMENTATION 1 // Fragmentation processing on: 1, off: 0
#define FRAG_HLST_CRFT 1 // if (FRAGMENTATION) then 1: Enables crafted packet processing (header missing, senseless flags etc); off: 1
#define FRAG_ERROR_DUMP 0 // if (FRAGMENTATION) then 1: Dumps flawed fragmented packet info on commandline for timebased identification ; WARNING: IF FRAG_HLST_CRFT != 0 THEN EVERY HEADERLESS FRAG PACKET IN A FLOW IS REPORTED!

// Aggregation modes
#define SUBNET	0x80
#define VLANID	0x20
#define SRCIP	0x10
#define DSTIP	0x08
#define SRCPORT	0x04
#define DSTPORT	0x02
#define L4PROT	0x01

// Flow Agreggation 
#define AGGREGATIONFLAG 0x00 // each bit: 1 : aggregation activated, def see aggregation mode local def above
#define SRCIPMASKn	0x00000000	// network order
#define DSTIPMASKn	0x00000000	// network oder
#define SRCPORTLW	53	//
#define SRCPORTHW	53			//
#define DSTPORTLW	53		//
#define DSTPORTHW	53	//

// The standard timeout for a flow in seconds
#define FLOW_TIMEOUT 183

// The sizes of the hash table
#define HASHTABLE_BASE_SIZE 240000UL
#define HASHCHAINTABLE_BASE_SIZE 120000UL
#define HASHTABLE_ENTRY_NOT_FOUND ULONG_MAX



// Jenkins golden number
#define JHASH_GOLDEN_RATIO 0x9e3779b9

// Network parameters
#define MAX_MTU 1552 // make this aligned! (divisible by 4)
#define MAX_PORTS 65535
#define MAX_IP_PRINT_LEN 16

// define if flow output should also be written into a char array
// currently not used
#define FLOW_OUTPUT_TO_STRING 0

// Transport protocol

#define L2_ETHERNET 1
// Switch for underlying layer 2 protocol.
#define L2PROTO L2_ETHERNET

#endif // __TRANALYZER_H__ 

