/*
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

/*
 * tranalyzer_utils.c
 */

/* includes */
#include <stdlib.h>
#include <stdio.h>
#include <string.h>

/* local includes */
#include "tranalyzer.h"
#include "tranalyzer_utils.h"
#include "main.h"

char* loadAscii(char fileName[]) {
	FILE *logoFile;
	char *logo;
	// open file
	if ((logoFile = fopen(fileName, "r")) == NULL) return NULL;

	struct stat logoStats;
	if (stat(fileName, &logoStats) != -1) {
		logo = malloc((logoStats.st_size + 1) * sizeof(char));
		memset(logo, '\0', (logoStats.st_size + 1) * sizeof(char));
	} else return NULL;

	// fill with logo
	char c;
	while ((c = fgetc(logoFile)) != EOF) strncat(logo, &c, 1);

	fclose(logoFile);

	return logo;
}

/*
 * Prints a packet into a stream. Useful for debugging.
 * Args:
 *
 * stream		A file descriptor, (Hint: stdout is also a file descriptor)
 * packet		A packet_t struct
 * width		The number of bytes printed in one row. The absolute amount of
 * 				chars in one row is then 4 times + 6 chars. E.g. on a screen
 * 				with 80 chars width, you should use a maximum value of 18,
 * 				better 16.
 */

void print_packet(FILE* stream, packet_t *packet, int width) {

	int i, offset;

	for (i = 0; i < (width * 3) + width + 6; i++) fprintf(stream, "-");
	fprintf(stream, "\n");

	fprintf(stream, "Packet No.: %lu\n", numOfPackets);

	// Print Layer 2 info
	if (packet->layer2Type == L2_ETHERNET) {
		uint8_t* tempMac = (uint8_t*) &(packet->layer2Header->ethernetHeader.ether_shost);
		fprintf(stream, "Mac Src: %02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8"\n", tempMac[0], tempMac[1], tempMac[2], tempMac[3], tempMac[4], tempMac[5]);

		tempMac = (uint8_t*) &packet->layer2Header->ethernetHeader.ether_dhost;
		fprintf(stream, "Mac Dst: %02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8":%02"PRIx8"\n", tempMac[0], tempMac[1], tempMac[2], tempMac[3], tempMac[4], tempMac[5]);

		if (packet->vlans)
			fprintf(stream, "VLAN ID: %"PRIu32"\n", packet->vlans[0]);
	}

	if (packet->layer3Header) {
		// Print Layer 3 info
		fprintf(stream, "Layer 3 Type: ");
		switch (packet->layer3Type) {
		case ETHERTYPE_IP:
			fprintf(stream, "IPv4");
			break;
		case ETHERTYPE_IPV6:
			fprintf(stream, "IPv6");
			break;
		case ETHERTYPE_ARP:
			fprintf(stream, "ARP");
			break;
		default:
			break;
		}
		fprintf(stream, " (0x%04X)\n", packet->layer3Type);

		if (packet->layer2Type == ETHERTYPE_IP) {
			//			uint8_t* tempIP = (uint8_t*) &(newPacket.layer3Header->ipHeader.ip_src.s_addr);
			//			fprintf(stream, "IP Src: %s", tempIP);
		}
	}

	// Start printing the raw packet
	for (i = 0; i < (width * 3) + width + 6; i++)
		fprintf(stream, "-");

	for (offset = 0; offset < packet->snapL2Length; offset += width) {

		fprintf(stream, "\n%04X|", offset);
		int j = (packet->snapL2Length - offset) < width ? (packet->snapL2Length - offset) : width;

		for (i = 0; i < j; i++) {
			fprintf(stream, "%02X ", packet->raw_packet[i + offset]);
		}

		// adjust position if less than 8 bytes (width) left
		for (; i < width; i++)
			fprintf(stream, "   ");

		fprintf(stream, "|");
		for (i = 0; i < j; i++) {
			if (PRINTABLE_ASCII(packet->raw_packet[i + offset])) {
				fprintf(stream, "%c", packet->raw_packet[i + offset]);
			} else {
				fprintf(stream, ".");
			}
		}
	}

	fprintf(stream, "\n");
	for (i = 0; i < (width * 3) + width + 6; i++)
		fprintf(stream, "-");
	fprintf(stream, "\n\n");
}
