#include "vector.h"

vector_t* initVector(unsigned long itemSize, unsigned long initSize, float incrementationStep) {
	vector_t *newVector = (vector_t*) malloc(sizeof(vector_t));
	if (!newVector) {
		fprintf(stderr, "ERROR: No memory left to allocate vector!\n");
		return NULL;
	}
	if (incrementationStep <= 1.0f) {
		fprintf(stderr, "ERROR: Incrementation step of vector is smaller than 1.0, which means the array gets smaller by inserting more elements!\n");
	}
	newVector->sizePerItem = itemSize;
	newVector->numAllocatedItems = initSize;
	newVector->numCurrentUsedItems = 0;
	newVector->incrementationStep = incrementationStep;
	newVector->data = (void*) malloc(itemSize*initSize);
	if (!newVector->data) {
		fprintf(stderr, "ERROR: No memory left to allocate vector elements!\n");
		return NULL;
	}
	return newVector;
}

void pushVector(vector_t *vector, void *newData) {
	// check if we have enough memory allocated to store new data
	if (vector->numCurrentUsedItems+1 > vector->numAllocatedItems) {
		// no, we have to expand the memory slice
		vector->numAllocatedItems = ceilf((float)vector->numAllocatedItems*vector->incrementationStep);
		vector->data = (void*) realloc(vector->data, vector->numAllocatedItems*vector->sizePerItem);
		
		if (!vector->data) {
			fprintf(stderr, "ERROR: No memory left to allocate vector elements!\n");
			return;
		}
	}

	// push the new data onto the vector struct
	memcpy(vector->data+vector->numCurrentUsedItems*vector->sizePerItem, newData, vector->sizePerItem);
	vector->numCurrentUsedItems++;
}

void destroyVector(vector_t *vector) {
	if (!vector) return;
	if (vector->data) free(vector->data); vector->data = NULL;
	vector = NULL;
}

void* linearLookupVector(vector_t *vector, void *data, unsigned long compareSize) {
	unsigned long i;
	for (i=0; i<vector->numCurrentUsedItems; i++) {
		if (memcmp(vector->data+i*vector->sizePerItem, data, compareSize) == 0) {
			return getVectorElement(vector, i);
		}
	}
	return NULL;
}


void* getVectorElement(vector_t *vector, unsigned long index) {
	if (index > vector->numCurrentUsedItems) return NULL;
	return vector->data+index*vector->sizePerItem;
}



