/*
 * tranalyzer_plugin_skeleton.c
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#include "tranalyzer_plugin_skeleton.h"

// Global variables 
#ifdef __APPLE__
extern hashMap_t* mainHashMap;
extern outputBuffer_t* main_output_buffer;
#endif // __APPLE__

// plugin variables
unsigned long plugin_PLUGINNAME_Index;
PLUGINNAME_flow_t *PLUGINNAME_flow;


// Tranalyzer functions

char* get_plugin_name() {
	return "tranalyzer_plugin_skeleton";
}

char* get_plugin_version() {
	return "0.0.0";
}

unsigned int get_supported_tranalyzer_version_major() {
	return 0;
}

unsigned int get_supported_tranalyzer_version_minor() {
	return 5;
}

char* get_dependencies() {
	return "";
}

void initialize(unsigned long pPluginIndex) {
	plugin_PLUGINNAME_Index = pPluginIndex;
#ifdef __APPLE__
        mainHashMap = getMainHashMap();
        baseFileName = getBaseFileName();
        specialOutput = getSpecialOutput();
#endif // __APPLE__

        PLUGINNAME_flow = malloc(sizeof(PLUGINNAME_flow_t) * mainHashMap->hashChainTableSize); // allocate struct for all flows
        memset(PLUGINNAME_flow, '\0', sizeof(PLUGINNAME_flow_t) * mainHashMap->hashChainTableSize); // set all to 0
}


binary_value_t* printHeader() {
	binary_value_t *main_bv = NULL;
        binary_value_t *act_bv = NULL;

	// sample header definition
	main_bv = bv_append_bv(main_bv, bv_new_bv("Single Var1", "Var1", 0, 1, bt_uint_64)); // 64 bit variable

        main_bv = bv_append_bv(main_bv, bv_new_bv("Single Var2", "Var2", 0, 1, bt_ip4_addr)); // ipv4 address or 32 bit

        main_bv = bv_append_bv(main_bv, bv_new_bv("Single Vector Var3;Var4", "Var3_Var4", 0, 2, bt_hex_32, bt_hex_16)); // two hex in a row

        main_bv = bv_append_bv(main_bv, bv_new_bv("Multiple Vector Var5;Var6", "Var5_Var6", 1, 2, bt_uint_8, bt_uint_8)); // multiple 2 * 8 bit groups in a row

	act_bv = bv_new_bv("Matrix/Multiple Vector Outut", "Nudelvector", 1, 1, 0); // a Matrix
        main_bv = bv_append_bv(main_bv, bv_add_sv_to_bv(act_bv , 0, 1, 1, bt_double));

}

void onFlowGenerated(packet_t* packet, unsigned long flowIndex) {
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *FlowP = &flows[flowIndex];
#if IPV6_ACTIVATE == 1
	ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE
	PLUGINNAME_flow_t *PLUGINNAME_P = &PLUGINNAME_flow[flowIndex];

// your code how you want to init
	memset( PLUGINNAME_P, '\0', sizeof(PLUGINNAME_flow_t) ); // set everything to 0

}


void claimLayer2Information(packet_t* packet) {
// sample code

ethernetHeader_t *ehP = (ethernetHeader_t*)packet->layer2Header;
mplsHdrh_t *mplsh = (mplsHdrh_t*)packet->mpls;

}


void claimLayer3Information(packet_t* packet) {
#if IPV6_ACTIVATE == 1
        ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
        ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

// your code

}


void claimLayer4Information(packet_t* packet, unsigned long flowIndex) {
	PLUGINNAME_flow_t *PLUGINNAME_P = &PLUGINNAME_flow[flowIndex];
#ifdef __APPLE__
	flows = getFlows();
#endif // __APPLE__
	flow_t *FlowP = &flows[flowIndex];

#if IPV6_ACTIVATE == 1
	ip6Header_t *ipHeader = (ip6Header_t*)packet->layer3Header;
#else // IPV6_ACTIVATE == 0
	ipHeader_t *ipHeader = (ipHeader_t*)packet->layer3Header;
#endif // IPV6_ACTIVATE

// your code


}

void onFlowTerminate(unsigned long flowIndex) {
	PLUGINNAME_flow_t *PLUGINNAME_flowP = &PLUGINNAME_flow[flowIndex];

#ifdef __APPLE__
	flows = getFlows();
	main_output_buffer = getMainOutputBuffer();
#endif // __APPLE__
	flow_t *FlowP = &flows[flowIndex];

// Sample code
	uint32_t i, j, cnt;

	// sample output
	// Single variable 64 bit variable
	outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var1, sizeof(uint64_t));

	// Single variable ip address: 32 bit
	outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var2, sizeof(uint32_t));

	// Single vector: 32 and 16 bit in a row separated by "_"
	outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var3, sizeof(uint32_t));
	outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var4, sizeof(uint16_t));

	// Multiple vector: uint8 groups separated by ";" and "_"
	cnt = NUM; 
	outputBuffer_append(main_output_buffer, (char*) &cnt, sizeof(uint32_t)); // number of rows

	outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var5, NUM * sizeof(uint8_t));
	outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var6, NUM * sizeof(uint8_t));

	// Matrix / Multiple vector: doubles separated by ";" and "_"
	cnt = NUM;
	outputBuffer_append(main_output_buffer, (char*) &cnt, sizeof(uint32_t)); // number of columns 
	
	for (i = 0; i < NUM; i++) {
		cnt = WURST; // number of rows
		outputBuffer_append(main_output_buffer, (char*) &cnt, sizeof(uint32_t)); 
		outputBuffer_append(main_output_buffer, (char*) &PLUGINNAME_flowP->var7[i][0], WURST * sizeof(double));
	}


}

void onApplicationTerminate() {

	if (PLUGINNAME_flow) {
		free(PLUGINNAME_flow);
		PLUGINNAME_flow = NULL;
	}

}

void bufferToSink(outputBuffer_t* buffer) {
// if you do not write an own outputBuffer routine, you dont need this here
}

