/*
 * tranalyzer_plugin_skeleton.h
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2, or (at your option)
 * any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
 */

#ifndef TRANALYZER_PLUGIN_SKELETON_H_
#define TRANALYZER_PLUGIN_SKELETON_H_

// global includes
#include <stdio.h>
#include <stdarg.h>
#include <stdlib.h>

// local includes
#include "main.h"
#include "outputBuffer.h"
#include "binaryValue.h"
#include "networkHeaders.h"

// user defines
#define NUM 5
#define WURST 10

// plugin defines

// sample plugin structures
typedef struct PLUGINNAME_Flow_s { // always large variables first to limit memory fragmentation
        double var7[NUM][WURST];
        uint64_t var1;
        struct in_addr var2; // or uint32_t is the same
        uint32_t var3;
        uint16_t var4;
        uint16_t var5[NUM];
        uint8_t var6[NUM];
} PLUGINNAME_flow_t;

// functions
char* get_plugin_name();
char* get_plugin_version();
unsigned int get_supported_tranalyzer_version_major();
unsigned int get_supported_tranalyzer_version_minor();
char* get_dependencies();
void initialize(unsigned long pPluginIndex);
binary_value_t* printHeader();
void onFlowGenerated(packet_t* packet, unsigned long flowIndex);
void claimLayer2Information(packet_t* packet);
void claimLayer3Information(packet_t* packet);
void claimLayer4Information(packet_t* packet, unsigned long flowIndex);
void onFlowTerminate(unsigned long flowIndex);
void onApplicationTerminate();
void bufferToSink(outputBuffer_t *buffer);

#endif // TRANALYZER_PLUGIN_SKELETON_H_

